#ifndef _ANTHOCNET_NETWORK_H_
#define _ANTHOCNET_NETWORK_H_


typedef struct struct_anthocnet_str AnthocnetData;

//UserCodeStart
#include <limits.h>
#include <math.h>
#include "buffer.h"
#include "random.h"
#include "routing_antsconstants.h"
#include "mac_802_11.h"
#include "partition.h"
#include "application.h"
#include "app_cbr.h"
#include "coordinates.h"
#include "mobility.h"
#include "mac_link.h"
#include <list>
#include "phy.h"
#include "phy_802_11.h"

typedef double Pheromone;
typedef AnthocnetData ANTHOCNETData; //Hack to deal with this QualNet ****

typedef enum ANTHOCNET_AntType_enum {
    ANTHOCNET_UNKNOWN = -3,
    ANTHOCNET_LINK_FAILURE = -2, //direct link failure
    ANTHOCNET_CLEAN_UP = -1,
    ANTHOCNET_REACTIVE_FORWARD = 0,
    ANTHOCNET_REACTIVE_BACKWARD = 1,
    ANTHOCNET_LINK_FAILURE_NOTIFICATION = 2,
    ANTHOCNET_PROACTIVE_FORWARD = 3,
    ANTHOCNET_REPAIR_FORWARD = 4,
    ANTHOCNET_REPAIR_BACKWARD = 5,
    ANTHOCNET_UNICAST_WARNING = 6,
    ANTHOCNET_HELLO = 7
} ANTHOCNETAntType;


typedef enum ANTHOCNET_table_entry_type_enum {
    ANTHOCNET_RT_DEST,
    ANTHOCNET_RT_DEST_NBR,
    ANTHOCNET_NT_NBR,
    ANTHOCNET_NT_NBR_DEST
} ANTHOCNETTableEntryType;


typedef enum ANTHOCNET_dest_status_type_enum {
    ANTHOCNET_NOT_INITIALIZED,
    ANTHOCNET_BOOTSTRAPPED,
    ANTHOCNET_FROM_HELLO_MSG,
    ANTHOCNET_TO_BE_CONFIRMED,
    ANTHOCNET_CONFIRMED,
    ANTHOCNET_PROBING,
    ANTHOCNET_IN_USE
} ANTHOCNETDestStatusType;



typedef struct ANTHOCNET_msg_id {
    short srcId;
    int seqNumber;
} ANTHOCNETMsgId;


typedef struct ANTHOCNET_list_entry_struct {
    struct ANTHOCNET_list_entry_struct* prev;
    struct ANTHOCNET_list_entry_struct* next;
    NodeAddress nodeAddress;
    void* data;
} ANTHOCNETListEntry;


typedef struct ANTHOCNET_list_struct {
    ANTHOCNETListEntry head;
    int numberOfEntries;
    void* data;
} ANTHOCNETList;


typedef struct ANTHOCNET_buffer_dest_struct {
    ANTHOCNETList messageList;
    clocktype oldestBufferTime;
    int numberOfAntGenerations;
    Message* timerMsg; //the timer running for reactive ants for this destination
    clocktype bufferDestCreationTime;
} ANTHOCNETBufferDest;


typedef struct ANTHOCNET_data_ptr_temp_data_struct {
    ANTHOCNETBufferDest* bufferDestData;
    NodeAddress destAddr;
    BOOL intermediateNodeFlag;
} ANTHOCNETDataPtrTempData;


typedef struct ANTHOCNET_ant_generation_id_struct {
    NodeAddress sourceAddress;
    unsigned short generationNumber;
} ANTHOCNETAntsGenerationID;


typedef struct ANTHOCNET_visited_node_struct {
    NodeAddress node; //A visited node
} ANTHOCNETVisitedNode;


//Data structure for a forward ant
typedef struct ANTHOCNET_Forward_Ant_struct {
    NodeAddress destinationAddress; //Thefinaldestination
    NodeAddress sourceAddress; //Thesource
    short numberOfVisitedNodes; //Number of nodes in list
    unsigned short generationNumber;
    float tripTime; //Accummulated virtual trip time
    ANTHOCNETVisitedNode* visitedNodes;
} ANTHOCNETForwardAnt;


//Data structure for a backward ant
typedef struct ANTHOCNET_Backward_Ant_struct {
    NodeAddress destinationAddress; //Thefinaldestination
    NodeAddress sourceAddress; //Thesource
    short numberOfVisitedNodes; //Number of nodes in list
    ANTHOCNETVisitedNode* visitedNodes;
    short currentPosition; //Current node's index into visitedNodes array
    unsigned short generationNumber;
    float tripTime; //Accummulated virtual trip time
    float prevSINR; //Hop time estimated at the mac layer of the previous node
    Pheromone pheromone;
} ANTHOCNETBackwardAnt;


typedef struct ANTHOCNET_RT_dest_next_hop_struct {
    Pheromone pheromone;
    clocktype lastUpdate;
    double hopsAvg;
    double timeAvg;
    double pathLengthSignalQualityAvg;
    unsigned short lastGeneration;
    Pheromone lastGenerationPheromone;
    Pheromone bootstrappedPheromone;
    clocktype bootstrappedTime;
    BOOLEAN bootstrappedForwarded;
    int interfaceIndex;
    NodeAddress interfaceAddress;
} ANTHOCNETRTDestNextHop;


typedef struct ANTHOCNET_RT_dest_predecessor_struct {
    int interfaceIndex;
    NodeAddress interfaceAddress;
} ANTHOCNETRTDestPredecessor;


typedef struct ANTHOCNET_RT_dest_struct {
    double hopsAvg;
    double timeAvg;
    ANTHOCNETList nextHops;
    ANTHOCNETList predecessors;
    int* hopsWindow;
    int windowIndex;
    int windowMin;
    int windowMax;
    ANTHOCNETAntsGenerationID idsWindow[ANTHOCNET_IDS_NUMBER];
    int idsIndex;
    clocktype lastUsed;
    clocktype waitingSince;
    clocktype lastDataTransported;
    short numberOfConfirmedEntries;
} ANTHOCNETRTDest;


typedef struct ANTHOCNET_NT_nbr_dest_struct {
    ANTHOCNETListEntry* RTDest;
} ANTHOCNETNTNbrDest;


typedef struct ANTHOCNET_NT_nbr_struct {
    ANTHOCNETList destinations;
    clocktype lastUpdate;
    int interfaceIndex;
    NodeAddress interfaceAddress;
    double SINR;
    double power;
} ANTHOCNETNTNbr;


typedef struct ANTHOCNET_buffer_message_struct {
    Message* msg;
    clocktype bufferTime;
} ANTHOCNETBufferMessage;


typedef struct ANTHOCNET_buffer_data_struct {
    int maxBufferSize;
    int currentBufferSize;
} ANTHOCNETBufferData;


typedef struct ANTHOCNET_reactive_ants_timer_info_struct {
    NodeAddress destAddr;
    BOOL intermediateNodeFlag; //Whether the node is an intermediate node for the packet (true), or the source node (false)
} ANTHOCNETReactiveAntsTimerInfo;


//Used in the error reporting ants to indicate the percentage of pheromone lost for a certain destination
typedef struct ANTHOCNET_lost_pheromone_struct {
    NodeAddress destAddr;
    double newBestHopsToGo;
    double newBestTimeToGo;
    Pheromone newBestPheromone;
} ANTHOCNETLostPheromone;


//Unicast Warning structure
typedef struct ANTHOCNET_Unicast_Warning_struct {
    NodeAddress nextHopAddr;
    NodeAddress destAddr;
    NodeAddress srcAddr;
} ANTHOCNETUnicastWarning;


typedef struct ANTHOCNET_Check_Paths_Visited_Node {
    NodeAddress nodeAddress;
    double probability;
} ANTHOCNETCheckPathsVisitedNode;


typedef struct ANTHOCNET_Check_Paths_Ant {
    NodeAddress destinationAddress;
    NodeAddress sourceAddress;
    NodeAddress previousAddress;
    int numberOfVisitedNodes;
    FILE* fp;
} ANTHOCNETCheckPathsAnt;


typedef struct ANTHOCNET_Start_Check_Paths {
    NodeAddress destAddr;
    FILE* fp;
} ANTHOCNETStartCheckPaths;


typedef struct ANTHOCNET_End_Check_Paths {
    Message* timerMsg;
    FILE* fp;
} ANTHOCNETEndCheckPaths;


typedef struct ANTHOCNET_hello_struct {
    unsigned char numberOfEntries;
} ANTHOCNETHello;


typedef struct ANTHOCNET_hello_entry_struct {
    NodeAddress destAddr;
    unsigned short discretePheromone;
    char bootstrapped;
} ANTHOCNETHelloEntry;


typedef struct ANTHOCNET_current_connections_data_struct {
    int counter;
    BOOL rejected;
} ANTHOCNETCurrentConnectionsData;



//UserCodeEnd

typedef struct {
    int totalReactiveAntsGenerated;
    int totalReactiveAntsGeneratedId;
    int totalReactiveBackwardAntsReceived;
    int totalReactiveBackwardAntsReceivedId;
    int totalLoopingAnts;
    int totalLoopingAntsId;
    int totalAntsForwarded;
    int totalAntsForwardedId;
    int totalReactiveForwardAntsResendings;
    int totalReactiveForwardAntsResendingsId;
    int totalLinkFailureNotificationsGenerated;
    int totalLinkFailureNotificationsGeneratedId;
    int totalProactiveAntsGenerated;
    int totalProactiveAntsGeneratedId;
    int totalRepairAntsGenerated;
    int totalRepairAntsGeneratedId;
    int totalRoutesRepaired;
    int totalRoutesRepairedId;
    int totalRoutesUnrepaired;
    int totalRoutesUnrepairedId;
    int totalCriticalLinkFailures;
    int totalCriticalLinkFailuresId;
    int totalReactiveInterUnacceptedAnts;
    int totalReactiveInterUnacceptedAntsId;
    int totalReactiveInterUnacceptedHops;
    int totalReactiveInterUnacceptedHopsId;
    int totalReactiveDestUnacceptedAnts;
    int totalReactiveDestUnacceptedAntsId;
    int totalReactiveDestUnacceptedHops;
    int totalReactiveDestUnacceptedHopsId;
    int totalReactiveAcceptedAnts;
    int totalReactiveAcceptedAntsId;
    int totalReactiveAcceptedHops;
    int totalReactiveAcceptedHopsId;
    int totalProactiveInterUnacceptedAnts;
    int totalProactiveInterUnacceptedAntsId;
    int totalProactiveInterUnacceptedHops;
    int totalProactiveInterUnacceptedHopsId;
    int totalProactiveDestUnacceptedAnts;
    int totalProactiveDestUnacceptedAntsId;
    int totalProactiveDestUnacceptedHops;
    int totalProactiveDestUnacceptedHopsId;
    int totalProactiveAcceptedAnts;
    int totalProactiveAcceptedAntsId;
    int totalProactiveAcceptedHops;
    int totalProactiveAcceptedHopsId;
    int totalRepairInterUnacceptedAnts;
    int totalRepairInterUnacceptedAntsId;
    int totalRepairInterUnacceptedHops;
    int totalRepairInterUnacceptedHopsId;
    int totalRepairDestUnacceptedAnts;
    int totalRepairDestUnacceptedAntsId;
    int totalRepairDestUnacceptedHops;
    int totalRepairDestUnacceptedHopsId;
    int totalRepairAcceptedAnts;
    int totalRepairAcceptedAntsId;
    int totalRepairAcceptedHops;
    int totalRepairAcceptedHopsId;
    int totalLoopingHops;
    int totalLoopingHopsId;
    int dropDataPacketsNoPheromone; //Counts the number of data packets dropped for the reason that there is no pheromone
    int dropDataPacketsNoPheromoneId;
    int totalPacketsDelivered;
    int totalPacketsDeliveredId;
    int totalHopsDelivered;
    int totalHopsDeliveredId;
    double avgHopsNumber;
    int avgHopsNumberId;
    int totalTtlKilledAnts;
    int totalTtlKilledAntsId;
    int totalProactiveDeadEndAnts;
    int totalProactiveDeadEndAntsId;
    int totalProactiveDeadEndHops;
    int totalProactiveDeadEndHopsId;
    int totalProactiveAntSendFailures;
    int totalProactiveAntSendFailuresId;
    int totalProactiveLoopingAnts;
    int totalProactiveLoopingAntsId;
    int totalProactiveLoopingHops;
    int totalProactiveLoopingHopsId;
    int dropDataPacketsMaxTtl;
    int dropDataPacketsMaxTtlId;
    int dropDataPacketsLoop;
    int dropDataPacketsLoopId;
    int totalHelloLinkBreaks;
    int totalHelloLinkBreaksId;
    int packetsDroppedAtDebuffering;
    int packetsDroppedAtDebufferingId;
    int packetsDroppedAtDebufferWithoutSend;
    int packetsDroppedAtDebufferWithoutSendId;
} AnthocnetStatsType;


enum {
    ANTHOCNET_GENERATE_REACTIVE_FORWARD_ANTS,
    ANTHOCNET_ROUTE_DATA_PACKETS,
    ANTHOCNET_PROCESS_REACTIVE_FORWARD_ANT,
    ANTHOCNET_PROCESS_REPAIR_BACKWARD_ANT,
    ANTHOCNET_PROCESS_REACTIVE_BACKWARD_ANT,
    ANTHOCNET_PROCESS_REPAIR_FORWARD_ANT,
    ANTHOCNET_INITIAL_STATE,
    ANTHOCNET_PROCESS_LINK_FAILURE_NOTIFICATION,
    ANTHOCNET_PROCESS_PROACTIVE_FORWARD_ANT,
    ANTHOCNET_HANDLE_LINK_FAILURE,
    ANTHOCNET_SEND_PROACTIVE_ANT,
    ANTHOCNET_CHECK_REACTIVE_ANTS_RETRANSMISSION,
    ANTHOCNET_PROCESS_UNICAST_WARNING,
    ANTHOCNET_PROCESS_RECEIVED_ANTS,
    ANTHOCNET_PROCESS_HELLO,
    ANTHOCNET_POST_NODE_INITIALISATION,
    ANTHOCNET_SEND_HELLO,
    ANTHOCNET_PROCESS_CHECK_PATHS_ANT,
    ANTHOCNET_END_CHECK_PATHS,
    ANTHOCNET_CHECK_PATHS,
    ANTHOCNET_CHECK_POSITIONS,
    ANTHOCNET_IDLE,
    ANTHOCNET_FINAL_STATE,
    ANTHOCNET_CLEAN_UP_ROUTING_TABLE,
    ANTHOCNET_ROUTE_REPAIR_TIME_OUT,
    ANTHOCNET_CHECK_PROACTIVE_ANTS,
    ANTHOCNET_DEFAULT_STATE
};



struct struct_anthocnet_str
{
    int state;
    NodeAddress myAddress; //My own address
    ANTHOCNETList routingTable;
    ANTHOCNETList neighbourTable;
    ANTHOCNETList messageBuffer;
    ANTHOCNETList relayMessageBuffer;
    clocktype reactiveAntsRetryInterval;
    int reactiveAntsNumberOfAttempts;
    ANTHOCNETDataPtrTempData tempData;
    unsigned short antGenerationNumber;
    double dataRoutingExponent;
    clocktype jitterRange;
    ANTHOCNETList currentConnectionsInfo;
    double proactiveAntRoutingExponent;
    clocktype RTCleanupInterval;
    int repairMaxBroadcasting;
    clocktype repairTimePerHop;
    double discountHops;
    double discountTripTime;
    double discountPheromone;
    double timeHopsWeight;
    double repairTimeFactor;
    clocktype helloInterval;
    int allowedHelloLoss;
    clocktype maxBufferTime;
    clocktype checkPathsInterval;
    FILE* positionsFp;
    int maxTtl;
    int maxHelloEntries;
    ANTHOCNETHelloEntry* tempHelloEntryArray;
    double proactiveTriggerPercentage;
    ANTHOCNETMsgId msgIds[ANTHOCNET_NUMBER_OF_MSG_IDS];
    short msgIdsIndex;
    double discountBootstrappedPheromone;
    double discountBootstrappedVirtualPheromone;
    int currentInterfaceIndex;
    NodeAddress* interfaceAddresses;
    NodeAddress currentInterfaceAddress;
    NodeAddress currentDefaultAddress;
    BOOL traceTime;
    BOOL tracePackets;
    BOOL notificationsAfterCleanup;
    BOOL printPathsAtAntReception;
    double* avgQTimeDbl;
    double discountQTime;
    double discountSendTime;
    double sensitivity;
    double currentSINR;
    double currentPower;
    double SINRTreshold;
    double SINRPheromoneFactor;
    int defaultSize;
    BOOL useSINR;
    BOOL useTime;
    BOOL useTimeOldWay;
    BOOL usePower;
    BOOL useBusyRatio;
    double maxPower;
    clocktype proactiveAntSendInterval;
    BOOL reSend;
    BOOL bidirectionalPaths;
    double busyLimit1;
    double busyLimit2;
    double busyFactor1;
    double busyFactor2;
    char nodeRole;
    BOOL useLabels;
    double fixedNodePheromoneFactor;
    BOOL useAversion;
    BOOL useSpeed;
    double speedPenalty;
    BOOL admissionControlFlag;
    double admissionControlHops;
    double admissionControlTime;
    AnthocnetStatsType stats;
    RandomSeed seed;
    int initStats;
    int printStats;
    BOOL statsPrinted;
};


void
AnthocnetInit(Node* node, AnthocnetData** dataPtr, const NodeInput* nodeInput, int interfaceIndex);


void
AnthocnetFinalize(Node* node);


void
AnthocnetProcessEvent(Node* node, Message* msg);


void AnthocnetRunTimeStat(Node* node, AnthocnetData* dataPtr);


void AnthocnetHandleProtocolPacket(Node* node,Message* msg, NodeAddress sourceAddr,
                                   NodeAddress destAddr, int interfaceIndex, unsigned ttl);
void AnthocnetRouterFunction(Node* node, Message* msg, NodeAddress destAddr,
                             NodeAddress previousHopAddr, BOOL* packetWasRouted);



//Interface Utility Functions


void ANTHOCNET_IPQueueTimeRegistration(Node* node,int interfaceIndex,clocktype qTime);


void ANTHOCNET_Mac802_11_UpdateSignalData(Node* node,NodeAddress nbrAddress,double currentPower,double currentSensitivity,double currentInterference,double currentNoise,double currentSINR);



#endif
