// Copyright (c) 2001-2006, Scalable Network Technologies, Inc.  All Rights Reserved.
//                          6701 Center Drive West
//                          Suite 520
//                          Los Angeles, CA 90045
//                          sales@scalable-networks.com
//
// This source code is licensed, not sold, and is subject to a written
// license agreement.  Among other things, no portion of this source
// code may be copied, transmitted, disclosed, displayed, distributed,
// translated, used as the basis for a derivative work, or used, in
// whole or in part, for any program or purpose other than its intended
// use in compliance with the license agreement as part of the QualNet
// software.  This source code and certain of the algorithms contained
// within it are confidential trade secrets of Scalable Network
// Technologies, Inc. and may not be used as the basis for any other
// software, hardware, product or service.



#ifndef MAC_802_11_H
#define MAC_802_11_H

#include "api.h"
#include "dvcs.h"
// ANTS01:
#include "routing_antsconstants.h"
#include "routing_anthocnet.h"
// ANTS01 end

//--------------------------------------------------------------------------
// #define's
//--------------------------------------------------------------------------

#define MAC_PHY_LOWEST_MANDATORY_RATE_TYPE 0

// FCS is accounted for in the header frame sizes
// (instead of after the data payload) for coding convenience.
#define M802_11_SHORT_CTRL_FRAME_SIZE     14
#define M802_11_LONG_CTRL_FRAME_SIZE      20

// Address4 field not considered.
#define M802_11_DATA_FRAME_HDR_SIZE       28



// Binary expon backoff lower bound and upper bound
#define M802_11a_CW_MIN          15
#define M802_11a_CW_MAX          1023
#define M802_11a_SLOT_TIME       (9 * MICRO_SECOND)
#define M802_11a_SIFS            (16 * MICRO_SECOND)
#define M802_11a_DELAY_UNTIL_SIGNAL_AIRBORN  (2 * MICRO_SECOND)

#define M802_11b_CW_MIN          31
#define M802_11b_CW_MAX          1023
#define M802_11b_SLOT_TIME       (20 * MICRO_SECOND)
#define M802_11b_SIFS            (10 * MICRO_SECOND)
#define M802_11b_DELAY_UNTIL_SIGNAL_AIRBORN  (5 * MICRO_SECOND)

#define M802_11_PROPAGATION_DELAY  (1 * MICRO_SECOND)



// This is the minimum time that a packet will be in the air
// for the purpose of computing (parallel simulation) lookahead.
// It also can include the SIFS delay for the time it takes
// the ACK to get in the air (small).  This is the minimum delay
// that a blindsiding packet can jump on the channel and demand
// an ACK.
//    ONLY need for NULL message protocol with incomplete
//    topology. Otherwise we can lie on lookahead.
//
// #define M802_11_MIN_PACKET_RECEIVE_DURATION (???*MICRO_SECOND)


// If packet size (bytes) of network layer is larger than this threshold,
// RTS-CTS method is used.  0 means always use RTS-CTS method.  Broadcast
// packets NEVER use RTS-CTS method.
#define M802_11_RTS_THRESH          0

// Retransmission limits.
#define M802_11_SHORT_RETRY_LIMIT   7
#define M802_11_LONG_RETRY_LIMIT    4


// FRAG_THRESH defaults to 2346.  If packet size (bytes) network layer is
// smaller than this threshold, no fragmentation is used.  Broadcast
// packets are NEVER fragmented.
#define M802_11_FRAG_THRESH         2346

// Masks used in frame headers in BSS.
#define M802_11_TO_DS                       0x80
#define M802_11_FROM_DS                     0x40

// Constant duration for CF frames.
#define M802_11_CF_FRAME_DURATION           32768


// CF Parameter Set
// Total length is 8 as per standard, though field lengths differ.
#define M802_11_CF_PARAMETER_SET_ID         4
#define M802_11_CF_PARAMETER_SET_LENGTH     6

#define M802_11_TIME_UNIT                   (1024 * MICRO_SECOND)
#define M802_11_BEACON_INTERVAL_MAX         32767
#define M802_11_BEACON_INTERVAL_DEFAULT     200
#define M802_11_BEACON_START_DEFAULT        1
#define M802_11_CFP_MAX_DURATION_DEFAULT    50

#define M802_11_STATION_POLL_TYPE_DEFAULT   M802_11_STA_POLLABLE_POLL
#define M802_11_PC_DELIVERY_MODE_DEFAULT    M802_11_PC_POLL_AND_DELIVER
#define M802_11_STATION_TYPE_DEFAULT        M802_11_STA_IBSS
#define M802_11_POLL_SAVE_MODE_DEFAULT      M802_11_POLL_SAVE_BYCOUNT
#define M802_11_POLL_SAVE_MIN_COUNT_DEFAULT 1
#define M802_11_POLL_SAVE_MAX_COUNT_DEFAULT 10

#define M802_11_RELAY_FRAMES_DEFAULT        TRUE
#define M802_11_BROADCASTS_HAVE_PRIORITY_DURING_CFP  TRUE
#define M802_11_PRINT_STATS_DEFAULT         TRUE
#define M802_11_TRACE_DEFAULT               M802_11_TRACE_NO




//--------------------------------------------------------------------------
// typedef's enums
//--------------------------------------------------------------------------


// MAC states
typedef enum {
    M802_11_S_IDLE,                 // 0
    M802_11_S_WFNAV,                // 1
    M802_11_S_WF_DIFS_OR_EIFS,      // 2
    M802_11_S_BO,                   // 3
    M802_11_S_NAV_RTS_CHECK_MODE,   // 4
    M802_11_S_HEADER_CHECK_MODE,    // 5

    // Waiting For Response States
    // Make consistent with Mac802_11InWaitingForResponseState()

    M802_11_S_WFCTS,                // 6 First State in range
    M802_11_S_WFDATA,               // 7
    M802_11_S_WFACK,                // 8
    M802_11_S_WF_CFP_BEACON,        // 9 Last State in range

    // Transmission States:
    // Make consistent with Mac802_11InTransmittingState().

    M802_11_X_RTS,                  // 10  First State in range
    M802_11_X_CTS,                  // 11
    M802_11_X_UNICAST,              // 12
    M802_11_X_BROADCAST,            // 13
    M802_11_X_ACK,                  // 14
    M802_11_X_CFP_BEACON,           // 15  Last State in range

    M802_11_CFP_START               // 16  State during CFP

} M802_11_MacStates;


// States during Contention Free Period
typedef enum {
    M802_11_S_CFP_NONE,             // Idle

    M802_11_S_CFP_WF_DATA,          // Sent poll, expecting data or null
    M802_11_S_CFP_WF_ACK,           // Sent data, waiting for ack
    M802_11_S_CFP_WF_DATA_ACK,      // Sent data+poll

    M802_11_X_CFP_BROADCAST,        // Transmit states
    M802_11_X_CFP_UNICAST,
    M802_11_X_CFP_UNICAST_POLL,
    M802_11_X_CFP_ACK,
    M802_11_X_CFP_POLL,
    M802_11_X_CFP_NULLDATA,
    M802_11_X_CFP_END

} M802_11_CfpStates;


// Type of MAC frame
typedef enum {
    M802_11_BEACON              = 0x08, // 00 1000

    M802_11_RTS                 = 0x1B, // 01 1011
    M802_11_CTS                 = 0x1C, // 01 1100
    M802_11_ACK                 = 0x1D, // 01 1101
    M802_11_CF_END              = 0x1E, // 01 1110
    M802_11_CF_END_ACK          = 0x1F, // 01 1111

    M802_11_DATA                = 0x20, // 10 0000
    M802_11_CF_DATA_ACK         = 0x21, // 10 0001
    M802_11_CF_DATA_POLL        = 0x22, // 10 0010
    M802_11_CF_DATA_POLL_ACK    = 0x23, // 10 0011
    M802_11_CF_NULLDATA         = 0x24, // 10 0100
    M802_11_CF_ACK              = 0x25, // 10 0101
    M802_11_CF_POLL             = 0x26, // 10 0110
    M802_11_CF_POLL_ACK         = 0x27, // 10 0111

    M802_11_CF_NONE             = 0x3F  // 11 1111 Reserved
} M802_11MacFrameType;


// Station type
typedef enum {
    M802_11_STA_UNKNOWN,
    M802_11_STA_IBSS,
    M802_11_STA_BSS,
    M802_11_STA_AP,
    M802_11_STA_PC
} M802_11_StationType;


// PC poll behavior
typedef enum {
    M802_11_PC_DELIVER_ONLY,            // PC only sends frames during CFP
    M802_11_PC_POLL_ONLY,               // PC polls and does not deliver
    M802_11_PC_POLL_AND_DELIVER         // PC polls then delivers, default
} M802_11_PcDeliveryMode;


// Poll save mode
typedef enum {
    M802_11_POLL_SAVE_NONE,
    M802_11_POLL_SAVE_BYCOUNT
} M802_11_PollSaveMode;


// Station poll behavior
typedef enum {
    M802_11_STA_NOT_POLLABLE,
    M802_11_STA_POLLABLE_NOT_IN_LIST,
    M802_11_STA_POLLABLE_DONT_POLL,
    M802_11_STA_POLLABLE_POLL
} M802_11_StaPollType;

// ASCII Trace
typedef enum {
    M802_11_TRACE_NO,                   // No trace generated, default
    M802_11_TRACE_YES                   // Internal ascii trace file format
} M802_11_TraceType;


//--------------------------------------------------------------------------
// typedef's struct
//--------------------------------------------------------------------------

#define M802_11_DATA_RATE_ADJUSTMENT_TIME  (60 * MILLI_SECOND)
//#define M802_11_DATA_RATE_ADJUSTMENT_TIME  (600 * SECOND)
#define M802_11_NUM_ACKS_FOR_RATE_INCREASE 10
#define M802_11_NUM_ACKS_FOR_RATE_DECREASE 2

typedef struct struct_mac_802_11_data_rate_entry_t {
    NodeAddress ipAddress;
    int       dataRateType;
    BOOL      firstTxAtNewRate;
    short     numAcksFailed;
    short     numAcksInSuccess;
    clocktype dataRateTimer;
    struct struct_mac_802_11_data_rate_entry_t* next;
} M802_11DataRateEntry;

// Keeps track of sequence numbers of frames.

typedef struct struct_mac_802_11_seqno_entry_t {
    NodeAddress ipAddress;
    unsigned short fromSeqNo;
    unsigned short toSeqNo;
    struct struct_mac_802_11_seqno_entry_t* next;
} M802_11SeqNoEntry;


//
// Following typedef's define MAC headers.
// NOTE:
//   The size of each structure is not used for computation of transmission
//   duration. Instead, size definitions (i.e., M802_11_DATA_FRAME_SIZE,
//   M802_11_SHORT_CTRL_FRAME_SIZE, M802_11_LONG_CTRL_FRAME_SIZE) are used
//   for such purposes. Therefore, the structure sizes do not have to be
//   identical to the sizes in the standard, but THEY MUST BE LESS THAN THE
//   ACTUAL SIZES.
//

//
// CTS and ACK frames.
// Note: - All frames types must match the short control (this one)
//         exactly for its first four (universal) fields.
//       - FCS is supposed to be after data payload.  However,
//         we include FCS within the header for coding convenience.
//       - Actual frame size need not match standard here since
//         we use M802_11_SHORT_CTRL_FRAME_SIZE to determine size.

typedef struct {
                                //  Should Be  Actually
    unsigned char frameType;    //      -         -
    unsigned char frameFlags;   //      2         2
    char Padding[2];            //      0         2
    int duration;               //      2         4
    NodeAddress destAddr;       //      6         4
    // char FCS;                //      4         0
} M802_11ShortControlFrame;     //---------------------
                                //     14        12


// RTS frames.
// Note: - Actual frame size need not match standard here since
//         we use M802_11_LONG_CTRL_FRAME_SIZE to determine frame size.
//       - FCS is supposed to be after data payload.  However,
//         we include FCS within the header for coding convenience.

typedef struct {
                                //  Should Be  Actually
    unsigned char frameType;    //      -         -
    unsigned char frameFlags;   //      2         2
    char Padding[2];            //      0         2
    int duration;               //      2         4
    NodeAddress destAddr;       //      6         4
    NodeAddress sourceAddr;     //      6         4
    char FCS[4];                //      4         4
} M802_11LongControlFrame;      //----------------------
                                //     20        20

// Data frame header.
// Note: - Actual frame size need not match standard here since
//         we use M802_11_DATA_FRAME_HDR_SIZE to determine frame size.
//       - FCS is supposed to be after data payload.  However,
//         we include FCS within the header for coding convenience.

typedef struct{
                                //  Should Be  Actually
    unsigned char frameType;    //      -         -
    unsigned char frameFlags;   //      2         2
    char Padding1[2];           //      0         2
    int duration;               //      2         4
    NodeAddress destAddr;       //      6         4
    NodeAddress sourceAddr;     //      6         4
    NodeAddress address3;       //      6         4
    unsigned short seqNo;       //      -         -
    unsigned short fragId;      //      2         4
    char FCS[4];                //      4         4
} M802_11FrameHdr;              //---------------------
                                //     28        28


// Data frames.
typedef struct {
    M802_11FrameHdr hdr;
    char payload[MAX_NW_PKT_SIZE];
} M802_11MacFrame;


// Management frame header.
// This is size 28 in the standard and matches the defined M802_11FrameHdr.
// Not defined separately.


// Beacon frame and related element structures

typedef struct struct_mac_802_11_CFParameterSet {
    unsigned char elementId;            // Set to CF_PARAMETER_SET_ID
    unsigned char length;               // Set to CF_PARAMETER_SET_LENGTH
    //char padding[2];                  // cfpCount and cfpPeriod
    unsigned short cfpMaxDuration;      // CFP Duration
    int cfpBalDuration;                 // CFP Balance duration
} M802_11_CFParameterSet;


typedef struct struct_mac_802_11_BeaconFrame {
    M802_11FrameHdr hdr;
    char padding1[4];
    clocktype timeStamp;
    unsigned short beaconInterval;
    char padding2[6];
    M802_11_CFParameterSet cfSet;
} M802_11_BeaconFrame;


// CF Pollable station list
typedef struct struct_mac_802_11_CfPollStation {
    NodeAddress macAddr;
    M802_11_StaPollType pollType;
    int nullDataCount;                  // Number of null data responses
    int pollCount;                      // Number of polls sent
    unsigned char flags;                // Can be used for various purposes
} M802_11_CfPollStation;


typedef struct struct_mac_802_11_CfPollListItem {
    M802_11_CfPollStation* data;
    struct struct_mac_802_11_CfPollListItem* prev;
    struct struct_mac_802_11_CfPollListItem* next;
} M802_11_CfPollListItem;


typedef struct struct_mac_802_11_CfPollList {
    int size;
    M802_11_CfPollListItem* first;      // First item in list
    M802_11_CfPollListItem* last;       // Last item in list
} M802_11_CfPollList;


// PC specific variables
typedef struct struct_mac_802_11_PcVars {
    M802_11_PcDeliveryMode deliveryMode;// Deliver only or poll+deliver
    BOOL broadcastsHavePriority;        // Priority of broadcasts over polls

    M802_11_CfPollList* cfPollList;     // List of stations
    int pollableCount;                  // Count of pollable stations
    M802_11_CfPollListItem* firstPolledItem;// First polled in a CFP
    M802_11_CfPollListItem* prevPolledItem; // Prev polled this CFP
    BOOL allStationsPolled;             // If all stations polled once

    M802_11_PollSaveMode pollSaveMode;  // Poll save mode
    int pollSaveMinCount;               // Min count for poll save
    int pollSaveMaxCount;               // Max count for poll save

    int dataRateTypeForAck;             // Data rate for ACK during PCF
} M802_11_PcVars;


// 802.11 data structure
typedef struct {
    MacData* myMacData;

    // DVCS (Directional Virtual Carrier Sensing) for directional transmission
    BOOL useDvcs;

    // Mac states
    M802_11_MacStates state;
    M802_11_MacStates prevState;

    BOOL IsInExtendedIfsMode;

    clocktype noResponseTimeoutDuration;

    // Backoff value at this station
    clocktype CW;
    clocktype BO;
    clocktype lastBOTimeStamp;

    // for random backoff
    RandomSeed seed;

    unsigned int timerSequenceNumber;

    char PartialFrame[MAX_NW_PKT_SIZE];

    M802_11DataRateEntry* dataRateHead;
    M802_11SeqNoEntry* seqNoHead;

    clocktype NAV;
    clocktype noOutgoingPacketsUntilTime;
    int shortRetryLimit;
    int longRetryLimit;
    int directionalShortRetryLimit;
    BOOL stopReceivingAfterHeaderMode;
    int rtsThreshold;
    int SSRC;
    int SLRC;
    NodeAddress waitingForAckOrCtsFromAddress;

    clocktype extraPropDelay;

    //
    int cwMin;
    int cwMax;
    clocktype slotTime;
    clocktype sifs;
    clocktype delayUntilSignalAirborn;
    clocktype txSifs;
    clocktype propagationDelay;
    clocktype difs;
    clocktype txDifs;
    clocktype pifs;
    clocktype txPifs;
    clocktype extendedIfsDelay;

    int lowestDataRateType;
    int highestDataRateType;
    int highestDataRateTypeForBC;
    clocktype rateAdjustmentTime;

    Message* currentMessage;
    NodeAddress currentNextHopAddress;
    M802_11DataRateEntry* dataRateInfo;

    // ANTS01 begin:
    clocktype currentStartTimeANTS01; // time at which we started trying to send the current packet
    clocktype avgSendTimeANTS01; // Average time for sending a packet from this node
    double avgSendTimeDiscountANTS01; // Discount factor for the average send time
    // variables for the physical transmission:
    double currentPower;
    double currentSensitivity;
    double currentInterference;
    double currentNoise;
    double currentSinr;
    // ANTS01 end

    const Message* potentialIncomingMessage;

    DirectionalAntennaMacInfoType* directionalInfo;


    // Infrastructure variables

    M802_11_StationType stationType;
    NodeAddress bssAddr;                // IP/Mac address of AP
    NodeAddress selfAddr;               // Own IP/Mac address of interface
    NodeAddress ipNextHopAddr;          // Next hop address given by Network
    NodeAddress ipSourceAddr;           // Originating source address
    NodeAddress ipDestAddr;             // Final destination address

    BOOL relayFrames;                   // Relay by AP/PC outside BSS?

    // CFP related variables

    M802_11_CfpStates cfpState;
    M802_11_CfpStates cfpPrevState;

    BOOL cfpBeaconIsDue;                // Beacon is pending
    unsigned int cfpBeaconSequenceNumber;//Timer sequence ID
    clocktype cfpMaxDuration;           // Max duration of CF period
    clocktype cfpBeaconInterval;        // Target beacon interval
    clocktype cfpBeaconTime;            // Last or next beacon time

    unsigned int cfpEndSequenceNumber;  // CFP end timer sequence ID
    clocktype cfpEndTime;               // CFP end time for stations

    clocktype cfpTimeout;               // CFP timer delay for ack/data/etc
    BOOL cfpAckIsDue;                   // Waiting for an ack?

    M802_11MacFrameType cfpLastFrameType;

    // PC specific variables
    M802_11_PcVars* pcVars;

    // Statistics collection variables.

    BOOL printCfpStatistics;            // Output PCF stats?

    int pktsToSend;                     // As received from Network layer
    //int pktsLostOverflow;             // Not used

    // Stats for DCF
    int unicastPacketsSentDcf;
    int broadcastPacketsSentDcf;

    int unicastPacketsGotDcf;
    int broadcastPacketsGotDcf;

    int ctsPacketsSentDcf;
    int rtsPacketsSentDcf;
    int ackPacketsSentDcf;

    int retxDueToCtsDcf;                // Retransmits due to CTS
    int retxDueToAckDcf;                // Retransmits due to ACK

    int retxAbortedRtsDcf;              // RTS retransmits aborted
    int pktsDroppedDcf;

    // Stats during PCF
    int beaconPacketsSentCfp;           // Beacons that start CFP
    int beaconPacketsGotCfp;

    int endPacketsSentCfp;
    int endPacketsGotCfp;

    int unicastPacketsSentCfp;
    int broadcastPacketsSentCfp;

    int unicastPacketsGotCfp;
    int broadcastPacketsGotCfp;

    int dataPacketsSentCfp;
    int dataPollPacketsSentCfp;
    int pollPacketsSentCfp;
    int nulldataPacketsSentCfp;
    //int ackPacketsSentCfp;            // Hard to isolate

    int dataPacketsGotCfp;
    int dataPollPacketsGotCfp;
    int pollPacketsGotCfp;
    int nulldataPacketsGotCfp;
    //int ackPacketsGotCfp;             // Hard to isolate

    int pollsSavedCfp;                  // Polls skipped by the PC

    // Trace related variables
    M802_11_TraceType trace;
} MacData802_11;



//--------------------------------------------------------------------------
// FUNCTION DECLARATIONS
//--------------------------------------------------------------------------

//--------------------------------------------------------------------------
// FUNCTION		Mac802_11UpdateCurrentSignalDataANTS01
// PURPOSE		Allow 802.11 to update the current value of physical layer 
//				data. These data can then later be used to update data at 
//				the routing layer, once we know which kind of reception we
//				are dealing with, and who is the sender
// PARAMETERS	Node *node;
//				int interfaceIndex
//				double power
//				double sensitivity
//				double interference
//				double noise
//				double sinr
// RETURN		None
// NOTES		None
//--------------------------------------------------------------------------

void Mac802_11UpdateCurrentSignalDataANTS01(Node *node, 
											int interfaceIndex,
											double power, 
											double sensitivity, 
											double interference, 
											double noise,
											double sinr);

//--------------------------------------------------------------------------
// FUNCTION     Mac802_11NotifySendReceiveANTS01
// PURPOSE		Notify the anthocnet protocol of a received acknowledgement
// PARAMETERS	Node *node
//				NodeAddress nextHop
// RETURN		None
// NOTES		None
//--------------------------------------------------------------------------

void Mac802_11NotifySendReceiveANTS01(Node *node, NodeAddress nextHop);

//--------------------------------------------------------------------------
// FUNCTION     Mac802_11UpdateAvgSendTimeANTS01
// PURPOSE      To Update the Average Send Time of unicast packets, used for
//              estimates in anthocnet.
// PARAMETERS   Node* node
//              MacData802_11* M802
// RETURN       None
// NOTES        None
//--------------------------------------------------------------------------

void Mac802_11UpdateAvgSendTimeANTS01(Node* node, MacData802_11* M802);


//--------------------------------------------------------------------------
// FUNCTION     Mac802_11InitAvgSendTimeANTS01
// PURPOSE	    Initialize avgSendTimeANTS01
// PARAMETERS	Node* node
//				int interfaceIndex
//              double initVal
//              double discount
// RETURN       None
// NOTES        None
//--------------------------------------------------------------------------
void Mac802_11InitAvgSendTimeANTS01(Node* node, int interfaceIndex, clocktype initVal, double discount);

//--------------------------------------------------------------------------
// FUNCTION     Mac802_11GetAvgSendTimeANTS01
// PURPOSE      To Get the Average Send Time of unicast packets, used for
//              estimates in anthocnet.
// PARAMETERS   Node* node, int interfaceIndex
// RETURN       the average send time per unicast packet
// NOTES        None
//--------------------------------------------------------------------------

clocktype Mac802_11GetAvgSendTimeANTS01(Node* node, int interfaceIndex);

//--------------------------------------------------------------------------
// FUNCTION     Mac802_11Init
// PURPOSE      Initialization function for 802_11 protocol of MAC layer
// PARAMETERS   Node* node
//                  Node being initialized.
//              NodeInput* nodeInput
//                  Structure containing contents of input file.
//              MacSubnetMemberData* subnetList
//                  Number of nodes in subnet.
//              int nodesInSubnet
//                  Number of nodes in subnet.
//              int subnetListIndex
//              NodeAddress subnetAddress
//                  Subnet address.
//              int numHostBits
//                  number of host bits.
// RETURN       None
// NOTES        None
//--------------------------------------------------------------------------
void Mac802_11Init(
   Node* node, int interfaceIndex, const NodeInput* nodeInput,
   PhyModel phyModel,
   MacSubnetMemberData* subnetList, int nodesInSubnet, int subnetListIndex,
   NodeAddress subnetAddress, int numHostBits,
   int networkType = NETWORK_IPV4,
   in6_addr *subnetAddr = 0,
    unsigned int subnetPrefixLen = 0);

//--------------------------------------------------------------------------
// FUNCTION     Mac802_11Layer
// PURPOSE      Models the behaviour of the MAC layer with the 802.11
//              protocol on receiving the message enclosed in msgHdr
// PARAMETERS   Node *node
//                  Node which received the message.
//              int interfaceIndex
//                  Interface index.
//              Message* msg
//                  Message received by the layer.
// RETURN       None
// NOTES        None
//--------------------------------------------------------------------------
void Mac802_11Layer(Node* node, int interfaceIndex, Message* msg);


//--------------------------------------------------------------------------
// FUNCTION     Mac802_11Finalize
// PURPOSE      Called at the end of simulation to collect the results of
//              the simulation of 802.11 protocol of the MAC Layer.
// PARAMETERS   Node* node
//                  Node which received the message.
//              int interfaceIndex
//                  Interface index.
// RETURN       None
// NOTES        None
//--------------------------------------------------------------------------
void Mac802_11Finalize(Node* node, int interfaceIndex);


//--------------------------------------------------------------------------
// NAME         Mac802_11NetworkLayerHasPacketToSend
// PURPOSE      To notify 802_11 that network has something to send
// PARAMETERS   Node* node
//                  Node which received the message.
//              MacData802_11* M802
//                  802.11 data structure
// RETURN       None
// NOTES        None
//--------------------------------------------------------------------------
void Mac802_11NetworkLayerHasPacketToSend(
   Node* node, MacData802_11* M802);


//--------------------------------------------------------------------------
// NAME         Mac802_11ReceivePacketFromPhy
// PURPOSE      To recieve packet from the physical layer
// PARAMETERS   Node* node
//                  Node which received the message.
//              MacData802_11* M802
//                  802.11 data structure
//              Message* msg
//                  Message received by the layer.
// RETURN       None.
// NOTES        None
//--------------------------------------------------------------------------
void Mac802_11ReceivePacketFromPhy(
   Node* node, MacData802_11* M802, Message* msg);


//--------------------------------------------------------------------------
// NAME         Mac802_11ReceivePhyStatusChangeNotification
// PURPOSE      Receive notification of status change in physical layer
// PARAMETERS   Node* node
//                  Node which received the message.
//              MacData802_11* M802
//                  802.11 data structure.
//              PhyStatusType oldPhyStatus
//                  The old physical status.
//              PhyStatusType newPhyStatus
//                  The new physical status.
//              clocktype receiveDuration
//                  The receiving duration.
//              Message* potentialIncomingPacket
//                  Incoming packet.
// RETURN       None.
// NOTES        None
//--------------------------------------------------------------------------
void Mac802_11ReceivePhyStatusChangeNotification(
   Node* node,
   MacData802_11* M802,
   PhyStatusType oldPhyStatus,
   PhyStatusType newPhyStatus,
   clocktype receiveDuration,
   const Message* potentialIncomingPacket);


//--------------------------------------------------------------------------
// STATIC FUNCTIONS
//--------------------------------------------------------------------------


//--------------------------------------------------------------------------
// NAME         Mac802_11IsWaitingForResponseState
// PURPOSE      Check if state is one that is waiting for response.
// PARAMETERS   M802_11_MacStates state
//                  The state to be validated.
// RETURN       BOOL
// NOTES        None
//-------------------------------------------------------------------------
static //inline
BOOL Mac802_11IsWaitingForResponseState(M802_11_MacStates state) {
   return((M802_11_S_WFCTS <= state) && (state <= M802_11_S_WFACK));
}


//--------------------------------------------------------------------------
// NAME         Mac802_11IsTransmittingState
// PURPOSE      Check if state is one of transmit states
// PARAMETERS   M802_11_MacStates state
//                  The state to be validated.
// RETURN       BOOL
// NOTES        None
//-------------------------------------------------------------------------
static //inline
BOOL Mac802_11IsTransmittingState(M802_11_MacStates state) {
   return((state >= M802_11_X_RTS) && (state <= M802_11_X_ACK));
}


//--------------------------------------------------------------------------
// NAME         Mac802_11CheckHeaderSizes
// PURPOSE      Check for valid header sizes.
// PARAMETERS   None
// RETURN       None
// NOTES        None
//-------------------------------------------------------------------------
static //inline
void Mac802_11CheckHeaderSizes() {
    int shortSize = sizeof(M802_11ShortControlFrame);
    int longSize = sizeof(M802_11LongControlFrame);
    int frameSize = sizeof(M802_11FrameHdr);

    assert(shortSize <= M802_11_SHORT_CTRL_FRAME_SIZE);
    assert(longSize == M802_11_LONG_CTRL_FRAME_SIZE);
    assert(frameSize == M802_11_DATA_FRAME_HDR_SIZE);
}


#endif /*MAC_802_11_H*/
