#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "api.h"
#include "network_ip.h"
#include "routing_anthocnet.h"

//UserCodeStartBodyData


//UserCodeEndBodyData

//Statistic Function Declarations
static void AnthocnetInitStats(Node* node, AnthocnetStatsType *stats);
void AnthocnetPrintStats(Node *node);

//Utility Function Declarations
static Pheromone ANTHOCNETBestPheromone(Node* node,ANTHOCNETList* nextHops);
static ANTHOCNETListEntry* ANTHOCNETBestPheromoneForChooseHelloEntries(Node* node,ANTHOCNETList* nextHops,Pheromone* bestPheromone,BOOLEAN* bootstrapped);
static BOOLEAN ANTHOCNETBetterBootstrappedPheromoneAvailable(Node* node,ANTHOCNETData* dataPtr,ANTHOCNETList* nextHops);
static PacketBuffer* ANTHOCNETBuildBackwardAntPacket(ANTHOCNETAntType antType,ANTHOCNETBackwardAnt* ant);
static PacketBuffer* ANTHOCNETBuildForwardAntPacket(ANTHOCNETAntType antType,ANTHOCNETForwardAnt* ant,NodeAddress myAddress);
static PacketBuffer* ANTHOCNETBuildLinkFailureAntPacket(ANTHOCNETAntType antType,ANTHOCNETLostPheromone* lostPheromoneArray,short arraySize);
static PacketBuffer* ANTHOCNETBuildProactiveAntPacket(ANTHOCNETAntType antType,ANTHOCNETForwardAnt* ant,NodeAddress myAddress);
static PacketBuffer* ANTHOCNETBuildRepairAntPacket(ANTHOCNETAntType antType,ANTHOCNETForwardAnt* ant,NodeAddress myAddress,NodeAddress lostNextHopAddress);
static PacketBuffer* ANTHOCNETBuildUnicastWarning(ANTHOCNETAntType antType,ANTHOCNETUnicastWarning* unicastWarning);
static Pheromone ANTHOCNETCalculatePheromone(double numberOfHops,double tripTime,ANTHOCNETData* dataPtr);
static Pheromone ANTHOCNETCalculatePheromoneOneHop(Node* node,ANTHOCNETData* dataPtr,Pheromone oldPheromone,float SINR,int interfaceIndex,int size,double numberOfHops,double tripTime);
static Pheromone ANTHOCNETCalculatePheromoneOneHopWithBusyRatioAndSINR(Node* node,ANTHOCNETData* dataPtr,Pheromone pheromone,float SINR,int interfaceIndex);
static Pheromone ANTHOCNETCalculatePheromoneOneHopWithBusyRatio(Node* node,ANTHOCNETData* dataPtr,Pheromone pheromone,int interfaceIndex);
static Pheromone ANTHOCNETCalculatePheromoneOneHopWithLabelsAndSINR(Node* node,ANTHOCNETData* dataPtr,Pheromone pheromone,float SINR,int interfaceIndex);
static Pheromone ANTHOCNETCalculatePheromoneOneHopWithLabels(Node* node,ANTHOCNETData* dataPtr,Pheromone pheromone,int interfaceIndex);
static Pheromone ANTHOCNETCalculatePheromoneOneHopWithPower(Node* node,ANTHOCNETData* dataPtr,Pheromone pheromone,float SINR,int interfaceIndex,int size);
static Pheromone ANTHOCNETCalculatePheromoneOneHopWithSINR(Node* node,ANTHOCNETData* dataPtr,Pheromone pheromone,float SINR,int interfaceIndex,int size);
static Pheromone ANTHOCNETCalculatePheromoneOneHopWithTime(Node* node,ANTHOCNETData* dataPtr,Pheromone pheromone,int interfaceIndex,int size);
static BOOL ANTHOCNETCheckIfForwardAntAlreadySeen(ANTHOCNETForwardAnt* FAnt,NodeAddress myAddress);
static unsigned char ANTHOCNETChooseHelloEntries(Node* node,ANTHOCNETData* dataPtr,ANTHOCNETHelloEntry** helloEntryArray,BOOLEAN* activeDest);
static ANTHOCNETBackwardAnt* ANTHOCNETCreateBackwardAnt(Node* node,ANTHOCNETForwardAnt* FAnt,NodeAddress myAddress,float SINR);
static ANTHOCNETForwardAnt* ANTHOCNETCreateForwardAnt(Node* node,NodeAddress destAddr,NodeAddress srcAddr,unsigned short generationNumber);
static ANTHOCNETListEntry* ANTHOCNETCreateNTNbrEntry(NodeAddress address,clocktype discoveryTime,ANTHOCNETListEntry* dest,int interfaceIndex,NodeAddress interfaceAddress,double currentSINR,double currentPower);
static ANTHOCNETListEntry* ANTHOCNETCreateNTNbrDestEntry(NodeAddress address,ANTHOCNETListEntry* destEntryPtr);
static ANTHOCNETListEntry* ANTHOCNETCreateRTDestEntry(NodeAddress address,double hopsAvg,double tripTime,ANTHOCNETListEntry* nextHop,ANTHOCNETData* dataPtr);
static ANTHOCNETListEntry* ANTHOCNETCreateRTDestNextHopEntry(Node* node,NodeAddress address,int numberOfHops,double tripTime,Pheromone pheromone,clocktype lastUpdate,unsigned short lastGeneration,Pheromone bootstrappedPheromone,int interfaceIndex,NodeAddress interfaceAddress);
static BOOL ANTHOCNETDecideIfProcessFAnt(Node* node,ANTHOCNETForwardAnt* FAnt,ANTHOCNETData* dataPtr,ANTHOCNETAntType antType);
static BOOL ANTHOCNETDecideIfProcessFAntAtDest(Node* node,ANTHOCNETForwardAnt* FAnt,ANTHOCNETData* dataPtr,ANTHOCNETAntType antType);
static Pheromone ANTHOCNETDediscretizePheromone(unsigned short discretePheromone);
static void ANTHOCNETDeleteBackwardAnt(ANTHOCNETBackwardAnt* ant);
static void ANTHOCNETDeleteForwardAnt(ANTHOCNETForwardAnt* ant);
static unsigned short ANTHOCNETDiscretizePheromone(Pheromone pheromone);
static void ANTHOCNETEncapsulatePacketInMessage(Node* node,Message* * msg,PacketBuffer* packetBuffer);
static ANTHOCNETListEntry* ANTHOCNETGetBestNextHopData(Node* node,NodeAddress destAddress,NodeAddress prevAddress,ANTHOCNETList* RT);
static ANTHOCNETListEntry* ANTHOCNETGetBestNextHopProactive(Node* node,NodeAddress destAddress,NodeAddress prevAddress,ANTHOCNETList* nextHops,int numberOfConfirmedEntries);
static ANTHOCNETListEntry* ANTHOCNETGetNextHopData(Node* node,NodeAddress destAddress,NodeAddress prevAddress,ANTHOCNETList* RT,double routingExponent);
static ANTHOCNETListEntry* ANTHOCNETGetNextHopProactive(Node* node,NodeAddress destAddress,NodeAddress prevAddress,ANTHOCNETList* nextHops,double routingExponent,int numberOfConfirmedEntries);
static float ANTHOCNETGetVirtualSendTime(Node* node,int interfaceIndex,int size);
static void ANTHOCNETInitiateForwardAnts(Node* node,ANTHOCNETAntType antType,NodeAddress myAddress,NodeAddress destAddress,unsigned short generationNumber,clocktype delay);
static void ANTHOCNETInitiateRepairForwardAnts(Node* node,NodeAddress myAddress,NodeAddress destAddress,unsigned short generationNumber,clocktype delay,NodeAddress lostNextHopAddress);
static BOOL ANTHOCNETInsertAndCheckBuffers(Node* node,ANTHOCNETData* dataPtr,Message* msg,NodeAddress destAddr,NodeAddress previousHop);
static ANTHOCNETListEntry* ANTHOCNETListCreateEntry(NodeAddress nodeAddress,void* data);
static void ANTHOCNETListInit(ANTHOCNETList* list);
static void ANTHOCNETListInsert(ANTHOCNETList* list,ANTHOCNETListEntry* newEntry,ANTHOCNETListEntry* insertAfter);
static void ANTHOCNETListPrint(Node* node,ANTHOCNETList* list,void (*printDataFunction)(void*),char * title);
static void* ANTHOCNETListRemoveEntry(ANTHOCNETListEntry* entry,ANTHOCNETList* list);
static int ANTHOCNETListSearch(ANTHOCNETListEntry** entry,ANTHOCNETList* list,NodeAddress lookupAddress);
static int ANTHOCNETListSearchFrom(ANTHOCNETListEntry** entry,ANTHOCNETListEntry* startPoint,NodeAddress lookupAddress);
static void ANTHOCNETMacAckReceived(Node* node,NodeAddress nextHop,int interfaceIndex);
static void ANTHOCNETMACLayerStatusEventHandler(Node* node,const Message* msg,const NodeAddress nextHop,const int interfaceIndex);
static ANTHOCNETBufferDest* ANTHOCNETMessageBufferInsert(Message* msg,ANTHOCNETList* messageBuffer,Node* node,NodeAddress destAddr);
static void ANTHOCNETMessageDebufferDestWithoutSend(Node* node,ANTHOCNETList* messageBuffer,ANTHOCNETListEntry* bufferDestEntry);
static void ANTHOCNETMessageDebufferDestWithSend(Node* node,ANTHOCNETList* messageBuffer,ANTHOCNETListEntry* bufferDestEntry,int interfaceIndex,NodeAddress interfaceAddress);
static void ANTHOCNETNTNbrPrintData(Node* node,void* data);
static void ANTHOCNETNTRemoveEntrySimple(ANTHOCNETList* NT,NodeAddress nbrAddr,NodeAddress destAddr);
static double ANTHOCNETNTRemoveNbrSplitArrays(ANTHOCNETListEntry* nbrEntry,ANTHOCNETList* NT,ANTHOCNETLostPheromone** lostBestArray,int* lostBestNumber,ANTHOCNETLostPheromone** lostLastArray,int* lostLastNumber,NodeAddress destAddr,BOOL wasSource,BOOL* wasLast,Node* node,ANTHOCNETData* dataPtr);
static BOOLEAN ANTHOCNETPacketInLoop(Node* node,ANTHOCNETData* dataPtr,Message* msg,NodeAddress prevHopAddr);
static void ANTHOCNETPrintPath(Node* node,ANTHOCNETVisitedNode* visitedNodes,int numberOfVisitedNodes,int seqno);
static void ANTHOCNETProcessAntAtDest(Node* node,ANTHOCNETData* dataPtr,ANTHOCNETForwardAnt* FAnt,ANTHOCNETAntType antType,int interfaceIndex,NodeAddress interfaceAddress);
static void ANTHOCNETRefreshTimeStampsAfterHello(ANTHOCNETData* dataPtr,NodeAddress nbrAddr,clocktype currentTime);
static void ANTHOCNETRemoveEmptyBufferDest(Node* node,ANTHOCNETList* messageBuffer,ANTHOCNETListEntry* bufferDestEntry);
static void ANTHOCNETResendOrBufferPacket(Node* node,ANTHOCNETData* dataPtr,NodeAddress destAddr,NodeAddress srcAddr,NodeAddress prevHopAddr,Message* originalMsg);
static void ANTHOCNETRTDestNextHopPrintData(void* data);
static void ANTHOCNETRTDestPrintData(Node* node,void* data);
static void ANTHOCNETRTRemoveEmptyDest(ANTHOCNETListEntry* RTDestEntry,ANTHOCNETList* routingTable);
static BOOL ANTHOCNETRTRemoveNextHop(ANTHOCNETListEntry* RTDest,NodeAddress nextHop,double* newBestHopsToGo,double* newBestTimeToGo,Pheromone* newBestPheromone);
static void ANTHOCNETRTRemoveNextHopSimple(ANTHOCNETListEntry* RTDestEntry,NodeAddress nextHop);
static void ANTHOCNETSendSimpleLinkFailureNotification(Node* node,ANTHOCNETData* dataPtr,NodeAddress destinationAddress);
static double ANTHOCNETTotalPheromone(ANTHOCNETList* nextHops,NodeAddress prevAddr,double exponent);
static short ANTHOCNETRTUpdateAfterFailureMsg(Node* node,ANTHOCNETData* dataPtr,ANTHOCNETList* RT,ANTHOCNETList* NT,ANTHOCNETLostPheromone* receivedArray,short receivedNumberOfElements,ANTHOCNETLostPheromone** returnArray,NodeAddress sourceAddr,int interfaceIndex,int size);
static void ANTHOCNETRTUpdateBootstrapping(Node* node,ANTHOCNETData* dataPtr,ANTHOCNETHelloEntry* helloEntryArray,ANTHOCNETHello* hello,NodeAddress nextHop,int interfaceIndex,NodeAddress interfaceAddress,int size);
static ANTHOCNETForwardAnt* ANTHOCNETUnpackForwardAntPacket(char* antPacket);
static void ANTHOCNETUpdateRTBackward(ANTHOCNETBackwardAnt* BAnt,ANTHOCNETData* dataPtr,Node* node,int interfaceIndex,NodeAddress interfaceAddress,BOOL hasPredecessor,NodeAddress predecessorAddress,int predecessorInterfaceIndex,NodeAddress predecessorInterfaceAddress);
static void ANTHOCNETUpdateRTDestBoth(Node* node,ANTHOCNETData* dataPtr,ANTHOCNETListEntry* RTDestEntry,NodeAddress destAddr);
static void ANTHOCNETUpdateRTDestLastDataTransported(Node* node,ANTHOCNETData* dataPtr,ANTHOCNETListEntry* RTDestEntry,NodeAddress destAddr);
static void ANTHOCNETUpdateRTDestLastUsed(Node* node,ANTHOCNETData* dataPtr,ANTHOCNETListEntry* RTDestEntry,NodeAddress destAddr);
static void ANTHOCNETUpdateRTDestWaitingSince(Node* node,ANTHOCNETData* dataPtr,ANTHOCNETListEntry* RTDestEntry,NodeAddress destAddr);
static ANTHOCNETListEntry* ANTHOCNETUpdateRTEntry(ANTHOCNETData* dataPtr,NodeAddress nextHop,NodeAddress destination,clocktype currentTime,int numberOfHops,Node* node,float virtualTripTime,Pheromone newPheromone,unsigned short generationNumber,int interfaceIndex,NodeAddress interfaceAddress);
static void ANTHOCNETUpdateRTForward(ANTHOCNETBackwardAnt* BAnt,ANTHOCNETData* dataPtr,Node* node);
static void ANTHOCNETUpdateNTTable(ANTHOCNETData* dataPtr,NodeAddress neighbour,NodeAddress destination,clocktype currentTime,ANTHOCNETListEntry* RTDestEntry,int interfaceIndex,NodeAddress interfaceAddress);
static void ANTHOCNETPrintLinkFailureAntData(Node* node,NodeAddress sourceAddr,short numberOfElements,ANTHOCNETLostPheromone* lostPheromoneArray);
static void* MEM_malloc2(int line,size_t size);
static void ANTHOCNETPrintPheromone(Node* node,ANTHOCNETData* dataPtr,NodeAddress destAddr,ANTHOCNETAntType antType);
static void ANTHOCNETBroadcastControlMessage(Node* node,Message* msg,TosType priority,clocktype delay,unsigned ttl);
static clocktype Mac802_11GetAvgSendTimeANTHOCNET(Node* node,int interfaceIndex);
static void ANTHOCNETUpdateAvgQueuingTime(Node* node,ANTHOCNETData* dataPtr,int interfaceIndex,clocktype qTime);
static double ANTHOCNETGetBusyRatio(Node* node,int interfaceIndex);
static Pheromone ANTHOCNETCalculatePheromoneOneHopWithLabelsAndTime(Node* node,ANTHOCNETData* dataPtr,Pheromone pheromone,float SINR,int interfaceIndex,int size);
static Pheromone ANTHOCNETCalculatePheromoneOneHopWithPheromoneAversion(Node* node,ANTHOCNETData* dataPtr,Pheromone pheromone,float SINR,int interfaceIndex,int size);
static Pheromone ANTHOCNETCalculatePheromoneOneHopWithSpeed(Node* node,ANTHOCNETData* dataPtr,Pheromone pheromone,float SINR,int interfaceIndex,int size);
static int ANTHOCNETRejectedSession(Node* node,ANTHOCNETData* dataPtr,NodeAddress destAddr);
static int ANTHOCNETRejectSession(Node* node,ANTHOCNETData* dataPtr,double hops,double time);

//State Function Declarations
static void enterAnthocnetGenerateReactiveForwardAnts(Node* node, AnthocnetData* dataPtr, Message* msg);
static void enterAnthocnetProcessReactiveForwardAnt(Node* node, AnthocnetData* dataPtr, Message* msg);
static void enterAnthocnetProcessRepairBackwardAnt(Node* node, AnthocnetData* dataPtr, Message* msg);
static void enterAnthocnetProcessReactiveBackwardAnt(Node* node, AnthocnetData* dataPtr, Message* msg);
static void enterAnthocnetProcessRepairForwardAnt(Node* node, AnthocnetData* dataPtr, Message* msg);
static void enterAnthocnetProcessLinkFailureNotification(Node* node, AnthocnetData* dataPtr, Message* msg);
static void enterAnthocnetProcessProactiveForwardAnt(Node* node, AnthocnetData* dataPtr, Message* msg);
static void enterAnthocnetHandleLinkFailure(Node* node, AnthocnetData* dataPtr, Message* msg);
static void enterAnthocnetSendProactiveAnt(Node* node, AnthocnetData* dataPtr, Message* msg);
static void enterAnthocnetCheckReactiveAntsRetransmission(Node* node, AnthocnetData* dataPtr, Message* msg);
static void enterAnthocnetProcessUnicastWarning(Node* node, AnthocnetData* dataPtr, Message* msg);
static void enterAnthocnetProcessHello(Node* node, AnthocnetData* dataPtr, Message* msg);
static void enterAnthocnetPostNodeInitialisation(Node* node, AnthocnetData* dataPtr, Message* msg);
static void enterAnthocnetSendHello(Node* node, AnthocnetData* dataPtr, Message* msg);
static void enterAnthocnetProcessCheckPathsAnt(Node* node, AnthocnetData* dataPtr, Message* msg);
static void enterAnthocnetEndCheckPaths(Node* node, AnthocnetData* dataPtr, Message* msg);
static void enterAnthocnetCheckPaths(Node* node, AnthocnetData* dataPtr, Message* msg);
static void enterAnthocnetCheckPositions(Node* node, AnthocnetData* dataPtr, Message* msg);
static void enterAnthocnetIdle(Node* node, AnthocnetData* dataPtr, Message* msg);
static void enterAnthocnetCleanUpRoutingTable(Node* node, AnthocnetData* dataPtr, Message* msg);
static void enterAnthocnetRouteRepairTimeOut(Node* node, AnthocnetData* dataPtr, Message* msg);
static void enterAnthocnetCheckProactiveAnts(Node* node, AnthocnetData* dataPtr, Message* msg);

//Utility Function Definitions
//UserCodeStart

static Pheromone ANTHOCNETBestPheromone(Node* node,ANTHOCNETList* nextHops) {
ANTHOCNETListEntry *nextHopEntry;
    ANTHOCNETRTDestNextHop *data;
    Pheromone bestPheromone = 0;
    for (nextHopEntry = nextHops->head.next;nextHopEntry != NULL;
         nextHopEntry = nextHopEntry->next)
    {
        data = (ANTHOCNETRTDestNextHop*)nextHopEntry->data;
        if (data->pheromone > bestPheromone)
        {
            bestPheromone = data->pheromone;
        }
        if (data->bootstrappedPheromone > bestPheromone)
        {
           {
               bestPheromone = data->bootstrappedPheromone;
           }
        }
    }
    return bestPheromone;
}


static ANTHOCNETListEntry* ANTHOCNETBestPheromoneForChooseHelloEntries(Node* node,ANTHOCNETList* nextHops,Pheromone* bestPheromone,BOOLEAN* bootstrapped) {
// We assume that if this function does not return NULL, the entry is sent
    ANTHOCNETListEntry *nextHopEntry, *bestNextHopEntry = NULL;
    ANTHOCNETRTDestNextHop *data;
    *bestPheromone = 0;
    for (nextHopEntry = nextHops->head.next;nextHopEntry != NULL;
         nextHopEntry = nextHopEntry->next)
    {
        data = (ANTHOCNETRTDestNextHop*)nextHopEntry->data;
        if (data->pheromone > *bestPheromone)
        {
            *bestPheromone = data->pheromone;
            bestNextHopEntry = nextHopEntry;
            *bootstrapped = FALSE;
        }
        if (!data->bootstrappedForwarded)
        {
            if ((data->bootstrappedPheromone > *bestPheromone) &&
                (data->bootstrappedTime > (getSimTime(node) - ANTHOCNET_MAX_TIME_TO_BOOTSTRAP)))
            {
                *bestPheromone = data->bootstrappedPheromone;
                bestNextHopEntry = nextHopEntry;
                *bootstrapped = TRUE;
            }
        }
    }
    // UPDATE BOOTSTRAPPEDFORWARDED FIELD!!!!!!!!!!!!!!!!!
    if (bestNextHopEntry != NULL)
    {
        if (*bootstrapped == TRUE)
        {
            ((ANTHOCNETRTDestNextHop*)bestNextHopEntry->data)->bootstrappedForwarded = TRUE;
        }
    }
    return bestNextHopEntry;
}


static BOOLEAN ANTHOCNETBetterBootstrappedPheromoneAvailable(Node* node,ANTHOCNETData* dataPtr,ANTHOCNETList* nextHops) {
Pheromone bestPheromone = 0, bestBootstrappedPheromone = 0;
    ANTHOCNETListEntry *RTDestNextHopEntry;
    ANTHOCNETRTDestNextHop *RTDestNextHopData;
    for (RTDestNextHopEntry = nextHops->head.next;
         RTDestNextHopEntry != NULL;
         RTDestNextHopEntry = RTDestNextHopEntry->next)
    {
        RTDestNextHopData = (ANTHOCNETRTDestNextHop*)RTDestNextHopEntry->data;
        if (RTDestNextHopData->pheromone > bestPheromone)
            bestPheromone = RTDestNextHopData->pheromone;
        if ((RTDestNextHopData->bootstrappedPheromone > bestBootstrappedPheromone) &&
            (RTDestNextHopData->bootstrappedTime > (getSimTime(node) - ANTHOCNET_MAX_BOOTSTRAPPED_PHEROMONE_AGE)))
            bestBootstrappedPheromone = RTDestNextHopData->bootstrappedPheromone;
    }
    if ((bestBootstrappedPheromone > (bestPheromone * (1 + dataPtr->proactiveTriggerPercentage))) &&
        (bestPheromone > 0))
        return TRUE;
    else
        return FALSE;
}


// FUNCTION	ANTHOCNETBuildBackwardAntPacket
// PURPOSE	//Put the ant and its type into a packetbuffer
// Parameters:
//	antType:
//	ant:
static PacketBuffer* ANTHOCNETBuildBackwardAntPacket(ANTHOCNETAntType antType,ANTHOCNETBackwardAnt* ant) {
char *packetContents = NULL;
        PacketBuffer* packetBuffer = BUFFER_AllocatePacketBuffer(
                    (sizeof(ANTHOCNETAntType)+
                     sizeof(ANTHOCNETBackwardAnt)+
                     (ant->numberOfVisitedNodes*sizeof(ANTHOCNETVisitedNode))),
                    0,FALSE,&packetContents);
        if (packetContents == NULL)
        {
            ERROR_Assert(FALSE, "Cannot allocate packet !!!");
            abort();
        }
        memcpy(packetContents,&antType,sizeof(ANTHOCNETAntType));
        packetContents += sizeof(ANTHOCNETAntType);
        memcpy(packetContents,ant,sizeof(ANTHOCNETBackwardAnt));
        packetContents += sizeof(ANTHOCNETBackwardAnt);
        memcpy(packetContents,ant->visitedNodes,sizeof(ANTHOCNETVisitedNode)*ant->numberOfVisitedNodes);
        return packetBuffer;
}


// FUNCTION	ANTHOCNETBuildForwardAntPacket
// PURPOSE	//Put the ant and its type into a packetbuffer
// Parameters:
//	antType:
//	ant:
//	myAddress:
static PacketBuffer* ANTHOCNETBuildForwardAntPacket(ANTHOCNETAntType antType,ANTHOCNETForwardAnt* ant,NodeAddress myAddress) {
char *packetContents = NULL;
        PacketBuffer* packetBuffer = BUFFER_AllocatePacketBuffer(
                    (sizeof(ANTHOCNETAntType)+
                     sizeof(ANTHOCNETForwardAnt)+
                     ((++ant->numberOfVisitedNodes)*sizeof(ANTHOCNETVisitedNode))),
                    0,FALSE,&packetContents);
        memcpy(packetContents,&antType,sizeof(ANTHOCNETAntType));
        packetContents += sizeof(ANTHOCNETAntType);
        memcpy(packetContents,ant,sizeof(ANTHOCNETForwardAnt));
        packetContents += sizeof(ANTHOCNETForwardAnt);
        memcpy(packetContents,ant->visitedNodes,sizeof(ANTHOCNETVisitedNode)*(ant->numberOfVisitedNodes-1));
        packetContents += sizeof(ANTHOCNETVisitedNode)*(ant->numberOfVisitedNodes-1);
        {
             ANTHOCNETVisitedNode currentNode; 
             currentNode.node = myAddress;
             memcpy(packetContents,&currentNode,sizeof(ANTHOCNETVisitedNode));
        }
        return packetBuffer;
}


// FUNCTION	ANTHOCNETBuildLinkFailureAntPacket
// PURPOSE	//Put the ant and its type into a packetbuffer
// Parameters:
//	antType:
//	lostPheromoneArray:
//	arraySize:
static PacketBuffer* ANTHOCNETBuildLinkFailureAntPacket(ANTHOCNETAntType antType,ANTHOCNETLostPheromone* lostPheromoneArray,short arraySize) {
char *packetContents = NULL;
    PacketBuffer* packetBuffer = BUFFER_AllocatePacketBuffer(
                (sizeof(ANTHOCNETAntType)+
                 sizeof(short)+
                 (arraySize*sizeof(ANTHOCNETLostPheromone))),
                0,FALSE,&packetContents);
    if (packetContents == NULL)
    {
        ERROR_Assert(FALSE, "Cannot allocate packet !!!");
        abort();
    }
    memcpy(packetContents,&antType,sizeof(ANTHOCNETAntType));
    packetContents += sizeof(ANTHOCNETAntType);
    memcpy(packetContents,&arraySize,sizeof(short));
    packetContents += sizeof(short);
    memcpy(packetContents,lostPheromoneArray,sizeof(ANTHOCNETLostPheromone)*arraySize);
    return packetBuffer;
}


static PacketBuffer* ANTHOCNETBuildProactiveAntPacket(ANTHOCNETAntType antType,ANTHOCNETForwardAnt* ant,NodeAddress myAddress) {
char *packetContents = NULL;
    PacketBuffer* packetBuffer = BUFFER_AllocatePacketBuffer(
                (sizeof(ANTHOCNETAntType)+
                 sizeof(ANTHOCNETForwardAnt)+
                 ((++ant->numberOfVisitedNodes)*sizeof(ANTHOCNETVisitedNode))),
                0,FALSE,&packetContents);
    if (packetContents == NULL)
    {
        ERROR_Assert(FALSE, "Cannot allocate packet !!!");
        abort();
    }
    memcpy(packetContents,&antType,sizeof(ANTHOCNETAntType));
    packetContents += sizeof(ANTHOCNETAntType);
    memcpy(packetContents,ant,sizeof(ANTHOCNETForwardAnt));
    packetContents += sizeof(ANTHOCNETForwardAnt);
    memcpy(packetContents,ant->visitedNodes,sizeof(ANTHOCNETVisitedNode)*(ant->numberOfVisitedNodes-1));
    packetContents += sizeof(ANTHOCNETVisitedNode)*(ant->numberOfVisitedNodes-1);
    {
         ANTHOCNETVisitedNode currentNode; 
         currentNode.node = myAddress;
         memcpy(packetContents,&currentNode,sizeof(ANTHOCNETVisitedNode));
    }
    return packetBuffer;
}


// FUNCTION	ANTHOCNETBuildRepairAntPacket
// PURPOSE	//Put the ant and its type into a packetbuffer
// Parameters:
//	antType:
//	ant:
//	myAddress:
//	lostNextHopAddress:
static PacketBuffer* ANTHOCNETBuildRepairAntPacket(ANTHOCNETAntType antType,ANTHOCNETForwardAnt* ant,NodeAddress myAddress,NodeAddress lostNextHopAddress) {
char *packetContents = NULL;
        PacketBuffer* packetBuffer = BUFFER_AllocatePacketBuffer(
                    (sizeof(ANTHOCNETAntType)+
                     sizeof(NodeAddress)+
                     sizeof(ANTHOCNETForwardAnt)+
                     ((++ant->numberOfVisitedNodes)*sizeof(ANTHOCNETVisitedNode))),
                    0,FALSE,&packetContents);
        memcpy(packetContents,&antType,sizeof(ANTHOCNETAntType));
        packetContents += sizeof(ANTHOCNETAntType);
        memcpy(packetContents,&lostNextHopAddress,sizeof(NodeAddress));
        packetContents += sizeof(NodeAddress);
        memcpy(packetContents,ant,sizeof(ANTHOCNETForwardAnt));
        packetContents += sizeof(ANTHOCNETForwardAnt);
        memcpy(packetContents,ant->visitedNodes,sizeof(ANTHOCNETVisitedNode)*(ant->numberOfVisitedNodes-1));
        packetContents += sizeof(ANTHOCNETVisitedNode)*(ant->numberOfVisitedNodes-1);
        {
             ANTHOCNETVisitedNode currentNode; 
             currentNode.node = myAddress;
             memcpy(packetContents,&currentNode,sizeof(ANTHOCNETVisitedNode));
        }
        return packetBuffer;
}


static PacketBuffer* ANTHOCNETBuildUnicastWarning(ANTHOCNETAntType antType,ANTHOCNETUnicastWarning* unicastWarning) {
char *packetContents = NULL;
    PacketBuffer* packetBuffer = BUFFER_AllocatePacketBuffer(
                (sizeof(ANTHOCNETAntType)+
                 sizeof(ANTHOCNETUnicastWarning)),
                0,FALSE,&packetContents);
    if (packetContents == NULL)
    {
        ERROR_Assert(FALSE, "Cannot allocate packet !!!");
        abort();
    }
    memcpy(packetContents,&antType,sizeof(ANTHOCNETAntType));
    packetContents += sizeof(ANTHOCNETAntType);
    memcpy(packetContents,unicastWarning,sizeof(ANTHOCNETUnicastWarning));
    return packetBuffer;
}


static Pheromone ANTHOCNETCalculatePheromone(double numberOfHops,double tripTime,ANTHOCNETData* dataPtr) {
return 1 / (dataPtr->timeHopsWeight * tripTime + 
                (1 - dataPtr->timeHopsWeight) * simTimeSec(ANTHOCNET_HOP_LATENCY) * numberOfHops);
}


static Pheromone ANTHOCNETCalculatePheromoneOneHop(Node* node,ANTHOCNETData* dataPtr,Pheromone oldPheromone,float SINR,int interfaceIndex,int size,double numberOfHops,double tripTime) {
if (dataPtr->useSINR) 
{
    if (dataPtr->useBusyRatio)
    {
        return ANTHOCNETCalculatePheromoneOneHopWithBusyRatioAndSINR(node,dataPtr,oldPheromone,SINR,interfaceIndex);
    } 
    else if (dataPtr->useLabels)
    {
        return ANTHOCNETCalculatePheromoneOneHopWithLabelsAndSINR(node,dataPtr,oldPheromone,SINR,interfaceIndex);
    } 
    else
    {
        return ANTHOCNETCalculatePheromoneOneHopWithSINR(node, dataPtr, oldPheromone, SINR, interfaceIndex, size);
    }
}
else if (dataPtr->useTime)
{
    if (dataPtr->useLabels)
    {
        return ANTHOCNETCalculatePheromoneOneHopWithLabelsAndTime(node,dataPtr,oldPheromone,SINR,interfaceIndex,size);
    }
    else
    {
        return ANTHOCNETCalculatePheromoneOneHopWithTime(node, dataPtr, oldPheromone, interfaceIndex, size);
    }
}
else if (dataPtr->useTimeOldWay)
{
    if (numberOfHops > -1)
    {
        return ANTHOCNETCalculatePheromone(numberOfHops,tripTime,dataPtr);
    }
    else
    {
        return ANTHOCNETCalculatePheromoneOneHopWithTime(node,dataPtr,oldPheromone,interfaceIndex,size);
    }
}
else if (dataPtr->usePower)
{
    return ANTHOCNETCalculatePheromoneOneHopWithPower(node,dataPtr,oldPheromone,SINR,interfaceIndex,size);
}
else if (dataPtr->useBusyRatio)
{
    return ANTHOCNETCalculatePheromoneOneHopWithBusyRatio(node,dataPtr,oldPheromone,interfaceIndex);
}
else if (dataPtr->useLabels)
{
    return ANTHOCNETCalculatePheromoneOneHopWithLabels(node,dataPtr,oldPheromone,interfaceIndex);
}
else if (dataPtr->useAversion)
{
    return ANTHOCNETCalculatePheromoneOneHopWithPheromoneAversion(node,dataPtr,oldPheromone,SINR,interfaceIndex,size);
}
else if (dataPtr->useSpeed)
{
    return ANTHOCNETCalculatePheromoneOneHopWithSpeed(node,dataPtr,oldPheromone,SINR,interfaceIndex,size);
}
else
{
    fprintf(stderr,"ANTHOCNETCalculatePheromoneOneHop: No pheromone measure specified\n");
    abort();
}
}


static Pheromone ANTHOCNETCalculatePheromoneOneHopWithBusyRatioAndSINR(Node* node,ANTHOCNETData* dataPtr,Pheromone pheromone,float SINR,int interfaceIndex) {
double busyRatio = ANTHOCNETGetBusyRatio(node,interfaceIndex);
double costOneHop = simTimeSec(ANTHOCNET_HOP_LATENCY);
if (busyRatio > dataPtr->busyLimit2)
{
    costOneHop *= dataPtr->busyFactor2;
}
else if (busyRatio > dataPtr->busyLimit1)
{
    costOneHop *= dataPtr->busyFactor1;
}
if (SINR < dataPtr->SINRTreshold)
{
    costOneHop *= dataPtr->SINRPheromoneFactor;
}
if (pheromone > 0) 
{
    return (1 / (costOneHop + (1 / pheromone)));
} 
else
{
    return (1 / costOneHop);
}
}


static Pheromone ANTHOCNETCalculatePheromoneOneHopWithBusyRatio(Node* node,ANTHOCNETData* dataPtr,Pheromone pheromone,int interfaceIndex) {
double busyRatio = ANTHOCNETGetBusyRatio(node,interfaceIndex);
double costOneHop = simTimeSec(ANTHOCNET_HOP_LATENCY);
if (busyRatio > dataPtr->busyLimit2)
{
    costOneHop *= dataPtr->busyFactor2;
}
else if (busyRatio > dataPtr->busyLimit1)
{
    costOneHop *= dataPtr->busyFactor1;
}
if (pheromone > 0) 
{
    return (1 / (costOneHop + (1 / pheromone)));
} 
else
{
    return (1 / costOneHop);
}
}


static Pheromone ANTHOCNETCalculatePheromoneOneHopWithLabelsAndSINR(Node* node,ANTHOCNETData* dataPtr,Pheromone pheromone,float SINR,int interfaceIndex) {
double costOneHop = simTimeSec(ANTHOCNET_HOP_LATENCY);
if (dataPtr->nodeRole == 'f')
{
    costOneHop *= dataPtr->fixedNodePheromoneFactor;
}
if (SINR < dataPtr->SINRTreshold)
{
    costOneHop *= dataPtr->SINRPheromoneFactor;
}
if (pheromone > 0) 
{
    return (1 / (costOneHop + (1 / pheromone)));
} 
else
{
    return (1 / costOneHop);
}
}


static Pheromone ANTHOCNETCalculatePheromoneOneHopWithLabels(Node* node,ANTHOCNETData* dataPtr,Pheromone pheromone,int interfaceIndex) {
double costOneHop = simTimeSec(ANTHOCNET_HOP_LATENCY);
if (dataPtr->nodeRole == 'f')
{
    costOneHop *= dataPtr->fixedNodePheromoneFactor;
}
if (pheromone > 0) 
{
    return (1 / (costOneHop + (1 / pheromone)));
} 
else
{
    return (1 / costOneHop);
}
}


static Pheromone ANTHOCNETCalculatePheromoneOneHopWithPower(Node* node,ANTHOCNETData* dataPtr,Pheromone pheromone,float SINR,int interfaceIndex,int size) {
double costOneHop = dataPtr->timeHopsWeight * ANTHOCNETGetVirtualSendTime(node, interfaceIndex, size) + 
                (1 - dataPtr->timeHopsWeight) * simTimeSec(ANTHOCNET_HOP_LATENCY);
if (node->macData[interfaceIndex]->macProtocol == MAC_PROTOCOL_802_11) {
    int phyNum = ((MacData802_11*)node->macData[interfaceIndex]->macVar)->myMacData->phyNumber;
    PhyData802_11* phy802_11 = (PhyData802_11*)((PhyData*)node->phyData[phyNum])->phyVar;
    costOneHop *= phy802_11->stats.energyConsumed / dataPtr->maxPower;//dataPtr->maxPower /  (dataPtr->maxPower - (phy802_11->stats.energyConsumed / 3600)); // not looking at RXpower, which is huge and the same for every node :)
}
if (SINR < dataPtr->SINRTreshold)
{
    costOneHop *= dataPtr->SINRPheromoneFactor;
}
if (pheromone > 0)
{
    return (1 / (costOneHop + (1 / pheromone)));
}
else
{
    return (1 / costOneHop);
}
}


static Pheromone ANTHOCNETCalculatePheromoneOneHopWithSINR(Node* node,ANTHOCNETData* dataPtr,Pheromone pheromone,float SINR,int interfaceIndex,int size) {
double costOneHop = dataPtr->timeHopsWeight * ANTHOCNETGetVirtualSendTime(node, interfaceIndex, size) + 
                (1 - dataPtr->timeHopsWeight) * simTimeSec(ANTHOCNET_HOP_LATENCY);
if (SINR < dataPtr->SINRTreshold)
{
    costOneHop *= dataPtr->SINRPheromoneFactor;
}
if (pheromone > 0) 
{
    return (1 / (costOneHop + (1 / pheromone)));
} 
else
{
    return (1 / costOneHop);
}
}


static Pheromone ANTHOCNETCalculatePheromoneOneHopWithTime(Node* node,ANTHOCNETData* dataPtr,Pheromone pheromone,int interfaceIndex,int size) {
double costOneHop = dataPtr->timeHopsWeight * ANTHOCNETGetVirtualSendTime(node, interfaceIndex, size) + 
                (1 - dataPtr->timeHopsWeight) * simTimeSec(ANTHOCNET_HOP_LATENCY);
if (pheromone > 0) 
{
    return (1 / (costOneHop + (1 / pheromone)));
} 
else
{
    return (1 / costOneHop);
}
}


static BOOL ANTHOCNETCheckIfForwardAntAlreadySeen(ANTHOCNETForwardAnt* FAnt,NodeAddress myAddress) {
int i;
    for(i=0;i<FAnt->numberOfVisitedNodes;i++)
       if (FAnt->visitedNodes[i].node == myAddress) 
          return (TRUE);
    return (FALSE);
}


static unsigned char ANTHOCNETChooseHelloEntries(Node* node,ANTHOCNETData* dataPtr,ANTHOCNETHelloEntry** helloEntryArray,BOOLEAN* activeDest) {
// TO BE IMPROVED!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    unsigned char numberOfEntries = 0;
    ANTHOCNETHelloEntry *tempHelloEntryArray = dataPtr->tempHelloEntryArray;
    ANTHOCNETListEntry *RTDestEntry;
    ANTHOCNETRTDest *RTDestData;
    Pheromone bestPheromone;
    BOOLEAN bootstrapped;
    *helloEntryArray = NULL;
    *activeDest = FALSE;
    if (dataPtr->routingTable.numberOfEntries == 0)
        return 0;
    if (dataPtr->routingTable.numberOfEntries <= dataPtr->maxHelloEntries)
    {
        for (RTDestEntry = dataPtr->routingTable.head.next;
             RTDestEntry != NULL; RTDestEntry = RTDestEntry->next)
        {
            RTDestData = (ANTHOCNETRTDest*)RTDestEntry->data;
            if ((RTDestData->lastUsed > (getSimTime(node) - ANTHOCNET_MAX_UNUSED_TIME)) &&
                (RTDestData->lastUsed > 0))
            {
                if (RTDestEntry->nodeAddress == dataPtr->myAddress)
                {
                    *activeDest = TRUE;
                }
                else if (RTDestData->nextHops.numberOfEntries > 0)
                {
                    if (ANTHOCNETBestPheromoneForChooseHelloEntries(node,&RTDestData->nextHops, &bestPheromone, &bootstrapped) != NULL)
                    {
                        tempHelloEntryArray[numberOfEntries].discretePheromone = 
                                ANTHOCNETDiscretizePheromone(bestPheromone);
                        tempHelloEntryArray[numberOfEntries].destAddr = RTDestEntry->nodeAddress;
                        tempHelloEntryArray[numberOfEntries++].bootstrapped = bootstrapped;
                    }
                }
            }
        }
        if (numberOfEntries > 0)
        {
            *helloEntryArray = (ANTHOCNETHelloEntry*)MEM_malloc2(__LINE__,numberOfEntries * sizeof(ANTHOCNETHelloEntry));
            memcpy(*helloEntryArray,tempHelloEntryArray,numberOfEntries * sizeof(ANTHOCNETHelloEntry));
        }
    }
    else
    {
        int startpoint = RANDOM_nrand(dataPtr->seed) % dataPtr->routingTable.numberOfEntries;
        int current = 0;
        int firstHalf = 0;
        ANTHOCNETListEntry *startRTDestEntry;
        for (startRTDestEntry = dataPtr->routingTable.head.next;current < startpoint;
             startRTDestEntry = startRTDestEntry->next, current++);
            
        for (RTDestEntry = startRTDestEntry;numberOfEntries < dataPtr->maxHelloEntries;)
        {
            RTDestData = (ANTHOCNETRTDest*)RTDestEntry->data;
            if ((RTDestData->lastUsed > (getSimTime(node) - ANTHOCNET_MAX_UNUSED_TIME)) && 
                (RTDestData->lastUsed > 0))
            {
                if (RTDestEntry->nodeAddress == dataPtr->myAddress)
                {
                    *activeDest = TRUE;
                }
                else if (RTDestData->nextHops.numberOfEntries > 0)
                {
                    if (ANTHOCNETBestPheromoneForChooseHelloEntries(node,&RTDestData->nextHops, &bestPheromone, &bootstrapped) != NULL)
                    {
                        tempHelloEntryArray[numberOfEntries].discretePheromone = 
                                ANTHOCNETDiscretizePheromone(bestPheromone);
                        tempHelloEntryArray[numberOfEntries].destAddr = RTDestEntry->nodeAddress;
                        tempHelloEntryArray[numberOfEntries++].bootstrapped = bootstrapped;
                    }
                }
            }
            if (RTDestEntry->next == NULL)
            {
                firstHalf = numberOfEntries;
                RTDestEntry = dataPtr->routingTable.head.next;
            }
            else
                RTDestEntry = RTDestEntry->next;
            if (RTDestEntry == startRTDestEntry)
                break;
        }
        if (numberOfEntries > 0)
        {
            *helloEntryArray = (ANTHOCNETHelloEntry*)MEM_malloc2(__LINE__,numberOfEntries * sizeof(ANTHOCNETHelloEntry));
            if ((firstHalf == 0) || (firstHalf == numberOfEntries))
            {
                memcpy(*helloEntryArray,tempHelloEntryArray,numberOfEntries * sizeof(ANTHOCNETHelloEntry));
            }
            else
            {
                memcpy(*helloEntryArray,tempHelloEntryArray+firstHalf,
                                        (numberOfEntries-firstHalf)*sizeof(ANTHOCNETHelloEntry));
                memcpy((*helloEntryArray)+(numberOfEntries-firstHalf),tempHelloEntryArray,
                                        firstHalf*sizeof(ANTHOCNETHelloEntry));
            }
        }
    }
    return numberOfEntries;
}


// FUNCTION	ANTHOCNETCreateBackwardAnt
// PURPOSE	//Create a new backward ant
// Parameters:
//	node:
//	FAnt:
//	myAddress:
//	SINR:
static ANTHOCNETBackwardAnt* ANTHOCNETCreateBackwardAnt(Node* node,ANTHOCNETForwardAnt* FAnt,NodeAddress myAddress,float SINR) {
ANTHOCNETVisitedNode currentNode;
        ANTHOCNETBackwardAnt *BAnt = (ANTHOCNETBackwardAnt*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETBackwardAnt));
        BAnt->destinationAddress = FAnt->destinationAddress;
        BAnt->sourceAddress = FAnt->sourceAddress;
        BAnt->numberOfVisitedNodes = FAnt->numberOfVisitedNodes + 1;
        BAnt->currentPosition = BAnt->numberOfVisitedNodes - 1;
        BAnt->generationNumber = FAnt->generationNumber;
        BAnt->tripTime = 0;
        BAnt->pheromone = 0;
        BAnt->prevSINR = SINR;
        BAnt->visitedNodes = (ANTHOCNETVisitedNode*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETVisitedNode)*BAnt->numberOfVisitedNodes);

        memcpy(BAnt->visitedNodes, FAnt->visitedNodes, 
               sizeof(ANTHOCNETVisitedNode)*FAnt->numberOfVisitedNodes);
        currentNode.node = myAddress;
        memcpy(&BAnt->visitedNodes[FAnt->numberOfVisitedNodes], 
               &currentNode, sizeof(ANTHOCNETVisitedNode));
        return BAnt;
}


// FUNCTION	ANTHOCNETCreateForwardAnt
// PURPOSE	//Create a new forward ant
// Parameters:
//	node:
//	destAddr:
//	srcAddr:
//	generationNumber:
static ANTHOCNETForwardAnt* ANTHOCNETCreateForwardAnt(Node* node,NodeAddress destAddr,NodeAddress srcAddr,unsigned short generationNumber) {
ANTHOCNETForwardAnt *FAnt = (ANTHOCNETForwardAnt*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETForwardAnt));
        FAnt->destinationAddress = destAddr;
        FAnt->sourceAddress = srcAddr;
        FAnt->numberOfVisitedNodes = 0;
        FAnt->generationNumber = generationNumber;
        FAnt->tripTime = 0;
        FAnt->visitedNodes = NULL;//MEM_malloc2(__LINE__,sizeof(ANTHOCNETVisitedNode));
//        FAnt->visitedNodes->node = srcAddr;
        return FAnt;
}


static ANTHOCNETListEntry* ANTHOCNETCreateNTNbrEntry(NodeAddress address,clocktype discoveryTime,ANTHOCNETListEntry* dest,int interfaceIndex,NodeAddress interfaceAddress,double currentSINR,double currentPower) {
ANTHOCNETListEntry* newEntry = (ANTHOCNETListEntry*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETListEntry));
    ANTHOCNETNTNbr* newData = (ANTHOCNETNTNbr*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETNTNbr));
    mem_error2(newEntry, newData, "Cannot allocate newEntry and newData in CreateNTNbrEntry!!");
    newEntry->nodeAddress = address;
    ANTHOCNETListInit(&(newData->destinations));
    newData->lastUpdate = discoveryTime;
    newData->interfaceIndex = interfaceIndex;
    newData->interfaceAddress = interfaceAddress;
    newData->SINR = currentSINR;
    newData->power = currentPower;
    if (dest != NULL)
    {
        ANTHOCNETListInsert(&(newData->destinations), dest, &(newData->destinations.head));
    }
    newEntry->data = newData;
    
    return newEntry;
}


static ANTHOCNETListEntry* ANTHOCNETCreateNTNbrDestEntry(NodeAddress address,ANTHOCNETListEntry* destEntryPtr) {
ANTHOCNETListEntry* newEntry = (ANTHOCNETListEntry*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETListEntry));
    ANTHOCNETNTNbrDest* newData = ( ANTHOCNETNTNbrDest*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETNTNbrDest));
    mem_error2(newEntry, newData, "Cannot allocate newEntry and newData!!");
    newEntry->nodeAddress = address;
    newData->RTDest = destEntryPtr;
    newEntry->data = newData;
    
    return newEntry;
}


static ANTHOCNETListEntry* ANTHOCNETCreateRTDestEntry(NodeAddress address,double hopsAvg,double tripTime,ANTHOCNETListEntry* nextHop,ANTHOCNETData* dataPtr) {
int i;
    ANTHOCNETListEntry* newEntry = (ANTHOCNETListEntry*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETListEntry));
    ANTHOCNETRTDest* newData = (ANTHOCNETRTDest*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETRTDest));
    newEntry->nodeAddress = address;
    
    ANTHOCNETListInit(&(newData->nextHops));
    ANTHOCNETListInit(&(newData->predecessors));
    
    newData->lastUsed = 0;
    newData->waitingSince = 0;
    newData->lastDataTransported = 0;
    newData->numberOfConfirmedEntries = 0;
    if (hopsAvg != ANTHOCNET_NO_DATA)
    {
        newData->hopsAvg = hopsAvg;
        newData->timeAvg = tripTime;
        for (i=0;i<ANTHOCNET_IDS_NUMBER;i++)
        {
            newData->idsWindow[i].generationNumber = 0;
        }
        newData->idsIndex = 0;
        ANTHOCNETListInsert(&(newData->nextHops), nextHop, &(newData->nextHops.head));
    }
    else
    {
        newData->hopsAvg = 0;
        newData->timeAvg = 0;
        for (i=0;i<ANTHOCNET_IDS_NUMBER;i++)
        {
            newData->idsWindow[i].generationNumber = 0;
        }
        newData->idsIndex = 0;
    }
    newEntry->data = newData;
    return newEntry;
}


static ANTHOCNETListEntry* ANTHOCNETCreateRTDestNextHopEntry(Node* node,NodeAddress address,int numberOfHops,double tripTime,Pheromone pheromone,clocktype lastUpdate,unsigned short lastGeneration,Pheromone bootstrappedPheromone,int interfaceIndex,NodeAddress interfaceAddress) {
ANTHOCNETListEntry* newEntry = (ANTHOCNETListEntry*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETListEntry));
    ANTHOCNETRTDestNextHop* newData = (ANTHOCNETRTDestNextHop*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETRTDestNextHop));
    
    newEntry->nodeAddress = address;
    newData->pheromone = pheromone;
    newData->bootstrappedPheromone = bootstrappedPheromone;
    newData->interfaceIndex = interfaceIndex;
    newData->interfaceAddress = interfaceAddress;
    if (pheromone != 0)
    {
        newData->hopsAvg = numberOfHops;
        newData->timeAvg = tripTime;
        newData->lastUpdate = lastUpdate;
        newData->lastGeneration = lastGeneration;
        newData->lastGenerationPheromone = pheromone;
    }
    if (bootstrappedPheromone != 0)
    {
        newData->bootstrappedTime = getSimTime(node);
        newData->bootstrappedForwarded = FALSE;
    }
    newEntry->data = newData;
    
    return newEntry;
}


static BOOL ANTHOCNETDecideIfProcessFAnt(Node* node,ANTHOCNETForwardAnt* FAnt,ANTHOCNETData* dataPtr,ANTHOCNETAntType antType) {
    double tripTime = FAnt->tripTime;
    BOOL found;
    ANTHOCNETListEntry *RTDestEntry;
    ANTHOCNETRTDest *RTDestData;
    found = ANTHOCNETListSearch(&RTDestEntry, &dataPtr->routingTable, FAnt->destinationAddress);
    if (found) // We have an entry for the destination
    {
        int i;
        RTDestData = (ANTHOCNETRTDest*)RTDestEntry->data;
        found = FALSE;
        for (i=0;(i<ANTHOCNET_IDS_NUMBER) && (!found);i++)
            found = (RTDestData->idsWindow[i].generationNumber == FAnt->generationNumber)
                      && (RTDestData->idsWindow[i].sourceAddress == FAnt->sourceAddress);
        i--;
        if (!found) // we have not seen this generation before
        {
            RTDestData->idsWindow[RTDestData->idsIndex].generationNumber = 
                                                            FAnt->generationNumber;
            RTDestData->idsWindow[RTDestData->idsIndex].sourceAddress = 
                                                            FAnt->sourceAddress;
            if (++RTDestData->idsIndex == ANTHOCNET_IDS_NUMBER)
                RTDestData->idsIndex = 0;
            return TRUE;
        }
        else // We have seen this generation
        {
            return FALSE;
        }
    }
    else // We do not have an entry for this destination
    {
        ANTHOCNETListEntry* insertAfter = RTDestEntry;
        RTDestEntry = ANTHOCNETCreateRTDestEntry(FAnt->destinationAddress, ANTHOCNET_NO_DATA, 
                                              ANTHOCNET_NO_DATA, NULL, dataPtr);
        RTDestData = (ANTHOCNETRTDest*)RTDestEntry->data;
        RTDestData->idsWindow[0].generationNumber = FAnt->generationNumber;
        RTDestData->idsWindow[0].sourceAddress = FAnt->sourceAddress;
        RTDestData->idsIndex++;
        ANTHOCNETListInsert(&dataPtr->routingTable, RTDestEntry, insertAfter);
        return TRUE;
    }
}


static BOOL ANTHOCNETDecideIfProcessFAntAtDest(Node* node,ANTHOCNETForwardAnt* FAnt,ANTHOCNETData* dataPtr,ANTHOCNETAntType antType) {
    double tripTime = FAnt->tripTime;
    BOOL found;
    ANTHOCNETListEntry *RTDestEntry;
    ANTHOCNETRTDest *RTDestData;
    found = ANTHOCNETListSearch(&RTDestEntry, &dataPtr->routingTable, FAnt->destinationAddress);
    if (found) // We have an entry for the destination
    {
        int i;
        RTDestData = (ANTHOCNETRTDest*)RTDestEntry->data;
        found = FALSE;
        for (i=0;(i<ANTHOCNET_IDS_NUMBER) && (!found);i++)
            found = (RTDestData->idsWindow[i].generationNumber == FAnt->generationNumber)
                      && (RTDestData->idsWindow[i].sourceAddress == FAnt->sourceAddress);
        i--;
        if (!found) // We have not seen this generation yet
        {
            RTDestData->idsWindow[RTDestData->idsIndex].generationNumber = 
                                                            FAnt->generationNumber;
            RTDestData->idsWindow[RTDestData->idsIndex].sourceAddress = 
                                                            FAnt->sourceAddress;
            if (++RTDestData->idsIndex == ANTHOCNET_IDS_NUMBER)
                RTDestData->idsIndex = 0;
            return TRUE;
        }
        else // We have seen ants of this generation before
        {
            return FALSE;
        }
    }
    else // We do not have an entry for this destination
    {
        ANTHOCNETListEntry* insertAfter = RTDestEntry;
        RTDestEntry = ANTHOCNETCreateRTDestEntry(FAnt->destinationAddress, ANTHOCNET_NO_DATA, 
                                              ANTHOCNET_NO_DATA, NULL, dataPtr);
        RTDestData = (ANTHOCNETRTDest*)RTDestEntry->data;
        RTDestData->idsWindow[0].generationNumber = FAnt->generationNumber;
        RTDestData->idsWindow[0].sourceAddress = FAnt->sourceAddress;
        RTDestData->idsIndex = 1;
        ANTHOCNETListInsert(&dataPtr->routingTable, RTDestEntry, insertAfter);
        return TRUE;
    }
}


static Pheromone ANTHOCNETDediscretizePheromone(unsigned short discretePheromone) {
    return 1.0 / (discretePheromone * ANTHOCNET_UNIT_DISCRETE_PHEROMONE);
}


static void ANTHOCNETDeleteBackwardAnt(ANTHOCNETBackwardAnt* ant) {
    MEM_free(ant->visitedNodes);
    MEM_free(ant);
}


static void ANTHOCNETDeleteForwardAnt(ANTHOCNETForwardAnt* ant) {
    MEM_free(ant->visitedNodes);
    MEM_free(ant);
}


static unsigned short ANTHOCNETDiscretizePheromone(Pheromone pheromone) {
    // added rounding on 12/5/2005:
    unsigned int discretePheromone = (unsigned int)(((1/pheromone) / ANTHOCNET_UNIT_DISCRETE_PHEROMONE) + 0.5); 
    if (discretePheromone < ANTHOCNET_MAX_DISCRETE_PHEROMONE)
        return (unsigned short)discretePheromone;
    else
        return ANTHOCNET_MAX_DISCRETE_PHEROMONE - 1;
}


// FUNCTION	ANTHOCNETEncapsulatePacketInMessage
// PURPOSE	//Encapsulate a packet in a message
// Parameters:
//	node:
//	msg:
//	packetBuffer:
static void ANTHOCNETEncapsulatePacketInMessage(Node* node,Message* * msg,PacketBuffer* packetBuffer) {
*msg = MESSAGE_Alloc(node, NETWORK_LAYER, ROUTING_PROTOCOL_ANTHOCNET,MSG_DEFAULT);
   MESSAGE_PacketAlloc(node, *msg, BUFFER_GetCurrentSize(packetBuffer),TRACE_ANTHOCNET);
   memcpy(MESSAGE_ReturnPacket((*msg)),BUFFER_GetData(packetBuffer),
                BUFFER_GetCurrentSize(packetBuffer));
}


static ANTHOCNETListEntry* ANTHOCNETGetBestNextHopData(Node* node,NodeAddress destAddress,NodeAddress prevAddress,ANTHOCNETList* RT) {
BOOL found;
    ANTHOCNETListEntry *destEntry;
    ANTHOCNETRTDest *RTDestData;
    found = ANTHOCNETListSearch(&destEntry, RT, destAddress);
    if (!found)
        return NULL;
    RTDestData = (ANTHOCNETRTDest*)destEntry->data;
    if (RTDestData->numberOfConfirmedEntries == 0)
        return NULL;
    else
    {
        ANTHOCNETListEntry *nextHopEntry, *bestNextHopEntry = NULL;
        Pheromone tempPheromone, bestPheromone = 0;
        for (nextHopEntry = RTDestData->nextHops.head.next;nextHopEntry != NULL;
                 nextHopEntry = nextHopEntry->next)
        {
            if (nextHopEntry->nodeAddress == prevAddress) 
                continue;
            tempPheromone = ((ANTHOCNETRTDestNextHop*)nextHopEntry->data)->pheromone;
            if (tempPheromone > bestPheromone)
            {
                bestPheromone = tempPheromone;
                bestNextHopEntry = nextHopEntry;
            }
        }
        if (bestNextHopEntry != NULL) // The destination has other next hops than the packet's previous hop and/or bootstrapped nexthops
        {
            // Every time we use a hop we update its lastUpdate time:
            ((ANTHOCNETRTDestNextHop*)bestNextHopEntry->data)->lastUpdate = getSimTime(node);
        }
        return bestNextHopEntry;
    }
}


static ANTHOCNETListEntry* ANTHOCNETGetBestNextHopProactive(Node* node,NodeAddress destAddress,NodeAddress prevAddress,ANTHOCNETList* nextHops,int numberOfConfirmedEntries) {
if (nextHops->numberOfEntries == 0)
        return NULL;
    else
    {
        Pheromone tempPheromone, bestPheromone = 0;
        ANTHOCNETListEntry *nextHopEntry, *bestNextHopEntry = NULL;
        ANTHOCNETListEntry* returnEntry = NULL;
        for (nextHopEntry = nextHops->head.next;nextHopEntry != NULL;
             nextHopEntry = nextHopEntry->next)
        {
            if (nextHopEntry->nodeAddress == prevAddress) 
                continue;
            tempPheromone = ((ANTHOCNETRTDestNextHop*)nextHopEntry->data)->pheromone;
            if (tempPheromone > bestPheromone)
            {
                bestPheromone = tempPheromone;
                bestNextHopEntry = nextHopEntry;
            }
            tempPheromone = ((ANTHOCNETRTDestNextHop*)nextHopEntry->data)->bootstrappedPheromone;
            if (tempPheromone > bestPheromone)
            {
                if (((ANTHOCNETRTDestNextHop*)nextHopEntry->data)->bootstrappedTime > 
                                (getSimTime(node) - ANTHOCNET_MAX_BOOTSTRAPPED_PHEROMONE_AGE))
                {
                    bestPheromone = tempPheromone;
                    bestNextHopEntry = nextHopEntry;
                }
            }
        }
        if (bestPheromone > 0) // false if the only next hop is the packet's previous hop or old bootstrapped pheromone
        {
            returnEntry = bestNextHopEntry;
            // Every time we use a hop we update its lastUpdate time:
            ((ANTHOCNETRTDestNextHop*)bestNextHopEntry->data)->lastUpdate = getSimTime(node);
        }
        return returnEntry;
    }
}


static ANTHOCNETListEntry* ANTHOCNETGetNextHopData(Node* node,NodeAddress destAddress,NodeAddress prevAddress,ANTHOCNETList* RT,double routingExponent) {
    ANTHOCNETData *dataPtr = (ANTHOCNETData*)NetworkIpGetRoutingProtocol(node,ROUTING_PROTOCOL_ANTHOCNET);
    if (routingExponent == -1)
        return ANTHOCNETGetBestNextHopData(node,destAddress,prevAddress,RT);
    BOOL found;
    ANTHOCNETListEntry* destEntry;
    ANTHOCNETListEntry* nextHopEntry;
    ANTHOCNETRTDest *RTDestData;
    ANTHOCNETRTDestNextHop* nextHopData;
    found = ANTHOCNETListSearch(&destEntry, RT, destAddress);
    if (!found)
        return NULL;
    RTDestData = (ANTHOCNETRTDest*)destEntry->data;
    if (RTDestData->numberOfConfirmedEntries == 0)
        return NULL;
    else
    {
        int i = 0;
        double pherSum = 0;
        double randValue;
        Pheromone tempPheromone;
        Pheromone* pheromoneArray;
        ANTHOCNETListEntry** entryArray;
        pheromoneArray = (Pheromone*)MEM_malloc2(__LINE__,sizeof(Pheromone) * 
                                        RTDestData->numberOfConfirmedEntries);
        entryArray = (ANTHOCNETListEntry**)MEM_malloc2(__LINE__,sizeof(ANTHOCNETListEntry*) *
                                        RTDestData->numberOfConfirmedEntries);
        for (nextHopEntry = RTDestData->nextHops.head.next;nextHopEntry != NULL;
                 nextHopEntry = nextHopEntry->next)
        {
            if (nextHopEntry->nodeAddress == prevAddress) 
                continue;
            tempPheromone = ((ANTHOCNETRTDestNextHop*)nextHopEntry->data)->pheromone;
            if (tempPheromone == 0)
                continue;
            entryArray[i] = nextHopEntry;
            pherSum += pow(tempPheromone,routingExponent);
            pheromoneArray[i] = pherSum;
            i++;
        }
        if (i == 0) // The only next hops are the packet's previous hop and/or bootstrapped nexthops
        {
            nextHopEntry = NULL;
        }
        else
        {
            randValue = RANDOM_erand(dataPtr->seed) * pherSum;
            for (i = 0; pheromoneArray[i] < randValue; i++);
            nextHopEntry = entryArray[i];
            // Every time we use a hop we update its lastUpdate time:
            ((ANTHOCNETRTDestNextHop*)nextHopEntry->data)->lastUpdate = getSimTime(node);
        }
        MEM_free(entryArray);
        MEM_free(pheromoneArray);
        return nextHopEntry;
    }
}


static ANTHOCNETListEntry* ANTHOCNETGetNextHopProactive(Node* node,NodeAddress destAddress,NodeAddress prevAddress,ANTHOCNETList* nextHops,double routingExponent,int numberOfConfirmedEntries) {
    ANTHOCNETData *dataPtr = (ANTHOCNETData*)NetworkIpGetRoutingProtocol(node,ROUTING_PROTOCOL_ANTHOCNET);
    if (routingExponent == -1)
        return ANTHOCNETGetBestNextHopProactive(node,destAddress,prevAddress,nextHops,numberOfConfirmedEntries);
    if (nextHops->numberOfEntries == 0)
        return NULL;
    else
    {
        int i = 0;
        double pherSum = 0;
        double randValue;
        Pheromone tempPheromone;
        Pheromone* pheromoneArray;
        ANTHOCNETListEntry** entryArray;
        ANTHOCNETListEntry *nextHopEntry;
        ANTHOCNETListEntry *returnEntry = NULL;

        pheromoneArray = (Pheromone*)MEM_malloc2(__LINE__,sizeof(Pheromone) * (nextHops->numberOfEntries +
                                                                      numberOfConfirmedEntries));
        entryArray = (ANTHOCNETListEntry**)MEM_malloc2(__LINE__,sizeof(ANTHOCNETListEntry*) * (nextHops->numberOfEntries +
                                                                      numberOfConfirmedEntries));
        for (nextHopEntry = nextHops->head.next;nextHopEntry != NULL;
             nextHopEntry = nextHopEntry->next)
        {
            if (nextHopEntry->nodeAddress == prevAddress) 
                continue;
            tempPheromone = ((ANTHOCNETRTDestNextHop*)nextHopEntry->data)->pheromone;
            if (tempPheromone > 0)
            {
                entryArray[i] = nextHopEntry;
                pherSum += pow(tempPheromone,routingExponent);
                pheromoneArray[i] = pherSum;
                i++;
            }
            tempPheromone = ((ANTHOCNETRTDestNextHop*)nextHopEntry->data)->bootstrappedPheromone;
            if (tempPheromone > 0)
            {
                if (((ANTHOCNETRTDestNextHop*)nextHopEntry->data)->bootstrappedTime > 
                                (getSimTime(node) - ANTHOCNET_MAX_BOOTSTRAPPED_PHEROMONE_AGE))
                {
                    entryArray[i] = nextHopEntry;
                    pherSum += pow(tempPheromone,routingExponent);
                    pheromoneArray[i] = pherSum;
                    i++;
                }
            }
        }
        if (i > 0) // false if the only next hop is the packet's previous hop
        {
            randValue = RANDOM_erand(dataPtr->seed) * pherSum;
            for (i = 0; pheromoneArray[i] < randValue; i++);
            returnEntry = entryArray[i];
            // Every time we use a hop we update its lastUpdate time:
            ((ANTHOCNETRTDestNextHop*)entryArray[i]->data)->lastUpdate = getSimTime(node);
        }
        MEM_free(entryArray);
        MEM_free(pheromoneArray);
        return returnEntry;
    }
}


static float ANTHOCNETGetVirtualSendTime(Node* node,int interfaceIndex,int size) {
    if (node->macData[interfaceIndex]->macProtocol == MAC_PROTOCOL_802_11)
    {
          ANTHOCNETData *dataPtr = (ANTHOCNETData*)NetworkIpGetRoutingProtocol(node,ROUTING_PROTOCOL_ANTHOCNET);
          return (dataPtr->avgQTimeDbl[interfaceIndex] + simTimeSec(Mac802_11GetAvgSendTimeANTHOCNET(node,interfaceIndex)));
    }
    else if (node->macData[interfaceIndex]->macProtocol == MAC_PROTOCOL_LINK)
    {
      double headerSizeInBytes = ((LinkData *)node->macData[interfaceIndex]->macVar)->headerSizeInBits / 8;
      double dataBytesInQueue = 512 * NetworkIpOutputQueueNumberInQueue(node,interfaceIndex,TRUE,IPTOS_PREC_ROUTINE);
      double controlBytesInQueue = 512 * NetworkIpOutputQueueNumberInQueue(node,interfaceIndex,TRUE,IPTOS_PREC_INTERNETCONTROL);
      double dataBytes = dataBytesInQueue + size + 
                   ((NetworkIpOutputQueueNumberInQueue(node,interfaceIndex,TRUE,IPTOS_PREC_ROUTINE) + 1) * 
                        headerSizeInBytes);
      double controlBytes = controlBytesInQueue + 
                    (NetworkIpOutputQueueNumberInQueue(node,interfaceIndex,TRUE,IPTOS_PREC_INTERNETCONTROL) * 
                        headerSizeInBytes);
      return simTimeSec((clocktype)((((dataBytes + controlBytes) * 8 * SECOND) / 
                                      node->macData[interfaceIndex]->bandwidth) + 0.5) + 
                          node->macData[interfaceIndex]->propDelay);
    }
    else
    {
        fprintf(stderr,"ANTNETGetVirtualSendTime: Unexpected MAC protocol\n");
        abort();
    }
}


static void ANTHOCNETInitiateForwardAnts(Node* node,ANTHOCNETAntType antType,NodeAddress myAddress,NodeAddress destAddress,unsigned short generationNumber,clocktype delay) {
    ANTHOCNETForwardAnt *FAnt = NULL;
    PacketBuffer* packetBuffer = NULL;
    Message *msgNew = NULL;
    
    /* Make an ant and indicate its type in a variable. */
    FAnt = ANTHOCNETCreateForwardAnt(node,destAddress, myAddress, generationNumber);

    /* Put the ant and its type in a packet */
    packetBuffer = ANTHOCNETBuildForwardAntPacket(antType, FAnt, myAddress);

    mem_error2(FAnt, packetBuffer, "Cannot allocate FAnt or packetBuffer");
    
    /* Put the ant packet in a message. */
    ANTHOCNETEncapsulatePacketInMessage(node, &msgNew, packetBuffer);
        
    /* Send the ant message (ttl at the moment set to 5) */
    ANTHOCNETBroadcastControlMessage(node, msgNew, IPTOS_PREC_INTERNETCONTROL, delay, 5);

    ANTHOCNETDeleteForwardAnt(FAnt);
}


static void ANTHOCNETInitiateRepairForwardAnts(Node* node,NodeAddress myAddress,NodeAddress destAddress,unsigned short generationNumber,clocktype delay,NodeAddress lostNextHopAddress) {
    ANTHOCNETForwardAnt *FAnt = NULL;
    PacketBuffer* packetBuffer = NULL;
    Message *msgNew = NULL;
    
    /* Make an ant and indicate its type in a variable. */
    FAnt = ANTHOCNETCreateForwardAnt(node,destAddress, myAddress, generationNumber);

    /* Put the ant and its type in a packet */
    packetBuffer = ANTHOCNETBuildRepairAntPacket(ANTHOCNET_REPAIR_FORWARD, FAnt, myAddress, lostNextHopAddress);

    /* Put the ant packet in a message. */
    ANTHOCNETEncapsulatePacketInMessage(node, &msgNew, packetBuffer);
        
    /* Send the ant message (ttl at the moment set to 5) */
    ANTHOCNETBroadcastControlMessage(node, msgNew, IPTOS_PREC_INTERNETCONTROL, delay, 5);

    ANTHOCNETDeleteForwardAnt(FAnt);
}


// FUNCTION	ANTHOCNETInsertAndCheckBuffers
// PURPOSE	//Insert a message into the right buffer, and return whether ants should be launched
// Parameters:
//	node:
//	dataPtr:
//	msg:
//	destAddr:
//	previousHop:
static BOOL ANTHOCNETInsertAndCheckBuffers(Node* node,ANTHOCNETData* dataPtr,Message* msg,NodeAddress destAddr,NodeAddress previousHop) {
    BOOL sendAnts = FALSE;
    ANTHOCNETBufferDest* bufferDestData[2];
    ANTHOCNETListEntry* bufferDestEntry;
    BOOL found;
    if (previousHop == ANY_IP) // I am the source of this packet
    {
        bufferDestData[0] = 
            ANTHOCNETMessageBufferInsert(msg, &dataPtr->messageBuffer, node, destAddr);
        found = 
            ANTHOCNETListSearch(&bufferDestEntry, &dataPtr->relayMessageBuffer, destAddr);
        dataPtr->tempData.intermediateNodeFlag = FALSE;
    }
    else // I am an intermediate node
    {
        bufferDestData[0] = 
            ANTHOCNETMessageBufferInsert(msg, &dataPtr->relayMessageBuffer, node, destAddr);
        found = 
            ANTHOCNETListSearch(&bufferDestEntry, &dataPtr->messageBuffer, destAddr);
        dataPtr->tempData.intermediateNodeFlag = TRUE;
    }
    if (found)
        bufferDestData[1] = (ANTHOCNETBufferDest*)bufferDestEntry->data;
    else
        bufferDestData[1] = NULL;
    if (bufferDestData[0]->numberOfAntGenerations == 0) // for this buffer no event is pending
    {
        bufferDestData[0]->timerMsg = NULL;
        if (bufferDestData[1] != NULL) // An event is pending for the destination in the other buffer
        {
            bufferDestData[0]->numberOfAntGenerations++;
        }
        else // also for the other buffer no ants are sent out
        {
            dataPtr->tempData.bufferDestData = bufferDestData[0];
            dataPtr->tempData.destAddr = destAddr;
            sendAnts = TRUE;
        }
    }
    return sendAnts;
}


static ANTHOCNETListEntry* ANTHOCNETListCreateEntry(NodeAddress nodeAddress,void* data) {
    ANTHOCNETListEntry *entry = (ANTHOCNETListEntry*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETListEntry));
    entry->nodeAddress = nodeAddress;
    entry->data = data;
    return entry;
}


static void ANTHOCNETListInit(ANTHOCNETList* list) {
    list->numberOfEntries = 0;
    list->data = NULL;
    list->head.prev = &(list->head); //prev of head points to head itself
    list->head.next = NULL;
    list->head.data = NULL;
}


static void ANTHOCNETListInsert(ANTHOCNETList* list,ANTHOCNETListEntry* newEntry,ANTHOCNETListEntry* insertAfter) {
    list->numberOfEntries++;
    newEntry->prev = insertAfter;
    newEntry->next = insertAfter->next;
    insertAfter->next = newEntry;
    if (newEntry->next != NULL)
        newEntry->next->prev = newEntry;
}


// FUNCTION	ANTHOCNETListPrint
// PURPOSE	//To use for debugging
// Parameters:
//	node:
//	list:
//	(*printDataFunction)(void*):
//	title:
static void ANTHOCNETListPrint(Node* node,ANTHOCNETList* list,void (*printDataFunction)(void*),char * title) {
    ANTHOCNETListEntry* entry;
    fprintf(stderr,"%s - Number of entries %d\n", title,list->numberOfEntries);
    for (entry = list->head.next; entry != NULL; entry = entry->next)
    {
        fprintf(stderr,"%u\t",MAPPING_GetNodeIdFromInterfaceAddress(node,entry->nodeAddress));
        if (printDataFunction != NULL)
            (*printDataFunction)(entry->data);
    }
    fprintf(stderr, "\n");
}


static void* ANTHOCNETListRemoveEntry(ANTHOCNETListEntry* entry,ANTHOCNETList* list) {
    void* data = entry->data;
    entry->prev->next = entry->next;
    if (entry->next != NULL)
        entry->next->prev = entry->prev;
    MEM_free(entry);
    list->numberOfEntries--;
    return data;
}


static int ANTHOCNETListSearch(ANTHOCNETListEntry** entry,ANTHOCNETList* list,NodeAddress lookupAddress) {
    int found = 0;
    ANTHOCNETListEntry* curr = list->head.prev;
    ANTHOCNETListEntry* next = list->head.next;
    
    while (next != NULL)
    {
        if (next->nodeAddress < lookupAddress) 
        {
            curr = next;
            next = next->next;
        }
        else 
        {
            if (next->nodeAddress == lookupAddress) 
                found = 1;
            break;
        }
    }
    if (found) 
        (*entry) = next;
    else
        (*entry) = curr;
    return found;
}


static int ANTHOCNETListSearchFrom(ANTHOCNETListEntry** entry,ANTHOCNETListEntry* startPoint,NodeAddress lookupAddress) {
    int found = 0;
    ANTHOCNETListEntry* curr = startPoint;
    ANTHOCNETListEntry* next = startPoint->next;
    
    while (next != NULL)
    {
        if (next->nodeAddress < lookupAddress) 
        {
            curr = next;
            next = next->next;
        }
        else 
        {
            if (next->nodeAddress == lookupAddress) 
                found = 1;
            break;
        }
    }
    if (found) 
        (*entry) = next;
    else
        (*entry) = curr;
    return found;
}


static void ANTHOCNETMacAckReceived(Node* node,NodeAddress nextHop,int interfaceIndex) {
    NodeAddress defaultAddressNextHop = MAPPING_GetDefaultInterfaceAddressFromNodeId(node,
                                            MAPPING_GetNodeIdFromInterfaceAddress(node,nextHop));
    ANTHOCNETData *dataPtr = (ANTHOCNETData*)NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_ANTHOCNET);
    ANTHOCNETListEntry* NTNbrEntry;
    BOOL found = ANTHOCNETListSearch(&NTNbrEntry, &(dataPtr->neighbourTable), defaultAddressNextHop);
    if (found) 
    {
        ((ANTHOCNETNTNbr*)NTNbrEntry->data)->lastUpdate = getSimTime(node);
        // We should in fact also update all the destination lastUpdate fields
    }
    else
    {
        ANTHOCNETUpdateRTEntry(dataPtr, defaultAddressNextHop, defaultAddressNextHop, getSimTime(node),
	                1, node, ANTHOCNETGetVirtualSendTime(node,interfaceIndex,dataPtr->defaultSize),
	                ANTHOCNETCalculatePheromoneOneHop(node,dataPtr,0,dataPtr->currentSINR, interfaceIndex,dataPtr->defaultSize,1,
                                                    ANTHOCNETGetVirtualSendTime(node,interfaceIndex,dataPtr->defaultSize)),
                  FROM_HELLO, interfaceIndex, nextHop);
    }
}


// FUNCTION	ANTHOCNETMACLayerStatusEventHandler
// PURPOSE	//Deal with mac layer status events; for the time being we only deal with packet drop notifications
// Parameters:
//	node:
//	msg:
//	nextHop:
//	interfaceIndex:
static void ANTHOCNETMACLayerStatusEventHandler(Node* node,const Message* msg,const NodeAddress nextHop,const int interfaceIndex) {
    static int j1 = 0;
    static clocktype t1 = 0;
    if (msg->eventType == MSG_NETWORK_PacketDropped)
    {
        ANTHOCNETData* dataPtr = (ANTHOCNETData*)NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_ANTHOCNET);
        IpHeaderType* ipHeader = (IpHeaderType *) MESSAGE_ReturnPacket(msg);
        unsigned char  ip_p = ipHeader->ip_p;
        unsigned char* ip_p_ptr;
        NodeAddress *nextHopInfo, *srcAddr, *destAddr;
        char* infoPtr;
        BOOL *fromCleanup;
        Message* linkFailureMsg = MESSAGE_Alloc(node,NETWORK_LAYER, 
                                ROUTING_PROTOCOL_ANTHOCNET, ANTHOCNET_LinkFailureNotification);
        Message* originalMsg = MESSAGE_Duplicate(node,msg);
        Message** originalMsgPtr;
        NodeAddress defaultAddressNextHop = MAPPING_GetDefaultInterfaceAddressFromNodeId(node,
                                                    MAPPING_GetNodeIdFromInterfaceAddress(node,nextHop));
        if (ip_p != IPPROTO_ANTHOCNET)
        {
            // Node %d fails to deliver DATA (seq %d) for dest %d\n",
        }
        else
        {
            // anttype send failure
            if (*((ANTHOCNETAntType*)(MESSAGE_ReturnPacket(originalMsg) + sizeof(IpHeaderType))) == ANTHOCNET_PROACTIVE_FORWARD)
                dataPtr->stats.totalProactiveAntSendFailures++;
        }
        MESSAGE_InfoAlloc(node, linkFailureMsg, sizeof(BOOL) + (3 * sizeof(NodeAddress)) + sizeof(char) + sizeof(Message*));
        infoPtr = MESSAGE_ReturnInfo(linkFailureMsg);
        fromCleanup = (BOOL *)infoPtr;
        *fromCleanup = FALSE;
        infoPtr += sizeof(BOOL);
        nextHopInfo = (NodeAddress *)infoPtr;    
        *nextHopInfo = defaultAddressNextHop;
        infoPtr += sizeof(NodeAddress);
        srcAddr = (NodeAddress *)infoPtr;
        *srcAddr = ipHeader->ip_src;
        infoPtr += sizeof(NodeAddress);
        destAddr = (NodeAddress *)infoPtr;
        *destAddr = ipHeader->ip_dst;
        infoPtr += sizeof(NodeAddress);
        ip_p_ptr = (unsigned char*)infoPtr;
        *ip_p_ptr = ip_p;
        infoPtr += sizeof(char);
        originalMsgPtr = (Message**)infoPtr;
        *originalMsgPtr = originalMsg;
        MESSAGE_Send(node,linkFailureMsg,0);
    }
    else if (msg->eventType == MAC_ANTHOCNET_SendReceive)
    {
        // msg is freed at the mac layer!
        NodeAddress defaultAddressNextHop = MAPPING_GetDefaultInterfaceAddressFromNodeId(node,
                                                MAPPING_GetNodeIdFromInterfaceAddress(node,nextHop));
        ANTHOCNETData *dataPtr = (ANTHOCNETData*)NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_ANTHOCNET);
        ANTHOCNETListEntry* NTNbrEntry;
        BOOL found = ANTHOCNETListSearch(&NTNbrEntry, &(dataPtr->neighbourTable), defaultAddressNextHop);
        if (found) 
        {
            ((ANTHOCNETNTNbr*)NTNbrEntry->data)->lastUpdate = getSimTime(node);
            // We should in fact also update all the destination lastUpdate fields
        }
        else
        {
            ANTHOCNETUpdateRTEntry(dataPtr,defaultAddressNextHop,defaultAddressNextHop,getSimTime(node),1,node, 
                                ANTHOCNETGetVirtualSendTime(node,interfaceIndex,dataPtr->defaultSize),
                                ANTHOCNETCalculatePheromoneOneHop(node,dataPtr,0,dataPtr->currentSINR,
                                                                                                   interfaceIndex,dataPtr->defaultSize,1,
                                                                                                   ANTHOCNETGetVirtualSendTime(node,interfaceIndex,dataPtr->defaultSize)),
                                FROM_HELLO, interfaceIndex, nextHop);
        }
    }
    else
    {
        ANTHOCNETData* dataPtr = (ANTHOCNETData*)NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_ANTHOCNET);
        IpHeaderType* ipHeader = (IpHeaderType *) MESSAGE_ReturnPacket(msg);
        unsigned char  ip_p = ipHeader->ip_p;
        Message* originalMsg = MESSAGE_Duplicate(node,msg);
        NodeAddress srcAddr = ipHeader->ip_src;
        NodeAddress destAddr = ipHeader->ip_dst;
        if (ip_p != IPPROTO_ANTHOCNET) // We don't re-send control packets
        {
            j1++;
            // callback and try resend DATA packet
            ANTHOCNETResendOrBufferPacket(node, dataPtr, destAddr, srcAddr, 
                             ANY_DEST, originalMsg);
        }
        else // it is a control packet
        {
            if (*((ANTHOCNETAntType*)(MESSAGE_ReturnPacket(originalMsg) + sizeof(IpHeaderType))) == ANTHOCNET_PROACTIVE_FORWARD)
                dataPtr->stats.totalProactiveAntSendFailures++;
            MESSAGE_Free(node,originalMsg);
        }
    }
}


// FUNCTION	ANTHOCNETMessageBufferInsert
// PURPOSE	//insert a new message into the message buffer
// Parameters:
//	msg:
//	messageBuffer:
//	node:
//	destAddr:
static ANTHOCNETBufferDest* ANTHOCNETMessageBufferInsert(Message* msg,ANTHOCNETList* messageBuffer,Node* node,NodeAddress destAddr) {
    ANTHOCNETBufferMessage* bufferMsgData = (ANTHOCNETBufferMessage*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETBufferMessage));
    ANTHOCNETBufferDest* bufferDestData;
    ANTHOCNETBufferDest* returnDestData;
    ANTHOCNETListEntry* bufferDestEntry;
    ANTHOCNETListEntry* bufferMsgEntry = (ANTHOCNETListEntry*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETListEntry));
    ANTHOCNETList* messageList;
    BOOL found;
    clocktype time = getSimTime(node);
    int* currentBufferSize = 
                     &((ANTHOCNETBufferData*)messageBuffer->data)->currentBufferSize;
    int maxBufferSize = ((ANTHOCNETBufferData*)messageBuffer->data)->maxBufferSize;
    bufferMsgData->bufferTime = time;
    bufferMsgData->msg = msg;
    bufferMsgEntry->data = bufferMsgData;
    bufferMsgEntry->nodeAddress = destAddr;
    
    found = ANTHOCNETListSearch(&bufferDestEntry, messageBuffer, destAddr);
    
    if (!found)
    {
        ANTHOCNETListEntry* prevBufferDestEntry = bufferDestEntry;
        bufferDestData = (ANTHOCNETBufferDest*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETBufferDest));
        bufferDestData->oldestBufferTime = time;
        bufferDestData->numberOfAntGenerations = 0;
        bufferDestData->bufferDestCreationTime = time;
        bufferDestData->timerMsg = NULL;
        ANTHOCNETListInit(&bufferDestData->messageList);
        bufferDestEntry = (ANTHOCNETListEntry*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETListEntry));
        bufferDestEntry->nodeAddress = destAddr;
        bufferDestEntry->data = bufferDestData;
        ANTHOCNETListInsert(messageBuffer,bufferDestEntry,prevBufferDestEntry);
    }
    else
    {
        bufferDestData = (ANTHOCNETBufferDest*)bufferDestEntry->data;
    }
    returnDestData = bufferDestData;
    if (msg != NULL)
    {
        ANTHOCNETListInsert(&bufferDestData->messageList,bufferMsgEntry,
                                            &bufferDestData->messageList.head);
        if (*currentBufferSize < maxBufferSize)
        {
            (*currentBufferSize)++;
        }
        else
        {
            clocktype oldest = time + 10;
            clocktype current;
            ANTHOCNETListEntry* oldestPtr = NULL;
            ANTHOCNETListEntry* prevBufferMsgEntry;
            ANTHOCNETBufferMessage* bufferMsgData;
            Message* oldMsg;
            for (bufferDestEntry = messageBuffer->head.next;bufferDestEntry != NULL;
                    bufferDestEntry = bufferDestEntry->next)
            {
                current = ((ANTHOCNETBufferDest*)bufferDestEntry->data)->oldestBufferTime;
                if (current < oldest)
                {
                    oldest = current;
                    oldestPtr = bufferDestEntry;
                }
            }
            bufferDestData = (ANTHOCNETBufferDest*)oldestPtr->data;
            for (bufferMsgEntry = &(bufferDestData->messageList.head);
                    bufferMsgEntry->next != NULL;bufferMsgEntry = bufferMsgEntry->next);
            prevBufferMsgEntry = bufferMsgEntry->prev;
            bufferMsgData = (ANTHOCNETBufferMessage*)ANTHOCNETListRemoveEntry(bufferMsgEntry,
                                                                        &bufferDestData->messageList);
            oldMsg = bufferMsgData->msg;
            MESSAGE_Free(node,oldMsg);
            MEM_free(bufferMsgData);
            if (bufferDestData->messageList.numberOfEntries > 0)
            {
                bufferDestData->oldestBufferTime = 
                        ((ANTHOCNETBufferMessage*)prevBufferMsgEntry->data)->bufferTime;
            }
            else
            {
                ANTHOCNETRemoveEmptyBufferDest(node,messageBuffer, oldestPtr);
            }
        }
    }
    return returnDestData;
}


// FUNCTION	ANTHOCNETMessageDebufferDestWithoutSend
// PURPOSE	//Destroy all the messages buffered for a certain destination
// Parameters:
//	node:
//	messageBuffer:
//	bufferDestEntry:
static void ANTHOCNETMessageDebufferDestWithoutSend(Node* node,ANTHOCNETList* messageBuffer,ANTHOCNETListEntry* bufferDestEntry) {
    int i = 0;
    ANTHOCNETData  *dataPtr = (ANTHOCNETData*)NetworkIpGetRoutingProtocol(node,ROUTING_PROTOCOL_ANTHOCNET);
    ANTHOCNETBufferData* bufferData = (ANTHOCNETBufferData*)messageBuffer->data;
    ANTHOCNETBufferDest* bufferDestData;
    ANTHOCNETListEntry *bufferMsgEntry, *tempBufferMsgEntry;
    ANTHOCNETBufferMessage* bufferMsgData;
    Message* msg;
    bufferDestData = (ANTHOCNETBufferDest*)bufferDestEntry->data;
    for (bufferMsgEntry = bufferDestData->messageList.head.next;bufferMsgEntry != NULL;)
    {
        bufferMsgData = (ANTHOCNETBufferMessage*)bufferMsgEntry->data;
        msg = bufferMsgData->msg;
        tempBufferMsgEntry = bufferMsgEntry;
        bufferMsgEntry = bufferMsgEntry->next;
        MESSAGE_Free(node,msg);
        MEM_free(bufferMsgData);
        ANTHOCNETListRemoveEntry(tempBufferMsgEntry,&(bufferDestData->messageList));
        i++;
        dataPtr->stats.packetsDroppedAtDebufferWithoutSend++;
    }
    bufferData->currentBufferSize -= i;
    ANTHOCNETRemoveEmptyBufferDest(node,messageBuffer, bufferDestEntry);
}


// FUNCTION	ANTHOCNETMessageDebufferDestWithSend
// PURPOSE	//Send all the messages buffered for a certain destination
// Parameters:
//	node:
//	messageBuffer:
//	bufferDestEntry:
//	interfaceIndex:
//	interfaceAddress:
static void ANTHOCNETMessageDebufferDestWithSend(Node* node,ANTHOCNETList* messageBuffer,ANTHOCNETListEntry* bufferDestEntry,int interfaceIndex,NodeAddress interfaceAddress) {
    ANTHOCNETData  *dataPtr = (ANTHOCNETData*)NetworkIpGetRoutingProtocol(node,ROUTING_PROTOCOL_ANTHOCNET);
    int i;
    ANTHOCNETBufferData* bufferData = (ANTHOCNETBufferData*)messageBuffer->data;
    ANTHOCNETBufferDest* bufferDestData;
    ANTHOCNETListEntry* bufferMsgEntry;
    ANTHOCNETBufferMessage* bufferMsgData;
    Message* msg;
    bufferDestData = (ANTHOCNETBufferDest*)bufferDestEntry->data;
    // Find the last message in the list (the oldest)
    for (bufferMsgEntry = &(bufferDestData->messageList.head),i = 0;
         bufferMsgEntry->next != NULL;
         bufferMsgEntry = bufferMsgEntry->next,i++);
    bufferData->currentBufferSize -= i;
    // Start sending, starting from the oldest
    for (;i>0;i--)
    {
        bufferMsgData = (ANTHOCNETBufferMessage*)bufferMsgEntry->data;
        msg = bufferMsgData->msg;
        if ((dataPtr->admissionControlFlag) && (msg->originatingNodeId == node->nodeId) && (ANTHOCNETRejectedSession(node, dataPtr, bufferDestEntry->nodeAddress)))
        {
            MESSAGE_Free(node,msg);
        }
        else if ((msg->packetCreationTime + dataPtr->maxBufferTime > getSimTime(node)))   
        {
            if (PRINT_DELAY_DETAILS)
            {
                IpHeaderType *ipHeader1 = (IpHeaderType *)msg->packet;
                fprintf(stderr,"[%g] node %u packet (%u,%u,%g,%d): debuffer and send from routing layer to IP\n",
                                simTimeSec(getSimTime(node)),node->nodeId,
                                MAPPING_GetNodeIdFromInterfaceAddress(node,ipHeader1->ip_src),
                                MAPPING_GetNodeIdFromInterfaceAddress(node,ipHeader1->ip_dst),
                                simTimeSec(msg->packetCreationTime),msg->sequenceNumber);
            }
            NetworkIpSendPacketToMacLayer(node,msg,interfaceIndex,interfaceAddress);
        }
        else
        {
            MESSAGE_Free(node,msg);
            dataPtr->stats.packetsDroppedAtDebuffering++;
        }
        bufferMsgEntry = bufferMsgEntry->prev;
        MEM_free(bufferMsgData);
        ANTHOCNETListRemoveEntry(bufferMsgEntry->next,&(bufferDestData->messageList));
    }
    ANTHOCNETRemoveEmptyBufferDest(node,messageBuffer, bufferDestEntry);
}


static void ANTHOCNETNTNbrPrintData(Node* node,void* data) {
    ANTHOCNETNTNbr* nbrData = (ANTHOCNETNTNbr*)(data);
    fprintf(stderr,"lastUpdate: %g\n\t",simTimeSec(nbrData->lastUpdate));
    ANTHOCNETListPrint(node,&nbrData->destinations, NULL, "destinations");
}


static void ANTHOCNETNTRemoveEntrySimple(ANTHOCNETList* NT,NodeAddress nbrAddr,NodeAddress destAddr) {
    ANTHOCNETListEntry *nbrEntry, *nbrDestEntry;
    ANTHOCNETNTNbrDest *nbrDestData;
    BOOL found = ANTHOCNETListSearch(&nbrEntry, NT, nbrAddr);
    if (found)
    {
        ANTHOCNETNTNbr* nbrData = (ANTHOCNETNTNbr*)nbrEntry->data;
        found = ANTHOCNETListSearch(&nbrDestEntry,&nbrData->destinations,destAddr);
        if (found)
        {
            nbrDestData = (ANTHOCNETNTNbrDest*)ANTHOCNETListRemoveEntry(nbrDestEntry,
                                                        &nbrData->destinations);
            ANTHOCNETRTRemoveNextHopSimple(nbrDestData->RTDest,nbrAddr);
            MEM_free(nbrDestData);
        }
    }
}


static double ANTHOCNETNTRemoveNbrSplitArrays(ANTHOCNETListEntry* nbrEntry,ANTHOCNETList* NT,ANTHOCNETLostPheromone** lostBestArray,int* lostBestNumber,ANTHOCNETLostPheromone** lostLastArray,int* lostLastNumber,NodeAddress destAddr,BOOL wasSource,BOOL* wasLast,Node* node,ANTHOCNETData* dataPtr) {
    int k;
    Pheromone newBestPheromone = 0;
    double newBestHopsToGo = 0;
    double newBestTimeToGo,maxAvgTime = 0;
    BOOL newBest;
    ANTHOCNETListEntry* nbrDestEntry;
    ANTHOCNETNTNbrDest* nbrDestData;
    ANTHOCNETRTDest* RTDestData;
    NodeAddress nbrAddress = nbrEntry->nodeAddress;
    ANTHOCNETNTNbr* nbrData = (ANTHOCNETNTNbr*)ANTHOCNETListRemoveEntry(nbrEntry,NT);
    ANTHOCNETLostPheromone* tempBestArray = (ANTHOCNETLostPheromone*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETLostPheromone) * 
                                        nbrData->destinations.numberOfEntries);
    ANTHOCNETLostPheromone* tempLastArray = (ANTHOCNETLostPheromone*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETLostPheromone) * 
                                        nbrData->destinations.numberOfEntries);
//    mem_error(tempBestArray, "Cannot allocate tempBestArray in NTRemoveNbr!!");                                        
//    mem_error(tempLastArray, "Cannot allocate tempLastArray in NTRemoveNbr!!");   
    *lostBestNumber = 0;
    *lostLastNumber = 0;                                     
    *wasLast = FALSE;
    for(nbrDestEntry = nbrData->destinations.head.next;nbrDestEntry != NULL;
        nbrDestEntry = nbrData->destinations.head.next)
    {
        nbrDestData = (ANTHOCNETNTNbrDest*)
                    ANTHOCNETListRemoveEntry(nbrDestEntry,&nbrData->destinations);
        newBest = ANTHOCNETRTRemoveNextHop(nbrDestData->RTDest,nbrAddress,
                                        &newBestHopsToGo,&newBestTimeToGo,&newBestPheromone);
        if (ANTHOCNET_PRINT_PHEROMONE_CHANGE) ANTHOCNETPrintPheromone(node, dataPtr, nbrDestData->RTDest->nodeAddress, ANTHOCNET_LINK_FAILURE);
        if (newBest)
        {
            RTDestData = (ANTHOCNETRTDest*)nbrDestData->RTDest->data;
            // 2nd part of if-clause is to let only the currently lost path be repaired
            // it makes the else-clause inside redundant, but we will fix this later
            if ((newBestHopsToGo == ANTHOCNET_NO_MORE_ENTRIES) && (nbrDestData->RTDest->nodeAddress == destAddr))
            {
                if (wasSource && (nbrDestData->RTDest->nodeAddress == destAddr))
                {
                    *wasLast = TRUE;
                }
                else
                {
                    if (maxAvgTime < RTDestData->timeAvg) 
                        maxAvgTime = RTDestData->timeAvg;
                    tempLastArray[*lostLastNumber].destAddr = nbrDestData->RTDest->nodeAddress;
                    tempLastArray[*lostLastNumber].newBestHopsToGo = newBestHopsToGo;
                    tempLastArray[*lostLastNumber].newBestTimeToGo = newBestTimeToGo;
                     tempLastArray[*lostLastNumber].newBestPheromone = newBestPheromone;
                    (*lostLastNumber)++;
                }
            }
            else // if (newBestHopsToGo == ANTHOCNET_NO_MORE_ENTRIES) ADAPTATION 03-01-2008: ONLY SEND LOST LAST, NOT LOST BEST... -  BUT CAREFUL!!!: LOST LAST AND LOST BEST ALSO USED TO DECIDE ABOUT REPAIR - SHOULD CHANGE THE NAMES OF THESE ARRAYS
            {
                tempBestArray[*lostBestNumber].destAddr = nbrDestData->RTDest->nodeAddress;
                tempBestArray[*lostBestNumber].newBestHopsToGo = newBestHopsToGo;
                tempBestArray[*lostBestNumber].newBestTimeToGo = newBestTimeToGo;
                 tempBestArray[*lostBestNumber].newBestPheromone = newBestPheromone;
                (*lostBestNumber)++;
            }
        }
        MEM_free(nbrDestData);
    }
    MEM_free(nbrData);   
    *lostLastArray = (ANTHOCNETLostPheromone*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETLostPheromone) * (*lostLastNumber));
//    mem_error(*lostLastArray, "Cannot allocate lostLastArray!!");
    memcpy(*lostLastArray,tempLastArray,sizeof(ANTHOCNETLostPheromone) * (*lostLastNumber));
    MEM_free(tempLastArray);
    *lostBestArray = (ANTHOCNETLostPheromone*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETLostPheromone) * (*lostBestNumber));
//    mem_error(*lostBestArray, "Cannot allocate lostBestArray!!");
    memcpy(*lostBestArray,tempBestArray,sizeof(ANTHOCNETLostPheromone) * (*lostBestNumber));
    MEM_free(tempBestArray);
    return maxAvgTime;
}


static BOOLEAN ANTHOCNETPacketInLoop(Node* node,ANTHOCNETData* dataPtr,Message* msg,NodeAddress prevHopAddr) {
    if (prevHopAddr == ANY_IP)
        return FALSE;
    else
    {
        int i;
        ANTHOCNETMsgId *msgIds = dataPtr->msgIds;
        for (i=0;i<ANTHOCNET_NUMBER_OF_MSG_IDS;i++)
        {
            if (msgIds[i].seqNumber == msg->sequenceNumber)
                if (msgIds[i].srcId == msg->originatingNodeId)
                    return TRUE;
        }
        return FALSE;
    }
}

static void ANTHOCNETPrintPath(Node* node,ANTHOCNETVisitedNode* visitedNodes,int numberOfVisitedNodes,int seqno) {
    int i;
    fprintf(stderr,"Source %d, Destination %u, Path length %d, Time %g, Seq no %d\n\t",
             MAPPING_GetNodeIdFromInterfaceAddress(node,visitedNodes[0].node),
             MAPPING_GetNodeIdFromInterfaceAddress(node,visitedNodes[numberOfVisitedNodes-1].node),
             numberOfVisitedNodes-1,simTimeSec(getSimTime(node)),seqno);
    for (i=0;i<numberOfVisitedNodes;i++)
    {
        fprintf(stderr,"%d - ",MAPPING_GetNodeIdFromInterfaceAddress(node,visitedNodes[i].node));
    }
    fprintf(stderr,"\n");
}


// FUNCTION	ANTHOCNETProcessAntAtDest
// PURPOSE	//Receiving a forward ant (proactive, reactive or path repair) and sending a backward one back.
// Parameters:
//	node:
//	dataPtr:
//	FAnt:
//	antType:
//	interfaceIndex:
//	interfaceAddress:
static void ANTHOCNETProcessAntAtDest(Node* node,ANTHOCNETData* dataPtr,ANTHOCNETForwardAnt* FAnt,ANTHOCNETAntType antType,int interfaceIndex,NodeAddress interfaceAddress) {
  ANTHOCNETBackwardAnt* BAnt;
  NodeAddress nextHop;
  PacketBuffer* packetBuffer; 
  Message* msgNew = NULL;
  ANTHOCNETListEntry *NTNbrEntry;
  float SINR = 0;
  if (ANTHOCNETListSearch(&NTNbrEntry,&dataPtr->neighbourTable,FAnt->visitedNodes[FAnt->numberOfVisitedNodes-1].node))
  {
      SINR = ((ANTHOCNETNTNbr*)NTNbrEntry->data)->SINR;
  }
                
//  ANTHOCNETUpdateRTForward(FAnt, dataPtr, node);

  BAnt = ANTHOCNETCreateBackwardAnt(node,FAnt, dataPtr->myAddress,SINR);
  nextHop = BAnt->visitedNodes[BAnt->currentPosition-1].node;
  if (dataPtr->bidirectionalPaths) 
  {
      ANTHOCNETUpdateRTForward(BAnt, dataPtr, node);
  }
  if (antType == ANTHOCNET_REPAIR_FORWARD)
  {
      packetBuffer = ANTHOCNETBuildBackwardAntPacket(ANTHOCNET_REPAIR_BACKWARD, BAnt);
  }
  else
  {
      packetBuffer = ANTHOCNETBuildBackwardAntPacket(ANTHOCNET_REACTIVE_BACKWARD, BAnt);
  }
  ANTHOCNETEncapsulatePacketInMessage(node, &msgNew, packetBuffer);
  NetworkIpSendRawMessageToMacLayer(node, msgNew,NetworkIpGetInterfaceAddress(node,interfaceIndex),
				    ANY_DEST, IPTOS_PREC_INTERNETCONTROL, IPPROTO_ANTHOCNET,
				    5, interfaceIndex,
				    interfaceAddress);

  TRACE(stderr,"[%g]Reactive Forward arrived and Generate Backward ant %u - %u -> %u (seq %d) at Node %u\n", simTimeSec(getSimTime(node)),
	    MAPPING_GetNodeIdFromInterfaceAddress(node,FAnt->sourceAddress),
        MAPPING_GetNodeIdFromInterfaceAddress(node,FAnt->visitedNodes[FAnt->numberOfVisitedNodes-1].node),
        MAPPING_GetNodeIdFromInterfaceAddress(node,FAnt->destinationAddress),
        FAnt->generationNumber, node->nodeId);
  ANTHOCNETDeleteBackwardAnt(BAnt);
}


static void ANTHOCNETRefreshTimeStampsAfterHello(ANTHOCNETData* dataPtr,NodeAddress nbrAddr,clocktype currentTime) {
    ANTHOCNETListEntry *NTNbrEntry, *next, *RTDestNextHopEntry;
    ANTHOCNETList *nextHops;
    ANTHOCNETRTDest *RTDestData;
    BOOL found = ANTHOCNETListSearch(&NTNbrEntry, &dataPtr->neighbourTable, nbrAddr);
    if (found)
    {
        ((ANTHOCNETNTNbr*)NTNbrEntry->data)->lastUpdate = currentTime;
        for (next = ((ANTHOCNETNTNbr*)NTNbrEntry->data)->destinations.head.next;
             next != NULL;next = next->next)
         {
             // if == nbrAddr, the updating is done inside ANTHOCNETUpdateRTEntry
             if (next->nodeAddress != nbrAddr)
             {
                 nextHops = &(((ANTHOCNETRTDest*)((ANTHOCNETNTNbrDest*)next->data)->RTDest->data)->nextHops);
                 found = ANTHOCNETListSearch(&RTDestNextHopEntry,nextHops,nbrAddr);
                 if (found)
                 {
                     ((ANTHOCNETRTDestNextHop*)RTDestNextHopEntry->data)->lastUpdate = currentTime;
                 }
                 else
                 {
                      ERROR_Assert(FALSE, "Neighbour table and routing table don't correspond!");
                      abort();
                 }
             }
         }
    }
}


// FUNCTION	ANTHOCNETRemoveEmptyBufferDest
// PURPOSE	//If there is an event pending for a destination, cancel it before deleting the destination
// Parameters:
//	node:
//	messageBuffer:
//	bufferDestEntry:
static void ANTHOCNETRemoveEmptyBufferDest(Node* node,ANTHOCNETList* messageBuffer,ANTHOCNETListEntry* bufferDestEntry) {
    ANTHOCNETBufferDest* bufferDestData = (ANTHOCNETBufferDest*)bufferDestEntry->data;
    if (bufferDestData->timerMsg != NULL)
        MESSAGE_CancelSelfMsg(node,bufferDestData->timerMsg);
    MEM_free(bufferDestData);
    ANTHOCNETListRemoveEntry(bufferDestEntry,messageBuffer);
}


static void ANTHOCNETResendOrBufferPacket(Node* node,ANTHOCNETData* dataPtr,NodeAddress destAddr,NodeAddress srcAddr,NodeAddress prevHopAddr,Message* originalMsg) {
    BOOL sent = FALSE,found;
    ANTHOCNETListEntry *destEntry, *nextHopEntry;
    
    if (PRINT_DELAY_DETAILS)
    {
        IpHeaderType *ipHeader1 = (IpHeaderType *)originalMsg->packet;
        fprintf(stderr,"[%g] node %u packet (%u,%u,%g,%d): failed sending\n",
                            simTimeSec(getSimTime(node)),node->nodeId,
                            MAPPING_GetNodeIdFromInterfaceAddress(node,ipHeader1->ip_src),
                            MAPPING_GetNodeIdFromInterfaceAddress(node,ipHeader1->ip_dst),
                            simTimeSec(originalMsg->packetCreationTime),
                            originalMsg->sequenceNumber);
    }

    found = ANTHOCNETListSearch(&destEntry, &dataPtr->routingTable, destAddr);
    if (found)
    {
        nextHopEntry = ANTHOCNETGetNextHopData(node, destAddr, prevHopAddr, 
                            &dataPtr->routingTable,dataPtr->dataRoutingExponent);
        if (nextHopEntry != NULL)
        {
            // REsend data
            ANTHOCNETRTDestNextHop *RTDestData = (ANTHOCNETRTDestNextHop*)nextHopEntry->data;
            ANTHOCNETUpdateRTDestLastUsed(node, dataPtr, destEntry, destAddr);
            ANTHOCNETUpdateRTDestLastDataTransported(node, dataPtr, destEntry, destAddr);
            if (PRINT_DELAY_DETAILS)
            {
                IpHeaderType *ipHeader1 = (IpHeaderType *)originalMsg->packet;
                fprintf(stderr,"[%g] node %u packet (%u,%u,%g,%d): resend from routing layer to IP\n",
                            simTimeSec(getSimTime(node)),node->nodeId,
                            MAPPING_GetNodeIdFromInterfaceAddress(node,ipHeader1->ip_src),
                            MAPPING_GetNodeIdFromInterfaceAddress(node,ipHeader1->ip_dst),
                            simTimeSec(originalMsg->packetCreationTime),
                            originalMsg->sequenceNumber);
            }
            NetworkIpSendPacketToMacLayer(node,originalMsg, RTDestData->interfaceIndex, 
                                          RTDestData->interfaceAddress);
            sent = TRUE;
        }
    }
    if (sent == FALSE)
    {
        ANTHOCNETListEntry* bufferDestEntry;
        found = ANTHOCNETListSearch(&bufferDestEntry, &dataPtr->messageBuffer, destAddr) ||
              ANTHOCNETListSearch(&bufferDestEntry, &dataPtr->relayMessageBuffer, destAddr);
        if (found) // I am waiting for ants searching this destination
        {
            // REbuffers DATA
            if (PRINT_DELAY_DETAILS)
            {
                IpHeaderType *ipHeader1 = (IpHeaderType *)originalMsg->packet;
                fprintf(stderr,"[%g] node %u packet (%u,%u,%g,%d): rebuffer at routing layer\n",
                            simTimeSec(getSimTime(node)),node->nodeId,
                            MAPPING_GetNodeIdFromInterfaceAddress(node,ipHeader1->ip_src),
                            MAPPING_GetNodeIdFromInterfaceAddress(node,ipHeader1->ip_dst),
                            simTimeSec(originalMsg->packetCreationTime),
                            originalMsg->sequenceNumber);
            }

            if (NetworkIpIsMyIP(node, srcAddr))
            {
                ANTHOCNETMessageBufferInsert(originalMsg, &dataPtr->messageBuffer, node,destAddr);
            }
            else
            {
                ANTHOCNETMessageBufferInsert(originalMsg, &dataPtr->relayMessageBuffer, node, destAddr);
            }
        }
        else // no ants are sent out, I just delete the packet
        {
            MESSAGE_Free(node,originalMsg);
        }
    }
}


static void ANTHOCNETRTDestNextHopPrintData(void* data) {
    ANTHOCNETRTDestNextHop* nextHopData = (ANTHOCNETRTDestNextHop*)(data);
    fprintf(stderr,"pheromone: %g   hops average : %g   last update: %g\n",
            nextHopData->pheromone,nextHopData->hopsAvg,simTimeSec(nextHopData->lastUpdate));
}


static void ANTHOCNETRTDestPrintData(Node* node,void* data) {
    ANTHOCNETRTDest* destData = (ANTHOCNETRTDest*)(data);
    ANTHOCNETListPrint(node,&destData->nextHops, ANTHOCNETRTDestNextHopPrintData, "nextHops");
}


// FUNCTION	ANTHOCNETRTRemoveEmptyDest
// PURPOSE	//Only for destinations which have no more next hops!
// Parameters:
//	RTDestEntry:
//	routingTable:
static void ANTHOCNETRTRemoveEmptyDest(ANTHOCNETListEntry* RTDestEntry,ANTHOCNETList* routingTable) {
    ANTHOCNETRTDest *RTDestData = (ANTHOCNETRTDest*)ANTHOCNETListRemoveEntry(RTDestEntry, routingTable);
    MEM_free(RTDestData);
}


// FUNCTION	ANTHOCNETRTRemoveNextHop
// PURPOSE	//Return the percentage of pheromone lost for this destination
// Parameters:
//	RTDest:
//	nextHop:
//	newBestHopsToGo:
//	newBestTimeToGo:
//	newBestPheromone:
static BOOL ANTHOCNETRTRemoveNextHop(ANTHOCNETListEntry* RTDest,NodeAddress nextHop,double* newBestHopsToGo,double* newBestTimeToGo,Pheromone* newBestPheromone) {
    double bestPheromone = 0, removedPheromone = 0;
    BOOL found = FALSE;
    ANTHOCNETRTDestNextHop* RTDestNextHopData;
    ANTHOCNETRTDest* RTDestData = (ANTHOCNETRTDest*)RTDest->data;
    
    ANTHOCNETListEntry* curr = RTDestData->nextHops.head.prev;
    ANTHOCNETListEntry* next = RTDestData->nextHops.head.next;
    
    *newBestHopsToGo = ANTHOCNET_NO_MORE_ENTRIES;
    *newBestTimeToGo = ANTHOCNET_NO_MORE_ENTRIES;
    *newBestPheromone = ANTHOCNET_NO_MORE_ENTRIES;
    
    while (next != NULL)
    {
        curr = next;
        next = next->next;
        if (curr->nodeAddress == nextHop) 
        {
            found = TRUE;
            RTDestNextHopData = (ANTHOCNETRTDestNextHop*)
                ANTHOCNETListRemoveEntry(curr, &(RTDestData->nextHops));
            removedPheromone = RTDestNextHopData->pheromone;
            if (removedPheromone > 0)
                RTDestData->numberOfConfirmedEntries--;
            MEM_free(RTDestNextHopData);
        }
        else
        {
            RTDestNextHopData = (ANTHOCNETRTDestNextHop*)curr->data;
            if  (bestPheromone < RTDestNextHopData->pheromone)
            {
                bestPheromone = RTDestNextHopData->pheromone;
                *newBestHopsToGo = RTDestNextHopData->hopsAvg;
                *newBestTimeToGo = RTDestNextHopData->timeAvg;
                *newBestPheromone = RTDestNextHopData->pheromone;
            }
        }
    }
    if (!found)
    {
        ERROR_Assert(FALSE,"Neighbour not there!!!!!!");
        abort();
    }
    return ((removedPheromone > bestPheromone) ? TRUE : FALSE);
}


static void ANTHOCNETRTRemoveNextHopSimple(ANTHOCNETListEntry* RTDestEntry,NodeAddress nextHop) {
    ANTHOCNETRTDest* RTDestData = (ANTHOCNETRTDest*)RTDestEntry->data;
    ANTHOCNETListEntry* RTDestNextHopEntry;
    ANTHOCNETRTDestNextHop* RTDestNextHopData;
    BOOL found = ANTHOCNETListSearch(&RTDestNextHopEntry, &RTDestData->nextHops, nextHop);
    if (found)
    {
        RTDestNextHopData = (ANTHOCNETRTDestNextHop*)
                    ANTHOCNETListRemoveEntry(RTDestNextHopEntry, &(RTDestData->nextHops));
        if (RTDestNextHopData->pheromone > 0)
            RTDestData->numberOfConfirmedEntries--;
        MEM_free(RTDestNextHopData);
    }
    else
    {
        ERROR_Assert(FALSE,"Neighbour not there!!!!!!");
        abort();
    }
}


static void ANTHOCNETSendSimpleLinkFailureNotification(Node* node,ANTHOCNETData* dataPtr,NodeAddress destinationAddress) {
    PacketBuffer *packetBuffer;
    Message *msg;
    ANTHOCNETLostPheromone *lostPheromone = (ANTHOCNETLostPheromone*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETLostPheromone));
    
    lostPheromone->destAddr = destinationAddress;
    lostPheromone->newBestHopsToGo = ANTHOCNET_NO_MORE_ENTRIES;
    lostPheromone->newBestTimeToGo = ANTHOCNET_NO_MORE_ENTRIES;
    lostPheromone->newBestPheromone = ANTHOCNET_NO_MORE_ENTRIES;
    
    packetBuffer = ANTHOCNETBuildLinkFailureAntPacket(ANTHOCNET_LINK_FAILURE_NOTIFICATION, lostPheromone, 1);
    ANTHOCNETEncapsulatePacketInMessage(node, &msg, packetBuffer);
    MEM_free(lostPheromone);
            
    /* Send the link failure message (ttl at the moment set to 1) */
    ANTHOCNETBroadcastControlMessage(node, msg, IPTOS_PREC_INTERNETCONTROL, 0, 1);
                                      
    dataPtr->stats.totalLinkFailureNotificationsGenerated++;
}


static double ANTHOCNETTotalPheromone(ANTHOCNETList* nextHops,NodeAddress prevAddr,double exponent) {
    double totalPheromone = 0;
    ANTHOCNETListEntry* nextHopEntry;
    for (nextHopEntry = nextHops->head.next;nextHopEntry != NULL;
         nextHopEntry = nextHopEntry->next)
    {
        if (nextHopEntry->nodeAddress == prevAddr) 
            continue;
        totalPheromone += pow(((ANTHOCNETRTDestNextHop*)nextHopEntry->data)->pheromone,exponent);
    }
    return totalPheromone;
}


static short ANTHOCNETRTUpdateAfterFailureMsg(Node* node,ANTHOCNETData* dataPtr,ANTHOCNETList* RT,ANTHOCNETList* NT,ANTHOCNETLostPheromone* receivedArray,short receivedNumberOfElements,ANTHOCNETLostPheromone** returnArray,NodeAddress sourceAddr,int interfaceIndex,int size) {
    short i, returnNumberOfElements = 0;
    ANTHOCNETListEntry* RTDestEntry = RT->head.next;
    ANTHOCNETListEntry *curr, *next;
    ANTHOCNETRTDest* RTDestData;
    ANTHOCNETRTDestNextHop* RTDestNextHopData;
    double lostPheromoneRatio;
    ANTHOCNETLostPheromone* tempArray = (ANTHOCNETLostPheromone*)MEM_malloc2(__LINE__,receivedNumberOfElements * 
                                                sizeof(ANTHOCNETLostPheromone));
    float SINR = 0;
    ANTHOCNETListEntry *NTNbrEntry;
    if (ANTHOCNETListSearch(&NTNbrEntry,&dataPtr->neighbourTable,sourceAddr))
    {
        SINR = ((ANTHOCNETNTNbr*)NTNbrEntry->data)->SINR;
    }
    for (i = 0;i < receivedNumberOfElements;i++)
    {
        while ((RTDestEntry != NULL) && 
               (RTDestEntry->nodeAddress < receivedArray[i].destAddr))
            RTDestEntry = RTDestEntry->next;
        if (RTDestEntry == NULL) break;
        // if: we have found one of the received destinations in the RT
        if (RTDestEntry->nodeAddress == receivedArray[i].destAddr)
        {
            // Important that pheromone values initialized to 0 and not -1 !!!
            // Otherwise we send updates about bootstrapped entries!!!!
            Pheromone bestPheromone = 0, updatedPheromone = 0; 
            double bestHopsToGo = ANTHOCNET_NO_MORE_ENTRIES, bestTimeToGo = ANTHOCNET_NO_MORE_ENTRIES;
            double newBestHopsToGo = receivedArray[i].newBestHopsToGo + 1;
            double newBestTimeToGo = receivedArray[i].newBestTimeToGo + 
                                     ANTHOCNETGetVirtualSendTime(node, interfaceIndex, size);
            Pheromone newBestPheromone = ANTHOCNETCalculatePheromoneOneHop(node,dataPtr,receivedArray[i].newBestPheromone,
                                                                                                              SINR,interfaceIndex,size,newBestHopsToGo,newBestTimeToGo);
            RTDestData = (ANTHOCNETRTDest*)RTDestEntry->data;
            curr = RTDestData->nextHops.head.prev;
            next = RTDestData->nextHops.head.next;
            while (next != NULL) // go through the nexthops list of a destination
            {
                curr = next;
                next = next->next;
                // if: the current nexthop is the one we got the ant from
                if (curr->nodeAddress == sourceAddr) 
                {
                    // If all the pheromone is lost, the entry is deleted from the
                    // routing table and from the neighbour table.
                    if (receivedArray[i].newBestHopsToGo == ANTHOCNET_NO_MORE_ENTRIES)
                    {
                        ANTHOCNETListEntry* NTNbrEntry;
                        ANTHOCNETListEntry* NTNbrDestEntry;
                        ANTHOCNETNTNbrDest* NTNbrDestData;
                        BOOL found;
                        RTDestNextHopData = (ANTHOCNETRTDestNextHop*)
                            ANTHOCNETListRemoveEntry(curr, &(RTDestData->nextHops));
                        updatedPheromone = RTDestNextHopData->pheromone;
                        if (updatedPheromone > 0)
                            RTDestData->numberOfConfirmedEntries--;
                        MEM_free(RTDestNextHopData);
                        found = ANTHOCNETListSearch(&NTNbrEntry, NT, sourceAddr);
                        if (found)
                            found = ANTHOCNETListSearch(&NTNbrDestEntry,
                                    &((ANTHOCNETNTNbr*)NTNbrEntry->data)->destinations,
                                    receivedArray[i].destAddr);
                        if (!found)
                        {
                            ERROR_Assert(FALSE, "entry not found in NT");
                            abort();
                        }
                        NTNbrDestData = (ANTHOCNETNTNbrDest*)ANTHOCNETListRemoveEntry(NTNbrDestEntry,
                                &((ANTHOCNETNTNbr*)NTNbrEntry->data)->destinations);
                        MEM_free(NTNbrDestData);
                    } // end if (receivedArray[i].newBestHopsToGo == ANTHOCNET_NO_MORE_ENTRIES)
                    else // receivedArray[i].newBestHopsToGo != ANTHOCNET_NO_MORE_ENTRIES
                    {
                        RTDestNextHopData = (ANTHOCNETRTDestNextHop*)curr->data;
                        updatedPheromone = RTDestNextHopData->pheromone;
                        if (RTDestNextHopData->pheromone != 0)
                        {
                            RTDestNextHopData->pheromone = newBestPheromone;
                            RTDestNextHopData->hopsAvg = newBestHopsToGo;
                            RTDestData->hopsAvg = newBestHopsToGo + dataPtr->discountHops *
                                                  (RTDestData->hopsAvg - newBestHopsToGo);
                            RTDestNextHopData->timeAvg = newBestTimeToGo;
                            RTDestData->timeAvg = newBestTimeToGo + dataPtr->discountTripTime *
                                                  (RTDestData->timeAvg - newBestTimeToGo);
                            if (bestPheromone < RTDestNextHopData->pheromone)
                            {
                                bestPheromone = RTDestNextHopData->pheromone;
                                bestHopsToGo = RTDestNextHopData->hopsAvg;
                                bestTimeToGo = RTDestNextHopData->timeAvg;
                            }
                        }
                        else // learn bootstrapped information from failure message, no further updates:
                        {
                            RTDestNextHopData->bootstrappedPheromone = newBestPheromone;
                            RTDestNextHopData->bootstrappedTime = getSimTime(node);
                            RTDestNextHopData->bootstrappedForwarded = FALSE;
                        }
                    }
                } // end if: we found the neighbour in the next hop list
                // else: the current nextHop is not the one we got the ant from
                else
                {
                    RTDestNextHopData = (ANTHOCNETRTDestNextHop*)curr->data;
                    if (bestPheromone < RTDestNextHopData->pheromone)
                    {
                        bestPheromone = RTDestNextHopData->pheromone;
                        bestHopsToGo = RTDestNextHopData->hopsAvg;
                        bestTimeToGo = RTDestNextHopData->timeAvg;
                    }
                }
            } // while: we go through the nexthop list
            if (bestPheromone < updatedPheromone)
            {
                tempArray[returnNumberOfElements].newBestHopsToGo = bestHopsToGo;
                tempArray[returnNumberOfElements].newBestTimeToGo = bestTimeToGo;
                tempArray[returnNumberOfElements].newBestPheromone = bestPheromone;
                tempArray[returnNumberOfElements++].destAddr = receivedArray[i].destAddr;
            }
            if (ANTHOCNET_PRINT_PHEROMONE_CHANGE) ANTHOCNETPrintPheromone(node,dataPtr,receivedArray[i].destAddr,ANTHOCNET_LINK_FAILURE_NOTIFICATION);
        }  // if we found a received destination in the routing table
    } // end of for through all the received destinations
    if (returnNumberOfElements > 0)
    {
        *returnArray = (ANTHOCNETLostPheromone*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETLostPheromone) * returnNumberOfElements);
        mem_error(*returnArray, "Cannot allocate returnArray");
        memcpy(*returnArray,tempArray,sizeof(ANTHOCNETLostPheromone) * returnNumberOfElements);
    }
    MEM_free(tempArray);
    return returnNumberOfElements;
}


static void ANTHOCNETRTUpdateBootstrapping(Node* node,ANTHOCNETData* dataPtr,ANTHOCNETHelloEntry* helloEntryArray,ANTHOCNETHello* hello,NodeAddress nextHop,int interfaceIndex,NodeAddress interfaceAddress,int size) {
int i;
    Pheromone bestPheromone, receivedPheromone, newPheromone;
    ANTHOCNETListEntry *RTDestEntry, *RTDestNextHopEntry, *newRTDestNextHopEntry, *NTNbrEntry, 
                            *NTNbrDestEntry, *newNTNbrDestEntry, *newRTDestEntry;
    ANTHOCNETRTDestNextHop *RTDestNextHopData;
    RTDestEntry = dataPtr->routingTable.head.prev;
    ANTHOCNETListSearch(&NTNbrEntry, &dataPtr->neighbourTable, nextHop);
    NTNbrDestEntry = ((ANTHOCNETNTNbr*)NTNbrEntry->data)->destinations.head.prev;
    for (i = 0;i < hello->numberOfEntries;i++)
    {
        // We do not update for ourselves:
        if (helloEntryArray[i].destAddr != dataPtr->myAddress)
        {
            receivedPheromone = ANTHOCNETDediscretizePheromone(helloEntryArray[i].discretePheromone);
            newPheromone = ANTHOCNETCalculatePheromoneOneHop(node,dataPtr,receivedPheromone,
                                                                                    ((ANTHOCNETNTNbr*)NTNbrEntry->data)->SINR,
                                                                                    interfaceIndex,size,-1,-1);
            // Destination already exists in routing table
            if (ANTHOCNETListSearchFrom(&RTDestEntry, RTDestEntry, helloEntryArray[i].destAddr))
            {
                // The nextHop exists
                if (ANTHOCNETListSearch(&RTDestNextHopEntry, &ANTHOCNET_nextHopsList(RTDestEntry), nextHop))
                {
                    RTDestNextHopData = (ANTHOCNETRTDestNextHop*)RTDestNextHopEntry->data;
                    // If this path was already sampled, bootstrapping is used for probing:
                    if ((RTDestNextHopData->pheromone > 0) && (!helloEntryArray[i].bootstrapped))
                    {
                        RTDestNextHopData->pheromone = newPheromone + 
                                                       ((RTDestNextHopData->pheromone - newPheromone)
                                                        * dataPtr->discountBootstrappedPheromone);
                        RTDestNextHopData->lastUpdate = getSimTime(node);
//                        RTDestNextHopData->bootstrappedPheromone = 0; // 19 May 2005 //rely on regular pheromone instead of bootstrapped
                        if (RTDestNextHopData->bootstrappedPheromone == 0)
                            RTDestNextHopData->bootstrappedPheromone = newPheromone;
                        else
                            RTDestNextHopData->bootstrappedPheromone = newPheromone + 
                                             ((RTDestNextHopData->bootstrappedPheromone - newPheromone)
                                              * dataPtr->discountBootstrappedVirtualPheromone);
                    }
                    // Otherwise, bootstrapping is done in the normal way, only to guide proactive ants
                    else
                    {
//                        RTDestNextHopData->bootstrappedPheromone = newPheromone; // 19 May 2005
                        if (RTDestNextHopData->bootstrappedPheromone == 0)
                            RTDestNextHopData->bootstrappedPheromone = newPheromone;
                        else
                            RTDestNextHopData->bootstrappedPheromone = newPheromone + 
                                             ((RTDestNextHopData->bootstrappedPheromone - newPheromone)
                                              * dataPtr->discountBootstrappedVirtualPheromone);
                        RTDestNextHopData->bootstrappedTime = getSimTime(node);
                        RTDestNextHopData->bootstrappedForwarded = FALSE;
                    }
                }
                // The nextHop does not exist
                else //if (ANTHOCNET_nextHopsList(RTDestEntry).numberOfEntries > 0)
                {
                    // Updates for NTNbrData are ok because this function is called after
                    // the update is done for the hello message.
                    // Therefore we don't call UpdateNTTable!!!
                    if (newPheromone > ANTHOCNETBestPheromone(node,&ANTHOCNET_nextHopsList(RTDestEntry)))
                    {
                        newRTDestNextHopEntry = ANTHOCNETCreateRTDestNextHopEntry(node, nextHop, 0, 0, 0,
                                                        getSimTime(node), FROM_HELLO, newPheromone,
                                                        interfaceIndex, interfaceAddress);
                        ANTHOCNETListInsert(&ANTHOCNET_nextHopsList(RTDestEntry), newRTDestNextHopEntry, 
                                                                     RTDestNextHopEntry);
                        ANTHOCNETListSearchFrom(&NTNbrDestEntry, NTNbrDestEntry, helloEntryArray[i].destAddr);
                        newNTNbrDestEntry = ANTHOCNETCreateNTNbrDestEntry(helloEntryArray[i].destAddr, 
                                                                       RTDestEntry);
                        ANTHOCNETListInsert(&((ANTHOCNETNTNbr*)NTNbrEntry->data)->destinations, 
                                            newNTNbrDestEntry, NTNbrDestEntry);
                    }
                }
            }
            // Destination is not present
            else
            {
                newRTDestNextHopEntry = ANTHOCNETCreateRTDestNextHopEntry(node, nextHop, 0, 0, 0,
                                                 getSimTime(node), FROM_HELLO, newPheromone,
                                                 interfaceIndex, interfaceAddress);
                newRTDestEntry = ANTHOCNETCreateRTDestEntry(helloEntryArray[i].destAddr, 0, 0,
                                                 newRTDestNextHopEntry, dataPtr);
                ANTHOCNETListInsert(&dataPtr->routingTable, newRTDestEntry, RTDestEntry);
                RTDestEntry = newRTDestEntry;
                ANTHOCNETListSearchFrom(&NTNbrDestEntry, NTNbrDestEntry, helloEntryArray[i].destAddr);
                newNTNbrDestEntry = ANTHOCNETCreateNTNbrDestEntry(helloEntryArray[i].destAddr, 
                                                               RTDestEntry);
                ANTHOCNETListInsert(&((ANTHOCNETNTNbr*)NTNbrEntry->data)->destinations, 
                                    newNTNbrDestEntry, NTNbrDestEntry);
            }
            if (RTDestEntry->nodeAddress == helloEntryArray[i].destAddr)
                ANTHOCNETUpdateRTDestLastUsed(node, dataPtr, RTDestEntry, helloEntryArray[i].destAddr);
        }
    }
}


static ANTHOCNETForwardAnt* ANTHOCNETUnpackForwardAntPacket(char* antPacket) {
    ANTHOCNETForwardAnt* FAnt = (ANTHOCNETForwardAnt*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETForwardAnt));
    if (FAnt == NULL)
    {
           ERROR_Assert(FALSE, "Cannot allocate ant !!!");
           abort();
    }
    memcpy(FAnt, antPacket, sizeof(ANTHOCNETForwardAnt));
    antPacket += sizeof(ANTHOCNETForwardAnt);
    FAnt->visitedNodes = (ANTHOCNETVisitedNode*)MEM_malloc2(__LINE__,(FAnt->numberOfVisitedNodes + 1) * 
                                           sizeof(ANTHOCNETVisitedNode));
    if (FAnt->visitedNodes == NULL)
    {
           ERROR_Assert(FALSE, "Cannot allocate visited nodes in ant !!!");
           abort();
    }
    memcpy(FAnt->visitedNodes, antPacket, FAnt->numberOfVisitedNodes *
                                           sizeof(ANTHOCNETVisitedNode));
    return FAnt;
}


static void ANTHOCNETUpdateRTBackward(ANTHOCNETBackwardAnt* BAnt,ANTHOCNETData* dataPtr,Node* node,int interfaceIndex,NodeAddress interfaceAddress,BOOL hasPredecessor,NodeAddress predecessorAddress,int predecessorInterfaceIndex,NodeAddress predecessorInterfaceAddress) {
    int i;
    NodeAddress nextHop = BAnt->visitedNodes[BAnt->currentPosition+1].node;
    NodeAddress destination;
    clocktype currentTime = getSimTime(node);
    int numberOfHops = 1;
    ANTHOCNETListEntry *RTDestEntry = ANTHOCNETUpdateRTEntry(dataPtr,nextHop, 
	  BAnt->visitedNodes[BAnt->numberOfVisitedNodes-1].node,
                            currentTime,BAnt->numberOfVisitedNodes - BAnt->currentPosition - 1,
                            node,BAnt->tripTime,BAnt->pheromone,BAnt->generationNumber, 
                            interfaceIndex, interfaceAddress);
}


// FUNCTION	ANTHOCNETUpdateRTDestBoth
// PURPOSE	//Update waitingSince and lastUsed field of the dest
// Parameters:
//	node:
//	dataPtr:
//	RTDestEntry:
//	destAddr:
static void ANTHOCNETUpdateRTDestBoth(Node* node,ANTHOCNETData* dataPtr,ANTHOCNETListEntry* RTDestEntry,NodeAddress destAddr) {
    // Update the waitingSince and the lastUsed field of the dest
    if (RTDestEntry == NULL)
    {
       BOOL found = ANTHOCNETListSearch(&RTDestEntry, &dataPtr->routingTable, destAddr);
       if (!found)
           return;
    }
    ((ANTHOCNETRTDest*)RTDestEntry->data)->waitingSince = 
          ((ANTHOCNETRTDest*)RTDestEntry->data)->lastUsed = 
          getSimTime(node);
}


static void ANTHOCNETUpdateRTDestLastDataTransported(Node* node,ANTHOCNETData* dataPtr,ANTHOCNETListEntry* RTDestEntry,NodeAddress destAddr) {
    // Update the lastDataTransported field of the dest
    if (RTDestEntry == NULL)
    {
       BOOL found = ANTHOCNETListSearch(&RTDestEntry, &dataPtr->routingTable, destAddr);
       if (!found)
           return;
    }
    ((ANTHOCNETRTDest*)RTDestEntry->data)->lastDataTransported = getSimTime(node);
}


static void ANTHOCNETUpdateRTDestLastUsed(Node* node,ANTHOCNETData* dataPtr,ANTHOCNETListEntry* RTDestEntry,NodeAddress destAddr) {
    // Update the lastUsed field of the dest
    if (RTDestEntry == NULL)
    {
       BOOL found = ANTHOCNETListSearch(&RTDestEntry, &dataPtr->routingTable, destAddr);
       if (!found)
           return;
    }
    ((ANTHOCNETRTDest*)RTDestEntry->data)->lastUsed = getSimTime(node);
}


static void ANTHOCNETUpdateRTDestWaitingSince(Node* node,ANTHOCNETData* dataPtr,ANTHOCNETListEntry* RTDestEntry,NodeAddress destAddr) {
    // Update the waitingSince field of the dest
    if (RTDestEntry == NULL)
    {
    BOOL found = ANTHOCNETListSearch(&RTDestEntry, &dataPtr->routingTable, destAddr);
    if (!found)
       return;
    }
    ((ANTHOCNETRTDest*)RTDestEntry->data)->waitingSince = getSimTime(node);
}


static ANTHOCNETListEntry* ANTHOCNETUpdateRTEntry(ANTHOCNETData* dataPtr,NodeAddress nextHop,NodeAddress destination,clocktype currentTime,int numberOfHops,Node* node,float virtualTripTime,Pheromone newPheromone,unsigned short generationNumber,int interfaceIndex,NodeAddress interfaceAddress) {
    // Used for updating when normal pheromone (not bootstrapped) is received
    int found;
    ANTHOCNETList* RT = &(dataPtr->routingTable);
    
    ANTHOCNETRTDest* RTDest;
    ANTHOCNETRTDestNextHop* RTDestNextHop;
    ANTHOCNETListEntry *listEntry, *returnEntry;
    
    found = ANTHOCNETListSearch(&listEntry, RT, destination);
    if (found) // Destination was found
    {
        ANTHOCNETListEntry* RTDestEntry = listEntry;
        returnEntry = listEntry;
        RTDest = (ANTHOCNETRTDest*)(RTDestEntry->data);
        if (RTDest->hopsAvg != 0)
        {
            RTDest->hopsAvg = numberOfHops + 
                              ((RTDest->hopsAvg - numberOfHops) * dataPtr->discountHops);
            RTDest->timeAvg = virtualTripTime + 
                              ((RTDest->timeAvg - virtualTripTime) * dataPtr->discountTripTime);
        }
        else // The case that the dest is there but no info has ever been gathered for it.
        {
            RTDest->hopsAvg = numberOfHops;
            RTDest->timeAvg = virtualTripTime;
        }
        found = ANTHOCNETListSearch(&listEntry, &RTDest->nextHops, nextHop);
        if (RTDest->numberOfConfirmedEntries <= 0) // there was no next hop for this destination
        {
            ANTHOCNETListEntry* bufferDestEntry;
            if (ANTHOCNETListSearch(&bufferDestEntry, &dataPtr->messageBuffer, destination))
                ANTHOCNETMessageDebufferDestWithSend(node, &dataPtr->messageBuffer, 
                                                  bufferDestEntry, interfaceIndex, interfaceAddress);
            if (ANTHOCNETListSearch(&bufferDestEntry, &dataPtr->relayMessageBuffer, destination))
                ANTHOCNETMessageDebufferDestWithSend(node, &dataPtr->relayMessageBuffer, 
                                                  bufferDestEntry, interfaceIndex, interfaceAddress);     
        }
        if (found) // Next Hop was found
        {
            ANTHOCNETListEntry* NTNbrEntry;
            ANTHOCNETListSearch(&NTNbrEntry, &(dataPtr->neighbourTable), nextHop);
            ((ANTHOCNETNTNbr*)NTNbrEntry->data)->lastUpdate = currentTime;
            RTDestNextHop = (ANTHOCNETRTDestNextHop*)(listEntry->data);
            // bootstrapping: entry exists based on bootstrapped information:
            if (RTDestNextHop->pheromone == 0)
            {
                if (generationNumber != FROM_HELLO)
                    RTDestNextHop->lastGeneration = generationNumber;
                RTDestNextHop->lastGenerationPheromone = newPheromone;
                RTDestNextHop->pheromone = newPheromone;
                RTDestNextHop->hopsAvg = numberOfHops;
                RTDestNextHop->timeAvg = virtualTripTime;
                RTDestNextHop->lastUpdate = currentTime;  
                
                RTDest->numberOfConfirmedEntries++;
            }
            // end bootstrapping
            else
            {
                if ((generationNumber != RTDestNextHop->lastGeneration) || 
                    (newPheromone > RTDestNextHop->lastGenerationPheromone) || 
                    (generationNumber == FROM_HELLO))
                {
                    if (generationNumber != FROM_HELLO)
                        RTDestNextHop->lastGeneration = generationNumber;
                    RTDestNextHop->lastGenerationPheromone = newPheromone;
                    RTDestNextHop->pheromone = newPheromone + 
						                       ((RTDestNextHop->pheromone - newPheromone)
                                                * dataPtr->discountPheromone);
                }
                RTDestNextHop->hopsAvg = numberOfHops + 
                            ((RTDestNextHop->hopsAvg - numberOfHops) * dataPtr->discountHops);
                RTDestNextHop->timeAvg = virtualTripTime + 
                            ((RTDestNextHop->timeAvg - virtualTripTime) * dataPtr->discountTripTime);
                RTDestNextHop->lastUpdate = currentTime;
            }
        }
        else // Next hop was not found
        {
            ANTHOCNETListEntry* insertAfter = listEntry;
            listEntry = ANTHOCNETCreateRTDestNextHopEntry(node,nextHop, numberOfHops, virtualTripTime,
                           newPheromone,currentTime,generationNumber,0,interfaceIndex,interfaceAddress);
            RTDest->numberOfConfirmedEntries++;
            
            ANTHOCNETListInsert(&RTDest->nextHops, listEntry, insertAfter);
            ANTHOCNETUpdateNTTable(dataPtr, nextHop, destination, currentTime, 
                      RTDestEntry, interfaceIndex, interfaceAddress);
        }
    }
    else // Destination was not found
    {
        ANTHOCNETListEntry* bufferDestEntry;
        ANTHOCNETListEntry* insertAfter = listEntry;
        ANTHOCNETListEntry* nextHopEntry = ANTHOCNETCreateRTDestNextHopEntry(node, nextHop, 
                    numberOfHops, virtualTripTime, newPheromone, currentTime,generationNumber,
                    0, interfaceIndex, interfaceAddress);
        listEntry = ANTHOCNETCreateRTDestEntry(destination, numberOfHops, virtualTripTime,
                                            nextHopEntry,dataPtr);
        ((ANTHOCNETRTDest*)listEntry->data)->numberOfConfirmedEntries++;
        
        returnEntry = listEntry;
        ANTHOCNETListInsert(RT, listEntry,insertAfter);
        ANTHOCNETUpdateNTTable(dataPtr,nextHop,destination,currentTime,listEntry,interfaceIndex,interfaceAddress);
        if (ANTHOCNETListSearch(&bufferDestEntry, &dataPtr->messageBuffer, destination))
            ANTHOCNETMessageDebufferDestWithSend(node, &dataPtr->messageBuffer, 
                                              bufferDestEntry, interfaceIndex, interfaceAddress);
        if (ANTHOCNETListSearch(&bufferDestEntry, &dataPtr->relayMessageBuffer, destination))
            ANTHOCNETMessageDebufferDestWithSend(node, &dataPtr->relayMessageBuffer, 
                                              bufferDestEntry, interfaceIndex, interfaceAddress);
    }
    return returnEntry;
}


static void ANTHOCNETUpdateRTForward(ANTHOCNETBackwardAnt* BAnt,ANTHOCNETData* dataPtr,Node* node) {
    NodeAddress nextHop = BAnt->visitedNodes[BAnt->currentPosition-1].node;
    clocktype currentTime = getSimTime(node);
    int numberOfHops = 1;
    ANTHOCNETListEntry *NTNbrEntry;
    BOOL found = ANTHOCNETListSearch(&NTNbrEntry, &dataPtr->neighbourTable, nextHop);

    if (found)
    {
        ANTHOCNETNTNbr *NTNbrData = (ANTHOCNETNTNbr*)NTNbrEntry->data;
        int interfaceIndex = NTNbrData->interfaceIndex;
        NodeAddress interfaceAddress = NTNbrData->interfaceAddress;
        double tripTime =  BAnt->currentPosition *
                                          ANTHOCNETGetVirtualSendTime(node,interfaceIndex,dataPtr->defaultSize); // maybe a bad idea...
        Pheromone pheromone = 0;
        int i;
        for (i = 0;i < BAnt->currentPosition;i++)
        {
            pheromone = ANTHOCNETCalculatePheromoneOneHop(node, dataPtr, pheromone, 
                                                    dataPtr->SINRTreshold + 1, // hack, so that SINR value always higher than minimum
                                                    interfaceIndex, dataPtr->defaultSize,
                                                    BAnt->currentPosition, tripTime);
        }
        ANTHOCNETListEntry *RTDestEntry = ANTHOCNETUpdateRTEntry(dataPtr,nextHop, BAnt->sourceAddress,currentTime, BAnt->currentPosition,node, 
                                                                 tripTime, pheromone,BAnt->generationNumber, interfaceIndex, interfaceAddress);
    }
}


static void ANTHOCNETUpdateNTTable(ANTHOCNETData* dataPtr,NodeAddress neighbour,NodeAddress destination,clocktype currentTime,ANTHOCNETListEntry* RTDestEntry,int interfaceIndex,NodeAddress interfaceAddress) {
    int found;
    int updated;
    ANTHOCNETList* NT = &(dataPtr->neighbourTable);
    
    ANTHOCNETNTNbr* NTNbr;
    ANTHOCNETNTNbrDest* NTNbrDest;
    ANTHOCNETListEntry* listEntry;
    
    found = ANTHOCNETListSearch(&listEntry, NT, neighbour);
    if (found) // Neighbour was found
    {
        NTNbr = (ANTHOCNETNTNbr*)(listEntry->data);
        NTNbr->lastUpdate = currentTime;
        found = ANTHOCNETListSearch(&listEntry, &NTNbr->destinations, destination);
        if (!found) // Destination was not found
        {
            ANTHOCNETListEntry* insertAfter = listEntry;
            listEntry = ANTHOCNETCreateNTNbrDestEntry(destination, RTDestEntry);
            ANTHOCNETListInsert(&NTNbr->destinations, listEntry, insertAfter);
        }
    }
    else // Neighbour was not found
    {
        ANTHOCNETListEntry* insertAfter = listEntry;
        ANTHOCNETListEntry* destinationEntry = 
                         ANTHOCNETCreateNTNbrDestEntry(destination, RTDestEntry);
        listEntry = ANTHOCNETCreateNTNbrEntry(neighbour, currentTime, destinationEntry, 
			interfaceIndex, interfaceAddress,
			dataPtr->currentSINR,dataPtr->currentPower);
        ANTHOCNETListInsert(NT, listEntry,insertAfter);
    }
}


static void ANTHOCNETPrintLinkFailureAntData(Node* node,NodeAddress sourceAddr,short numberOfElements,ANTHOCNETLostPheromone* lostPheromoneArray) {
    int i;
    fprintf(stderr,"[%g] %u receives Link Failure Ant from: %u\n",
             simTimeSec(getSimTime(node)),node->nodeId,
             MAPPING_GetNodeIdFromInterfaceAddress(node,sourceAddr));
    for (i=0;i<numberOfElements;i++)
    {
        fprintf(stderr,"%u (%g,%g) \t",
                 MAPPING_GetNodeIdFromInterfaceAddress(node,lostPheromoneArray[i].destAddr),
                 lostPheromoneArray[i].newBestHopsToGo, lostPheromoneArray[i].newBestTimeToGo);
    }
    fprintf(stderr,"\n");
}


static void* MEM_malloc2(int line,size_t size) {
    void *ptr = malloc(size);

    if (ptr == NULL) {
        fprintf(stderr, "Ran out of Memory at Line %d (size %d)\n", line, size);
        exit(-1);
    }

    return ptr;
}


// FUNCTION	ANTHOCNETPrintPheromone
// PURPOSE	//Print the pheromone values for a certain destination
// Parameters:
//	node:
//	dataPtr:
//	destAddr:
//	antType:
static void ANTHOCNETPrintPheromone(Node* node,ANTHOCNETData* dataPtr,NodeAddress destAddr,ANTHOCNETAntType antType) {
BOOL found;
    ANTHOCNETListEntry *RTDestEntry, *RTDestNextHopEntry;
    ANTHOCNETRTDest *RTDestData;
    ANTHOCNETRTDestNextHop *RTDestNextHopData;
    found = ANTHOCNETListSearch(&RTDestEntry, &dataPtr->routingTable, destAddr);
    if (found)
    {
        fprintf(stderr,"[%g] Node %u Dest %u AntType %d\n", simTimeSec(getSimTime(node)),
                node->nodeId, MAPPING_GetNodeIdFromInterfaceAddress(node,destAddr), antType);
        RTDestData = (ANTHOCNETRTDest*)RTDestEntry->data;
        fprintf(stderr,"numberOfEntries: %d numberOfConfirmedEntries: %d\n",
                    RTDestData->nextHops.numberOfEntries, RTDestData->numberOfConfirmedEntries);
        for (RTDestNextHopEntry = RTDestData->nextHops.head.next;RTDestNextHopEntry != NULL;
                        RTDestNextHopEntry = RTDestNextHopEntry->next)
        {
            RTDestNextHopData = (ANTHOCNETRTDestNextHop*)RTDestNextHopEntry->data;
            fprintf(stderr,"NextHop %u Pheromone %g bootstrappedPheromone %g - ",
                    MAPPING_GetNodeIdFromInterfaceAddress(node,RTDestNextHopEntry->nodeAddress),
                    RTDestNextHopData->pheromone, RTDestNextHopData->bootstrappedPheromone);
        }
        fprintf(stderr,"\n");
    }
}


static void ANTHOCNETBroadcastControlMessage(Node* node,Message* msg,TosType priority,clocktype delay,unsigned ttl) {
    NetworkDataIp* ip = (NetworkDataIp *) node->networkData.networkVar;
    int i;
    for (i = 0; i < node->numberInterfaces; i++)
    {
        IpInterfaceInfoType* intfInfo = ip->interfaceInfo[i];
        if (intfInfo->routingProtocolType
            == ROUTING_PROTOCOL_ANTHOCNET)
        {
            if (delay > 0)
            {
                NetworkIpSendRawMessageToMacLayerWithDelay(
                    node,
                    MESSAGE_Duplicate(node, msg),
                    intfInfo->ipAddress,//NetworkIpGetInterfaceAddress(node, DEFAULT_INTERFACE),
                    ANY_DEST,//NetworkIpGetInterfaceBroadcastAddress(node,i),
                    priority,
                    IPPROTO_ANTHOCNET,
                    ttl,
                    i,
                    ANY_DEST,//NetworkIpGetInterfaceBroadcastAddress(node,i),
                    delay);//(RANDOM_nrand(dataPtr->seed) % (delay)));//
            }
            else
            {
                NetworkIpSendRawMessageToMacLayer(
                    node,
                    MESSAGE_Duplicate(node, msg),
                    intfInfo->ipAddress,//NetworkIpGetInterfaceAddress(node, DEFAULT_INTERFACE),
                    ANY_DEST,//NetworkIpGetInterfaceBroadcastAddress(node,i),//
                    priority,
                    IPPROTO_ANTHOCNET,
                    ttl,
                    i,
                    ANY_DEST);//NetworkIpGetInterfaceBroadcastAddress(node,i));//
            }                    
        }
    }
    MESSAGE_Free(node,msg);
}


static clocktype Mac802_11GetAvgSendTimeANTHOCNET(Node* node,int interfaceIndex) {
   if (node->macData[interfaceIndex]->macProtocol == MAC_PROTOCOL_802_11)
   {
       return Mac802_11GetAvgSendTimeANTS01(node,interfaceIndex);
   }
   else
   {
       return (clocktype)(ANTHOCNET_HOP_LATENCY);
   }
}


void ANTHOCNET_IPQueueTimeRegistration(Node* node,int interfaceIndex,clocktype qTime) {
  ANTHOCNETData *dataPtr = (ANTHOCNETData*)NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_ANTHOCNET);
  ANTHOCNETUpdateAvgQueuingTime(node,dataPtr,interfaceIndex,qTime);
}


static void ANTHOCNETUpdateAvgQueuingTime(Node* node,ANTHOCNETData* dataPtr,int interfaceIndex,clocktype qTime) {
  dataPtr->avgQTimeDbl[interfaceIndex] *= dataPtr->discountQTime;
  dataPtr->avgQTimeDbl[interfaceIndex] += (1 - dataPtr->discountQTime) * qTime;
}


void ANTHOCNET_Mac802_11_UpdateSignalData(Node* node,NodeAddress nbrAddress,double currentPower,double currentSensitivity,double currentInterference,double currentNoise,double currentSINR) {
  ANTHOCNETData *dataPtr = (ANTHOCNETData*)NetworkIpGetRoutingProtocol(node,
                                                          ROUTING_PROTOCOL_ANTHOCNET);
  BOOL found = FALSE;
  ANTHOCNETListEntry *NTNbrEntry;
  dataPtr->sensitivity = currentSensitivity;
  dataPtr->currentSINR = currentSINR;
  dataPtr->currentPower = currentPower;
  found = ANTHOCNETListSearch(&NTNbrEntry,&dataPtr->neighbourTable,
      MAPPING_GetDefaultInterfaceAddressFromNodeId(node,
                                                  MAPPING_GetNodeIdFromInterfaceAddress(node,nbrAddress)));
  if (found)
  {
      ANTHOCNETNTNbr *NTNbrData = (ANTHOCNETNTNbr*)NTNbrEntry->data;
      NTNbrData->SINR = currentSINR;
      NTNbrData->power = currentPower;
  }
}


static double ANTHOCNETGetBusyRatio(Node* node,int interfaceIndex) {
  if (node->macData[interfaceIndex]->macProtocol == MAC_PROTOCOL_802_11)
  {
      MacData802_11 *M802 = (MacData802_11*)node->macData[interfaceIndex]->macVar;
      PhyData *thisPhy = node->phyData[M802->myMacData->phyNumber];
      return ((PhyData802_11 *)thisPhy->phyVar)->avgBusy;
  }
  else
  {
      return 0;
  }
}


static Pheromone ANTHOCNETCalculatePheromoneOneHopWithLabelsAndTime(Node* node,ANTHOCNETData* dataPtr,Pheromone pheromone,float SINR,int interfaceIndex,int size) {
  double costOneHop = dataPtr->timeHopsWeight * ANTHOCNETGetVirtualSendTime(node, interfaceIndex, size) + 
                  (1 - dataPtr->timeHopsWeight) * simTimeSec(ANTHOCNET_HOP_LATENCY);
  if (dataPtr->nodeRole == 'f')
  {
      costOneHop *= dataPtr->fixedNodePheromoneFactor;
  }
  if (pheromone > 0) 
  {
      return (1 / (costOneHop + (1 / pheromone)));
  } 
  else
  {
      return (1 / costOneHop);
  }
}


static Pheromone ANTHOCNETCalculatePheromoneOneHopWithPheromoneAversion(Node* node,ANTHOCNETData* dataPtr,Pheromone pheromone,float SINR,int interfaceIndex,int size) {
  double costOneHop = 1;
  if (SINR < dataPtr->SINRTreshold)
  {
      costOneHop *= dataPtr->SINRPheromoneFactor;
  }

  //for loop through the routing table...
  fprintf(stderr, "Warning: this function is not working yet...");

  if (pheromone > 0) 
  {
      return (1 / (costOneHop + (1 / pheromone)));
  } 
  else
  {
      return (1 / costOneHop);
  }
}


static Pheromone ANTHOCNETCalculatePheromoneOneHopWithSpeed(Node* node,ANTHOCNETData* dataPtr,Pheromone pheromone,float SINR,int interfaceIndex,int size) {
  double costOneHop;
  double speed;
  MOBILITY_ReturnInstantaneousSpeed(node, &speed);
  costOneHop  = simTimeSec(ANTHOCNET_HOP_LATENCY);
  if (speed > 0)
  {
      costOneHop *= dataPtr->speedPenalty;
  }
  if (SINR < dataPtr->SINRTreshold)
  {
      costOneHop *= dataPtr->SINRPheromoneFactor;
  }
  if (pheromone > 0) 
  {
      return (1 / (costOneHop + (1 / pheromone)));
  } 
  else
  {
      return (1 / costOneHop);
  }
}


static int ANTHOCNETRejectedSession(Node* node,ANTHOCNETData* dataPtr,NodeAddress destAddr) {
  ANTHOCNETListEntry *connEntry;
  BOOL found = ANTHOCNETListSearch(&connEntry, &dataPtr->currentConnectionsInfo, destAddr);
  if (found)
  {
      return ((ANTHOCNETCurrentConnectionsData*)connEntry->data)->rejected;
  }
  else
  {
      return FALSE;
  }
}


static int ANTHOCNETRejectSession(Node* node,ANTHOCNETData* dataPtr,double hops,double time) {
  return ((hops > dataPtr->admissionControlHops) || (time > dataPtr->admissionControlTime));
}



//UserCodeEnd

//State Function Definitions
void AnthocnetHandleProtocolPacket(Node* node,Message* msg, NodeAddress sourceAddr,
                                   NodeAddress destAddr, int interfaceIndex, unsigned ttl) {
    //UserCodeStartProcessReceivedAnts
    ANTHOCNETData *dataPtr = (ANTHOCNETData*)NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_ANTHOCNET);
    char* antPacket = MESSAGE_ReturnPacket(msg);   
    ANTHOCNETAntType* antType = (ANTHOCNETAntType*)antPacket;
    BOOL reactiveForwardAnt = FALSE, 
         reactiveBackwardAnt = FALSE, 
         linkFailureAnt = FALSE,
         proactiveForwardAnt = FALSE,
         repairForwardAnt = FALSE,
         repairBackwardAnt = FALSE,
         unicastWarning = FALSE,
         hello = FALSE;

    TRACE(stderr,"[%g] Process received ANT at Node %u\n", simTimeSec(getSimTime(node)),
                 node->nodeId);
                        
    dataPtr->currentInterfaceIndex = interfaceIndex;
    dataPtr->currentInterfaceAddress = sourceAddr;
    dataPtr->currentDefaultAddress = MAPPING_GetDefaultInterfaceAddressFromNodeId(node,
                                                MAPPING_GetNodeIdFromInterfaceAddress(node,sourceAddr));

    switch(*antType)
    {
      case ANTHOCNET_REACTIVE_FORWARD:
        reactiveForwardAnt = TRUE;
        break;
      case ANTHOCNET_REACTIVE_BACKWARD:
        reactiveBackwardAnt = TRUE;
        break;
     case ANTHOCNET_LINK_FAILURE_NOTIFICATION:
        linkFailureAnt = TRUE;
        break;   
     case ANTHOCNET_PROACTIVE_FORWARD:
         proactiveForwardAnt = TRUE;
         break;
     case ANTHOCNET_REPAIR_FORWARD:
         repairForwardAnt = TRUE;
         break;
     case ANTHOCNET_REPAIR_BACKWARD:
         repairBackwardAnt = TRUE;
         break;
     case ANTHOCNET_UNICAST_WARNING:
         unicastWarning = TRUE;
         break;
     case ANTHOCNET_HELLO:
         hello = TRUE;
         break;
     default:
      ERROR_Assert(FALSE, "AntType does not exist");
      abort();
    }

    //UserCodeEndProcessReceivedAnts
    if (reactiveForwardAnt) {
        if (reactiveBackwardAnt ||
            linkFailureAnt ||
            proactiveForwardAnt ||
            repairBackwardAnt ||
            repairForwardAnt ||
            unicastWarning ||
            hello ) {
            ERROR_ReportError("Multiple true guards found in transitions from same state.");
        }
        dataPtr->state = ANTHOCNET_PROCESS_REACTIVE_FORWARD_ANT;
        enterAnthocnetProcessReactiveForwardAnt(node, dataPtr, msg);
        return;
    } else if (reactiveBackwardAnt) {
        if (reactiveForwardAnt ||
            linkFailureAnt ||
            proactiveForwardAnt ||
            repairBackwardAnt ||
            repairForwardAnt ||
            unicastWarning ||
            hello ) {
            ERROR_ReportError("Multiple true guards found in transitions from same state.");
        }
        dataPtr->state = ANTHOCNET_PROCESS_REACTIVE_BACKWARD_ANT;
        enterAnthocnetProcessReactiveBackwardAnt(node, dataPtr, msg);
        return;
    } else if (linkFailureAnt) {
        if (reactiveForwardAnt ||
            reactiveBackwardAnt ||
            proactiveForwardAnt ||
            repairBackwardAnt ||
            repairForwardAnt ||
            unicastWarning ||
            hello ) {
            ERROR_ReportError("Multiple true guards found in transitions from same state.");
        }
        dataPtr->state = ANTHOCNET_PROCESS_LINK_FAILURE_NOTIFICATION;
        enterAnthocnetProcessLinkFailureNotification(node, dataPtr, msg);
        return;
    } else if (proactiveForwardAnt) {
        if (reactiveForwardAnt ||
            reactiveBackwardAnt ||
            linkFailureAnt ||
            repairBackwardAnt ||
            repairForwardAnt ||
            unicastWarning ||
            hello ) {
            ERROR_ReportError("Multiple true guards found in transitions from same state.");
        }
        dataPtr->state = ANTHOCNET_PROCESS_PROACTIVE_FORWARD_ANT;
        enterAnthocnetProcessProactiveForwardAnt(node, dataPtr, msg);
        return;
    } else if (repairBackwardAnt) {
        if (reactiveForwardAnt ||
            reactiveBackwardAnt ||
            linkFailureAnt ||
            proactiveForwardAnt ||
            repairForwardAnt ||
            unicastWarning ||
            hello ) {
            ERROR_ReportError("Multiple true guards found in transitions from same state.");
        }
        dataPtr->state = ANTHOCNET_PROCESS_REPAIR_BACKWARD_ANT;
        enterAnthocnetProcessRepairBackwardAnt(node, dataPtr, msg);
        return;
    } else if (repairForwardAnt) {
        if (reactiveForwardAnt ||
            reactiveBackwardAnt ||
            linkFailureAnt ||
            proactiveForwardAnt ||
            repairBackwardAnt ||
            unicastWarning ||
            hello ) {
            ERROR_ReportError("Multiple true guards found in transitions from same state.");
        }
        dataPtr->state = ANTHOCNET_PROCESS_REPAIR_FORWARD_ANT;
        enterAnthocnetProcessRepairForwardAnt(node, dataPtr, msg);
        return;
    } else if (unicastWarning) {
        if (reactiveForwardAnt ||
            reactiveBackwardAnt ||
            linkFailureAnt ||
            proactiveForwardAnt ||
            repairBackwardAnt ||
            repairForwardAnt ||
            hello ) {
            ERROR_ReportError("Multiple true guards found in transitions from same state.");
        }
        dataPtr->state = ANTHOCNET_PROCESS_UNICAST_WARNING;
        enterAnthocnetProcessUnicastWarning(node, dataPtr, msg);
        return;
    } else if (hello) {
        if (reactiveForwardAnt ||
            reactiveBackwardAnt ||
            linkFailureAnt ||
            proactiveForwardAnt ||
            repairBackwardAnt ||
            repairForwardAnt ||
            unicastWarning ) {
            ERROR_ReportError("Multiple true guards found in transitions from same state.");
        }
        dataPtr->state = ANTHOCNET_PROCESS_HELLO;
        enterAnthocnetProcessHello(node, dataPtr, msg);
        return;
    } else {
        dataPtr->state = ANTHOCNET_IDLE;
        enterAnthocnetIdle(node, dataPtr, msg);
        return;
    }
}


void AnthocnetRouterFunction(Node* node, Message* msg, NodeAddress destAddr,
                             NodeAddress previousHopAddr, BOOL* packetWasRouted) {
        //UserCodeStartRouteDataPackets
        NodeAddress nextHop = ANY_DEST;  /* Broadcast message*/
        ANTHOCNETListEntry* nextHopEntry = NULL;
        BOOL sendAnts = FALSE;
        static Node *node1 = MAPPING_GetNodePtrFromHash(node->partitionData->nodeIdHash, 1);  

        ANTHOCNETData *dataPtr = (ANTHOCNETData*)NetworkIpGetRoutingProtocol(
                            node,
                            ROUTING_PROTOCOL_ANTHOCNET);
        IpHeaderType *ipHeader1 = (IpHeaderType *) msg->packet;
        NodeAddress srcAddr = ipHeader1->ip_src;
        NodeAddress previousHopAddrDefault =  
                MAPPING_GetDefaultInterfaceAddressFromNodeId(node,
                                       MAPPING_GetNodeIdFromInterfaceAddress(node,previousHopAddr));
/*        if (ipHeader1->ip_p != IPPROTO_ANTHOCNET) the received packet is normally a data packet*/

        /* Inspect if the packet needs to be routed (necessary for ip.DeliverPacket(...)) */
        if (NetworkIpIsMyIP(node, destAddr) == TRUE)
        {
            *packetWasRouted = FALSE;      
            if (destAddr != ANY_IP) {
                IpHeaderType *ipHeader = (IpHeaderType *) msg->packet;
                if (ipHeader->ip_p != IPPROTO_ANTHOCNET)
                {
// for killing duplicate packets at destination
                    if (ANTHOCNETPacketInLoop(node,dataPtr,msg,previousHopAddr))
                    {
                        msg->packetCreationTime = getSimTime(node) + 1;
                        dataPtr->stats.dropDataPacketsLoop++;
                    }
                    else
                    {
                        dataPtr->msgIds[dataPtr->msgIdsIndex].srcId = msg->originatingNodeId;
                        dataPtr->msgIds[dataPtr->msgIdsIndex].seqNumber = msg->sequenceNumber;
                        if (++dataPtr->msgIdsIndex == ANTHOCNET_NUMBER_OF_MSG_IDS)
                            dataPtr->msgIdsIndex = 0;
// end of taking care of duplicate packets at destination
                        dataPtr->stats.totalPacketsDelivered++;
                        dataPtr->stats.totalHopsDelivered += IPDEFTTL - ipHeader->ip_ttl;
                        ANTHOCNETUpdateRTDestLastUsed(node, dataPtr, NULL, destAddr);
                    }
                    if (PRINT_DELAY_DETAILS)
                    {
                        IpHeaderType *ipHeader1 = (IpHeaderType *)msg->packet;
                        fprintf(stderr,"[%g] node %u packet (%u,%u,%g,%d): arrival reception\n",
                                simTimeSec(getSimTime(node)),node->nodeId,
                                MAPPING_GetNodeIdFromInterfaceAddress(node,ipHeader1->ip_src),
                                MAPPING_GetNodeIdFromInterfaceAddress(node,ipHeader1->ip_dst),
                                simTimeSec(msg->packetCreationTime),
                                msg->sequenceNumber);
                    }
                }
            }
        }
        else if ((ipHeader1->ip_p != IPPROTO_ANTHOCNET) &&
                 ((IPDEFTTL - ipHeader1->ip_ttl) >= dataPtr->maxTtl))
        {
            MESSAGE_Free(node,msg);
            dataPtr->stats.dropDataPacketsMaxTtl++;
            *packetWasRouted = TRUE;
        }
        else if (ANTHOCNETPacketInLoop(node,dataPtr,msg,previousHopAddr))
        {
            MESSAGE_Free(node,msg);
            dataPtr->stats.dropDataPacketsLoop++;
            *packetWasRouted = TRUE;
        }
        else if ((dataPtr->admissionControlFlag) && (previousHopAddr == ANY_IP) && (ANTHOCNETRejectedSession(node, dataPtr, destAddr)))
        {
            MESSAGE_Free(node, msg);
            *packetWasRouted = TRUE;
        }
        else
        {
            dataPtr->msgIds[dataPtr->msgIdsIndex].srcId = msg->originatingNodeId;
            dataPtr->msgIds[dataPtr->msgIdsIndex].seqNumber = msg->sequenceNumber;
            if (++dataPtr->msgIdsIndex == ANTHOCNET_NUMBER_OF_MSG_IDS)
                dataPtr->msgIdsIndex = 0;
            ANTHOCNETUpdateRTDestLastUsed(node, dataPtr, NULL, destAddr);
            nextHopEntry = ANTHOCNETGetNextHopData(node, destAddr, previousHopAddrDefault, 
                                    &dataPtr->routingTable,dataPtr->dataRoutingExponent);
                                    
            if (PRINT_DELAY_DETAILS)
            {
                IpHeaderType *ipHeader1 = (IpHeaderType *)msg->packet;
                if (previousHopAddr == ANY_IP)
                {
                    fprintf(stderr,"[%g] node %u packet (%u,%u,%g,%d): start\n",
                            simTimeSec(getSimTime(node)),node->nodeId,
                            MAPPING_GetNodeIdFromInterfaceAddress(node,ipHeader1->ip_src),
                            MAPPING_GetNodeIdFromInterfaceAddress(node,ipHeader1->ip_dst),
                            simTimeSec(msg->packetCreationTime),
                            msg->sequenceNumber);
                }
                else
                {
                    fprintf(stderr,"[%g] node %u packet (%u,%u,%g,%d): reception\n",
                            simTimeSec(getSimTime(node)),node->nodeId,
                            MAPPING_GetNodeIdFromInterfaceAddress(node,ipHeader1->ip_src),
                            MAPPING_GetNodeIdFromInterfaceAddress(node,ipHeader1->ip_dst),
                            simTimeSec(msg->packetCreationTime),
                            msg->sequenceNumber);
                }
            }
            if (nextHopEntry != NULL)
            {
                BOOL sendMessage = TRUE;
                ANTHOCNETUpdateRTDestLastDataTransported(node,dataPtr,NULL,destAddr);
                if (previousHopAddr == ANY_IP) // I am the source of this packet, so I check for proactive forward ants
                {
                    ANTHOCNETListEntry *connEntry;
                    BOOL found = ANTHOCNETListSearch(&connEntry, &dataPtr->currentConnectionsInfo, destAddr);
                    if (found)
                    {
                        ANTHOCNETCurrentConnectionsData *ccdata = (ANTHOCNETCurrentConnectionsData*)connEntry->data;
                        (ccdata->counter)++;
                    }
                    else
                    {
                        ANTHOCNETRTDestNextHop *RTDestNextHopData = (ANTHOCNETRTDestNextHop*)nextHopEntry->data;
                        BOOL rejectedSession = ANTHOCNETRejectSession(node, dataPtr, RTDestNextHopData->hopsAvg, RTDestNextHopData->timeAvg);
                        ANTHOCNETListEntry *newEntry = (ANTHOCNETListEntry*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETListEntry));
                        newEntry->nodeAddress = destAddr;
                        // newEntry->data = (int*)MEM_malloc2(__LINE__,sizeof(int));
                        newEntry->data = (ANTHOCNETCurrentConnectionsData*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETCurrentConnectionsData));
                        // *((int*)newEntry->data) = 1;
                        ((ANTHOCNETCurrentConnectionsData*)newEntry->data)->counter = 1;
                        ((ANTHOCNETCurrentConnectionsData*)newEntry->data)->rejected = rejectedSession;
                        ANTHOCNETListInsert(&dataPtr->currentConnectionsInfo, newEntry, connEntry);
                        if ((rejectedSession) && (dataPtr->admissionControlFlag))
                        {
                            MESSAGE_Free(node, msg);
                            sendMessage = FALSE;
                        }
                    }
                }
                else // I am not the source
                {
                }

                if (ANTHOCNET_PRINT_HOP_CHOSEN) 
                {
                    ANTHOCNETPrintPheromone(node, dataPtr, destAddr, ANTHOCNET_UNKNOWN);
                    fprintf(stderr,"Data Packet - Chosen hop: %u\n",
                             MAPPING_GetNodeIdFromInterfaceAddress(node,nextHopEntry->nodeAddress));
                }
              
                if (PRINT_DELAY_DETAILS)
                {
                    IpHeaderType *ipHeader1 = (IpHeaderType *)msg->packet;
                    fprintf(stderr,"[%g] node %u packet (%u,%u,%g,%d): send from routing layer to IP\n",
                            simTimeSec(getSimTime(node)),node->nodeId,
                            MAPPING_GetNodeIdFromInterfaceAddress(node,ipHeader1->ip_src),
                            MAPPING_GetNodeIdFromInterfaceAddress(node,ipHeader1->ip_dst),
                            simTimeSec(msg->packetCreationTime),
                            msg->sequenceNumber);
                }

                if (sendMessage)
                {
                    NetworkIpSendPacketToMacLayer(node,msg,
                                 ((ANTHOCNETRTDestNextHop*)nextHopEntry->data)->interfaceIndex,
                                 ((ANTHOCNETRTDestNextHop*)nextHopEntry->data)->interfaceAddress);
      
                    TRACE(stderr,"[%g]Sending DATA packet coming from %u: %u -> %u (seq %d) ...\n",
                        simTimeSec(getSimTime(node)), msg->originatingNodeId, node->nodeId,
                        MAPPING_GetNodeIdFromInterfaceAddress(node,nextHopEntry->nodeAddress),
                        msg->sequenceNumber);
                }
            }
            else // I don't have a next hop for this destination
            {
                if (previousHopAddr == ANY_IP) // I am the source, so I might have to start looking for the destination
                {
                    // buffer data and send out reactive forward ants (sendAnts == true) or just wait (sendAnts == false)
                    if (PRINT_DELAY_DETAILS)
                    {
                        IpHeaderType *ipHeader1 = (IpHeaderType *)msg->packet;
                        fprintf(stderr,"[%g] node %u packet (%u,%u,%g,%d): buffer at routing layer\n",
                                simTimeSec(getSimTime(node)),node->nodeId,
                                MAPPING_GetNodeIdFromInterfaceAddress(node,ipHeader1->ip_src),
                                MAPPING_GetNodeIdFromInterfaceAddress(node,ipHeader1->ip_dst),
                                simTimeSec(msg->packetCreationTime),
                                msg->sequenceNumber);
                    }
                    sendAnts = ANTHOCNETInsertAndCheckBuffers(node, dataPtr, msg, destAddr, 
                                            previousHopAddr);
                }
                else // I am not the source
                {
                    ANTHOCNETListEntry* bufferDestEntry = NULL;
                    BOOL found = ANTHOCNETListSearch(&bufferDestEntry, &dataPtr->messageBuffer, destAddr) ||
                                    ANTHOCNETListSearch(&bufferDestEntry, &dataPtr->relayMessageBuffer, destAddr);
                    if (found) // I am waiting for ants searching this destination
                    {
                        if (PRINT_DELAY_DETAILS)
                        {
                            IpHeaderType *ipHeader1 = (IpHeaderType *)msg->packet;
                            fprintf(stderr,"[%g] node %u packet (%u,%u,%g,%d): buffer at routing layer\n",
                                    simTimeSec(getSimTime(node)),node->nodeId,
                                    MAPPING_GetNodeIdFromInterfaceAddress(node,ipHeader1->ip_src),
                                    MAPPING_GetNodeIdFromInterfaceAddress(node,ipHeader1->ip_dst),
                                    simTimeSec(msg->packetCreationTime),
                                    msg->sequenceNumber);
                        }
                        ANTHOCNETInsertAndCheckBuffers(node,dataPtr,msg,destAddr,previousHopAddr);
                    }
                    else // no ants are sent out, I delete the packet + send unicast warning
                    {
                        ANTHOCNETListEntry *previousHopEntry;
                        BOOL found = ANTHOCNETListSearch(&previousHopEntry,&dataPtr->neighbourTable,
                                                        previousHopAddrDefault);
                        ANTHOCNETNTNbr *NTNbrData = (ANTHOCNETNTNbr*)previousHopEntry->data;
                        PacketBuffer* packetBuffer = NULL;
                        Message* msgNew = NULL;
                        ANTHOCNETUnicastWarning* unicastWarning = (ANTHOCNETUnicastWarning*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETUnicastWarning));
                        unicastWarning->nextHopAddr = dataPtr->myAddress;
                        unicastWarning->destAddr = destAddr;
                        unicastWarning->srcAddr = srcAddr;
                        packetBuffer = ANTHOCNETBuildUnicastWarning(ANTHOCNET_UNICAST_WARNING, unicastWarning);
                        ANTHOCNETEncapsulatePacketInMessage(node, &msgNew, packetBuffer);
                        if (found)
                        {
                            NetworkIpSendRawMessageToMacLayer(node, msgNew,
                                      NetworkIpGetInterfaceAddress(node,NTNbrData->interfaceIndex),
                                      ANY_DEST, IPTOS_PREC_INTERNETCONTROL, IPPROTO_ANTHOCNET, 5, 
                                      NTNbrData->interfaceIndex,
                                      previousHopAddr);                        
                        }
                        else
                        {
                            NetworkIpSendRawMessageToMacLayer(node, msgNew,
                                      dataPtr->myAddress, ANY_DEST, IPTOS_PREC_INTERNETCONTROL, IPPROTO_ANTHOCNET, 5, 
                                      DEFAULT_INTERFACE, previousHopAddr);
                        }
                        MESSAGE_Free(node,msg);
                        dataPtr->stats.dropDataPacketsNoPheromone++;
                        // Send back notification
                    }
                }
            }
            *packetWasRouted = TRUE;
        }

    //UserCodeEndRouteDataPackets
    if (sendAnts) {
        dataPtr->state = ANTHOCNET_GENERATE_REACTIVE_FORWARD_ANTS;
        enterAnthocnetGenerateReactiveForwardAnts(node, dataPtr, msg);
        return;
    } else {
        dataPtr->state = ANTHOCNET_IDLE;
        enterAnthocnetIdle(node, dataPtr, msg);
        return;
    }
}


static void enterAnthocnetGenerateReactiveForwardAnts(Node* node,
                                                      AnthocnetData* dataPtr,
                                                      Message *msg) {

    //UserCodeStartGenerateReactiveForwardAnts
    Message* timerMsg = MESSAGE_Alloc(node, NETWORK_LAYER, ROUTING_PROTOCOL_ANTHOCNET,
                                                          ANTHOCNET_ReactiveAntsTimer);
    ANTHOCNETReactiveAntsTimerInfo* timerInfo;
       
    ANTHOCNETInitiateForwardAnts(node, ANTHOCNET_REACTIVE_FORWARD, dataPtr->myAddress, 
                    dataPtr->tempData.destAddr, dataPtr->antGenerationNumber++,0);
    // The lastUsed is updated in route data packets
    ANTHOCNETUpdateRTDestWaitingSince(node, dataPtr, NULL, dataPtr->tempData.destAddr);

    if (dataPtr->printPathsAtAntReception)
    { // checkpaths ant sent with fileptr NULL so that startcheckpaths function would know that it is special case
      ANTHOCNETStartCheckPaths* startCP;
      Message* checkPathsStartMsg = MESSAGE_Alloc(node, NETWORK_LAYER, 
                                                  ROUTING_PROTOCOL_ANTHOCNET,
                                                  ANTHOCNET_CheckPaths);
      MESSAGE_InfoAlloc(node, checkPathsStartMsg, sizeof(ANTHOCNETStartCheckPaths));
      startCP = (ANTHOCNETStartCheckPaths*)MESSAGE_ReturnInfo(checkPathsStartMsg);
      startCP->destAddr = dataPtr->tempData.destAddr;
      startCP->fp = NULL; 
      fprintf(stderr,"[%g] node %u destination %u start reactive forward ant\n",
                  simTimeSec(getSimTime(node)), node->nodeId,
                  MAPPING_GetNodeIdFromInterfaceAddress(node,dataPtr->tempData.destAddr));
      MESSAGE_Send(node,checkPathsStartMsg,0);
    }

    TRACE(stderr,"[%g]Generated reactive forward ants: %u -> %u\n", simTimeSec(getSimTime(node)),
            node->nodeId, MAPPING_GetNodeIdFromInterfaceAddress(node, dataPtr->tempData.destAddr));
    fflush(stderr);
                                                              
    MESSAGE_InfoAlloc(node, timerMsg, sizeof(ANTHOCNETReactiveAntsTimerInfo));
    timerInfo = (ANTHOCNETReactiveAntsTimerInfo*)MESSAGE_ReturnInfo(timerMsg);
    timerInfo->destAddr = dataPtr->tempData.destAddr;
    timerInfo->intermediateNodeFlag = dataPtr->tempData.intermediateNodeFlag;
    dataPtr->tempData.bufferDestData->timerMsg = timerMsg;
    MESSAGE_Send(node,timerMsg,dataPtr->reactiveAntsRetryInterval);
    dataPtr->tempData.bufferDestData->numberOfAntGenerations++;
    dataPtr->stats.totalReactiveAntsGenerated++;

    //UserCodeEndGenerateReactiveForwardAnts
    dataPtr->state = ANTHOCNET_IDLE;
    enterAnthocnetIdle(node, dataPtr, msg);
}

static void enterAnthocnetProcessReactiveForwardAnt(Node* node,
                                                    AnthocnetData* dataPtr,
                                                    Message *msg) {
    //UserCodeStartProcessReactiveForwardAnt
    clocktype jitter;
    ANTHOCNETForwardAnt* FAnt;
    char* antPacket = MESSAGE_ReturnPacket(msg);   
    char* movingPtr = antPacket + sizeof(ANTHOCNETAntType);
    /*char maxBroadcasts = *((char*)movingPtr); // maxBroadcasts
    movingPtr += sizeof(char); // maxBroadcasts*/
    FAnt = (ANTHOCNETForwardAnt*)movingPtr;
    movingPtr += sizeof(ANTHOCNETForwardAnt);
    FAnt->visitedNodes = (ANTHOCNETVisitedNode*)movingPtr;
    FAnt->tripTime += ANTHOCNETGetVirtualSendTime(node,
                                 dataPtr->currentInterfaceIndex,MESSAGE_ReturnPacketSize(msg));
    if (dataPtr->jitterRange >= NANO_SECOND) 
        jitter = RANDOM_nrand(dataPtr->seed) % (dataPtr->jitterRange);
    else
    {
        jitter = 0;
        RANDOM_nrand(dataPtr->seed);
    }
    // I am the final destination of the forward ant
    if (FAnt->destinationAddress == dataPtr->myAddress)
    {
        ANTHOCNETUpdateRTDestWaitingSince(node, dataPtr, NULL, FAnt->destinationAddress);
        if (ANTHOCNETDecideIfProcessFAntAtDest(node, FAnt, dataPtr, ANTHOCNET_REACTIVE_FORWARD))
        {
            dataPtr->stats.totalReactiveAcceptedAnts++;
            dataPtr->stats.totalReactiveAcceptedHops += FAnt->numberOfVisitedNodes;
            ANTHOCNETProcessAntAtDest(node, dataPtr, FAnt, ANTHOCNET_REACTIVE_FORWARD, 
                                    dataPtr->currentInterfaceIndex, dataPtr->currentInterfaceAddress);
            if (ANTHOCNET_PRINT_PHEROMONE_CHANGE) ANTHOCNETPrintPheromone(node, dataPtr, FAnt->sourceAddress, ANTHOCNET_REACTIVE_FORWARD);
        }
        else
        {
            dataPtr->stats.totalReactiveDestUnacceptedAnts++;
            dataPtr->stats.totalReactiveDestUnacceptedHops += FAnt->numberOfVisitedNodes;
        }
    }
    // I have already seen this ant, it is in a loop
    else if (ANTHOCNETCheckIfForwardAntAlreadySeen(FAnt, dataPtr->myAddress))
    {
      TRACE(stderr,"[%g]Kill already seen reactive forward ant %u - %u -> %u (seq %d) at Node %u\n", 
            simTimeSec(getSimTime(node)),
            MAPPING_GetNodeIdFromInterfaceAddress(node,FAnt->sourceAddress),
            MAPPING_GetNodeIdFromInterfaceAddress(node,FAnt->visitedNodes[FAnt->numberOfVisitedNodes-1].node),
            MAPPING_GetNodeIdFromInterfaceAddress(node,FAnt->destinationAddress),
            FAnt->generationNumber, node->nodeId);
      
      dataPtr->stats.totalLoopingAnts++;
      dataPtr->stats.totalLoopingHops += FAnt->numberOfVisitedNodes;
    }
    // This ant has violated the ttl limits !
    else if (FAnt->numberOfVisitedNodes > dataPtr->maxTtl)
    {
        dataPtr->stats.totalTtlKilledAnts++;
    }
    // I am on the path of the ant, and forward it.
    else
    {                    
      PacketBuffer* packetBuffer;
      Message* msgNew = NULL;
                   
      if (ANTHOCNETDecideIfProcessFAnt(node,FAnt,dataPtr,ANTHOCNET_REACTIVE_FORWARD))
        {
          ANTHOCNETListEntry *nextHopEntry = ANTHOCNETGetNextHopData(node, FAnt->destinationAddress, // maxBroadcasts
                                              FAnt->visitedNodes[FAnt->numberOfVisitedNodes-1].node, // maxBroadcasts : -1!!!!!!!!
                                              &dataPtr->routingTable, dataPtr->dataRoutingExponent); // maxBroadcasts
           
          // here code rearanged for maxBroadcasts:
          if (nextHopEntry != NULL)
          {
              ANTHOCNETRTDestNextHop *RTDestData = (ANTHOCNETRTDestNextHop*)nextHopEntry->data;
    //          ANTHOCNETUpdateRTForward(FAnt, dataPtr, node);
              ANTHOCNETUpdateRTDestWaitingSince(node, dataPtr, NULL, FAnt->destinationAddress);
        
              packetBuffer = ANTHOCNETBuildForwardAntPacket(ANTHOCNET_REACTIVE_FORWARD, FAnt, dataPtr->myAddress);
              ANTHOCNETEncapsulatePacketInMessage(node, &msgNew, packetBuffer);
              
              msgNew->packetCreationTime = msg->packetCreationTime; // Passing on the creation time to the next ant

              NetworkIpSendRawMessageToMacLayer(node, msgNew,
                     NetworkIpGetInterfaceAddress(node,RTDestData->interfaceIndex), //dataPtr->myAddress, 
                               ANY_DEST, IPTOS_PREC_INTERNETCONTROL, IPPROTO_ANTHOCNET, 5,
                               RTDestData->interfaceIndex,
                               RTDestData->interfaceAddress);
              dataPtr->stats.totalAntsForwarded++;
          }
          else
          {
    //          ANTHOCNETUpdateRTForward(FAnt, dataPtr, node);
              ANTHOCNETUpdateRTDestWaitingSince(node, dataPtr, NULL, FAnt->destinationAddress);
        
              packetBuffer = ANTHOCNETBuildForwardAntPacket(ANTHOCNET_REACTIVE_FORWARD, FAnt, dataPtr->myAddress);
              
              ANTHOCNETEncapsulatePacketInMessage(node, &msgNew, packetBuffer);
              
              msgNew->packetCreationTime = msg->packetCreationTime; // Passing on the creation time to the next ant

              ANTHOCNETBroadcastControlMessage(node, msgNew, IPTOS_PREC_INTERNETCONTROL, jitter, 5);
              dataPtr->stats.totalAntsForwarded++;
          }
                       
          TRACE(stderr,"[%g]Forward reactive forward ant %u - %u -> %u (seq %d) at Node %u\n", 
                simTimeSec(getSimTime(node)),
                MAPPING_GetNodeIdFromInterfaceAddress(node,FAnt->sourceAddress),
                (FAnt->numberOfVisitedNodes == 1)? 
                    MAPPING_GetNodeIdFromInterfaceAddress(node,FAnt->sourceAddress) :
                    MAPPING_GetNodeIdFromInterfaceAddress(node,FAnt->visitedNodes[FAnt->numberOfVisitedNodes-2].node),
                    MAPPING_GetNodeIdFromInterfaceAddress(node,FAnt->destinationAddress),
                    FAnt->generationNumber, node->nodeId);
        }
      else
        {
          dataPtr->stats.totalReactiveInterUnacceptedAnts++;
          dataPtr->stats.totalReactiveInterUnacceptedHops += FAnt->numberOfVisitedNodes;

          TRACE(stderr,"[%g] Kill bad reactive forward ant %u - %u -> %u (seq %d) at Node %u\n", 
                simTimeSec(getSimTime(node)),
                MAPPING_GetNodeIdFromInterfaceAddress(node,FAnt->sourceAddress), 
                MAPPING_GetNodeIdFromInterfaceAddress(node,FAnt->visitedNodes[FAnt->numberOfVisitedNodes-1].node),
                MAPPING_GetNodeIdFromInterfaceAddress(node,FAnt->destinationAddress), 
                FAnt->generationNumber, node->nodeId);
        }
    }
    ANTHOCNETUpdateRTDestLastUsed(node, dataPtr, NULL, FAnt->destinationAddress);
    MESSAGE_Free(node,msg);

    //UserCodeEndProcessReactiveForwardAnt
    dataPtr->state = ANTHOCNET_IDLE;
    enterAnthocnetIdle(node, dataPtr, msg);
}

static void enterAnthocnetProcessRepairBackwardAnt(Node* node,
                                                   AnthocnetData* dataPtr,
                                                   Message *msg) {
    //UserCodeStartProcessRepairBackwardAnt
    char* antPacket = MESSAGE_ReturnPacket(msg);
    char* movingPtr = antPacket + sizeof(ANTHOCNETAntType);
    ANTHOCNETBackwardAnt* BAnt = (ANTHOCNETBackwardAnt*)movingPtr;

    movingPtr += sizeof(ANTHOCNETBackwardAnt);
    BAnt->visitedNodes = (ANTHOCNETVisitedNode*)movingPtr;
    BAnt->tripTime += ANTHOCNETGetVirtualSendTime(node,
                                                  dataPtr->currentInterfaceIndex,
                                                  MESSAGE_ReturnPacketSize(msg));
    BAnt->currentPosition--;
    BAnt->pheromone = ANTHOCNETCalculatePheromoneOneHop(node, dataPtr, BAnt->pheromone, 
                                                        BAnt->prevSINR,
                                                        dataPtr->currentInterfaceIndex,
                                                        MESSAGE_ReturnPacketSize(msg),
                                                        BAnt->numberOfVisitedNodes-BAnt->currentPosition-1,
                                                        BAnt->tripTime);
    // I am the ant's final destination
    if (BAnt->currentPosition == 0) // -- sets it to current node's position
    {
      if (BAnt->visitedNodes[BAnt->currentPosition].node != dataPtr->myAddress)
      {
        ERROR_Assert(FALSE, "Wrong backward ant forwarding.");
        abort();
      }
      ANTHOCNETUpdateRTBackward(BAnt, dataPtr, node, dataPtr->currentInterfaceIndex,
                                dataPtr->currentInterfaceAddress,FALSE, ANTHOCNET_NO_DATA, ANTHOCNET_NO_DATA, ANTHOCNET_NO_DATA);

      if (ANTHOCNET_PRINT_PHEROMONE_CHANGE) ANTHOCNETPrintPheromone(node, dataPtr, BAnt->destinationAddress, ANTHOCNET_REPAIR_BACKWARD);
            
      dataPtr->stats.totalReactiveBackwardAntsReceived++;

      TRACE(stderr,"[%g]Backward ant %u - %u -> %u arrived at destination Node %u\n", 
            simTimeSec(getSimTime(node)),
            MAPPING_GetNodeIdFromInterfaceAddress(node,BAnt->sourceAddress),
            MAPPING_GetNodeIdFromInterfaceAddress(node,BAnt->visitedNodes[BAnt->currentPosition+1].node),
            MAPPING_GetNodeIdFromInterfaceAddress(node,BAnt->destinationAddress),
            node->nodeId);
    }
    // I am on the path of the ant and forward it.
    else
    {
      NodeAddress nextHop;
      PacketBuffer* packetBuffer;
      Message* msgNew = NULL;
      ANTHOCNETListEntry* NTNbrEntry;
      BOOL found;
      nextHop = BAnt->visitedNodes[BAnt->currentPosition-1].node; 
      found = ANTHOCNETListSearch(&NTNbrEntry, &dataPtr->neighbourTable, nextHop);
      if (found)
      {  
          ANTHOCNETNTNbr *NTNbrData = (ANTHOCNETNTNbr*)NTNbrEntry->data;
          ANTHOCNETUpdateRTBackward(BAnt, dataPtr, node, dataPtr->currentInterfaceIndex,
                                dataPtr->currentInterfaceAddress, TRUE, nextHop, NTNbrData->interfaceIndex, NTNbrData->interfaceAddress);
          if (dataPtr->bidirectionalPaths) 
          {
              ANTHOCNETUpdateRTForward(BAnt, dataPtr, node);
          }
          if (ANTHOCNET_PRINT_PHEROMONE_CHANGE) ANTHOCNETPrintPheromone(node, dataPtr, BAnt->destinationAddress, ANTHOCNET_REPAIR_BACKWARD);

          BAnt->prevSINR = NTNbrData->SINR;

          packetBuffer = ANTHOCNETBuildBackwardAntPacket(ANTHOCNET_REPAIR_BACKWARD, BAnt);
          ANTHOCNETEncapsulatePacketInMessage(node, &msgNew, packetBuffer);
                 
          msgNew->packetCreationTime = msg->packetCreationTime; // Passing on the creation time to the next ant

          NetworkIpSendRawMessageToMacLayer(node, msgNew,
                            NetworkIpGetInterfaceAddress(node,NTNbrData->interfaceIndex), //dataPtr->myAddress, 
                            NTNbrData->interfaceAddress,//nextHop,
                            IPTOS_PREC_INTERNETCONTROL, IPPROTO_ANTHOCNET, 5, 
                            NTNbrData->interfaceIndex,
                            NTNbrData->interfaceAddress);//nextHop);
                                          
          TRACE(stderr,"[%g]Forward Backward ant %u - %u -> %u at Node %u to NextHop %u\n",
                simTimeSec(getSimTime(node)),
                MAPPING_GetNodeIdFromInterfaceAddress(node,BAnt->sourceAddress),
                MAPPING_GetNodeIdFromInterfaceAddress(node,BAnt->visitedNodes[BAnt->currentPosition+1].node),
                MAPPING_GetNodeIdFromInterfaceAddress(node,BAnt->destinationAddress),
                node->nodeId, MAPPING_GetNodeIdFromInterfaceAddress(node, nextHop));
      }
      else
      {
          ANTHOCNETUpdateRTBackward(BAnt, dataPtr, node, dataPtr->currentInterfaceIndex,
                                dataPtr->currentInterfaceAddress, FALSE, ANTHOCNET_NO_DATA, ANTHOCNET_NO_DATA, ANTHOCNET_NO_DATA);
          if (ANTHOCNET_PRINT_PHEROMONE_CHANGE) ANTHOCNETPrintPheromone(node, dataPtr, BAnt->destinationAddress, ANTHOCNET_REPAIR_BACKWARD);
      }
    }
           
    MESSAGE_Free(node,msg);

    //UserCodeEndProcessRepairBackwardAnt
    dataPtr->state = ANTHOCNET_IDLE;
    enterAnthocnetIdle(node, dataPtr, msg);
}

static void enterAnthocnetProcessReactiveBackwardAnt(Node* node,
                                                     AnthocnetData* dataPtr,
                                                     Message *msg) {
    //UserCodeStartProcessReactiveBackwardAnt
    char* antPacket = MESSAGE_ReturnPacket(msg);
    char* movingPtr = antPacket + sizeof(ANTHOCNETAntType);
    ANTHOCNETBackwardAnt* BAnt = (ANTHOCNETBackwardAnt*)movingPtr;

    movingPtr += sizeof(ANTHOCNETBackwardAnt);
    BAnt->visitedNodes = (ANTHOCNETVisitedNode*)movingPtr;

    BAnt->tripTime += ANTHOCNETGetVirtualSendTime(node,
                                                  dataPtr->currentInterfaceIndex,
                                                  MESSAGE_ReturnPacketSize(msg));
    BAnt->currentPosition--;
    BAnt->pheromone = ANTHOCNETCalculatePheromoneOneHop(node, dataPtr, BAnt->pheromone, 
                                                        BAnt->prevSINR,
                                                        dataPtr->currentInterfaceIndex,
                                                        MESSAGE_ReturnPacketSize(msg),
                                                        BAnt->numberOfVisitedNodes-BAnt->currentPosition-1,
                                                        BAnt->tripTime);

    // I am the ant's final destination
    if (BAnt->currentPosition == 0) // -- sets it to current node's position
    {
        if (BAnt->visitedNodes[BAnt->currentPosition].node != dataPtr->myAddress)
        {
          ERROR_Assert(FALSE, "Wrong backward ant forwarding.");
          abort();
        }
        if (dataPtr->admissionControlFlag) 
        {
            ANTHOCNETListEntry *connEntry;
            BOOL connFound = ANTHOCNETListSearch(&connEntry, &dataPtr->currentConnectionsInfo, BAnt->visitedNodes[BAnt->numberOfVisitedNodes-1].node);
            if (!connFound)
            {
                ANTHOCNETListEntry *newEntry = (ANTHOCNETListEntry*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETListEntry));
                newEntry->nodeAddress = BAnt->visitedNodes[BAnt->numberOfVisitedNodes-1].node;
                newEntry->data = (ANTHOCNETCurrentConnectionsData*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETCurrentConnectionsData));
                ((ANTHOCNETCurrentConnectionsData*)newEntry->data)->counter = 0;
                ((ANTHOCNETCurrentConnectionsData*)newEntry->data)->rejected = ANTHOCNETRejectSession(node, dataPtr, (double)(BAnt->numberOfVisitedNodes - 1), (double)BAnt->tripTime);
                ANTHOCNETListInsert(&dataPtr->currentConnectionsInfo, newEntry, connEntry);
            }
        }
        // prevSINR: use directly from BAnt - is only changed later, at forwarding...
        ANTHOCNETUpdateRTBackward(BAnt, dataPtr, node, dataPtr->currentInterfaceIndex, 
                                dataPtr->currentInterfaceAddress, FALSE, ANTHOCNET_NO_DATA, ANTHOCNET_NO_DATA, ANTHOCNET_NO_DATA);
     
        if (ANTHOCNET_PRINT_PHEROMONE_CHANGE) ANTHOCNETPrintPheromone(node, dataPtr, BAnt->destinationAddress, ANTHOCNET_REACTIVE_BACKWARD);
      
        dataPtr->stats.totalReactiveBackwardAntsReceived++;
      
      if (dataPtr->printPathsAtAntReception)
      { // checkpaths ant sent with fileptr NULL so that startcheckpaths function would know that it is special case
          ANTHOCNETStartCheckPaths* startCP;
          Message* checkPathsStartMsg = MESSAGE_Alloc(node, NETWORK_LAYER, 
                                                      ROUTING_PROTOCOL_ANTHOCNET,
                                                      ANTHOCNET_CheckPaths);
          MESSAGE_InfoAlloc(node, checkPathsStartMsg, sizeof(ANTHOCNETStartCheckPaths));
          startCP = (ANTHOCNETStartCheckPaths*)MESSAGE_ReturnInfo(checkPathsStartMsg);
          startCP->destAddr = BAnt->destinationAddress;
          startCP->fp = NULL; 
          fprintf(stderr,"[%g] node %u destination %u backward ant received\n",
                      simTimeSec(getSimTime(node)), node->nodeId,
                      MAPPING_GetNodeIdFromInterfaceAddress(node,BAnt->destinationAddress));
          MESSAGE_Send(node,checkPathsStartMsg,0);
      }
      TRACE(stderr,"[%g]Backward ant %u - %u -> %u arrived at destination Node %u\n", 
            simTimeSec(getSimTime(node)),
          MAPPING_GetNodeIdFromInterfaceAddress(node,BAnt->sourceAddress),
          MAPPING_GetNodeIdFromInterfaceAddress(node,BAnt->visitedNodes[BAnt->currentPosition+1].node),
          MAPPING_GetNodeIdFromInterfaceAddress(node,BAnt->destinationAddress),
            node->nodeId);
    }
    // I am on the path of the ant and forward it.
    else
    {
      NodeAddress nextHop;
      PacketBuffer* packetBuffer;
      Message* msgNew = NULL;
      BOOL found;
      ANTHOCNETListEntry *NTNbrEntry;

      nextHop = BAnt->visitedNodes[BAnt->currentPosition-1].node;
      found = ANTHOCNETListSearch(&NTNbrEntry, &dataPtr->neighbourTable, nextHop);
      if (found)
      {      
          ANTHOCNETNTNbr *NTNbrData = (ANTHOCNETNTNbr*)NTNbrEntry->data;

          ANTHOCNETUpdateRTBackward(BAnt, dataPtr, node, dataPtr->currentInterfaceIndex,
                                dataPtr->currentInterfaceAddress, TRUE, nextHop, NTNbrData->interfaceIndex, NTNbrData->interfaceAddress);
          if (dataPtr->bidirectionalPaths) 
          {
              ANTHOCNETUpdateRTForward(BAnt, dataPtr, node);
          }
          if (ANTHOCNET_PRINT_PHEROMONE_CHANGE) ANTHOCNETPrintPheromone(node, dataPtr, BAnt->destinationAddress, ANTHOCNET_REACTIVE_BACKWARD);

          BAnt->prevSINR = NTNbrData->SINR; 

          packetBuffer = ANTHOCNETBuildBackwardAntPacket(ANTHOCNET_REACTIVE_BACKWARD, BAnt);
          ANTHOCNETEncapsulatePacketInMessage(node, &msgNew, packetBuffer);
                 
          msgNew->packetCreationTime = msg->packetCreationTime; // Passing on the creation time to the next ant

          NetworkIpSendRawMessageToMacLayer(node, msgNew,
                  NetworkIpGetInterfaceAddress(node,NTNbrData->interfaceIndex),//dataPtr->myAddress, 
                  NTNbrData->interfaceAddress,//nextHop,
                  IPTOS_PREC_INTERNETCONTROL, IPPROTO_ANTHOCNET, 5,
                  NTNbrData->interfaceIndex,
                  NTNbrData->interfaceAddress);//nextHop);

          TRACE(stderr,"[%g]Forward Backward ant %u - %u -> %u at Node %u to NextHop %u\n", 
                simTimeSec(getSimTime(node)),
                MAPPING_GetNodeIdFromInterfaceAddress(node,BAnt->sourceAddress),
                MAPPING_GetNodeIdFromInterfaceAddress(node,BAnt->visitedNodes[BAnt->currentPosition+1].node),
                MAPPING_GetNodeIdFromInterfaceAddress(node,BAnt->destinationAddress),
                node->nodeId, MAPPING_GetNodeIdFromInterfaceAddress(node,nextHop));
      }
      else
      {
          ANTHOCNETUpdateRTBackward(BAnt, dataPtr, node, dataPtr->currentInterfaceIndex,
                                dataPtr->currentInterfaceAddress, FALSE, ANTHOCNET_NO_DATA, ANTHOCNET_NO_DATA, ANTHOCNET_NO_DATA);
          if (ANTHOCNET_PRINT_PHEROMONE_CHANGE) ANTHOCNETPrintPheromone(node, dataPtr, BAnt->destinationAddress, ANTHOCNET_REACTIVE_BACKWARD);
      }

    }
           
    MESSAGE_Free(node,msg);

    //UserCodeEndProcessReactiveBackwardAnt
    dataPtr->state = ANTHOCNET_IDLE;
    enterAnthocnetIdle(node, dataPtr, msg);
}

static void enterAnthocnetProcessRepairForwardAnt(Node* node,
                                                  AnthocnetData* dataPtr,
                                                  Message *msg) {
    //UserCodeStartProcessRepairForwardAnt
    char* antPacket = MESSAGE_ReturnPacket(msg);
    char* movingPtr = antPacket + sizeof(ANTHOCNETAntType);
    NodeAddress lostNextHopAddress = *(NodeAddress*)movingPtr;
    ANTHOCNETForwardAnt* FAnt;
    clocktype jitter;
    movingPtr += sizeof(NodeAddress);
    FAnt = (ANTHOCNETForwardAnt*)movingPtr;
    movingPtr += sizeof(ANTHOCNETForwardAnt);
    FAnt->visitedNodes = (ANTHOCNETVisitedNode*)movingPtr;
    FAnt->tripTime += ANTHOCNETGetVirtualSendTime(node,
                                                  dataPtr->currentInterfaceIndex,
                                                  MESSAGE_ReturnPacketSize(msg));
    if (dataPtr->jitterRange >= NANO_SECOND) 
        jitter = RANDOM_nrand(dataPtr->seed) % (dataPtr->jitterRange);
    else
    {
        jitter = 0;
        RANDOM_nrand(dataPtr->seed);
    }
 
    // I am the final destination of the forward ant
    if (FAnt->destinationAddress == dataPtr->myAddress)
    {
        ANTHOCNETUpdateRTDestWaitingSince(node, dataPtr, NULL, FAnt->destinationAddress);
        if (ANTHOCNETDecideIfProcessFAntAtDest(node, FAnt, dataPtr, ANTHOCNET_REPAIR_FORWARD))
        {
            dataPtr->stats.totalRepairAcceptedAnts++;
            dataPtr->stats.totalRepairAcceptedHops += FAnt->numberOfVisitedNodes;
            ANTHOCNETProcessAntAtDest(node, dataPtr, FAnt, ANTHOCNET_REPAIR_FORWARD, 
                                    dataPtr->currentInterfaceIndex,dataPtr->currentInterfaceAddress);
            if (ANTHOCNET_PRINT_PHEROMONE_CHANGE) ANTHOCNETPrintPheromone(node, dataPtr, FAnt->sourceAddress, ANTHOCNET_REPAIR_FORWARD);
        }
        else
        {
            dataPtr->stats.totalRepairDestUnacceptedAnts++;
            dataPtr->stats.totalRepairDestUnacceptedHops += FAnt->numberOfVisitedNodes;
        }
    }
    // I have already seen this ant, it is in a loop
    else if (ANTHOCNETCheckIfForwardAntAlreadySeen(FAnt, dataPtr->myAddress))
    {
      dataPtr->stats.totalLoopingAnts++;
      dataPtr->stats.totalLoopingHops += FAnt->numberOfVisitedNodes;
    }
    // I am on the path of the ant and decide to forward it
    else if (ANTHOCNETDecideIfProcessFAnt(node,FAnt,dataPtr,ANTHOCNET_REPAIR_FORWARD))
    {
      ANTHOCNETListEntry *nextHopEntry = ANTHOCNETGetNextHopData(node, FAnt->destinationAddress, 
                                        FAnt->visitedNodes[FAnt->numberOfVisitedNodes-1].node,
                                        &dataPtr->routingTable, dataPtr->dataRoutingExponent);
      PacketBuffer* packetBuffer;
      Message* msgNew = NULL;
      // If we have a nextHop we do unicast forwarding:
      if (nextHopEntry != NULL)
      {
          ANTHOCNETRTDestNextHop *RTDestData = (ANTHOCNETRTDestNextHop*)nextHopEntry->data;
          ANTHOCNETUpdateRTDestWaitingSince(node, dataPtr, NULL, FAnt->destinationAddress);
//          ANTHOCNETUpdateRTForward(FAnt, dataPtr, node);
          
          if (ANTHOCNET_PRINT_PHEROMONE_CHANGE) ANTHOCNETPrintPheromone(node, dataPtr, FAnt->sourceAddress, ANTHOCNET_REPAIR_FORWARD);
                       
          packetBuffer = ANTHOCNETBuildRepairAntPacket(ANTHOCNET_REPAIR_FORWARD, FAnt, 
                                                        dataPtr->myAddress,lostNextHopAddress);
          ANTHOCNETEncapsulatePacketInMessage(node, &msgNew, packetBuffer);
            
          msgNew->packetCreationTime = msg->packetCreationTime; // Passing on the creation time to the next ant
          
          NetworkIpSendRawMessageToMacLayerWithDelay(node, msgNew,
                        NetworkIpGetInterfaceAddress(node,RTDestData->interfaceIndex), //dataPtr->myAddress, 
                        ANY_DEST, IPTOS_PREC_INTERNETCONTROL, IPPROTO_ANTHOCNET, 5,
                        RTDestData->interfaceIndex,
                        RTDestData->interfaceAddress,jitter);
                                          
          dataPtr->stats.totalAntsForwarded++;
      }
      // Otherwise we broadcast, if at least we have not exceeded the maximum number of broadcasts:
      else if (FAnt->numberOfVisitedNodes < dataPtr->repairMaxBroadcasting)
      {
          ANTHOCNETUpdateRTDestWaitingSince(node, dataPtr, NULL, FAnt->destinationAddress);
//          ANTHOCNETUpdateRTForward(FAnt, dataPtr, node);

          if (ANTHOCNET_PRINT_PHEROMONE_CHANGE) ANTHOCNETPrintPheromone(node, dataPtr, FAnt->sourceAddress, ANTHOCNET_REPAIR_FORWARD);
                       
          packetBuffer = ANTHOCNETBuildRepairAntPacket(ANTHOCNET_REPAIR_FORWARD, FAnt, dataPtr->myAddress,
                                                    lostNextHopAddress);
          ANTHOCNETEncapsulatePacketInMessage(node, &msgNew, packetBuffer);
            
          msgNew->packetCreationTime = msg->packetCreationTime; // Passing on the creation time to the next ant
                        
          ANTHOCNETBroadcastControlMessage(node, msgNew, IPTOS_PREC_INTERNETCONTROL, jitter, 5);
                                          
          dataPtr->stats.totalAntsForwarded++;                                                 
      }
      // bad ant: max number of broadcast hops exceeded
      else
      {
          dataPtr->stats.totalRepairInterUnacceptedAnts++;
          dataPtr->stats.totalRepairInterUnacceptedHops += FAnt->numberOfVisitedNodes;
      }
    }
    // We decide not to forward it: bad ant compared to others of same generation
    else
    {
      dataPtr->stats.totalRepairInterUnacceptedAnts++;
      dataPtr->stats.totalRepairInterUnacceptedHops += FAnt->numberOfVisitedNodes;
    }      
    ANTHOCNETUpdateRTDestLastUsed(node, dataPtr, NULL, FAnt->destinationAddress);
    MESSAGE_Free(node,msg);  

    //UserCodeEndProcessRepairForwardAnt
    dataPtr->state = ANTHOCNET_IDLE;
    enterAnthocnetIdle(node, dataPtr, msg);
}

static void enterAnthocnetProcessLinkFailureNotification(Node* node,
                                                         AnthocnetData* dataPtr,
                                                         Message *msg) {
    //UserCodeStartProcessLinkFailureNotification
    char* antPacket = MESSAGE_ReturnPacket(msg);   
    char* movingPtr = antPacket + sizeof(ANTHOCNETAntType);
    ANTHOCNETLostPheromone* receivedArray;
    ANTHOCNETLostPheromone* sendArray;
    short sendNumberOfElements;
    short receivedNumberOfElements;
    clocktype jitter;
    int i;

    if (dataPtr->jitterRange >= NANO_SECOND) 
        jitter = RANDOM_nrand(dataPtr->seed) % (dataPtr->jitterRange);
    else
    {
        jitter = 0;
        RANDOM_nrand(dataPtr->seed);
    }
    TRACE(stderr,"[%g]Link Failure ANT at Node %u\n", simTimeSec(getSimTime(node)), node->nodeId);
             
    receivedNumberOfElements = *(short*)movingPtr;
    movingPtr += sizeof(short);
    receivedArray = (ANTHOCNETLostPheromone*)movingPtr;

    sendNumberOfElements = ANTHOCNETRTUpdateAfterFailureMsg(node, dataPtr, &dataPtr->routingTable, 
                        &dataPtr->neighbourTable, receivedArray, receivedNumberOfElements, 
                        &sendArray, dataPtr->currentDefaultAddress, dataPtr->currentInterfaceIndex,
                        MESSAGE_ReturnPacketSize(msg));

    if (sendNumberOfElements > 0)
    {
      PacketBuffer* packetBuffer;
      Message* msgNew;

      packetBuffer = ANTHOCNETBuildLinkFailureAntPacket(ANTHOCNET_LINK_FAILURE_NOTIFICATION,sendArray, 
                 sendNumberOfElements);
      ANTHOCNETEncapsulatePacketInMessage(node, &msgNew, packetBuffer);

      MEM_free(sendArray);
            
      /* Send the link failure message (ttl at the moment set to 1) */
      ANTHOCNETBroadcastControlMessage(node, msgNew, IPTOS_PREC_INTERNETCONTROL, jitter, 1);
    }
    MESSAGE_Free(node,msg);

    //UserCodeEndProcessLinkFailureNotification
    dataPtr->state = ANTHOCNET_IDLE;
    enterAnthocnetIdle(node, dataPtr, msg);
}

static void enterAnthocnetProcessProactiveForwardAnt(Node* node,
                                                     AnthocnetData* dataPtr,
                                                     Message *msg) {
    //UserCodeStartProcessProactiveForwardAnt
    ANTHOCNETForwardAnt* FAnt;
    char* antPacket = MESSAGE_ReturnPacket(msg);   
    char* movingPtr = antPacket + sizeof(ANTHOCNETAntType);
    FAnt = (ANTHOCNETForwardAnt*)movingPtr;
    movingPtr += sizeof(ANTHOCNETForwardAnt);
    FAnt->visitedNodes = (ANTHOCNETVisitedNode*)movingPtr;
    FAnt->tripTime += ANTHOCNETGetVirtualSendTime(node,
                                                  dataPtr->currentInterfaceIndex,
                                                  MESSAGE_ReturnPacketSize(msg));
              
    // I am the final destination of the forward ant
    if (FAnt->destinationAddress == dataPtr->myAddress)
    {
        if (ANTHOCNETDecideIfProcessFAntAtDest(node, FAnt, dataPtr, ANTHOCNET_PROACTIVE_FORWARD)) //&& (!existingLink)) // existingLink
        {
            dataPtr->stats.totalProactiveAcceptedAnts++;
            dataPtr->stats.totalProactiveAcceptedHops += FAnt->numberOfVisitedNodes;
            ANTHOCNETProcessAntAtDest(node, dataPtr, FAnt, ANTHOCNET_PROACTIVE_FORWARD, 
                                    dataPtr->currentInterfaceIndex, dataPtr->currentInterfaceAddress);
            if (ANTHOCNET_PRINT_PHEROMONE_CHANGE) ANTHOCNETPrintPheromone(node, dataPtr, FAnt->sourceAddress, ANTHOCNET_PROACTIVE_FORWARD);
        }
        else
        {
            dataPtr->stats.totalProactiveDestUnacceptedAnts++;
            dataPtr->stats.totalProactiveDestUnacceptedHops += FAnt->numberOfVisitedNodes;
        }
        MESSAGE_Free(node,msg);
    }
    // I have already seen this ant, it is in a loop
    else if (ANTHOCNETCheckIfForwardAntAlreadySeen(FAnt, dataPtr->myAddress))
    {
      TRACE(stderr,"[%g]Kill already seen reactive forward ant %u - %u -> %u (seq %d) at Node %u\n", 
            simTimeSec(getSimTime(node)),
          MAPPING_GetNodeIdFromInterfaceAddress(node,FAnt->sourceAddress),
            MAPPING_GetNodeIdFromInterfaceAddress(node,FAnt->visitedNodes[FAnt->numberOfVisitedNodes-1].node),
            MAPPING_GetNodeIdFromInterfaceAddress(node,FAnt->destinationAddress),
            FAnt->generationNumber, node->nodeId);
      
      dataPtr->stats.totalProactiveLoopingAnts++;
      dataPtr->stats.totalProactiveLoopingHops += FAnt->numberOfVisitedNodes;
      
      MESSAGE_Free(node,msg);
    }
    // This ant has violated the ttl limits !
    else if (FAnt->numberOfVisitedNodes > dataPtr->maxTtl)
    {
        dataPtr->stats.totalTtlKilledAnts++;
        MESSAGE_Free(node,msg);
    }
    // I am on the path of the ant, and have to make some serious decisions
    else
    {         
        ANTHOCNETListEntry *nextHopEntry;
        if (ANTHOCNETDecideIfProcessFAnt(node,FAnt,dataPtr,ANTHOCNET_PROACTIVE_FORWARD))
        {
            ANTHOCNETListEntry* destEntry;
            BOOL found = ANTHOCNETListSearch(&destEntry,&dataPtr->routingTable,FAnt->destinationAddress);
            if (found)
            {
                if (((ANTHOCNETRTDest*)destEntry->data)->nextHops.numberOfEntries > 0)
                {
                    nextHopEntry = ANTHOCNETGetNextHopProactive(node,FAnt->destinationAddress,
                                            FAnt->visitedNodes[FAnt->numberOfVisitedNodes-1].node, 
                                            &((ANTHOCNETRTDest*)destEntry->data)->nextHops, 
                                            dataPtr->proactiveAntRoutingExponent,
                                            ((ANTHOCNETRTDest*)destEntry->data)->numberOfConfirmedEntries);
                    if (nextHopEntry != NULL)
                    {
                        ANTHOCNETRTDestNextHop *RTDestData = (ANTHOCNETRTDestNextHop*)nextHopEntry->data;
                        PacketBuffer* packetBuffer;
                        Message* msgNew = NULL;
                        clocktype jitter;
    //                    ANTHOCNETUpdateRTForward(FAnt, dataPtr, node);
                        packetBuffer = ANTHOCNETBuildProactiveAntPacket(ANTHOCNET_PROACTIVE_FORWARD, FAnt, 
                                              dataPtr->myAddress);//,(existingLink && newExistingLink)); // existingLink
                        ANTHOCNETEncapsulatePacketInMessage(node, &msgNew, packetBuffer);
                        msgNew->packetCreationTime = msg->packetCreationTime; // Passing on the creation time to the next ant
                        if (dataPtr->jitterRange >= NANO_SECOND) 
                            jitter = RANDOM_nrand(dataPtr->seed) % (dataPtr->jitterRange);
                        else
                        {
                            jitter = 0;
                            RANDOM_nrand(dataPtr->seed);
                        }
                        NetworkIpSendRawMessageToMacLayerWithDelay(node, msgNew,
                                NetworkIpGetInterfaceAddress(node,RTDestData->interfaceIndex), 
                                ANY_DEST, IPTOS_PREC_INTERNETCONTROL, IPPROTO_ANTHOCNET,
                                5, RTDestData->interfaceIndex,
                                RTDestData->interfaceAddress, //nextHopEntry->nodeAddress,
                                jitter);
                        dataPtr->stats.totalAntsForwarded++;
                                                                         
                        TRACE(stderr,
                              "[%g]Forward proactive forward ant %u - %u -> %u (seq %d) at Node %u\n", 
                              simTimeSec(getSimTime(node)),
                              MAPPING_GetNodeIdFromInterfaceAddress(node,FAnt->sourceAddress),
                              (FAnt->numberOfVisitedNodes == 1)? 
                                MAPPING_GetNodeIdFromInterfaceAddress(node,FAnt->sourceAddress) :
                                MAPPING_GetNodeIdFromInterfaceAddress(node,FAnt->visitedNodes[FAnt->numberOfVisitedNodes-2].node),
                              MAPPING_GetNodeIdFromInterfaceAddress(node,FAnt->destinationAddress),
                              FAnt->generationNumber, node->nodeId);
                    }
                    else
                    {
                        dataPtr->stats.totalProactiveDeadEndAnts++;
                        dataPtr->stats.totalProactiveDeadEndHops += FAnt->numberOfVisitedNodes;
                    }
                }
                else
                {
                    dataPtr->stats.totalProactiveDeadEndAnts++;
                    dataPtr->stats.totalProactiveDeadEndHops += FAnt->numberOfVisitedNodes;
                }
            }
            else
            {
                dataPtr->stats.totalProactiveDeadEndAnts++;
                dataPtr->stats.totalProactiveDeadEndHops += FAnt->numberOfVisitedNodes;
            }
        }
        else
        {
            dataPtr->stats.totalProactiveInterUnacceptedAnts++;
            dataPtr->stats.totalProactiveInterUnacceptedHops += FAnt->numberOfVisitedNodes;
            TRACE(stderr,
                  "[%g]Kill bad proactive forward ant %u - %u -> %u (seq %d) at Node %u\n", 
                  simTimeSec(getSimTime(node)),
                  MAPPING_GetNodeIdFromInterfaceAddress(node,FAnt->sourceAddress),
                  MAPPING_GetNodeIdFromInterfaceAddress(node,FAnt->visitedNodes[FAnt->numberOfVisitedNodes-1].node),
                  MAPPING_GetNodeIdFromInterfaceAddress(node,FAnt->destinationAddress), 
                  FAnt->generationNumber, node->nodeId);
        }
        MESSAGE_Free(node,msg);  
    }

    //UserCodeEndProcessProactiveForwardAnt
    dataPtr->state = ANTHOCNET_IDLE;
    enterAnthocnetIdle(node, dataPtr, msg);
}

static void enterAnthocnetHandleLinkFailure(Node* node,
                                            AnthocnetData* dataPtr,
                                            Message *msg) {
    //UserCodeStartHandleLinkFailure
    BOOL found, fromCleanup, sendReactiveAnts = FALSE;
    ANTHOCNETListEntry *nbrEntry, *destEntry, *nextHopEntry;
    char* infoField = MESSAGE_ReturnInfo(msg);
    NodeAddress nextHop, srcAddr, destAddr;
    unsigned char ip_p;
    Message* originalMsg;
    int hops = 0;
    fromCleanup = *(BOOL*)infoField;
    infoField += sizeof(BOOL);
    nextHop = *(NodeAddress*)infoField;
    if (!fromCleanup)
    {
        infoField += sizeof(NodeAddress);
        srcAddr = *(NodeAddress*)infoField;
        infoField += sizeof(NodeAddress);
        destAddr = *(NodeAddress*)infoField;
        infoField += sizeof(NodeAddress);
        ip_p = *(char*)infoField;
        infoField += sizeof(char);
        originalMsg = *(Message**)infoField;
        originalMsg->eventType = 10000; // To avoid that we get it back as a drop packet event
        hops = IPDEFTTL - ((IpHeaderType *)originalMsg->packet)->ip_ttl;

    }
    found = ANTHOCNETListSearch(&nbrEntry, &dataPtr->neighbourTable, nextHop);
    if (found) 
    {
        NodeAddress nextHopInterfaceAddress = ((ANTHOCNETNTNbr*)nbrEntry->data)->interfaceAddress;
        ANTHOCNETLostPheromone* lostLastArray;
        ANTHOCNETLostPheromone* lostBestArray;
        int lostLastNumber, lostBestNumber;
        BOOL wasLast;
        double maxAvgTime;
        BOOL currentNodeIsSrc;
        if (fromCleanup)
        {
            destAddr = dataPtr->myAddress;
            currentNodeIsSrc = FALSE;
        }
        else
        {
            currentNodeIsSrc = NetworkIpIsMyIP(node, srcAddr);
        }
        maxAvgTime = ANTHOCNETNTRemoveNbrSplitArrays(nbrEntry, &dataPtr->neighbourTable,
                             &lostBestArray,&lostBestNumber,&lostLastArray,&lostLastNumber,
                             destAddr,currentNodeIsSrc,&wasLast,node,dataPtr);
        if ((dataPtr->repairMaxBroadcasting > 0) && // if max broadcasting is 0, it means we don't do repair
            (ip_p != IPPROTO_ANTHOCNET) &&  // we don't repair if a control packet lost its path
            (!fromCleanup)) // We don't repair if the failure is detected via missed hello's
        {
            if (lostBestNumber > 0)
            {
                PacketBuffer* packetBuffer;
                Message* msgNew;
                int i;

                packetBuffer = ANTHOCNETBuildLinkFailureAntPacket(ANTHOCNET_LINK_FAILURE_NOTIFICATION,
                                             lostBestArray,(short)lostBestNumber);
                ANTHOCNETEncapsulatePacketInMessage(node, &msgNew, packetBuffer);

                MEM_free(lostBestArray);
            
                /* Send the link failure message (ttl at the moment set to 1) */
                ANTHOCNETBroadcastControlMessage(node, msgNew, IPTOS_PREC_INTERNETCONTROL, 0, 1);
                dataPtr->stats.totalLinkFailureNotificationsGenerated++;
            }
            if (lostLastNumber > 0)
            {
                int i;
                clocktype delay = (clocktype)(0.0005 * SECOND);
                ANTHOCNETLostPheromone* timerArray;
                Message* timerMsg = MESSAGE_Alloc(node, NETWORK_LAYER, ROUTING_PROTOCOL_ANTHOCNET,
                                                                ANTHOCNET_RepairTimeOut);  
                MESSAGE_InfoAlloc(node,timerMsg,lostLastNumber * sizeof(ANTHOCNETLostPheromone));
                timerArray = (ANTHOCNETLostPheromone*)MESSAGE_ReturnInfo(timerMsg);
                memcpy(timerArray,lostLastArray,lostLastNumber * sizeof(ANTHOCNETLostPheromone));
                for (i=0;i<lostLastNumber;i++)
                {
                    // Could also be the other buffer, but in the end it does not really matter.
                    ANTHOCNETMessageBufferInsert(NULL, &dataPtr->relayMessageBuffer, node, 
                                                                lostLastArray[i].destAddr);
                                                                
                    ANTHOCNETUpdateRTDestBoth(node, dataPtr, NULL, lostLastArray[i].destAddr);
                    
                    ANTHOCNETInitiateRepairForwardAnts(node, dataPtr->myAddress, 
                                                    lostLastArray[i].destAddr, 
                                                    dataPtr->antGenerationNumber++, 
                                                    delay, nextHop);
                    delay += (clocktype)(0.0005 * SECOND);
                    dataPtr->stats.totalRepairAntsGenerated++;
                }
                if (wasLast)
                {
                    sendReactiveAnts = ANTHOCNETInsertAndCheckBuffers(node, dataPtr, NULL, destAddr, 
                                                                                            ANY_IP);
                }
                dataPtr->stats.totalCriticalLinkFailures++;
                MESSAGE_Send(node,timerMsg,secSimTime(maxAvgTime*dataPtr->repairTimeFactor)+delay);
            }
        }
        else
        {
            int lostNumber = lostBestNumber + lostLastNumber;
            if (lostNumber > 0)
            {
                PacketBuffer* packetBuffer;
                Message* msgNew;
                ANTHOCNETLostPheromone* sendArray;
                int i = 0, j = 0, k=0;
                if (lostLastNumber == 0)
                    sendArray = lostBestArray;
                else if (lostBestNumber == 0)
                    sendArray = lostLastArray;
                else
                {
                    sendArray = (ANTHOCNETLostPheromone*)MEM_malloc2(__LINE__,lostNumber * sizeof(ANTHOCNETLostPheromone));
                    while (i + j < lostNumber)
                    {
                        if (lostBestArray[i].destAddr < lostLastArray[j].destAddr)
                        {
                            sendArray[i+j] = lostBestArray[i];
                            if (++i == lostBestNumber)
                                for (;j<lostLastNumber;j++)
                                    sendArray[i+j] = lostLastArray[j];
                        }
                        else
                        {
                            sendArray[i+j] = lostLastArray[j];
                            if (++j == lostLastNumber)
                                for (;i<lostBestNumber;i++)
                                    sendArray[i+j] = lostBestArray[i];
                        }
                    }
                }

                packetBuffer = ANTHOCNETBuildLinkFailureAntPacket(
                                                ANTHOCNET_LINK_FAILURE_NOTIFICATION,
                                                sendArray,(short)lostNumber);
                ANTHOCNETEncapsulatePacketInMessage(node, &msgNew, packetBuffer);

                MEM_free(lostBestArray);
                MEM_free(lostLastArray);
                if ((lostLastNumber != 0) && (lostBestNumber != 0))
                    MEM_free(sendArray);
            
                /* Send the link failure message (ttl at the moment set to 1) */
                if (fromCleanup && (!dataPtr->notificationsAfterCleanup))
                {
                    MESSAGE_Free(node,msgNew);
                }
                else
                {
                    ANTHOCNETBroadcastControlMessage(node, msgNew, IPTOS_PREC_INTERNETCONTROL, 0, 1);
                    dataPtr->stats.totalLinkFailureNotificationsGenerated++;
                }
            }
        }        
        // Re-sending of the failed packet
        if (!fromCleanup) // There is no failed packet if the failure was triggered by missed hellos
        {
            if ((ip_p != IPPROTO_ANTHOCNET)&&(dataPtr->reSend == TRUE)) // We don't re-send control packets
            {
                ANTHOCNETResendOrBufferPacket(node, dataPtr, destAddr, srcAddr, 
                                 ANY_DEST, originalMsg);
            }
            else // it is a control packet
            {
                MESSAGE_Free(node,originalMsg);
            }
        }
        NetworkIpDeleteOutboundPacketsToANode(node,nextHopInterfaceAddress,ANY_IP,TRUE);
    }

    MESSAGE_Free(node,msg);

    //UserCodeEndHandleLinkFailure
    if (sendReactiveAnts) {
        dataPtr->state = ANTHOCNET_GENERATE_REACTIVE_FORWARD_ANTS;
        enterAnthocnetGenerateReactiveForwardAnts(node, dataPtr, msg);
        return;
    } else {
        dataPtr->state = ANTHOCNET_IDLE;
        enterAnthocnetIdle(node, dataPtr, msg);
        return;
    }
}

static void enterAnthocnetSendProactiveAnt(Node* node,
                                           AnthocnetData* dataPtr,
                                           Message *msg) {
    //UserCodeStartSendProactiveAnt
    NodeAddress destAddr = *(NodeAddress*)MESSAGE_ReturnInfo(msg);
    ANTHOCNETListEntry* RTDestEntry;
    BOOL found = ANTHOCNETListSearch(&RTDestEntry, &dataPtr->routingTable, destAddr);
    if (found)
    {
        if (((ANTHOCNETRTDest*)RTDestEntry->data)->nextHops.numberOfEntries > 0)
        {
            ANTHOCNETListEntry *nextHopEntry = ANTHOCNETGetNextHopProactive(node, destAddr, dataPtr->myAddress, 
                                          &((ANTHOCNETRTDest*)RTDestEntry->data)->nextHops, 
                                          dataPtr->proactiveAntRoutingExponent,
                                          ((ANTHOCNETRTDest*)RTDestEntry->data)->numberOfConfirmedEntries);
            if (nextHopEntry != NULL)
            {
                ANTHOCNETForwardAnt *FAnt = NULL;
                PacketBuffer* packetBuffer = NULL;
                Message *msgNew = NULL;
                clocktype jitter;
                ANTHOCNETRTDestNextHop *RTDestData = (ANTHOCNETRTDestNextHop*)nextHopEntry->data;

                /* Make an ant and indicate its type in a variable. */
                FAnt = ANTHOCNETCreateForwardAnt(node,destAddr, dataPtr->myAddress, 
                                              dataPtr->antGenerationNumber++);
                
                /* Put the ant and its type in a packet */
                packetBuffer = ANTHOCNETBuildProactiveAntPacket(ANTHOCNET_PROACTIVE_FORWARD,FAnt,dataPtr->myAddress);
            
                /* Put the ant packet in a message. */
                ANTHOCNETEncapsulatePacketInMessage(node, &msgNew, packetBuffer);
                
                /* Send the ant message (ttl at the moment set to 5) */
                if (dataPtr->jitterRange >= NANO_SECOND) 
                    jitter = dataPtr->jitterRange + RANDOM_nrand(dataPtr->seed) % (dataPtr->jitterRange);
                else
                {
                    jitter = 0;
                    RANDOM_nrand(dataPtr->seed);
                }
                NetworkIpSendRawMessageToMacLayerWithDelay(node, msgNew,
                                     NetworkIpGetInterfaceAddress(node,RTDestData->interfaceIndex),//dataPtr->myAddress,
                                     ANY_DEST, IPTOS_PREC_INTERNETCONTROL, IPPROTO_ANTHOCNET, 5, 
                                     RTDestData->interfaceIndex, RTDestData->interfaceAddress,jitter);

                dataPtr->stats.totalProactiveAntsGenerated++;

                if (dataPtr->printPathsAtAntReception)
                { // checkpaths ant sent with fileptr NULL so that startcheckpaths function would know that it is special case
                    ANTHOCNETStartCheckPaths* startCP;
                    Message* checkPathsStartMsg = MESSAGE_Alloc(node, NETWORK_LAYER, 
                                                                ROUTING_PROTOCOL_ANTHOCNET,
                                                                ANTHOCNET_CheckPaths);
                    MESSAGE_InfoAlloc(node, checkPathsStartMsg, sizeof(ANTHOCNETStartCheckPaths));
                    startCP = (ANTHOCNETStartCheckPaths*)MESSAGE_ReturnInfo(checkPathsStartMsg);
                    startCP->destAddr = destAddr;
                    startCP->fp = NULL; 
                    fprintf(stderr,"[%g] node %u destination %u start proactive forward ant\n",
                                simTimeSec(getSimTime(node)), node->nodeId,
                                MAPPING_GetNodeIdFromInterfaceAddress(node,destAddr));
                    MESSAGE_Send(node,checkPathsStartMsg,0);
                }
                ANTHOCNETDeleteForwardAnt(FAnt);
            }
            else
            {
                dataPtr->stats.totalProactiveDeadEndAnts++;
            }
        }
        else
        {
            dataPtr->stats.totalProactiveDeadEndAnts++;
        }
    }
    else
    {
        dataPtr->stats.totalProactiveDeadEndAnts++;
    }
    MESSAGE_Free(node,msg);

    //UserCodeEndSendProactiveAnt
    dataPtr->state = ANTHOCNET_IDLE;
    enterAnthocnetIdle(node, dataPtr, msg);
}

static void enterAnthocnetCheckReactiveAntsRetransmission(Node* node,
                                                          AnthocnetData* dataPtr,
                                                          Message *msg) {
    //UserCodeStartCheckReactiveAntsRetransmission
    ANTHOCNETReactiveAntsTimerInfo* info = (ANTHOCNETReactiveAntsTimerInfo*)MESSAGE_ReturnInfo(msg);
    ANTHOCNETListEntry* bufferDestEntry;
    BOOL found;
    BOOL sendAnts = FALSE;
        
    found = ANTHOCNETListSearch(&bufferDestEntry, &dataPtr->relayMessageBuffer, info->destAddr);
    
    if (found)
        ANTHOCNETMessageDebufferDestWithoutSend(node, &dataPtr->relayMessageBuffer, bufferDestEntry);
    
    found = ANTHOCNETListSearch(&bufferDestEntry, &dataPtr->messageBuffer, info->destAddr);
    
    if (found)
    {
        ANTHOCNETBufferDest* bufferDestData = (ANTHOCNETBufferDest*)bufferDestEntry->data;
        if (bufferDestData->numberOfAntGenerations  < dataPtr->reactiveAntsNumberOfAttempts)
        {
            dataPtr->tempData.bufferDestData = bufferDestData;
            dataPtr->tempData.destAddr = info->destAddr;
            dataPtr->tempData.intermediateNodeFlag = FALSE;
            sendAnts = TRUE; // Go on to next state where new event is generated.
            
            dataPtr->stats.totalReactiveForwardAntsResendings++;
     
           TRACE(stderr,"Re-send reactive forward ants ... ");
        }
        else
        {
            ANTHOCNETMessageDebufferDestWithoutSend(node, &dataPtr->messageBuffer, bufferDestEntry);
        }
    }
    MESSAGE_Free(node,msg); 

    //UserCodeEndCheckReactiveAntsRetransmission
    if (sendAnts) {
        dataPtr->state = ANTHOCNET_GENERATE_REACTIVE_FORWARD_ANTS;
        enterAnthocnetGenerateReactiveForwardAnts(node, dataPtr, msg);
        return;
    } else {
        dataPtr->state = ANTHOCNET_IDLE;
        enterAnthocnetIdle(node, dataPtr, msg);
        return;
    }
}

static void enterAnthocnetProcessUnicastWarning(Node* node,
                                                AnthocnetData* dataPtr,
                                                Message *msg) {
    //UserCodeStartProcessUnicastWarning
    char* unicastWarningPacket = MESSAGE_ReturnPacket(msg) + sizeof(ANTHOCNETAntType);
    ANTHOCNETUnicastWarning* unicastWarning = (ANTHOCNETUnicastWarning*)unicastWarningPacket;
    ANTHOCNETListEntry *RTDestEntry;
    // Remove this neighbour - destination combination
    ANTHOCNETNTRemoveEntrySimple(&dataPtr->neighbourTable, unicastWarning->nextHopAddr, 
                                                        unicastWarning->destAddr);
    if (ANTHOCNET_PRINT_PHEROMONE_CHANGE) ANTHOCNETPrintPheromone(node, dataPtr, unicastWarning->destAddr, ANTHOCNET_UNICAST_WARNING);
    // Throw away all packets already routed for that destination with this next hop
    NetworkIpDeleteOutboundPacketsToANode(node,dataPtr->currentInterfaceAddress,//unicastWarning->nextHopAddr,
                                                        unicastWarning->destAddr,FALSE);
    // Send Link Failure Notification if there are no other nextHops for this destination and I am not the source:
    if (!NetworkIpIsMyIP(node,unicastWarning->srcAddr)) //(unicastWarning->srcAddr != dataPtr->myAddress)
    {
        if (ANTHOCNETListSearch(&RTDestEntry, &dataPtr->routingTable, unicastWarning->destAddr))
        {
            if (((ANTHOCNETRTDest*)RTDestEntry->data)->numberOfConfirmedEntries == 0)
            {
                ANTHOCNETSendSimpleLinkFailureNotification(node,dataPtr,unicastWarning->destAddr);
            }
        }
        else
        {
            ANTHOCNETSendSimpleLinkFailureNotification(node,dataPtr,unicastWarning->destAddr);
        }
    }
    MESSAGE_Free(node,msg);

    //UserCodeEndProcessUnicastWarning
    dataPtr->state = ANTHOCNET_IDLE;
    enterAnthocnetIdle(node, dataPtr, msg);
}

static void enterAnthocnetProcessHello(Node* node,
                                       AnthocnetData* dataPtr,
                                       Message *msg) {
    //UserCodeStartProcessHello
    int i;
    float SINR;
    ANTHOCNETListEntry *RTDestEntry, *NTNbrEntry;
    NodeAddress sender = dataPtr->currentDefaultAddress;
    char* helloPacket = MESSAGE_ReturnPacket(msg);
    ANTHOCNETHello *hello = (ANTHOCNETHello*)(helloPacket + sizeof(ANTHOCNETAntType));
    ANTHOCNETHelloEntry *helloEntryArray = (ANTHOCNETHelloEntry*)(helloPacket + sizeof(ANTHOCNETAntType) + sizeof(ANTHOCNETHello));
    BOOLEAN activeDest = FALSE;
    if (hello->numberOfEntries >= ANTHOCNET_ACTIVE_DESTINATION)
    {
        activeDest = TRUE;
        hello->numberOfEntries -= ANTHOCNET_ACTIVE_DESTINATION;
    }
    if (ANTHOCNETListSearch(&NTNbrEntry,&dataPtr->neighbourTable,sender))
    {
        SINR = ((ANTHOCNETNTNbr*)NTNbrEntry->data)->SINR;
    }
    else
    {
        SINR = dataPtr->currentSINR;
    }
    // THE FOLLOWING TWO FUNCTIONS SHOULD BE CALLED TOGETHER!
    ANTHOCNETRefreshTimeStampsAfterHello(dataPtr, sender, getSimTime(node));
    RTDestEntry = ANTHOCNETUpdateRTEntry(dataPtr, sender, sender, getSimTime(node),1, node, 
                                      ANTHOCNETGetVirtualSendTime(node,
                                                                  dataPtr->currentInterfaceIndex,
                                                                  MESSAGE_ReturnPacketSize(msg)), 
                                      ANTHOCNETCalculatePheromoneOneHop(node,dataPtr,0,SINR,
                                                                  dataPtr->currentInterfaceIndex,
                                                                  MESSAGE_ReturnPacketSize(msg),1,
                                                                  ANTHOCNETGetVirtualSendTime(node,
                                                                  dataPtr->currentInterfaceIndex,
                                                                  MESSAGE_ReturnPacketSize(msg))),
                                      FROM_HELLO, dataPtr->currentInterfaceIndex,
                                      dataPtr->currentInterfaceAddress);
                       
    if (activeDest)
        ANTHOCNETUpdateRTDestLastUsed(node,dataPtr,RTDestEntry,sender);
    ANTHOCNETRTUpdateBootstrapping(node,dataPtr,helloEntryArray,hello,sender,
                                   dataPtr->currentInterfaceIndex, 
                                   dataPtr->currentInterfaceAddress, 
                                   MESSAGE_ReturnPacketSize(msg));
    MESSAGE_Free(node,msg);

    //UserCodeEndProcessHello
    dataPtr->state = ANTHOCNET_IDLE;
    enterAnthocnetIdle(node, dataPtr, msg);
}

static void enterAnthocnetPostNodeInitialisation(Node* node,
                                                 AnthocnetData* dataPtr,
                                                 Message *msg) {
    //UserCodeStartPostNodeInitialisation
    if (dataPtr->checkPathsInterval < TIME_ReturnMaxSimClock(node))
    {
      AppInfo *appList = node->appData.appPtr;
      for (;appList != NULL;appList = appList->appNext)
      {

          if (appList->appType == APP_CBR_CLIENT)
          {
              ANTHOCNETStartCheckPaths* startCP;
              ANTHOCNETEndCheckPaths* endCP;
              Message* checkPathsStartMsg = MESSAGE_Alloc(node, NETWORK_LAYER, 
                                                          ROUTING_PROTOCOL_ANTHOCNET,
                                                          ANTHOCNET_CheckPaths);
              Message* checkPathsEndMsg = MESSAGE_Alloc(node, NETWORK_LAYER, 
                                                          ROUTING_PROTOCOL_ANTHOCNET,
                                                          ANTHOCNET_EndCheckPaths);
              AppDataCbrClient *cbrClient = (AppDataCbrClient*) appList->appDetail;
              FILE* fp;
              char fn[100];
              sprintf(fn,"anthocnet_%d_%d_%g.paths",
                         MAPPING_GetNodeIdFromInterfaceAddress(node,GetIPv4Address(cbrClient->localAddr)),
                         MAPPING_GetNodeIdFromInterfaceAddress(node,GetIPv4Address(cbrClient->remoteAddr)),
                         simTimeSec(cbrClient->sessionStart));
              fp = fopen(fn,"w");
              MESSAGE_InfoAlloc(node, checkPathsStartMsg, sizeof(ANTHOCNETStartCheckPaths));
              startCP = (ANTHOCNETStartCheckPaths*)MESSAGE_ReturnInfo(checkPathsStartMsg);
              startCP->destAddr = GetIPv4Address(cbrClient->remoteAddr);
              startCP->fp = fp;
              MESSAGE_InfoAlloc(node, checkPathsEndMsg, sizeof(ANTHOCNETEndCheckPaths));
              endCP = (ANTHOCNETEndCheckPaths*)MESSAGE_ReturnInfo(checkPathsEndMsg);
              endCP->timerMsg = checkPathsStartMsg;
              endCP->fp = fp;
              MESSAGE_Send(node,checkPathsStartMsg,cbrClient->sessionStart + 
                                                   dataPtr->checkPathsInterval -
                                                   getSimTime(node));
              MESSAGE_Send(node,checkPathsEndMsg,cbrClient->endTime +
                                                 dataPtr->checkPathsInterval -
                                                 getSimTime(node));
          }
          
      }
    
      // Start check positions:
      if (node->nodeId == 1)
      {
          Message* checkPositionsMsg = MESSAGE_Alloc(node,NETWORK_LAYER,ROUTING_PROTOCOL_ANTHOCNET,
                                                              ANTHOCNET_CheckPositions);
          // We assume checkPathsInterval is longer than the current time:
          MESSAGE_Send(node,checkPositionsMsg,dataPtr->checkPathsInterval - getSimTime(node)); // (dataPtr->checkPathsInterval / 2)
      }
    }

    //UserCodeEndPostNodeInitialisation
    dataPtr->state = ANTHOCNET_IDLE;
    enterAnthocnetIdle(node, dataPtr, msg);
}

static void enterAnthocnetSendHello(Node* node,
                                    AnthocnetData* dataPtr,
                                    Message *msg) {

    //UserCodeStartSendHello
   Message* helloMsg = MESSAGE_Alloc(node, NETWORK_LAYER, ROUTING_PROTOCOL_ANTHOCNET,MSG_DEFAULT);
   char *packetContents = NULL;
   PacketBuffer* packetBuffer;
   
   ANTHOCNETAntType antType = ANTHOCNET_HELLO;
   ANTHOCNETHello hello;
   ANTHOCNETHelloEntry *helloEntryArray;
   int helloArrayByteSize;
   BOOLEAN activeDest;

   hello.numberOfEntries = ANTHOCNETChooseHelloEntries(node,dataPtr,&helloEntryArray,&activeDest);

   helloArrayByteSize = hello.numberOfEntries * sizeof(ANTHOCNETHelloEntry);
   if (activeDest)
       hello.numberOfEntries += ANTHOCNET_ACTIVE_DESTINATION;
   packetBuffer = BUFFER_AllocatePacketBuffer(sizeof(ANTHOCNETAntType)+sizeof(ANTHOCNETHello)+
                                              helloArrayByteSize,0,FALSE,&packetContents);
   memcpy(packetContents,&antType,sizeof(ANTHOCNETAntType));
   packetContents += sizeof(ANTHOCNETAntType);
   memcpy(packetContents,&hello,sizeof(ANTHOCNETHello));
   packetContents += sizeof(ANTHOCNETHello);
   if ((hello.numberOfEntries != 0) && (hello.numberOfEntries != ANTHOCNET_ACTIVE_DESTINATION))
       memcpy(packetContents,helloEntryArray,helloArrayByteSize);
   MESSAGE_PacketAlloc(node,helloMsg,BUFFER_GetCurrentSize(packetBuffer),TRACE_ANTHOCNET);
   memcpy(MESSAGE_ReturnPacket(helloMsg),BUFFER_GetData(packetBuffer),
                                         BUFFER_GetCurrentSize(packetBuffer));
   ANTHOCNETBroadcastControlMessage(node, helloMsg, IPTOS_PREC_INTERNETCONTROL, 0, 5); // (3 * MILLI_SECOND)
   
   MESSAGE_Send(node,msg,dataPtr->helloInterval);
   if ((hello.numberOfEntries != 0) && (hello.numberOfEntries != ANTHOCNET_ACTIVE_DESTINATION))
       MEM_free(helloEntryArray);

    //UserCodeEndSendHello
    dataPtr->state = ANTHOCNET_IDLE;
    enterAnthocnetIdle(node, dataPtr, msg);
}

static void enterAnthocnetProcessCheckPathsAnt(Node* node,
                                               AnthocnetData* dataPtr,
                                               Message *msg) {

    //UserCodeStartProcessCheckPathsAnt
    int i;
    ANTHOCNETListEntry *RTDestEntry, *RTDestNextHopEntry;
    BOOL found = FALSE, forwarded = FALSE, loopPrinted = FALSE;
    ANTHOCNETCheckPathsAnt* oldCheckPathsAnt = (ANTHOCNETCheckPathsAnt*)MESSAGE_ReturnInfo(msg);
    NodeAddress destAddr = oldCheckPathsAnt->destinationAddress;
    NodeAddress prevAddr = oldCheckPathsAnt->previousAddress;
    NodeAddress sourceAddr = oldCheckPathsAnt->sourceAddress;
    int oldNumberOfVisitedNodes = oldCheckPathsAnt->numberOfVisitedNodes;
    FILE* fp = oldCheckPathsAnt->fp;
    ANTHOCNETCheckPathsVisitedNode* oldCheckPathsVisitedNodes = 
                    (ANTHOCNETCheckPathsVisitedNode*)(MESSAGE_ReturnInfo(msg) + sizeof(ANTHOCNETCheckPathsAnt));
    
    if (destAddr  != dataPtr->myAddress)
    {
        found = ANTHOCNETListSearch(&RTDestEntry, &dataPtr->routingTable, destAddr);
        if (found)
        {
            double totalPheromone = 
                        ANTHOCNETTotalPheromone(&((ANTHOCNETRTDest*)RTDestEntry->data)->nextHops, 
                                             prevAddr,dataPtr->dataRoutingExponent);
            for (RTDestNextHopEntry = ((ANTHOCNETRTDest*)RTDestEntry->data)->nextHops.head.next;
                 RTDestNextHopEntry != NULL;
                 RTDestNextHopEntry = RTDestNextHopEntry->next)
             {
                 if (((ANTHOCNETRTDestNextHop*)RTDestNextHopEntry->data)->pheromone) // instead of only bootstrapped e.g.
                 {
                     BOOL loop = FALSE;
                     for (i=0;i<oldNumberOfVisitedNodes;i++)
                     {
                         if (RTDestNextHopEntry->nodeAddress == oldCheckPathsVisitedNodes[i].nodeAddress)
                         {
                             loop = TRUE;
                             break;
                         }
                     }
                     if (!loop)
                     {
                         ANTHOCNETCheckPathsAnt* checkPathsAnt;
                         ANTHOCNETCheckPathsVisitedNode* checkPathsVisitedNodes;
                         NodeAddress destId = MAPPING_GetNodeIdFromInterfaceAddress(node,
                                                                RTDestNextHopEntry->nodeAddress);
                         Node *destNode = MAPPING_GetNodePtrFromHash(node->partitionData->nodeIdHash,
                                                                     destId);    
                         Message *checkPathsMsg = MESSAGE_Alloc(node, NETWORK_LAYER, 
                                                                ROUTING_PROTOCOL_ANTHOCNET,
                                                                ANTHOCNET_CheckPathsAnt);
                         MESSAGE_InfoAlloc(node,checkPathsMsg,sizeof(ANTHOCNETCheckPathsAnt) +
                                             ((oldNumberOfVisitedNodes+1) * 
                                              sizeof(ANTHOCNETCheckPathsVisitedNode)));
                         checkPathsAnt = (ANTHOCNETCheckPathsAnt*)MESSAGE_ReturnInfo(checkPathsMsg);
                         checkPathsAnt->destinationAddress = destAddr;
                         checkPathsAnt->sourceAddress = sourceAddr;
                         checkPathsAnt->previousAddress = dataPtr->myAddress;
                         checkPathsAnt->numberOfVisitedNodes = oldNumberOfVisitedNodes + 1;
                         checkPathsAnt->fp = fp;
                         checkPathsVisitedNodes = (ANTHOCNETCheckPathsVisitedNode*)(MESSAGE_ReturnInfo(checkPathsMsg) + 
                                                    sizeof(ANTHOCNETCheckPathsAnt));
                         memcpy(checkPathsVisitedNodes,oldCheckPathsVisitedNodes,
                                oldNumberOfVisitedNodes * sizeof(ANTHOCNETCheckPathsVisitedNode));
                         checkPathsVisitedNodes[oldNumberOfVisitedNodes].nodeAddress = 
                                                            RTDestNextHopEntry->nodeAddress;
                         checkPathsVisitedNodes[oldNumberOfVisitedNodes].probability = 
                                    pow(((ANTHOCNETRTDestNextHop*)RTDestNextHopEntry->data)->pheromone,
                                        dataPtr->dataRoutingExponent) / totalPheromone;
                         assert(destNode!= NULL);
                         MESSAGE_Send(destNode,checkPathsMsg,0);
                         forwarded = TRUE;
                     }
                     else // loop
                     {
                         double p = 1;
                         fprintf(fp,"[%g] Loop: ",simTimeSec(getSimTime(node)));
                         for (i=0;i<oldNumberOfVisitedNodes;i++)
                         {
                             fprintf(fp,"%d(%g) ", MAPPING_GetNodeIdFromInterfaceAddress(node, 
                                                       oldCheckPathsVisitedNodes[i].nodeAddress),
                                                   oldCheckPathsVisitedNodes[i].probability);
                             p *= oldCheckPathsVisitedNodes[i].probability;
                         }
                         if (totalPheromone == 0) totalPheromone = 1; // totalPheromone is 0 if only next hop is previous address
                         p *= pow(((ANTHOCNETRTDestNextHop*)RTDestNextHopEntry->data)->pheromone,
                                        dataPtr->dataRoutingExponent) / totalPheromone;
                         fprintf(fp,"%d(%g) - %g\n", 
                                    MAPPING_GetNodeIdFromInterfaceAddress(node,
                                                                RTDestNextHopEntry->nodeAddress),
                                    pow(((ANTHOCNETRTDestNextHop*)RTDestNextHopEntry->data)->pheromone,
                                        dataPtr->dataRoutingExponent) / totalPheromone,p);
                         loopPrinted = TRUE;
                     }
                 }
             }
        }
        if (!(forwarded || loopPrinted))
        {
             double p = 1;
             fprintf(fp,"[%g] Broken: ",simTimeSec(getSimTime(node)));
             for (i=0;i<oldNumberOfVisitedNodes;i++)
             {
                 fprintf(fp,"%d(%g) ", MAPPING_GetNodeIdFromInterfaceAddress(node, 
                                               oldCheckPathsVisitedNodes[i].nodeAddress),
                                      oldCheckPathsVisitedNodes[i].probability);
                 p *= oldCheckPathsVisitedNodes[i].probability;
             }
             fprintf(fp," - %g\n",p);
        }
    }
    else // I am the destination
    {
         double p = 1;
         fprintf(fp,"[%g] Complete: ",simTimeSec(getSimTime(node)));
         for (i=0;i<oldNumberOfVisitedNodes;i++)
         {
             fprintf(fp,"%d(%g) ", MAPPING_GetNodeIdFromInterfaceAddress(node, 
                                           oldCheckPathsVisitedNodes[i].nodeAddress),
                                  oldCheckPathsVisitedNodes[i].probability);
             p *= oldCheckPathsVisitedNodes[i].probability;
         }
         fprintf(fp," - %g\n",p);
    }

    MESSAGE_Free(node, msg);

    //UserCodeEndProcessCheckPathsAnt
    dataPtr->state = ANTHOCNET_IDLE;
    enterAnthocnetIdle(node, dataPtr, msg);
}

static void enterAnthocnetEndCheckPaths(Node* node,
                                        AnthocnetData* dataPtr,
                                        Message *msg) {
    //UserCodeStartEndCheckPaths
    ANTHOCNETEndCheckPaths* endCP = (ANTHOCNETEndCheckPaths*)MESSAGE_ReturnInfo(msg);    
    MESSAGE_CancelSelfMsg(node,endCP->timerMsg);
    fclose(endCP->fp);
    MESSAGE_Free(node, msg);

    //UserCodeEndEndCheckPaths
    dataPtr->state = ANTHOCNET_IDLE;
    enterAnthocnetIdle(node, dataPtr, msg);
}

static void enterAnthocnetCheckPaths(Node* node,
                                     AnthocnetData* dataPtr,
                                     Message *msg) {
    //UserCodeStartCheckPaths
    ANTHOCNETStartCheckPaths* startCP = (ANTHOCNETStartCheckPaths*)MESSAGE_ReturnInfo(msg);
    ANTHOCNETCheckPathsAnt* checkPathsAnt;
    ANTHOCNETCheckPathsVisitedNode* checkPathsVisitedNode;
    Message *checkPathsMsg = MESSAGE_Alloc(node, NETWORK_LAYER, ROUTING_PROTOCOL_ANTHOCNET,
                                           ANTHOCNET_CheckPathsAnt);    
    MESSAGE_InfoAlloc(node,checkPathsMsg,sizeof(ANTHOCNETCheckPathsAnt) +
                                         sizeof(ANTHOCNETCheckPathsVisitedNode));
    checkPathsAnt = (ANTHOCNETCheckPathsAnt*)MESSAGE_ReturnInfo(checkPathsMsg);
    checkPathsAnt->destinationAddress = startCP->destAddr;
    checkPathsAnt->sourceAddress = dataPtr->myAddress;
    checkPathsAnt->previousAddress = ANY_IP;
    checkPathsAnt->numberOfVisitedNodes = 1;

// for printPathsAtAntReception
    if (startCP->fp == NULL)
    {
        checkPathsAnt->fp = stderr;
    }
    else
    {
        checkPathsAnt->fp = startCP->fp;
    }
// for printPathsAtAntReception end

    checkPathsVisitedNode = (ANTHOCNETCheckPathsVisitedNode*)(MESSAGE_ReturnInfo(checkPathsMsg) + 
                            sizeof(ANTHOCNETCheckPathsAnt));
    checkPathsVisitedNode->nodeAddress = dataPtr->myAddress;
    checkPathsVisitedNode->probability = 1;
    MESSAGE_Send(node,checkPathsMsg,0);
// for printPathsAtAntReception
    if (startCP->fp == NULL)
    {
        MESSAGE_Free(node, msg);
    }
    else
    {
        MESSAGE_Send(node,msg,dataPtr->checkPathsInterval);
    }
// for printPathsAtAntReception end

    //UserCodeEndCheckPaths
    dataPtr->state = ANTHOCNET_IDLE;
    enterAnthocnetIdle(node, dataPtr, msg);
}

static void enterAnthocnetCheckPositions(Node* node,
                                         AnthocnetData* dataPtr,
                                         Message *msg) {
    //UserCodeStartCheckPositions
    if ((node->nodeId == 1)&&(dataPtr->checkPathsInterval < TIME_ReturnMaxSimClock(node)))
    {
        int i;
        fprintf(dataPtr->positionsFp,"[%g]\n",simTimeSec(getSimTime(node)));
        fprintf(dataPtr->positionsFp,"%d %.2lf %.2lf %.3f\n",1,
                            node->mobilityData->current->position.common.c1,
                            node->mobilityData->current->position.common.c2,
                            ANTHOCNETGetBusyRatio(node,0));
        for (i=2;i<node->numNodes+1;i++)
        {
            Node *node1 = MAPPING_GetNodePtrFromHash(node->partitionData->nodeIdHash,i);
            fprintf(dataPtr->positionsFp,"%d %.2lf %.2lf %.3f\n",i,
                            node1->mobilityData->current->position.common.c1,
                            node1->mobilityData->current->position.common.c2,
                            ANTHOCNETGetBusyRatio(node1,0));
        }
    }   

    MESSAGE_Send(node,msg,dataPtr->checkPathsInterval); // dataPtr->checkPathsInterval/2

    //UserCodeEndCheckPositions
    dataPtr->state = ANTHOCNET_IDLE;
    enterAnthocnetIdle(node, dataPtr, msg);
}

static void enterAnthocnetIdle(Node* node,
                               AnthocnetData* dataPtr,
                               Message *msg) {

    //UserCodeStartIdle

    //UserCodeEndIdle

}

static void enterAnthocnetCleanUpRoutingTable(Node* node,
                                              AnthocnetData* dataPtr,
                                              Message *msg) {
    //UserCodeStartCleanUpRoutingTable
     ANTHOCNETListEntry *connEntry;
     ANTHOCNETListEntry* NTNbrEntry;
     ANTHOCNETListEntry* NTNbrDestEntry;
     ANTHOCNETNTNbrDest* NTNbrDestData;
     ANTHOCNETNTNbr* NTNbrData;
     ANTHOCNETRTDestNextHop* RTDestNextHopData;
     ANTHOCNETRTDest* RTDestData;
     ANTHOCNETListEntry *RTDestEntry, *TempRTDestEntry;
     ANTHOCNETListEntry* RTDestNextHopEntry;
     ANTHOCNETListEntry *TempNTNbrDestEntry, *TempNTNbrEntry;
     ANTHOCNETList* nextHops;
     BOOL found;
     int i = 0;
     int *dataPacketCounter;
     clocktype jitter;
     if (dataPtr->jitterRange >= NANO_SECOND) 
         jitter = RANDOM_nrand(dataPtr->seed) % (dataPtr->jitterRange);
     else
     {
         jitter = 0;
         RANDOM_nrand(dataPtr->seed);
     }
     
     for (NTNbrEntry = dataPtr->neighbourTable.head.next;NTNbrEntry != NULL;
            NTNbrEntry = NTNbrEntry->next)
     {
         if (((ANTHOCNETNTNbr*)NTNbrEntry->data)->lastUpdate < 
                                (getSimTime(node) - dataPtr->RTCleanupInterval))
         {
             char* infoPtr;
             BOOL *fromCleanup;
             NodeAddress *nextHopInfo;
             Message* linkFailureMsg = MESSAGE_Alloc(node,NETWORK_LAYER, 
                                                     ROUTING_PROTOCOL_ANTHOCNET, 
                                                     ANTHOCNET_LinkFailureNotification);
             MESSAGE_InfoAlloc(node, linkFailureMsg, sizeof(BOOL) + sizeof(NodeAddress));
             infoPtr = MESSAGE_ReturnInfo(linkFailureMsg);
             fromCleanup = (BOOL *)infoPtr;
             *fromCleanup = TRUE;
             infoPtr += sizeof(BOOL);
             nextHopInfo = (NodeAddress *)infoPtr;    
             *nextHopInfo = NTNbrEntry->nodeAddress;
             MESSAGE_Send(node,linkFailureMsg,i*jitter);
             i++;
             dataPtr->stats.totalHelloLinkBreaks++;
         }
     }

     MESSAGE_Send(node,msg,dataPtr->RTCleanupInterval);

     connEntry = dataPtr->currentConnectionsInfo.head.prev;
     for (RTDestEntry = dataPtr->routingTable.head.next;RTDestEntry != NULL;
            RTDestEntry = RTDestEntry->next)
     {
         RTDestData = (ANTHOCNETRTDest*)RTDestEntry->data;
         if (RTDestData->nextHops.numberOfEntries == 0)
         {
             if (RTDestEntry->nodeAddress != dataPtr->myAddress)
             {
                 if ((RTDestData->waitingSince < (getSimTime(node) - ANTHOCNET_MAX_WAITING_TIME)) ||
                     (RTDestData->waitingSince == 0))
                 {
                     TempRTDestEntry = RTDestEntry;
                     RTDestEntry = TempRTDestEntry->prev;
                     ANTHOCNETRTRemoveEmptyDest(TempRTDestEntry, &dataPtr->routingTable);
                 }
             }
             else
             {
                 if ((RTDestData->waitingSince < (getSimTime(node) - ANTHOCNET_MAX_WAITING_TIME)) &&
                      (RTDestData->lastUsed < (getSimTime(node) - ANTHOCNET_MAX_UNUSED_TIME)))
                 {
                     TempRTDestEntry = RTDestEntry;
                     RTDestEntry = TempRTDestEntry->prev;
                     ANTHOCNETRTRemoveEmptyDest(TempRTDestEntry, &dataPtr->routingTable);
                 }
             }
         }
     }

    //UserCodeEndCleanUpRoutingTable
    dataPtr->state = ANTHOCNET_IDLE;
    enterAnthocnetIdle(node, dataPtr, msg);
}

static void enterAnthocnetRouteRepairTimeOut(Node* node,
                                             AnthocnetData* dataPtr,
                                             Message *msg) {
    //UserCodeStartRouteRepairTimeOut
    ANTHOCNETLostPheromone* lostPheromone = (ANTHOCNETLostPheromone*)MESSAGE_ReturnInfo(msg);
    int n = MESSAGE_ReturnInfoSize(msg) / sizeof(ANTHOCNETLostPheromone);
    ANTHOCNETLostPheromone* tempNotRepaired = (ANTHOCNETLostPheromone*)MEM_malloc2(__LINE__,n * sizeof(ANTHOCNETLostPheromone));
    short notRepairedNumber = 0;
    int i;
    ANTHOCNETListEntry* bufferEntry;
    BOOL found;

    for (i=0;i<n;i++)
    {
        found = ANTHOCNETListSearch(&bufferEntry, &dataPtr->relayMessageBuffer, 
                                        lostPheromone[i].destAddr);
        if (found)
        {
            ANTHOCNETMessageDebufferDestWithoutSend(node, &dataPtr->relayMessageBuffer, bufferEntry);
            tempNotRepaired[notRepairedNumber++] = lostPheromone[i];
        }
    }
    dataPtr->stats.totalRoutesRepaired += n - notRepairedNumber;
    dataPtr->stats.totalRoutesUnrepaired += notRepairedNumber;

    if (notRepairedNumber > 0)
    {
        ANTHOCNETLostPheromone* notRepaired;
        PacketBuffer* packetBuffer;
        Message* msgNew;
        notRepaired = (ANTHOCNETLostPheromone*)MEM_malloc2(__LINE__,notRepairedNumber * sizeof(ANTHOCNETLostPheromone));
        memcpy(notRepaired,tempNotRepaired,notRepairedNumber * sizeof(ANTHOCNETLostPheromone));

        packetBuffer = ANTHOCNETBuildLinkFailureAntPacket(ANTHOCNET_LINK_FAILURE_NOTIFICATION,
                                           notRepaired,notRepairedNumber);
        ANTHOCNETEncapsulatePacketInMessage(node, &msgNew, packetBuffer);
        
        MEM_free(notRepaired);
        
        /* Send the link failure message (ttl at the moment set to 1) */
        ANTHOCNETBroadcastControlMessage(node, msgNew, IPTOS_PREC_INTERNETCONTROL, 0, 1);
        dataPtr->stats.totalLinkFailureNotificationsGenerated++;
    }
    MEM_free(tempNotRepaired);
    MESSAGE_Free(node,msg);  

    //UserCodeEndRouteRepairTimeOut
    dataPtr->state = ANTHOCNET_IDLE;
    enterAnthocnetIdle(node, dataPtr, msg);
}

static void enterAnthocnetCheckProactiveAnts(Node* node,
                                             AnthocnetData* dataPtr,
                                             Message *msg) {
    //UserCodeStartCheckProactiveAnts
    ANTHOCNETListEntry *connEntry;
    ANTHOCNETListEntry *RTDestEntry = dataPtr->routingTable.head.prev;
    ANTHOCNETRTDest *RTDestData;
    ANTHOCNETCurrentConnectionsData *ccdata;
    for (connEntry = dataPtr->currentConnectionsInfo.head.next;
         connEntry != NULL;connEntry = connEntry->next)
    {
        ccdata = (ANTHOCNETCurrentConnectionsData*)connEntry->data;
        if (ccdata->counter > 0)
        {
            if (ANTHOCNETListSearchFrom(&RTDestEntry, RTDestEntry, connEntry->nodeAddress))
            {
                RTDestData = (ANTHOCNETRTDest*)RTDestEntry->data;
                if (RTDestData->numberOfConfirmedEntries > 0)
                {
                    if (ANTHOCNETBetterBootstrappedPheromoneAvailable(node,dataPtr,&RTDestData->nextHops))
                    {
                        Message* newMsg = MESSAGE_Alloc(node, NETWORK_LAYER, 
                                                    ROUTING_PROTOCOL_ANTHOCNET, 
                                                    ANTHOCNET_SendProactiveAnt);
                        MESSAGE_InfoAlloc(node, newMsg, sizeof(NodeAddress));
                        *(NodeAddress*)MESSAGE_ReturnInfo(newMsg) = RTDestEntry->nodeAddress;
                        ccdata->counter = 0;
                        MESSAGE_Send(node,newMsg,1);
                    }
                }
                else
                {
                    ccdata->counter = 0;
                }
            }
        }
    }
    MESSAGE_Send(node,msg,dataPtr->proactiveAntSendInterval);

    //UserCodeEndCheckProactiveAnts
    dataPtr->state = ANTHOCNET_IDLE;
    enterAnthocnetIdle(node, dataPtr, msg);
}


//Statistic Function Definitions
static void AnthocnetInitStats(Node* node, AnthocnetStatsType *stats) { 
    if (node->guiOption) {
        stats->totalReactiveAntsGeneratedId = GUI_DefineMetric("Total Reactive Ants Generated", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalReactiveBackwardAntsReceivedId = GUI_DefineMetric("Total Reactive BAnts Received", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalLoopingAntsId = GUI_DefineMetric("Total Looping Ants", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalAntsForwardedId = GUI_DefineMetric("Total Forwarded Ants", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalReactiveForwardAntsResendingsId = GUI_DefineMetric("Total Sending Attempts for Forward Ants", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalLinkFailureNotificationsGeneratedId = GUI_DefineMetric("Total Link Failure Notifications Generated", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalProactiveAntsGeneratedId = GUI_DefineMetric("Total Proactive Ants Generated", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalRepairAntsGeneratedId = GUI_DefineMetric("Total Repair Ants Generated", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalRoutesRepairedId = GUI_DefineMetric("Total Routes Repaired", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalRoutesUnrepairedId = GUI_DefineMetric("Total Routes Unrepaired", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalCriticalLinkFailuresId = GUI_DefineMetric("Total Critical Link Failures", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalReactiveInterUnacceptedAntsId = GUI_DefineMetric("Total Reactive Inter Unaccepted Ants", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalReactiveInterUnacceptedHopsId = GUI_DefineMetric("Total Reactive Inter Unaccepted Hops", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalReactiveDestUnacceptedAntsId = GUI_DefineMetric("Total Reactive Dest Unaccepted Ants", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalReactiveDestUnacceptedHopsId = GUI_DefineMetric("Total Reactive Dest Unaccepted Hops", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalReactiveAcceptedAntsId = GUI_DefineMetric("Total Reactive Accepted Ants", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalReactiveAcceptedHopsId = GUI_DefineMetric("Total Reactive Accepted Hops", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalProactiveInterUnacceptedAntsId = GUI_DefineMetric("Total Proactive Inter Unaccepted Ants", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalProactiveInterUnacceptedHopsId = GUI_DefineMetric("Total Proactive Inter Unaccepted Hops", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalProactiveDestUnacceptedAntsId = GUI_DefineMetric("Total Proactive Dest Unaccepted Ants", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalProactiveDestUnacceptedHopsId = GUI_DefineMetric("Total Proactive Dest Unaccepted Hops", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalProactiveAcceptedAntsId = GUI_DefineMetric("Total Proactive Accepted Ants", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalProactiveAcceptedHopsId = GUI_DefineMetric("Total Proactive Accepted Hops", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalRepairInterUnacceptedAntsId = GUI_DefineMetric("Total Repair Inter Unaccepted Ants", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalRepairInterUnacceptedHopsId = GUI_DefineMetric("Total Repair Inter Unaccepted Hops", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalRepairDestUnacceptedAntsId = GUI_DefineMetric("Total Repair Dest Unaccepted Ants", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalRepairDestUnacceptedHopsId = GUI_DefineMetric("Total Repair Dest Unaccepted Hops", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalRepairAcceptedAntsId = GUI_DefineMetric("Total Repair Accepted Ants", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalRepairAcceptedHopsId = GUI_DefineMetric("Total Repair Accepted Hops", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalLoopingHopsId = GUI_DefineMetric("Total Looping Hops", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->dropDataPacketsNoPheromoneId = GUI_DefineMetric("Drop Data Packets No Pheromone", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalPacketsDeliveredId = GUI_DefineMetric("Total Packets Delivered", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalHopsDeliveredId = GUI_DefineMetric("Total Hops Delivered", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->avgHopsNumberId = GUI_DefineMetric("Average Hops Number", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_DOUBLE_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalTtlKilledAntsId = GUI_DefineMetric("Total Ttl Killed Ants", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalProactiveDeadEndAntsId = GUI_DefineMetric("Total Proactive Dead End Ants", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalProactiveDeadEndHopsId = GUI_DefineMetric("Total Proactive Dead End Hops", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalProactiveAntSendFailuresId = GUI_DefineMetric("Total Proactive Ant Send Failures", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalProactiveLoopingAntsId = GUI_DefineMetric("Total Proactive Looping Ants", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalProactiveLoopingHopsId = GUI_DefineMetric("Total Proactive Looping Hops", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->dropDataPacketsMaxTtlId = GUI_DefineMetric("Drop Data Packets Max Ttl", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->dropDataPacketsLoopId = GUI_DefineMetric("Drop Data Packets Loop", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->totalHelloLinkBreaksId = GUI_DefineMetric("Total Hello Link Breaks", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->packetsDroppedAtDebufferingId = GUI_DefineMetric("Packets Dropped At Debuffering", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
        stats->packetsDroppedAtDebufferWithoutSendId = GUI_DefineMetric("Packets Dropped At Debuffer Without Send", node->nodeId, GUI_NETWORK_LAYER, 0, GUI_INTEGER_TYPE,GUI_CUMULATIVE_METRIC);
    }

    stats->totalReactiveAntsGenerated = 0;
    stats->totalReactiveBackwardAntsReceived = 0;
    stats->totalLoopingAnts = 0;
    stats->totalAntsForwarded = 0;
    stats->totalReactiveForwardAntsResendings = 0;
    stats->totalLinkFailureNotificationsGenerated = 0;
    stats->totalProactiveAntsGenerated = 0;
    stats->totalRepairAntsGenerated = 0;
    stats->totalRoutesRepaired = 0;
    stats->totalRoutesUnrepaired = 0;
    stats->totalCriticalLinkFailures = 0;
    stats->totalReactiveInterUnacceptedAnts = 0;
    stats->totalReactiveInterUnacceptedHops = 0;
    stats->totalReactiveDestUnacceptedAnts = 0;
    stats->totalReactiveDestUnacceptedHops = 0;
    stats->totalReactiveAcceptedAnts = 0;
    stats->totalReactiveAcceptedHops = 0;
    stats->totalProactiveInterUnacceptedAnts = 0;
    stats->totalProactiveInterUnacceptedHops = 0;
    stats->totalProactiveDestUnacceptedAnts = 0;
    stats->totalProactiveDestUnacceptedHops = 0;
    stats->totalProactiveAcceptedAnts = 0;
    stats->totalProactiveAcceptedHops = 0;
    stats->totalRepairInterUnacceptedAnts = 0;
    stats->totalRepairInterUnacceptedHops = 0;
    stats->totalRepairDestUnacceptedAnts = 0;
    stats->totalRepairDestUnacceptedHops = 0;
    stats->totalRepairAcceptedAnts = 0;
    stats->totalRepairAcceptedHops = 0;
    stats->totalLoopingHops = 0;
    stats->dropDataPacketsNoPheromone = 0;
    stats->totalPacketsDelivered = 0;
    stats->totalHopsDelivered = 0;
    stats->avgHopsNumber = 0;
    stats->totalTtlKilledAnts = 0;
    stats->totalProactiveDeadEndAnts = 0;
    stats->totalProactiveDeadEndHops = 0;
    stats->totalProactiveAntSendFailures = 0;
    stats->totalProactiveLoopingAnts = 0;
    stats->totalProactiveLoopingHops = 0;
    stats->dropDataPacketsMaxTtl = 0;
    stats->dropDataPacketsLoop = 0;
    stats->totalHelloLinkBreaks = 0;
    stats->packetsDroppedAtDebuffering = 0;
    stats->packetsDroppedAtDebufferWithoutSend = 0;
}

void AnthocnetPrintStats(Node* node) {
    AnthocnetStatsType *stats = &((AnthocnetData *) NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_ANTHOCNET))->stats;
    char buf[MAX_STRING_LENGTH];

    char statProtocolName[MAX_STRING_LENGTH];
    strcpy(statProtocolName,"ANTHOCNET");
    sprintf(buf, "Total Reactive Ants Generated = %d", stats->totalReactiveAntsGenerated);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Reactive BAnts Received = %d", stats->totalReactiveBackwardAntsReceived);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Looping Ants = %d", stats->totalLoopingAnts);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Forwarded Ants = %d", stats->totalAntsForwarded);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Sending Attempts for Forward Ants = %d", stats->totalReactiveForwardAntsResendings);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Link Failure Notifications Generated = %d", stats->totalLinkFailureNotificationsGenerated);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Proactive Ants Generated = %d", stats->totalProactiveAntsGenerated);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Repair Ants Generated = %d", stats->totalRepairAntsGenerated);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Routes Repaired = %d", stats->totalRoutesRepaired);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Routes Unrepaired = %d", stats->totalRoutesUnrepaired);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Critical Link Failures = %d", stats->totalCriticalLinkFailures);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Reactive Inter Unaccepted Ants = %d", stats->totalReactiveInterUnacceptedAnts);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Reactive Inter Unaccepted Hops = %d", stats->totalReactiveInterUnacceptedHops);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Reactive Dest Unaccepted Ants = %d", stats->totalReactiveDestUnacceptedAnts);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Reactive Dest Unaccepted Hops = %d", stats->totalReactiveDestUnacceptedHops);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Reactive Accepted Ants = %d", stats->totalReactiveAcceptedAnts);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Reactive Accepted Hops = %d", stats->totalReactiveAcceptedHops);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Proactive Inter Unaccepted Ants = %d", stats->totalProactiveInterUnacceptedAnts);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Proactive Inter Unaccepted Hops = %d", stats->totalProactiveInterUnacceptedHops);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Proactive Dest Unaccepted Ants = %d", stats->totalProactiveDestUnacceptedAnts);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Proactive Dest Unaccepted Hops = %d", stats->totalProactiveDestUnacceptedHops);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Proactive Accepted Ants = %d", stats->totalProactiveAcceptedAnts);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Proactive Accepted Hops = %d", stats->totalProactiveAcceptedHops);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Repair Inter Unaccepted Ants = %d", stats->totalRepairInterUnacceptedAnts);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Repair Inter Unaccepted Hops = %d", stats->totalRepairInterUnacceptedHops);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Repair Dest Unaccepted Ants = %d", stats->totalRepairDestUnacceptedAnts);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Repair Dest Unaccepted Hops = %d", stats->totalRepairDestUnacceptedHops);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Repair Accepted Ants = %d", stats->totalRepairAcceptedAnts);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Repair Accepted Hops = %d", stats->totalRepairAcceptedHops);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Looping Hops = %d", stats->totalLoopingHops);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Drop Data Packets No Pheromone = %d", stats->dropDataPacketsNoPheromone);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Packets Delivered = %d", stats->totalPacketsDelivered);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Hops Delivered = %d", stats->totalHopsDelivered);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Average Hops Number = %f", stats->avgHopsNumber);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Ttl Killed Ants = %d", stats->totalTtlKilledAnts);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Proactive Dead End Ants = %d", stats->totalProactiveDeadEndAnts);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Proactive Dead End Hops = %d", stats->totalProactiveDeadEndHops);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Proactive Ant Send Failures = %d", stats->totalProactiveAntSendFailures);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Proactive Looping Ants = %d", stats->totalProactiveLoopingAnts);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Proactive Looping Hops = %d", stats->totalProactiveLoopingHops);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Drop Data Packets Max Ttl = %d", stats->dropDataPacketsMaxTtl);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Drop Data Packets Loop = %d", stats->dropDataPacketsLoop);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Total Hello Link Breaks = %d", stats->totalHelloLinkBreaks);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Packets Dropped At Debuffering = %d", stats->packetsDroppedAtDebuffering);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
    sprintf(buf, "Packets Dropped At Debuffer Without Send = %d", stats->packetsDroppedAtDebufferWithoutSend);
    IO_PrintStat(node, "Network", statProtocolName, ANY_DEST, -1, buf);
}

void AnthocnetRunTimeStat(Node* node, AnthocnetData* dataPtr) {
    clocktype now = getSimTime(node);

    if (node->guiOption) {
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalReactiveAntsGeneratedId, dataPtr->stats.totalReactiveAntsGenerated, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalReactiveBackwardAntsReceivedId, dataPtr->stats.totalReactiveBackwardAntsReceived, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalLoopingAntsId, dataPtr->stats.totalLoopingAnts, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalAntsForwardedId, dataPtr->stats.totalAntsForwarded, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalReactiveForwardAntsResendingsId, dataPtr->stats.totalReactiveForwardAntsResendings, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalLinkFailureNotificationsGeneratedId, dataPtr->stats.totalLinkFailureNotificationsGenerated, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalProactiveAntsGeneratedId, dataPtr->stats.totalProactiveAntsGenerated, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalRepairAntsGeneratedId, dataPtr->stats.totalRepairAntsGenerated, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalRoutesRepairedId, dataPtr->stats.totalRoutesRepaired, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalRoutesUnrepairedId, dataPtr->stats.totalRoutesUnrepaired, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalCriticalLinkFailuresId, dataPtr->stats.totalCriticalLinkFailures, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalReactiveInterUnacceptedAntsId, dataPtr->stats.totalReactiveInterUnacceptedAnts, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalReactiveInterUnacceptedHopsId, dataPtr->stats.totalReactiveInterUnacceptedHops, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalReactiveDestUnacceptedAntsId, dataPtr->stats.totalReactiveDestUnacceptedAnts, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalReactiveDestUnacceptedHopsId, dataPtr->stats.totalReactiveDestUnacceptedHops, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalReactiveAcceptedAntsId, dataPtr->stats.totalReactiveAcceptedAnts, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalReactiveAcceptedHopsId, dataPtr->stats.totalReactiveAcceptedHops, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalProactiveInterUnacceptedAntsId, dataPtr->stats.totalProactiveInterUnacceptedAnts, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalProactiveInterUnacceptedHopsId, dataPtr->stats.totalProactiveInterUnacceptedHops, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalProactiveDestUnacceptedAntsId, dataPtr->stats.totalProactiveDestUnacceptedAnts, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalProactiveDestUnacceptedHopsId, dataPtr->stats.totalProactiveDestUnacceptedHops, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalProactiveAcceptedAntsId, dataPtr->stats.totalProactiveAcceptedAnts, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalProactiveAcceptedHopsId, dataPtr->stats.totalProactiveAcceptedHops, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalRepairInterUnacceptedAntsId, dataPtr->stats.totalRepairInterUnacceptedAnts, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalRepairInterUnacceptedHopsId, dataPtr->stats.totalRepairInterUnacceptedHops, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalRepairDestUnacceptedAntsId, dataPtr->stats.totalRepairDestUnacceptedAnts, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalRepairDestUnacceptedHopsId, dataPtr->stats.totalRepairDestUnacceptedHops, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalRepairAcceptedAntsId, dataPtr->stats.totalRepairAcceptedAnts, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalRepairAcceptedHopsId, dataPtr->stats.totalRepairAcceptedHops, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalLoopingHopsId, dataPtr->stats.totalLoopingHops, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.dropDataPacketsNoPheromoneId, dataPtr->stats.dropDataPacketsNoPheromone, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalPacketsDeliveredId, dataPtr->stats.totalPacketsDelivered, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalHopsDeliveredId, dataPtr->stats.totalHopsDelivered, now);
        GUI_SendRealData(node->nodeId, dataPtr->stats.avgHopsNumberId, dataPtr->stats.avgHopsNumber, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalTtlKilledAntsId, dataPtr->stats.totalTtlKilledAnts, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalProactiveDeadEndAntsId, dataPtr->stats.totalProactiveDeadEndAnts, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalProactiveDeadEndHopsId, dataPtr->stats.totalProactiveDeadEndHops, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalProactiveAntSendFailuresId, dataPtr->stats.totalProactiveAntSendFailures, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalProactiveLoopingAntsId, dataPtr->stats.totalProactiveLoopingAnts, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalProactiveLoopingHopsId, dataPtr->stats.totalProactiveLoopingHops, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.dropDataPacketsMaxTtlId, dataPtr->stats.dropDataPacketsMaxTtl, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.dropDataPacketsLoopId, dataPtr->stats.dropDataPacketsLoop, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.totalHelloLinkBreaksId, dataPtr->stats.totalHelloLinkBreaks, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.packetsDroppedAtDebufferingId, dataPtr->stats.packetsDroppedAtDebuffering, now);
        GUI_SendIntegerData(node->nodeId, dataPtr->stats.packetsDroppedAtDebufferWithoutSendId, dataPtr->stats.packetsDroppedAtDebufferWithoutSend, now);
    }
}

void AnthocnetInit(Node* node, AnthocnetData** dataPtr, const NodeInput* nodeInput, int interfaceIndex) {

    Message *msg = NULL;

    //UserCodeStartInit
    char buf[MAX_STRING_LENGTH];
    char paramRole;
    double paramFixedNodePheromoneFactor;
    double paramSINRTreshold;
    double paramSINRPheromoneFactor;
    double paramSpeedPenalty;
    BOOL paramAdmissionControlFlag;
    double paramAdmissionControlHops;
    double paramAdmissionControlTime;
    BOOL paramUseAversion;
    BOOL paramUseSpeed;
    BOOL paramUseLabels;
    BOOL paramUsePower;
    BOOL paramUseTimeOldWay;
    BOOL paramUseTime;
    BOOL paramUseSINR;
    BOOL paramUseBusyRatio;
    BOOL paramBidirectionalPaths;
    double paramDiscountQTime;
    double paramDiscountSendTime;
    BOOL paramTraceTime;
    BOOL paramTracePackets;
    BOOL paramNotificationsAfterCleanup;
    BOOL wasFound;
    clocktype paramReactiveAntsRetryTimeInterval;
    int paramMessageBufferSize;
    int paramReactiveAntsNumberOfAttempts;
    double paramDataRoutingExponent;
    clocktype paramJitterRange;
    double paramProactiveAntRoutingExponent;
    clocktype paramProactiveAntSendInterval;
    double paramRepairTimeFactor;
    int paramRepairMaxBroadcasting;
    int paramMaxTtl;
    int paramMaxHelloEntries;
    clocktype paramRepairTimePerHop;
    clocktype paramMaxBufferTime;
    clocktype paramHelloInterval;
    int i, paramAlgorithmSeeds;
    double paramProactiveTriggerPercentage;
    double paramTimeHopsWeight;
    clocktype paramCheckPathsInterval;
    double paramDiscountBootstrappedPheromone;
    double paramDiscountBootstrappedVirtualPheromone;
    int paramReSend;
    double paramBusyLimit1;
    double paramBusyLimit2;
    double paramBusyFactor1;
    double paramBusyFactor2;

    Message* cleanupTimerMsg = MESSAGE_Alloc(node,NETWORK_LAYER, ROUTING_PROTOCOL_ANTHOCNET,
                                                            ANTHOCNET_RTCleanup);
    
    Message* helloTimerMsg = MESSAGE_Alloc(node, NETWORK_LAYER, ROUTING_PROTOCOL_ANTHOCNET,
                                                            ANTHOCNET_SendHello);
    Message* postNodeInitTimerMsg = MESSAGE_Alloc(node, NETWORK_LAYER, ROUTING_PROTOCOL_ANTHOCNET,
                                                            ANTHOCNET_PostNodeInit);
    Message* checkProactiveAntsMsg = MESSAGE_Alloc(node,NETWORK_LAYER,ROUTING_PROTOCOL_ANTHOCNET,
                                                            ANTHOCNET_CheckProactiveAnts);


    IO_ReadDouble(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-FIXED-NODE-PHEROMONE-FACTOR", &wasFound, &paramFixedNodePheromoneFactor);

    if (!wasFound) {
        paramFixedNodePheromoneFactor = 0.5;
    }
    
    IO_ReadDouble(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-BUSY-LIMIT1", &wasFound, &paramBusyLimit1);

    if (!wasFound) {
        paramBusyLimit1 = 1;
    }
    
    IO_ReadDouble(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-BUSY-LIMIT2", &wasFound, &paramBusyLimit2);

    if (!wasFound) {
        paramBusyLimit2 = 1;
    }
    
    IO_ReadDouble(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-BUSY-FACTOR1", &wasFound, &paramBusyFactor1);

    if (!wasFound) {
        paramBusyFactor1 = 1;
    }
    
    IO_ReadDouble(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-BUSY-FACTOR2", &wasFound, &paramBusyFactor2);

    if (!wasFound) {
        paramBusyFactor2 = 1;
    }
    
    IO_ReadDouble(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-SINR-TRESHOLD", &wasFound, &paramSINRTreshold);

    if (!wasFound) {
        paramSINRTreshold = 50;
    } 

    IO_ReadDouble(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-SINR-PHEROMONE-FACTOR", &wasFound, &paramSINRPheromoneFactor);

    if (!wasFound) {
        paramSINRPheromoneFactor = 3;
    } 

    IO_ReadDouble(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-SPEED-PENALTY", &wasFound, &paramSpeedPenalty);

    if (!wasFound) {
        paramSpeedPenalty = 2;
    } 

    IO_ReadDouble(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-DISCOUNT-Q-TIME", &wasFound, &paramDiscountQTime);

    if (!wasFound) {
        paramDiscountQTime = 0.7;
    } 

    IO_ReadDouble(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-DISCOUNT-SEND-TIME", &wasFound, &paramDiscountSendTime);

    if (!wasFound) {
        paramDiscountSendTime = 0.7;
    } 

    IO_ReadDouble(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-DISCOUNT-BOOTSTRAPPED-VIRTUAL-PHEROMONE", &wasFound, &paramDiscountBootstrappedVirtualPheromone);

    if (!wasFound) {
        paramDiscountBootstrappedVirtualPheromone = 0.0;
    } 

    IO_ReadDouble(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-DISCOUNT-BOOTSTRAPPED-PHEROMONE", &wasFound, &paramDiscountBootstrappedPheromone);

    if (!wasFound) {
        paramDiscountBootstrappedPheromone = 0.0;
    } 
                                                            
    IO_ReadDouble(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-TIME-HOPS-WEIGHT", &wasFound, &paramTimeHopsWeight);

    if (!wasFound) {
        paramTimeHopsWeight = 0;
    } 
    
    IO_ReadDouble(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-PROACTIVE-TRIGGER-PERCENTAGE", &wasFound, &paramProactiveTriggerPercentage);

    if (!wasFound) {
        paramProactiveTriggerPercentage = 0.1;
    }   

    IO_ReadDouble(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-PROACTIVE-ANT-ROUTING-EXPONENT", &wasFound, &paramProactiveAntRoutingExponent);

    if (!wasFound) {
        paramProactiveAntRoutingExponent = 20;
    }   

    IO_ReadDouble(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-REPAIR-TIME-FACTOR", &wasFound, &paramRepairTimeFactor);

    if (!wasFound) {
        paramRepairTimeFactor = 5;
    }   

    IO_ReadDouble(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-DATA-ROUTING-EXPONENT", &wasFound, &paramDataRoutingExponent);

    if (!wasFound) {
        paramDataRoutingExponent = 20.0;
    }   

    IO_ReadTime(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-MAX-BUFFER-TIME", &wasFound, &paramMaxBufferTime);

    if (!wasFound) {
        paramMaxBufferTime = 1 * SECOND;
    }

    IO_ReadTime(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-PROACTIVE-ANT-SEND-INTERVAL", &wasFound, &paramProactiveAntSendInterval);

    if (!wasFound) {
        paramProactiveAntSendInterval = 2 * SECOND;
    }

    IO_ReadTime(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-REPAIR-TIME-PER-HOP", &wasFound, &paramRepairTimePerHop);

    if (!wasFound) {
        paramRepairTimePerHop = (clocktype)(0.05 * SECOND);
    }
    
    IO_ReadTime(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-JITTER-RANGE", &wasFound, &paramJitterRange);

    if (!wasFound) {
        paramJitterRange = 10 * MILLI_SECOND; 
    }
    
    IO_ReadTime(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-HELLO-INTERVAL", &wasFound, &paramHelloInterval);
    
    if (!wasFound) {
        paramHelloInterval = 1 * SECOND;
    }

    IO_ReadInt(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-MAX-HELLO-ENTRIES", &wasFound, &paramMaxHelloEntries);

    if (!wasFound) {
        paramMaxHelloEntries = 10;
    }

    IO_ReadInt(node->nodeId, ANY_ADDRESS, nodeInput, "ALGORITHM-SEEDS", &wasFound, &paramAlgorithmSeeds);

    if (!wasFound) {
        paramAlgorithmSeeds = 10;
    }

    IO_ReadTime(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-REACTIVE-ANTS-RETRY-INTERVAL", &wasFound, &paramReactiveAntsRetryTimeInterval);

    if (!wasFound) {
        paramReactiveAntsRetryTimeInterval = 1 * SECOND;
    }
    
    IO_ReadTime(node->nodeId, ANY_ADDRESS, nodeInput, "CHECK-PATHS-INTERVAL", &wasFound, &paramCheckPathsInterval);

    if (!wasFound) {
        paramCheckPathsInterval = TIME_ReturnMaxSimClock(node);
    }
    
    IO_ReadInt(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-MAX-TTL", &wasFound, &paramMaxTtl);
    
    if(!wasFound) {
        paramMaxTtl = 20;
    }

    IO_ReadInt(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-REPAIR-MAX-BROADCASTING", &wasFound, &paramRepairMaxBroadcasting);
    
    if(!wasFound) {
        paramRepairMaxBroadcasting = 2;
    }

   IO_ReadInt(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-PACKET-BUFFER-SIZE", &wasFound, &paramMessageBufferSize);

   if (!wasFound) {
        paramMessageBufferSize = 50;
   }
   
   IO_ReadInt(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-REACTIVE-ANTS-NUMBER-OF-ATTEMPTS", &wasFound, &paramReactiveAntsNumberOfAttempts);

   if (!wasFound) {
        paramReactiveAntsNumberOfAttempts = 3;
   }
   
   IO_ReadInt(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-TRACE-TIME", &wasFound, &paramTraceTime);

   if (!wasFound) {
        paramTraceTime = FALSE;
   }
   
   IO_ReadInt(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-TRACE-PACKETS", &wasFound, &paramTracePackets);

   if (!wasFound) {
        paramTracePackets = FALSE;
   }

   IO_ReadInt(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-NOTIFICATIONS-AFTER-CLEANUP", &wasFound, &paramNotificationsAfterCleanup);

   if (!wasFound) {
        paramNotificationsAfterCleanup = TRUE;
   }

   IO_ReadInt(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-USE-OLD-WAY-TIME", &wasFound, &paramUseTimeOldWay);

   if (!wasFound) {
        paramUseTimeOldWay = FALSE;
   }

   IO_ReadInt(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-USE-TIME", &wasFound, &paramUseTime);

   if (!wasFound) {
        paramUseTime = FALSE;
   }

   IO_ReadInt(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-USE-SINR", &wasFound, &paramUseSINR);

   if (!wasFound) {
        paramUseSINR = TRUE;
   }
 
   IO_ReadInt(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-USE-BUSY-RATIO", &wasFound, &paramUseBusyRatio);

   if (!wasFound) {
        paramUseBusyRatio = FALSE;
   }
 
   IO_ReadInt(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-USE-LABELS", &wasFound, &paramUseLabels);

   if (!wasFound) {
        paramUseLabels = FALSE;
   }
 
   IO_ReadInt(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-USE-AVERSION", &wasFound, &paramUseAversion);

   if (!wasFound) {
        paramUseAversion = FALSE;
   }
 
   IO_ReadInt(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-USE-SPEED", &wasFound, &paramUseSpeed);

   if (!wasFound) {
        paramUseSpeed = FALSE;
   }
 
   IO_ReadInt(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-BIDIRECTIONAL-PATHS", &wasFound, &paramBidirectionalPaths);

   if (!wasFound) {
        paramBidirectionalPaths = FALSE;
   }
 
   IO_ReadInt(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-USE-POWER", &wasFound, &paramUsePower);

   if (!wasFound) {
        paramUsePower = FALSE;
   }

   IO_ReadInt(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-RESEND", &wasFound, &paramReSend);

   if (!wasFound) {
        paramReSend = 1;
   }

   IO_ReadInt(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-ADMISSION-CONTROL-FLAG", &wasFound, &paramAdmissionControlFlag);

   if (!wasFound) {
        paramAdmissionControlFlag = 0;
   }

   IO_ReadDouble(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-ADMISSION-CONTROL-HOPS", &wasFound, &paramAdmissionControlHops);

   if (!wasFound) {
        paramAdmissionControlHops = 1000;
   }

   IO_ReadDouble(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-ADMISSION-CONTROL-TIME", &wasFound, &paramAdmissionControlTime);

   if (!wasFound) {
        paramAdmissionControlTime = 1000;
   }

   IO_ReadString(node->nodeId, ANY_ADDRESS, nodeInput, "ANTHOCNET-LABELS-FILE", &wasFound, buf);
   if (wasFound) {
       int id = 0;
       FILE *labelsFp;
       char line[100];
       labelsFp = fopen(buf,"rb");
       while (id != node->nodeId) {
           fgets(line, 100, labelsFp);
           sscanf(line, "%d %c", &id, &paramRole);
       }
   } else {
       paramRole = 'm';
   }

   (*dataPtr) = (ANTHOCNETData*) MEM_malloc2(__LINE__,sizeof(ANTHOCNETData));
   mem_error(*dataPtr, "Cannot allocate dataPtr!!");
   memset((*dataPtr), 0, sizeof(ANTHOCNETData)); 
   
   if (paramAlgorithmSeeds <= 0)
       paramAlgorithmSeeds = 10;
   RANDOM_SetSeed((*dataPtr)->seed, node->globalSeed + paramAlgorithmSeeds, node->nodeId, ROUTING_PROTOCOL_ANTHOCNET, interfaceIndex);
   
   (*dataPtr)->nodeRole = paramRole;

   (*dataPtr)->msgIdsIndex = 0;
   for (i=0;i<ANTHOCNET_NUMBER_OF_MSG_IDS;i++)
   {
       (*dataPtr)->msgIds[i].seqNumber = -1;
   }
   
   (*dataPtr)->myAddress = NetworkIpGetInterfaceAddress(node, DEFAULT_INTERFACE);
   
   (*dataPtr)->interfaceAddresses = (NodeAddress*)MEM_malloc2(__LINE__,node->numberInterfaces*sizeof(NodeAddress));
   for (i = 0; i < node->numberInterfaces; i++)
   {
       (*dataPtr)->interfaceAddresses[i] = NetworkIpGetInterfaceAddress(node,i);
   }
    
   (*dataPtr)->maxPower = 3;

   (*dataPtr)->initStats = TRUE;

   (*dataPtr)->printPathsAtAntReception = FALSE;

   (*dataPtr)->reSend = (paramReSend > 0);
   
   (*dataPtr)->admissionControlFlag = paramAdmissionControlFlag;
   (*dataPtr)->admissionControlHops = paramAdmissionControlHops;
   (*dataPtr)->admissionControlTime = paramAdmissionControlTime;

   (*dataPtr)->fixedNodePheromoneFactor = paramFixedNodePheromoneFactor;
   (*dataPtr)->busyFactor1 = paramBusyFactor1;
   (*dataPtr)->busyFactor2 = paramBusyFactor2;
   (*dataPtr)->busyLimit1 = paramBusyLimit1;
   (*dataPtr)->busyLimit2 = paramBusyLimit2;
   (*dataPtr)->bidirectionalPaths = paramBidirectionalPaths;
   (*dataPtr)->useLabels = paramUseLabels;
   (*dataPtr)->useAversion = paramUseAversion;
   (*dataPtr)->useSpeed = paramUseSpeed;
   (*dataPtr)->speedPenalty = paramSpeedPenalty;
   (*dataPtr)->useBusyRatio = paramUseBusyRatio;
   (*dataPtr)->usePower = paramUsePower;
   (*dataPtr)->useTimeOldWay = paramUseTimeOldWay;
   (*dataPtr)->useTime = paramUseTime;
   (*dataPtr)->useSINR = paramUseSINR;
   (*dataPtr)->traceTime = paramTraceTime;   
   (*dataPtr)->tracePackets = paramTracePackets;
   (*dataPtr)->notificationsAfterCleanup = paramNotificationsAfterCleanup;

   (*dataPtr)->SINRTreshold = paramSINRTreshold;
   (*dataPtr)->SINRPheromoneFactor = paramSINRPheromoneFactor;

   (*dataPtr)->discountPheromone = 0.7;
   (*dataPtr)->discountBootstrappedPheromone = paramDiscountBootstrappedPheromone;
   (*dataPtr)->discountBootstrappedVirtualPheromone = paramDiscountBootstrappedVirtualPheromone;
   (*dataPtr)->discountHops = (*dataPtr)->discountTripTime = (*dataPtr)->discountPheromone;
   (*dataPtr)->timeHopsWeight = paramTimeHopsWeight;
   
   (*dataPtr)->checkPathsInterval = paramCheckPathsInterval;
   
   (*dataPtr)->helloInterval = paramHelloInterval;
   (*dataPtr)->allowedHelloLoss = 2;
   (*dataPtr)->RTCleanupInterval = (*dataPtr)->helloInterval * (*dataPtr)->allowedHelloLoss;
   
   (*dataPtr)->proactiveTriggerPercentage = paramProactiveTriggerPercentage;
   (*dataPtr)->maxHelloEntries = paramMaxHelloEntries;
   (*dataPtr)->maxTtl = paramMaxTtl;
   (*dataPtr)->maxBufferTime = paramMaxBufferTime;
   (*dataPtr)->repairTimeFactor = paramRepairTimeFactor;
   (*dataPtr)->repairTimePerHop = paramRepairTimePerHop;
   (*dataPtr)->repairMaxBroadcasting = paramRepairMaxBroadcasting;
   (*dataPtr)->proactiveAntRoutingExponent = paramProactiveAntRoutingExponent;
   (*dataPtr)->proactiveAntSendInterval = paramProactiveAntSendInterval;
   (*dataPtr)->jitterRange = paramJitterRange;
   (*dataPtr)->dataRoutingExponent = paramDataRoutingExponent;
   (*dataPtr)->reactiveAntsRetryInterval = paramReactiveAntsRetryTimeInterval;
   (*dataPtr)->reactiveAntsNumberOfAttempts = paramReactiveAntsNumberOfAttempts;
   (*dataPtr)->antGenerationNumber = 1; // 0 is used in the routing table to indicate "uninitialised", or FROM_HELLO
   ANTHOCNETListInit(&((*dataPtr)->routingTable));
   ANTHOCNETListInit(&((*dataPtr)->neighbourTable));
   ANTHOCNETListInit(&((*dataPtr)->messageBuffer));
   ANTHOCNETListInit(&((*dataPtr)->currentConnectionsInfo));
   (*dataPtr)->messageBuffer.data = (ANTHOCNETBufferData*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETBufferData));
   mem_error((*dataPtr)->messageBuffer.data, "Cannot allocate data buffer");
   ((ANTHOCNETBufferData*)((*dataPtr)->messageBuffer.data))->maxBufferSize = paramMessageBufferSize;
   ((ANTHOCNETBufferData*)((*dataPtr)->messageBuffer.data))->currentBufferSize = 0;
   ANTHOCNETListInit(&((*dataPtr)->relayMessageBuffer));
   (*dataPtr)->relayMessageBuffer.data = (ANTHOCNETBufferData*)MEM_malloc2(__LINE__,sizeof(ANTHOCNETBufferData));
   mem_error((*dataPtr)->relayMessageBuffer.data,"Cannot allocate relay buffer");
   ((ANTHOCNETBufferData*)((*dataPtr)->relayMessageBuffer.data))->maxBufferSize = paramMessageBufferSize;
   ((ANTHOCNETBufferData*)((*dataPtr)->relayMessageBuffer.data))->currentBufferSize = 0;
   
   (*dataPtr)->tempHelloEntryArray = (ANTHOCNETHelloEntry*)MEM_malloc2(__LINE__,(*dataPtr)->maxHelloEntries * 
                                                 sizeof(ANTHOCNETHelloEntry));
   
   NetworkIpSetRouterFunction(node, &AnthocnetRouterFunction,interfaceIndex);

   NetworkIpSetMacLayerStatusEventHandlerFunction(node,ANTHOCNETMACLayerStatusEventHandler,
                                                    interfaceIndex);

    (*dataPtr)->defaultSize = 64; // used in macackreceived to call getvirtualsendtime
                                  // is actually pointless: the packet size is only used to calculate the time in wired environment, but macackreceived will not be called if 802.11 is not used
    {
       int i;
       (*dataPtr)->avgQTimeDbl = (double*)MEM_malloc2(__LINE__,node->numberInterfaces * sizeof(double));
       (*dataPtr)->discountQTime = paramDiscountQTime;
       (*dataPtr)->discountSendTime = paramDiscountSendTime;

       for (i = 0; i < node->numberInterfaces; i++)
       {
           (*dataPtr)->avgQTimeDbl[i] = 0;
           if (node->macData[i]->macProtocol == MAC_PROTOCOL_802_11)
           {
               Mac802_11InitAvgSendTimeANTS01(node,i,(clocktype)ANTHOCNET_HOP_LATENCY,(*dataPtr)->discountSendTime);
           }
       }
   }
                                                    
   MESSAGE_Send(node,cleanupTimerMsg,RANDOM_erand((*dataPtr)->seed) * (*dataPtr)->RTCleanupInterval);
   MESSAGE_Send(node,helloTimerMsg,RANDOM_erand((*dataPtr)->seed) * (*dataPtr)->helloInterval);
   MESSAGE_Send(node,postNodeInitTimerMsg,1 * NANO_SECOND);
   MESSAGE_Send(node,checkProactiveAntsMsg,RANDOM_erand((*dataPtr)->seed) * (*dataPtr)->proactiveAntSendInterval);

   (*dataPtr)->stats.totalTtlKilledAnts = 0;
   (*dataPtr)->stats.totalReactiveAntsGenerated = 0;
   (*dataPtr)->stats.totalReactiveBackwardAntsReceived = 0;
   (*dataPtr)->stats.totalLinkFailureNotificationsGenerated = 0;
   (*dataPtr)->stats.totalAntsForwarded = 0;
   (*dataPtr)->stats.totalReactiveForwardAntsResendings = 0;
   (*dataPtr)->stats.totalLoopingAnts = 0;
   (*dataPtr)->stats.totalLoopingHops = 0;
   (*dataPtr)->stats.totalProactiveAntsGenerated = 0;
   (*dataPtr)->stats.totalRepairAntsGenerated = 0;
   (*dataPtr)->stats.totalRoutesRepaired = 0;
   (*dataPtr)->stats.totalRoutesUnrepaired = 0;
   (*dataPtr)->stats.totalCriticalLinkFailures = 0;
   (*dataPtr)->stats.totalReactiveInterUnacceptedAnts = 0;
   (*dataPtr)->stats.totalReactiveInterUnacceptedHops = 0;
   (*dataPtr)->stats.totalReactiveDestUnacceptedAnts = 0;
   (*dataPtr)->stats.totalReactiveDestUnacceptedHops = 0;
   (*dataPtr)->stats.totalReactiveAcceptedAnts = 0;
   (*dataPtr)->stats.totalReactiveAcceptedHops = 0;
   (*dataPtr)->stats.totalProactiveInterUnacceptedAnts = 0;
   (*dataPtr)->stats.totalProactiveInterUnacceptedHops = 0;
   (*dataPtr)->stats.totalProactiveDestUnacceptedAnts = 0;
   (*dataPtr)->stats.totalProactiveDestUnacceptedHops = 0;
   (*dataPtr)->stats.totalProactiveAcceptedAnts = 0;
   (*dataPtr)->stats.totalProactiveAcceptedHops = 0;
   (*dataPtr)->stats.totalRepairInterUnacceptedAnts = 0;
   (*dataPtr)->stats.totalRepairInterUnacceptedHops = 0;
   (*dataPtr)->stats.totalRepairDestUnacceptedAnts = 0;
   (*dataPtr)->stats.totalRepairDestUnacceptedHops = 0;
   (*dataPtr)->stats.totalRepairAcceptedAnts = 0;
   (*dataPtr)->stats.totalRepairAcceptedHops = 0;
   (*dataPtr)->stats.dropDataPacketsNoPheromone = 0;
   (*dataPtr)->stats.totalPacketsDelivered = 0;
   (*dataPtr)->stats.totalHopsDelivered = 0;
   (*dataPtr)->stats.totalProactiveDeadEndAnts = 0;
   (*dataPtr)->stats.totalProactiveDeadEndHops = 0;
   (*dataPtr)->stats.totalProactiveAntSendFailures = 0;
   (*dataPtr)->stats.totalProactiveLoopingAnts = 0;
   (*dataPtr)->stats.totalProactiveLoopingHops = 0;
   (*dataPtr)->stats.dropDataPacketsMaxTtl = 0;
   (*dataPtr)->stats.dropDataPacketsLoop = 0;
   (*dataPtr)->stats.totalHelloLinkBreaks = 0;
   (*dataPtr)->stats.packetsDroppedAtDebuffering = 0;
   (*dataPtr)->stats.packetsDroppedAtDebufferWithoutSend = 0;
  
   // print out node positions
   if ((node->nodeId == 1)&&((*dataPtr)->checkPathsInterval < TIME_ReturnMaxSimClock(node))) {
       (*dataPtr)->positionsFp = fopen("anthocnet_positions.dat","w");
       fprintf((*dataPtr)->positionsFp,"%ld %g %.2lf %.2lf %.2lf %.2lf\n",node->numNodes,
                simTimeSec(node->partitionData->maxSimClock),
                node->partitionData->terrainData.origin.common.c1,
                node->partitionData->terrainData.dimensions.common.c1,
                node->partitionData->terrainData.origin.common.c2,
                node->partitionData->terrainData.dimensions.common.c2);
   }
       

   (*dataPtr)->printStats = TRUE;
   
    /* The following line of code should be executed only
    // for the *first* initialization of an interface that
    // shares this data structure.*/
    (*dataPtr)->statsPrinted = FALSE;
    //UserCodeEndInit
    if ((*dataPtr)->initStats) {
        AnthocnetInitStats(node, &((*dataPtr)->stats));
    }
    (*dataPtr)->state = ANTHOCNET_IDLE;
    enterAnthocnetIdle(node, *dataPtr, msg);
}


void AnthocnetFinalize(Node* node) {

    AnthocnetData *dataPtr = (AnthocnetData*) NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_ANTHOCNET);

    //UserCodeStartFinalize
    if (dataPtr->stats.totalPacketsDelivered > 0)
    {
        dataPtr->stats.avgHopsNumber = dataPtr->stats.totalHopsDelivered / 
                                        (double)dataPtr->stats.totalPacketsDelivered;
    }
    else
    {
        dataPtr->stats.avgHopsNumber = 0;
    }
    
   if ((node->nodeId == 1)&&(dataPtr->checkPathsInterval < TIME_ReturnMaxSimClock(node)))
        fclose(dataPtr->positionsFp);
        
    MEM_free(dataPtr->tempHelloEntryArray);

    //UserCodeEndFinalize

    if (dataPtr->statsPrinted) {
        return;
    }

    if (dataPtr->printStats) {
        dataPtr->statsPrinted = TRUE;
        AnthocnetPrintStats(node);
    }
}


void AnthocnetProcessEvent(Node* node,   Message* msg) {

    AnthocnetData* dataPtr = (AnthocnetData*)  NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_ANTHOCNET);

    int event = msg->eventType;
    switch (dataPtr->state) {
        case ANTHOCNET_GENERATE_REACTIVE_FORWARD_ANTS:
            assert(FALSE);
            break;
        case ANTHOCNET_PROCESS_REACTIVE_FORWARD_ANT:
            assert(FALSE);
            break;
        case ANTHOCNET_PROCESS_REPAIR_BACKWARD_ANT:
            assert(FALSE);
            break;
        case ANTHOCNET_PROCESS_REACTIVE_BACKWARD_ANT:
            assert(FALSE);
            break;
        case ANTHOCNET_PROCESS_REPAIR_FORWARD_ANT:
            assert(FALSE);
            break;
        case ANTHOCNET_PROCESS_LINK_FAILURE_NOTIFICATION:
            assert(FALSE);
            break;
        case ANTHOCNET_PROCESS_PROACTIVE_FORWARD_ANT:
            assert(FALSE);
            break;
        case ANTHOCNET_HANDLE_LINK_FAILURE:
            assert(FALSE);
            break;
        case ANTHOCNET_SEND_PROACTIVE_ANT:
            assert(FALSE);
            break;
        case ANTHOCNET_CHECK_REACTIVE_ANTS_RETRANSMISSION:
            assert(FALSE);
            break;
        case ANTHOCNET_PROCESS_UNICAST_WARNING:
            assert(FALSE);
            break;
        case ANTHOCNET_PROCESS_HELLO:
            assert(FALSE);
            break;
        case ANTHOCNET_POST_NODE_INITIALISATION:
            assert(FALSE);
            break;
        case ANTHOCNET_SEND_HELLO:
            assert(FALSE);
            break;
        case ANTHOCNET_PROCESS_CHECK_PATHS_ANT:
            assert(FALSE);
            break;
        case ANTHOCNET_END_CHECK_PATHS:
            assert(FALSE);
            break;
        case ANTHOCNET_CHECK_PATHS:
            assert(FALSE);
            break;
        case ANTHOCNET_CHECK_POSITIONS:
            assert(FALSE);
            break;
        case ANTHOCNET_IDLE:
            switch(event) {
                case ANTHOCNET_ReactiveAntsTimer:
                    dataPtr->state = ANTHOCNET_CHECK_REACTIVE_ANTS_RETRANSMISSION;
                    enterAnthocnetCheckReactiveAntsRetransmission(node, dataPtr, msg);
                    break;
                case ANTHOCNET_LinkFailureNotification:
                    dataPtr->state = ANTHOCNET_HANDLE_LINK_FAILURE;
                    enterAnthocnetHandleLinkFailure(node, dataPtr, msg);
                    break;
                case ANTHOCNET_SendProactiveAnt:
                    dataPtr->state = ANTHOCNET_SEND_PROACTIVE_ANT;
                    enterAnthocnetSendProactiveAnt(node, dataPtr, msg);
                    break;
                case ANTHOCNET_RTCleanup:
                    dataPtr->state = ANTHOCNET_CLEAN_UP_ROUTING_TABLE;
                    enterAnthocnetCleanUpRoutingTable(node, dataPtr, msg);
                    break;
                case ANTHOCNET_RepairTimeOut:
                    dataPtr->state = ANTHOCNET_ROUTE_REPAIR_TIME_OUT;
                    enterAnthocnetRouteRepairTimeOut(node, dataPtr, msg);
                    break;
                case ANTHOCNET_SendHello:
                    dataPtr->state = ANTHOCNET_SEND_HELLO;
                    enterAnthocnetSendHello(node, dataPtr, msg);
                    break;
                case ANTHOCNET_CheckPaths:
                    dataPtr->state = ANTHOCNET_CHECK_PATHS;
                    enterAnthocnetCheckPaths(node, dataPtr, msg);
                    break;
                case ANTHOCNET_PostNodeInit:
                    dataPtr->state = ANTHOCNET_POST_NODE_INITIALISATION;
                    enterAnthocnetPostNodeInitialisation(node, dataPtr, msg);
                    break;
                case ANTHOCNET_EndCheckPaths:
                    dataPtr->state = ANTHOCNET_END_CHECK_PATHS;
                    enterAnthocnetEndCheckPaths(node, dataPtr, msg);
                    break;
                case ANTHOCNET_CheckPathsAnt:
                    dataPtr->state = ANTHOCNET_PROCESS_CHECK_PATHS_ANT;
                    enterAnthocnetProcessCheckPathsAnt(node, dataPtr, msg);
                    break;
                case ANTHOCNET_CheckPositions:
                    dataPtr->state = ANTHOCNET_CHECK_POSITIONS;
                    enterAnthocnetCheckPositions(node, dataPtr, msg);
                    break;
                case ANTHOCNET_CheckProactiveAnts:
                    dataPtr->state = ANTHOCNET_CHECK_PROACTIVE_ANTS;
                    enterAnthocnetCheckProactiveAnts(node, dataPtr, msg);
                    break;
                default:
                    assert(FALSE);
            }
            break;
        case ANTHOCNET_CLEAN_UP_ROUTING_TABLE:
            assert(FALSE);
            break;
        case ANTHOCNET_ROUTE_REPAIR_TIME_OUT:
            assert(FALSE);
            break;
        case ANTHOCNET_CHECK_PROACTIVE_ANTS:
            assert(FALSE);
            break;
        default:
            ERROR_ReportError("Illegal transition");
            break;
    }

}

