// Copyright (c) 2001-2006, Scalable Network Technologies, Inc.  All Rights Reserved.
//                          6701 Center Drive West
//                          Suite 520
//                          Los Angeles, CA 90045
//                          sales@scalable-networks.com
//
// This source code is licensed, not sold, and is subject to a written
// license agreement.  Among other things, no portion of this source
// code may be copied, transmitted, disclosed, displayed, distributed,
// translated, used as the basis for a derivative work, or used, in
// whole or in part, for any program or purpose other than its intended
// use in compliance with the license agreement as part of the QualNet
// software.  This source code and certain of the algorithms contained
// within it are confidential trade secrets of Scalable Network
// Technologies, Inc. and may not be used as the basis for any other
// software, hardware, product or service.

//
// Objectives: IP (Internet Protocol)
//             A very simple IP that does multiplexing and demultiplexing.
// References: RFC 791
// Date: 8/20/1999
//


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "api.h"
#include "partition.h"

#include "adaptation_aal5.h"
#include "ipv6.h"
#include "ip6_icmp.h"
#include "ip6_output.h"
#include "network_ip.h"
#include "network_dualip.h"
#include "network_icmp.h"
#include "multicast_static.h"
#include "multicast_igmp.h"

#include "if_queue.h"
#include "queue_red.h"
#include "queue_wred_ecn.h"
#include "queue_rio_ecn.h"

#include "sch_graph.h"
#include "sch_strictprio.h"
#include "sch_wfq.h"
#include "sch_roundrobin.h"
#include "sch_scfq.h"
#include "sch_wrr.h"
#include "resource_manager_cbq.h"

#include "transport_tcp_hdr.h"
#include "transport_udp.h"

#ifdef WIRELESS_LIB
#include "multicast_odmrp.h"
#include "network_ipv4_ndp.h"
#include "routing_aodv.h"
#include "routing_dsr.h"
#include "routing_fsrl.h"
#include "routing_iarp.h"
#include "routing_ierp.h"
#include "routing_brp.h" // brp needs to follow ierp
#include "routing_lar1.h"
#include "routing_star.h"
#include "routing_zrp.h"
#include "routing_dymo.h"
//StartANTHOCNET
#include "routing_anthocnet.h"
//EndANTHOCNET
#include "manet_packet.h"
#endif // WIRELESS_LIB

#ifdef ENTERPRISE_LIB
#include "mf_traffic_conditioner.h"
#include "mpls.h"
#include "multicast_dvmrp.h"
#include "multicast_mospf.h"
#include "multicast_pim.h"
#include "network_mobileip.h"
#include "network_access_list.h"
#include "route_parse_util.h"
#include "route_map.h"
#include "routing_eigrp.h"
#include "routing_igrp.h"
#include "routing_ospfv2.h"
#include "routing_policy_routing.h"
#include "routing_qospf.h"
#include "sch_diffserv.h"
#endif // ENTERPRISE_LIB

#ifdef MILITARY_RADIOS_LIB
#include "routing_odr.h"
#endif

#ifdef ADDON_MAODV
#include "multicast_maodv.h"
#endif // ADDON_MAODV

#ifdef CELLULAR_LIB
#include "cellular_gsm.h"
#include "layer3_gsm.h"
#include "cellular_layer3.h"
#endif // CELLULAR_LIB

#ifdef STK_INTERFACE
#include "stk_interface.h"
#endif /* STK_INTERFACE */

#ifdef ADVANCED_WIRELESS_LIB
#include "dot16_backbone.h"
#endif // ADVANCED_WIRELESS_LIB

#ifdef IPNE_INTERFACE
#include "ipnetworkemulator.h"
#endif // IPNE_INTERFACE

//InsertPatch HEADER_FILES

#ifdef MPI_DEBUG
#include "parallel.h"
#endif

//-----------------------------------------------------------------------------
// DEFINES
//-----------------------------------------------------------------------------
#define INTERFACE_DEBUG 0

#define rtDEBUG 0
//-----------------------------------------------------------------------------
// IP header
//-----------------------------------------------------------------------------

//
// Additional number of bytes to add to a source route option field,
// to make up for the fact that the option field header is 3 bytes.
// This is so alignment is preserved, making extraction of IP packet
// data easier (casts to structs instead of memcpy()'s).
// (non-standard, may be looked at -- the only supported option
// is strict source routes)
//

#define IP_SOURCE_ROUTE_OPTION_PADDING  1

//-----------------------------------------------------------------------------
// MACRO        IpHeaderHasSourceRoute()
// PURPOSE      Returns boolean depending on whether IP header has a
//              source route.  (Just calls IpHeaderSourceRouteOptionField()
//              and checks if what's returned is NULL.)
// PARAMETERS   IpHeaderType *IpHeader
//                  Pointer to IP header.
// RETURNS      true, if ipHeader has source route.
//              false, if ipHeader does not have source route.
//-----------------------------------------------------------------------------

#define IpHeaderHasSourceRoute(ipHeader) \
    (IpHeaderSourceRouteOptionField(ipHeader) != NULL)

//-----------------------------------------------------------------------------
// Per hop behavior (PHB)
//-----------------------------------------------------------------------------

#define NUM_INITIAL_PHB_INFO_ENTRIES 4

//-----------------------------------------------------------------------------
// Routing table (forwarding table)
//-----------------------------------------------------------------------------

//
// Initial allocated entries for routing table.
//

#define FORWARDING_TABLE_ROW_START_SIZE 8

//-----------------------------------------------------------------------------
// Router info
//-----------------------------------------------------------------------------

#define NUM_INITIAL_ROUTER_INFO_ENTRIES 2

//-----------------------------------------------------------------------------
// Maximum value of IP-QUEUE-NUM-PRIORITIES at any interface
//-----------------------------------------------------------------------------

#define NUM_MAX_IP_QUEUE        256

//-----------------------------------------------------------------------------
// ECN debug test
//-----------------------------------------------------------------------------

#define NoECN_DEBUG_TEST

/*
 * Change this value to the number of drops required
 * e.g. #define ECN_TEST_PKT_MARK 3 for 3 drops
 * The value should be 0 (zero) for no marks
 */
#define ECN_TEST_PKT_MARK 0

#include "external_socket.h"

void D_IpPrint::ExecuteAsString(const char *in, char *out)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    NetworkForwardingTable *rt = &ip->forwardTable;

    int i;
    char clockStr[MAX_CLOCK_STRING_LENGTH];
    char str[MAX_STRING_LENGTH];
    EXTERNAL_VarArray v;

    EXTERNAL_VarArrayInit(&v, 400);


    ctoa((getSimTime(node) / SECOND), clockStr);

    EXTERNAL_VarArrayConcatString(&v, "Forwarding Table\n");
    EXTERNAL_VarArrayConcatString(&v, "---------------------------------------------------------------"
        "--------------------\n");
    EXTERNAL_VarArrayConcatString(&v, "          dest          mask        intf       nextHop    protocol"
        "    admin    Flag\n");
    EXTERNAL_VarArrayConcatString(&v, "---------------------------------------------------------------"
        "--------------------\n");
    for (i = 0; i < rt->size; i++)
    {
        char address[20];
        IO_ConvertIpAddressToString(rt->row[i].destAddress, address);
        sprintf(str, "%15s  ", address);
        EXTERNAL_VarArrayConcatString(&v, str);
        IO_ConvertIpAddressToString(rt->row[i].destAddressMask, address);
        sprintf(str, "%15s  ", address);
        EXTERNAL_VarArrayConcatString(&v, str);
        sprintf(str, "%5u", rt->row[i].interfaceIndex);
        EXTERNAL_VarArrayConcatString(&v, str);
        IO_ConvertIpAddressToString(rt->row[i].nextHopAddress, address);
        sprintf(str, "%15s   ", address);
        EXTERNAL_VarArrayConcatString(&v, str);
        sprintf(str, "%5u      ", rt->row[i].protocolType);
        EXTERNAL_VarArrayConcatString(&v, str);
        sprintf(str, "%5u", rt->row[i].adminDistance);
        EXTERNAL_VarArrayConcatString(&v, str);

        if (rt->row[i].interfaceIsEnabled) {
            EXTERNAL_VarArrayConcatString(&v, "       U\n");
        } else {
            EXTERNAL_VarArrayConcatString(&v, "       D\n");
        }
    }

    EXTERNAL_VarArrayConcatString(&v, "\n");

    strcpy(out, v.data);
    EXTERNAL_VarArrayFree(&v);
}

//-----------------------------------------------------------------------------
// STRUCTS, ENUMS, AND TYPEDEFS
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// PROTOTYPES FOR FUNCTIONS WITH INTERNAL LINKAGE
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Packet delivery and forwarding
//-----------------------------------------------------------------------------

static void //inline//
DeliverPacket(Node *node, Message *msg,
              int interfaceIndex, NodeAddress previousHopAddress);

static void //inline//
ForwardPacket(
    Node *node,
    Message *msg,
    int incomingInterface,
    NodeAddress previousHopAddress);

static void //inline//
QueueUpIpFragmentForMacLayer(
    Node *node,
    Message *msg,
    int interfaceIndex,
    NodeAddress nextHop);

static void //inline//
ProcessDelayedSendToMac(Node *node, Message *msg);

//-----------------------------------------------------------------------------
// Source route
//-----------------------------------------------------------------------------

static IpOptionsHeaderType * //inline//
IpHeaderSourceRouteOptionField(IpHeaderType *ipHeader);

static void //inline//
SourceRouteThePacket(Node *node, Message *msg);

void
ExtractIpSourceAndRecordedRoute(
    Message *msg,
    NodeAddress RouteAddresses[],
    int *NumAddresses,
    int *RouteAddressIndex);

//-----------------------------------------------------------------------------
// Boolean utility routines for packet forwarding process
//-----------------------------------------------------------------------------

static BOOL //inline//
IsMyPacket(Node *node, NodeAddress destAddress);

BOOL
IsOutgoingBroadcast(
    Node *node,
    NodeAddress destAddress,
    int *outgoingInterface,
    NodeAddress *outgoingBroadcastAddress);

//-----------------------------------------------------------------------------
// Per hop behavior (PHB)
//-----------------------------------------------------------------------------

/*static void //inline//
IpInitPerHopBehaviors(
    Node *node,
    const NodeInput *nodeInput);*/

static void //inline//
AddPHBEntry(
    Node *node,
    unsigned char ds, //dscp value
    QueuePriorityType priority);

static unsigned //inline//
GetQueuePriorityFromUserTos(
    Node *node,
    TosType userTos,
    int numQueues);

//-----------------------------------------------------------------------------
// Callbacks into IP made by the MAC layer, helper functions
//-----------------------------------------------------------------------------

//ANTS01 begin: comment out following function declaration; moved to network_ip.h
//static void //inline//
//HandleSpecialMacLayerStatusEvents(Node *node,
//                                  Message *msg,
//                                  const NodeAddress nextHopAddress,
//                                  int interfaceIndex);
//ANTS01 end

//-----------------------------------------------------------------------------
// IP header
//-----------------------------------------------------------------------------

static void
AddIpHeader(
    Node *node,
    Message *msg,
    NodeAddress sourceAddress,
    NodeAddress destinationAddress,
    TosType priority,
    unsigned char protocol,
    unsigned ttl);

static void //inline//
ExpandOrShrinkIpHeader(
    Node *node,
    Message *msg,
    int newHeaderSize);

//-----------------------------------------------------------------------------
// IP header option field
//-----------------------------------------------------------------------------

static void //inline//
ExpandOrShrinkIpOptionField(
    Node *node,
    Message *msg,
    const int optionCode,
    const int newIpOptionSize);

//-----------------------------------------------------------------------------
// Statistics
//-----------------------------------------------------------------------------

static void //inline//
NetworkIpInitStats(Node* node, NetworkIpStatsType *stats);

void
NetworkIpPrintStats(Node *node);

//-----------------------------------------------------------------------------
// FUNCTIONS WITH EXTERNAL LINKAGE
//-----------------------------------------------------------------------------



//---------------------------------------------------------------------------
// FUNCTION   NetworkIpCheckMulticastRoutingProtocol
//
// PURPOSE    Check if this node running the given multicast routing protocol
//
// PARAMETERS node - this node.
//            routingProtcolType - the multicast protocol to get.
//            interfaceId - specific interface index or ANY_INTERFACE
//
// RETURN     TRUE if match found, FALSE otherwise.
//---------------------------------------------------------------------------
static
BOOL NetworkIpCheckMulticastRoutingProtocol(
    Node* node,
    NetworkRoutingProtocolType routingProtocolType,
    int interfaceId)
{
    int i;
    BOOL retVal = FALSE;
    NetworkDataIp* ip = (NetworkDataIp*) node->networkData.networkVar;

    if (interfaceId == ANY_INTERFACE)
    {
        for (i = 0; i < node->numberInterfaces; i++)
        {
            if (ip->interfaceInfo[i]->multicastProtocolType ==
                    routingProtocolType)
            {
                retVal =  TRUE;
                break;
            }
        }
    }
    else if (ip->interfaceInfo[interfaceId]->multicastProtocolType ==
                routingProtocolType)
    {
        retVal = TRUE;
    }

    return retVal;
}



//-----------------------------------------------------------------------------
// Init functions
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpPreInit()
// PURPOSE      IP initialization required before any of the other
//              layers are initialized.
// PARAMETERS   Node *node
//                  Pointer to node.
// RETURN       None.
// NOTES        This is mainly for MAC initialization, which requires
//              certain IP structures be pre-initialized.
//-----------------------------------------------------------------------------

void
NetworkIpPreInit(Node *node)
{
    NetworkDataIp *ip =
        (NetworkDataIp *) MEM_malloc(sizeof(NetworkDataIp));
    memset(ip, 0, sizeof(NetworkDataIp));

    int i;

    node->networkData.networkVar = ip;

    ip->packetIdCounter = 0;
    ip->checkMessagePeekFunction = FALSE;
    ip->checkMacAckHandler = FALSE;

    NetworkInitForwardingTable(node);
    NetworkInitMulticastForwardingTable(node);

    // Initialize the multicast group list
    ListInit(node, &ip->multicastGroupList);

    ip->routeUpdateFunction = NULL;

    ip->ipFragUnit = MAX_NW_PKT_SIZE;

    ip->maxPacketLength = ip->ipFragUnit;

    NetworkIpInitStats(node, &(ip->stats));

    for (i = 0; i < MAX_NUM_INTERFACES; i++)
    {
        ip->interfaceInfo[i] = NULL;
    }

    ip->reflexTimeout = -1;
    ip->ipv6 = NULL;

#ifdef ENTERPRISE_LIB
    ip->allNameValueLists = NULL;
    ip->accessListName = NULL;
    ip->nestReflex = NULL;
    ip->accessListSession = NULL;
    ip->isACLStatOn = FALSE;
    ip->accessListTrace = FALSE;
    ip->routeMapList = NULL;
    ip->bufferLastEntryRMap = NULL;
    ip->pbrTrace = FALSE;
    ip->isPBRStatOn = FALSE;
    ip->local = FALSE;
    ip->rMapForPbr = FALSE;
    ip->pbrStat.packetsPolicyRoutedLocal = 0;
    ip->pbrStat.packetsNotPolicyRouted = 0;
    ip->pbrStat.packetPrecSet = 0;
    ip->pbrStat.packetsPolicyRouted = 0;

    //Initialize for route redistribution
    ip->rtRedistributeIsEnabled = FALSE;
    ip->rtRedistributeIsEnabled = FALSE;
    ip->rtRedistributeInfo = NULL;
#endif // ENTERPRISE_LIB

    // By default Loopback is kept enabled
    ip->isLoopbackEnabled = TRUE;

    NetworkIpLoopbackForwardingTableInit(node);

    char path[MAX_STRING_LENGTH];
    D_Hierarchy *h = &node->partitionData->dynamicHierarchy;

    if (h->CreateNetworkPath(
            node,
            "ip",
            "ipInReceives",
            path))
    {
        h->AddObject(
            path,
            new D_UInt32Obj(&ip->stats.ipInReceives));
    }

    if (h->CreateNetworkPath(
            node,
            "ip",
            "ipReasmReqds",
            path))
    {
        h->AddObject(
            path,
            new D_UInt32Obj(&ip->stats.ipReasmReqds));
    }

    if (h->CreateNetworkPath(
            node,
            "ip",
            "ipFragsCreated",
            path))
    {
        h->AddObject(
            path,
            new D_UInt32Obj(&ip->stats.ipFragsCreated));
    }

    if (h->CreateNetworkPath(
            node,
            "ip",
            "ipFragUnit",
            path))
    {
        h->AddObject(
            path,
            new D_Int32Obj(&ip->ipFragUnit));
        node->partitionData->dynamicHierarchy.SetWriteable(
            path,
            FALSE);
    }

    if (h->CreateNetworkPath(
            node,
            "ip",
            "print",
            path))
    {
        D_IpPrint *ipPrint = new D_IpPrint(node);
        h->AddObject(
            path,
            ipPrint);
    }
}

// /**
// FUNCTION   :: NetworkIpPrintTraceXML
// LAYER      :: NETWORK
// PURPOSE    :: Print packet trace information in XML format
// PARAMETERS ::
// + node     : Node*    : Pointer to node
// + msg      : Message* : Pointer to packet to print headers from
// RETURN     ::  void   : NULL
// **/

void NetworkIpPrintTraceXML(Node* node, Message* msg)
{
    char buf[MAX_STRING_LENGTH];
    IpHeaderType *ipHeader = (IpHeaderType *) MESSAGE_ReturnPacket(msg);


    char addr1[20];
    char addr2[20];


    sprintf(buf, "<ipv4>");
    TRACE_WriteToBufferXML(node, buf);

    sprintf(buf, "%hu %hu %hu %hu %hu %hu %hu",
            ipHeader->ip_v,
            ipHeader->ip_hl,
            ipHeader->ip_tos >> 2,  // R SFT for ECN
            ipHeader->ip_tos & 0x00000010,  // ECN ECT
            ipHeader->ip_tos & 0x00000001,  // ECN CE
            MESSAGE_ReturnPacketSize(msg),
            ipHeader->ip_id);
    TRACE_WriteToBufferXML(node, buf);

    sprintf(buf, " <flags>%hu %hu %hu</flags>",
            ipHeader->ip_reserved,
            ipHeader->ip_dont_fragment,
            ipHeader->ip_more_fragments);
    TRACE_WriteToBufferXML(node, buf);

    IO_ConvertIpAddressToString(ipHeader->ip_src, addr1);
    IO_ConvertIpAddressToString(ipHeader->ip_dst, addr2);
    sprintf(buf, " %hu %hu %hu %hu %s %s",
            ipHeader->ip_fragment_offset,
            ipHeader->ip_ttl,
            ipHeader->ip_p,
            ipHeader->ip_sum,
            addr1,
            addr2);
    TRACE_WriteToBufferXML(node, buf);

    sprintf(buf, "</ipv4>");
    TRACE_WriteToBufferXML(node, buf);
}


// /**
// FUNCTION   :: NetworkIpInitTrace
// LAYER      :: NETWORK
// PURPOSE    :: IP initialization  for tracing
// PARAMETERS ::
// + node : Node* : Pointer to node
// + nodeInput    : const NodeInput* : Pointer to NodeInput
// RETURN ::  void : NULL
// **/

static
void NetworkIpInitTrace(Node* node, const NodeInput* nodeInput)
{
    char buf[MAX_STRING_LENGTH];
    BOOL retVal;
    BOOL traceAll = TRACE_IsTraceAll(node);
    BOOL trace = FALSE;
    static BOOL writeMap = TRUE;

    IO_ReadString(
        node->nodeId,
        ANY_ADDRESS,
        nodeInput,
        "TRACE-IP",
        &retVal,
        buf);

    if (retVal)
    {
        if (strcmp(buf, "YES") == 0)
        {
            trace = TRUE;
        }
        else if (strcmp(buf, "NO") == 0)
        {
            trace = FALSE;
        }
        else
        {
            ERROR_ReportError(
                "TRACE-IP should be either \"YES\" or \"NO\".\n");
        }
    }
    else
    {
        if (traceAll)
        {
            trace = TRUE;
        }
    }

    if (trace)
    {
            TRACE_EnableTraceXML(node, TRACE_IP,
                "IPv4", NetworkIpPrintTraceXML, writeMap);
    }
    else
    {
            TRACE_DisableTraceXML(node, TRACE_IP,
                "IPv4", writeMap);
    }
    writeMap = FALSE;
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpInit
// PURPOSE      Initialize IP variables, and all network-layer IP
//              protocols.
// PARAMETERS   Node *node
//                  Pointer to node.
//              const NodeInput *nodeInput
//                  Pointer to node input.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkIpInit(Node *node, const NodeInput *nodeInput)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    BOOL retVal;
    char protocolString[MAX_STRING_LENGTH];
    char forwardingEnabledString[MAX_STRING_LENGTH];
    char buf[MAX_STRING_LENGTH];
    int i;

    int ipFragUnit = 0;

    double throughput = 0.0;
    float rtPerformVar; // router performance variation
    char rtBackType[MAX_STRING_LENGTH]; //router backplane type
    char backplaneThroughput[MAX_STRING_LENGTH];

// For Dymo init default value
    ip->isManetGateway = FALSE;
    ip->manetPrefixlength = 0;
// End for Dymo

    ip->backplaneType = BACKPLANE_TYPE_DISTRIBUTED;
    ip->backplaneStatus = NETWORK_IP_BACKPLANE_IDLE;

    ip->backplaneThroughputCapacity =
        NETWORK_IP_UNLIMITED_BACKPLANE_THROUGHPUT;

    // IGMP is not enabled by default.
    ip->isIgmpEnable = FALSE;

    // ICMP is not enabled by default.
    ip->isIcmpEnable = FALSE;

#ifdef ENTERPRISE_LIB
    // Route Redistribution is disabled by default.
    ip->isRtRedistributeStatOn = FALSE;
#endif // ENTERPRISE_LIB

    // packet is not ECNcapable by default.
    ip->isPacketEcnCapable = FALSE;

    // BEGIN NDP start
    ip->isNdpEnable = FALSE;   // NDP is not enabled by default
    ip->ndpData = NULL;
    // END NDP start.

    // MOBILE-IP is not enabled by default.
    ip->mobileIpStruct = NULL;

    // dualIp is not enabled by default.
    ip->dualIp = NULL;

    NetworkIpInitTrace(node, nodeInput);

    IO_ReadInt(ANY_NODEID,
               ANY_ADDRESS,
               nodeInput,
               "IP-FRAGMENTATION-UNIT",
               &retVal,
               &ipFragUnit);

    if (retVal)
    {
        ERROR_Assert(ipFragUnit >= 256 &&
                     ipFragUnit <= MAX_NW_PKT_SIZE  && ipFragUnit <= 65535,
                     "Fragmentation unit should be greater than 0 "
                     " and less than MAX_NW_PKT_SIZE  and 65535");

        ip->ipFragUnit = ipFragUnit;
        ip->maxPacketLength = ip->ipFragUnit;
    }

    IO_ReadString(
        node->nodeId,
        ANY_ADDRESS,
        nodeInput,
        "ROUTER-BACKPLANE-THROUGHPUT",
        &retVal,
        backplaneThroughput);

    char *p;

    if (retVal == FALSE ||
        strcmp(backplaneThroughput, "UNLIMITED") == 0 ||
        strcmp(backplaneThroughput, "0") == 0)
    {
        // If not specified, we assume infinite backplane throughput.
        ip->backplaneThroughputCapacity =
            NETWORK_IP_UNLIMITED_BACKPLANE_THROUGHPUT;
    }
    else if ((throughput = strtod(backplaneThroughput , &p)) > 0.0)
    {
        IO_ReadFloat(
            node->nodeId,
            ANY_ADDRESS,
            nodeInput,
            "ROUTER-PERFORMANCE-VARIATION",
            &retVal,
            &rtPerformVar);

        if (retVal)
        {
            throughput += ((throughput * rtPerformVar) / 100);
        }

        ip->backplaneThroughputCapacity = (clocktype)throughput;

        for (i = 0; i < node->numberInterfaces; i++)
        {
            ip->interfaceInfo[i]->disBackplaneCapacity =
                (clocktype)(throughput / (node->numberInterfaces + 1));
        }

        IO_ReadString(
            node->nodeId,
            ANY_ADDRESS,
            nodeInput,
            "ROUTER-BACKPLANE-TYPE",
            &retVal,
            rtBackType);

        if (strcmp(rtBackType, "CENTRAL") == 0)
        {
            ip->backplaneType = BACKPLANE_TYPE_CENTRAL;
        }
        else if (strcmp(rtBackType, "DISTRIBUTED"))
        {
            ERROR_ReportError("ROUTER-BACKPLANE-TYPE should be "
                              "either \"CENTRAL\" or \"DISTRIBUTED\".\n");
        }
        else
        {
            ip->backplaneThroughputCapacity =
                (clocktype)(throughput / (node->numberInterfaces + 1));
        }
    }
    else if (throughput == 0.0)
    {
        // If not specified, we assume infinite backplane throughput.
        ip->backplaneThroughputCapacity =
            NETWORK_IP_UNLIMITED_BACKPLANE_THROUGHPUT;
    }
    else
    {
        ERROR_ReportError("Wrong ROUTER-BACKPLANE-THROUGHPUT value.\n");
    }

    // Create buffer for CPU
    NetworkIpInitCpuQueueConfiguration(node,
                                       nodeInput);
    IO_ReadString(
        node->nodeId,
        ANY_ADDRESS,
        nodeInput,
        "IP-FORWARDING",
        &retVal,
        forwardingEnabledString);

    if (retVal)
    {
        if (strcmp(forwardingEnabledString, "YES") == 0)
        {
            ip->ipForwardingEnabled = TRUE;
        }
        else if (strcmp(forwardingEnabledString, "NO") == 0)
        {
            ip->ipForwardingEnabled = FALSE;
        }
        else
        {
            ERROR_ReportError("IP-FORWARDING should be either \"YES\" or \""
                              "NO\".\n");
        }
    }
    else
    {
        ip->ipForwardingEnabled = TRUE;
    }

    // Loopback Init
    NetworkIpLoopbackInit(node, nodeInput);

    // added for gateway

    char gatewayStr[MAX_ADDRESS_STRING_LENGTH];

    IO_ReadString(
        node->nodeId,
        ANY_ADDRESS, // subnet address need to be taken
        nodeInput,
        "DEFAULT-GATEWAY",
        &retVal,
        gatewayStr);

    if (!retVal)
    {
        ip->gatewayConfigured = FALSE;
        ip->defaultGatewayId = 0;
    }
    else
    {
        NodeAddress gatewayRtrId;
        int numHostBits;
        BOOL isNodeId;

        // Parse IP Address
        IO_ParseNodeIdHostOrNetworkAddress(
            gatewayStr,
            &gatewayRtrId,
            &numHostBits,
            &isNodeId);

        if (isNodeId)
        {
            ip->gatewayConfigured = TRUE;

            // consider default address only
            ip->defaultGatewayId =
                MAPPING_GetInterfaceAddressForInterface(
                    node, gatewayRtrId, DEFAULT_INTERFACE);
        }
        else
        {
            ip->gatewayConfigured = TRUE;
            ip->defaultGatewayId = gatewayRtrId;
        }
    }
    // end for gateway

    IO_ReadString(
        node->nodeId,
        ANY_ADDRESS,
        nodeInput,
        "IPSEC-ENABLED",
        &retVal,
        buf);

    if (!retVal || !strcmp(buf, "NO"))
    {
        ip->isIPsecEnabled = FALSE;
    }
    else if (!strcmp(buf, "YES"))
    {
        ip->isIPsecEnabled = TRUE;
        IPsecInit(node, nodeInput);
    }
    else
    {
        ERROR_ReportError("IPSEC-ENABLED expects YES or NO");
    }

#ifdef MILITARY_RADIOS_LIB
    for (i = 0; i < node->numberInterfaces; ++i)
    {
        IO_ReadString(node->nodeId,
                      NetworkIpGetInterfaceAddress(node, i),
                      nodeInput,
                      "INC-TYPE",
                      &retVal,
                      protocolString);

        if (retVal && strcmp(protocolString, "EPLRS") == 0)
        {
            EplrsInit(node, nodeInput, i);
        }
    }
#endif
    for (i = 0; i < node->numberInterfaces; i++)
    {
#ifdef ENTERPRISE_LIB
        IO_ReadString(
            node->nodeId,
            NetworkIpGetInterfaceAddress(node, i),
            nodeInput,
            "MPLS-PROTOCOL",
            &retVal,
            protocolString);

        if (!retVal || strcmp(protocolString, "NO") == 0)
        {
            node->macData[i]->mplsVar = NULL;
        }
        else
        {
            if (strcmp(protocolString, "YES") == 0)
            {
                MplsInit(node, nodeInput, i,
                         (MplsData **) &node->macData[i]->mplsVar);
            }
            else
            {
                fprintf(stderr,
                        "Expecting YES or NO for MPLS-PROTOCOL parameter\n");
                abort();
            }
        }
#endif // ENTERPRISE_LIB

        if (NetworkIpGetInterfaceType(node, i) == NETWORK_IPV4)
        {
            IO_ReadString(
                node->nodeId,
                NetworkIpGetInterfaceAddress(node, i),
                nodeInput,
                "ROUTING-PROTOCOL",
                &retVal,
                protocolString);
        }
        else
        {
            Address address;

            NetworkGetInterfaceInfo(
                node,
                i,
                &address);

            IO_ReadString(
                node->nodeId,
                &address.interfaceAddr.ipv6,
                nodeInput,
                "ROUTING-PROTOCOL",
                &retVal,
                protocolString);
        }

        if (retVal)
        {
#ifdef WIRELESS_LIB
            if (strcmp(protocolString, "LAR1") == 0)
            {
                NetworkIpAddUnicastRoutingProtocolType(
                    node,
                    ROUTING_PROTOCOL_LAR1,
                    i);

                if (!NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_LAR1))
                {
                    Lar1Init(
                        node,
                        (Lar1Data **) &ip->interfaceInfo[i]->routingProtocol,
                        nodeInput,
                        i);
                }
                else
                {
                    NetworkIpUpdateUnicastRoutingProtocolAndRouterFunction(
                        node,
                        ROUTING_PROTOCOL_LAR1,
                        i);
                }
            }
            else if (strcmp(protocolString, "AODV") == 0)
            {
                NetworkRoutingProtocolType protocolType;
                AodvData** routingProtocol;

                if(NetworkIpGetInterfaceType(node, i) == NETWORK_IPV4)
                {
                    protocolType = ROUTING_PROTOCOL_AODV;
                    routingProtocol =
                        (AodvData**)&ip->interfaceInfo[i]->routingProtocol;

                    NetworkIpAddUnicastRoutingProtocolType(
                        node,
                        protocolType,
                        i);

                    if (!NetworkIpGetRoutingProtocol(node, protocolType))
                    {
                        AodvInit(
                            node,
                            routingProtocol,
                            nodeInput,
                            i,
                            protocolType);
                    }
                    else
                    {
                        NetworkIpUpdateUnicastRoutingProtocolAndRouterFunction(
                            node,
                            protocolType,
                            i);
                    }
                }
                else
                {
                    // Let IPv6 handle the initialization
                }
            }
            else if (strcmp(protocolString, "DYMO") == 0)
            {
                NetworkRoutingProtocolType protocolType;
                DymoData** routingProtocol;

                if(NetworkIpGetInterfaceType(node, i) == NETWORK_IPV4)
                {
                    protocolType = ROUTING_PROTOCOL_DYMO;
                    routingProtocol =
                        (DymoData**)&ip->interfaceInfo[i]->routingProtocol;

                    NetworkIpAddUnicastRoutingProtocolType(
                        node,
                        protocolType,
                        i);

                    if (!NetworkIpGetRoutingProtocol(
                        node,
                        protocolType))
                    {
                        DymoInit(
                            node,
                            routingProtocol,
                            nodeInput,
                            i,
                            protocolType);
                    }
                    else
                    {
                        NetworkIpUpdateUnicastRoutingProtocolAndRouterFunction(
                            node,
                            protocolType,
                            i);
                    }
                }
                else
                {
                    // Let IPv6 handle the initialization
                }
            }
            else if (strcmp(protocolString, "DSR") == 0)
            {
                NetworkIpAddUnicastRoutingProtocolType(
                        node,
                        ROUTING_PROTOCOL_DSR,
                        i);

                if (!NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_DSR))
                {
                    DsrInit(
                        node,
                        (DsrData **) &ip->interfaceInfo[i]->routingProtocol,
                        nodeInput,
                        i);
                }
                else
                {
                    NetworkIpUpdateUnicastRoutingProtocolAndRouterFunction(
                                              node,
                                              ROUTING_PROTOCOL_DSR,
                                              i);
                }
            }
            else
            if (strcmp(protocolString, "FSRL") == 0)
            {
                NetworkIpAddUnicastRoutingProtocolType(
                        node,
                        ROUTING_PROTOCOL_FSRL,
                        i);

                if (!NetworkIpGetRoutingProtocol(node,
                                                 ROUTING_PROTOCOL_FSRL))
                {
                    FsrlInit(
                        node,
                        (FsrlData **) &ip->interfaceInfo[i]->routingProtocol,
                        nodeInput,
                        i);
                }
                else
                {
                    NetworkIpUpdateUnicastRoutingProtocolAndRouterFunction(
                                              node,
                                              ROUTING_PROTOCOL_FSRL,
                                              i);
                }
            }
            else
            if (strcmp(protocolString, "STAR") == 0)
            {
                NetworkIpAddUnicastRoutingProtocolType(
                        node,
                        ROUTING_PROTOCOL_STAR,
                        i);

                if (!NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_STAR))
                {
                    StarInit(
                        node,
                        (StarData **) &ip->interfaceInfo[i]->routingProtocol,
                        nodeInput,
                        i);
                }
                else
                {
                    NetworkIpUpdateUnicastRoutingProtocolAndRouterFunction(
                                              node,
                                              ROUTING_PROTOCOL_STAR,
                                              i);
                }
            }
//StartIARP
            else if (strcmp(protocolString, "IARP") == 0) {
                NetworkIpAddUnicastRoutingProtocolType(
                    node,
                    ROUTING_PROTOCOL_IARP,
                    i);
                if (!NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_IARP)) {
                    IarpInit(node, (IarpData**) &ip->interfaceInfo[i]->routingProtocol, nodeInput, i);
                } else {
                    NetworkIpUpdateUnicastRoutingProtocolAndRouterFunction(
                        node, ROUTING_PROTOCOL_IARP, i);
                }
            }
//EndIARP
//StartZRP
            else if (strcmp(protocolString, "ZRP") == 0) {
                NetworkIpAddUnicastRoutingProtocolType(
                    node,
                    ROUTING_PROTOCOL_ZRP,
                    i);
                if (!NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_ZRP)) {
                    ZrpInit(node, (ZrpData**) &ip->interfaceInfo[i]->routingProtocol,
                            nodeInput, i);
                } else {
                    NetworkIpUpdateUnicastRoutingProtocolAndRouterFunction(
                        node, ROUTING_PROTOCOL_ZRP, i);
                }
            }
//EndZRP
//StartIARP
            else if (strcmp(protocolString, "IARP") == 0) {
                NetworkIpAddUnicastRoutingProtocolType(
                    node,
                    ROUTING_PROTOCOL_IARP,
                    i);
                if (!NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_IARP)) {
                    IarpInit(node, (IarpData**) &ip->interfaceInfo[i]->routingProtocol, nodeInput, i);
                } else {
                    NetworkIpUpdateUnicastRoutingProtocolAndRouterFunction(
                        node, ROUTING_PROTOCOL_IARP, i);
                }
            }
//EndIARP
//StartIERP
            else if (strcmp(protocolString, "IERP") == 0) {
                NetworkIpAddUnicastRoutingProtocolType(
                    node,
                    ROUTING_PROTOCOL_IERP,
                    i);
                if (!NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_IERP)) {
                    IerpInit(node, (IerpData**) &ip->interfaceInfo[i]->routingProtocol, nodeInput, i);
                } else {
                    NetworkIpUpdateUnicastRoutingProtocolAndRouterFunction(
                        node, ROUTING_PROTOCOL_IERP, i);
                }
            }
//EndIERP
//StartANTHOCNET
            else if (strcmp(protocolString, "ANTHOCNET") == 0)
            {
                if(NetworkIpGetInterfaceType(node, i) == NETWORK_IPV4)
                {
                    NetworkIpAddUnicastRoutingProtocolType(
                        node,
                        ROUTING_PROTOCOL_ANTHOCNET,
                        i);

                    if (!NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_ANTHOCNET))
                    {
                        AnthocnetInit(node, (AnthocnetData**)&ip->interfaceInfo[i]->routingProtocol, nodeInput, i);
                    }
                    else
                    {
                        NetworkIpUpdateUnicastRoutingProtocolAndRouterFunction(node, ROUTING_PROTOCOL_ANTHOCNET, i);
                    }
                }
                else
                {
                    // Let IPv6 handle the initialization
                }
            }
//EndANTHOCNET
            else
#endif // WIRELESS_LIB
#ifdef ENTERPRISE_LIB
            if (strcmp(protocolString, "OSPFv2") == 0)
            {
                ERROR_Assert(
                    NetworkIpGetInterfaceType(node, i) == NETWORK_IPV4,
                    "OSPFv2 enabled on non-IPv4 interface\n");

                NetworkIpAddUnicastRoutingProtocolType(
                        node,
                        ROUTING_PROTOCOL_OSPFv2,
                        i);

                MAC_SetInterfaceStatusHandlerFunction(
                                        node,
                                        i,
                                        &Ospfv2InterfaceStatusHandler);

                if (!NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_OSPFv2))
                {
                    Ospfv2Init(
                        node,
                        (Ospfv2Data **) &ip->interfaceInfo[i]->routingProtocol,
                        nodeInput,
                        i);
                }
                else
                {
                    NetworkIpUpdateUnicastRoutingProtocolAndRouterFunction(
                                              node,
                                              ROUTING_PROTOCOL_OSPFv2,
                                              i);
                }
            }
            else
            if (strcmp(protocolString, "IGRP") == 0)
            {
                NetworkIpAddUnicastRoutingProtocolType(
                    node,
                    ROUTING_PROTOCOL_IGRP,
                    i);

                if (!NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_IGRP))
                {
                    IgrpInit(
                        node,
                        (RoutingIgrp **)
                        &ip->interfaceInfo[i]->routingProtocol,
                        nodeInput,
                        i);
                }
                else
                {
                    NetworkIpUpdateUnicastRoutingProtocolAndRouterFunction(
                        node,
                        ROUTING_PROTOCOL_IGRP,
                        i);
                }
            }
            else
            if (strcmp(protocolString, "EIGRP") == 0)
            {
                NetworkIpAddUnicastRoutingProtocolType(
                    node,
                    ROUTING_PROTOCOL_EIGRP,
                    i);

                MAC_SetInterfaceStatusHandlerFunction(
                    node,
                    i,
                    EigrpInterfaceStatusHandler);

                if (!NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_EIGRP))
                {
                    EigrpInit(
                        node,
                        (RoutingEigrp **)
                        &ip->interfaceInfo[i]->routingProtocol,
                        nodeInput,
                        i);
                }
                else
                {
                    NetworkIpUpdateUnicastRoutingProtocolAndRouterFunction(
                        node,
                        ROUTING_PROTOCOL_EIGRP,
                        i);
                }
            }
            else
#endif // ENTERPRISE_LIB
            if (strcmp(protocolString, "ODR") == 0)
            {
#ifdef MILITARY_RADIOS_LIB
                NetworkIpAddUnicastRoutingProtocolType(
                    node,
                    ROUTING_PROTOCOL_ODR,
                    i);

                if (!NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_ODR))
                {
                    OdrInit(
                        node,
                        (OdrData **) &ip->interfaceInfo[i]->routingProtocol,
                        nodeInput,
                        i);
                }
                else
                {
                    NetworkIpUpdateUnicastRoutingProtocolAndRouterFunction(
                        node,
                        ROUTING_PROTOCOL_ODR,
                        i);
                }
#else // MILITARY_RADIOS_LIB
                ERROR_ReportMissingLibrary(protocolString, "Military Radios\n");
#endif // MILITARY_RADIOS_LIB
            }
            else if (retVal && strcmp(protocolString, "SDR") == 0)
            {
#ifdef MILITARY_RADIOS_LIB
                NetworkIpAddUnicastRoutingProtocolType(
                        node,
                        ROUTING_PROTOCOL_SDR,
                        i);
                if (!NetworkIpGetRoutingProtocol(node, ROUTING_PROTOCOL_SDR))
                {
                    SdrInit(node, nodeInput, i);
                }
                else
                {
                    NetworkIpUpdateUnicastRoutingProtocolAndRouterFunction(
                        node,
                        ROUTING_PROTOCOL_SDR,
                        i);
                }
#else // MILITARY_RADIOS_LIB
                ERROR_ReportMissingLibrary(protocolString, "Military Radios\n");
#endif // MILITARY_RADIOS_LIB
            }
            //InsertPatch NETWORK_INIT_CODE
            else
            if ((strcmp(protocolString, "BELLMANFORD") == 0) ||
                (strcmp(protocolString, "OPNET_RIP") == 0) ||
                (strcmp(protocolString, "OLSR-INRIA") == 0) ||
                (strcmp(protocolString, "OLSRv2-NIIGATA") == 0) ||
                (strcmp(protocolString, "FISHEYE") == 0))
            {
                // These routing protocols are initialized in the app layer.
            }

//StartRIP
            else
            if (strcmp(protocolString, "RIP") == 0)
            {
                //The protocol is initialized in the app layer
            }
//EndRIP
//StartRIPng
            else
            if (strcmp(protocolString, "RIPng") == 0)
            {
                //The protocol is initialized in the app layer
            }
//EndRIPng
//InsertPatch ROUTING_INIT_IP
            else
            if (strcmp(protocolString, "NONE") == 0)
            {
                // Allow a node to specify explicitly that routing
                // protocols are not used.
            }

            else
            if (strcmp(protocolString, "OSPFv3") == 0)
            {
                // The protocol is initialized in the Ipv6Init Function.
            }

            else
            {
                fprintf(stderr, "Unknown ROUTING-PROTOCOL %s\n",
                        protocolString);
                abort();
            }
        }

        IO_ReadString(node->nodeId,
                     NetworkIpGetInterfaceAddress(node, i),
                     nodeInput,
                     "GROUP-MANAGEMENT-PROTOCOL",
                     &retVal,
                     protocolString);

        if (retVal)
        {
            if (strcmp(protocolString, "IGMP") == 0)
            {
                IgmpInit(node,
                            nodeInput,
                            &ip->igmpDataPtr,
                            i);
            }
        }

        IO_ReadString(
                     node->nodeId,
                     NetworkIpGetInterfaceAddress(node, i),
                     nodeInput,
                     "MULTICAST-PROTOCOL",
                     &retVal,
                     protocolString);

        if (retVal)
        {
#ifdef ENTERPRISE_LIB
            if (strcmp(protocolString, "DVMRP") == 0)
            {
                NetworkIpAddMulticastRoutingProtocolType(node,
                                                     MULTICAST_PROTOCOL_DVMRP,
                                                     i);

                if (!NetworkIpGetMulticastRoutingProtocol(node,
                                                     MULTICAST_PROTOCOL_DVMRP))
                {
                    RoutingDvmrpInit(node, nodeInput, i);
                }
                else
                {
                    NetworkIpUpdateMulticastRoutingProtocolAndRouterFunction(
                                              node,
                                              MULTICAST_PROTOCOL_DVMRP,
                                              i);

                    /* Inform IGMP about multicast routing protocol */
                    if (ip->isIgmpEnable == TRUE)
                    {
                        IgmpSetMulticastProtocolInfo(
                                       node,
                                       i,
                                       &RoutingDvmrpLocalMembersJoinOrLeave);
                    }
                }
            } // if (strcmp...)

            else if (strcmp(protocolString, "MOSPF") == 0)
            {
                BOOL tempRetVal = FALSE;
                char tempProtocolString[MAX_STRING_LENGTH];

                IO_ReadString(
                    node->nodeId,
                    NetworkIpGetInterfaceAddress(node, i),
                    nodeInput,
                    "ROUTING-PROTOCOL",
                    &tempRetVal,
                    tempProtocolString);

                if (!tempRetVal)
                {
                    NetworkIpAddUnicastRoutingProtocolType(
                            node,
                            ROUTING_PROTOCOL_OSPFv2,
                            i);

                    MAC_SetInterfaceStatusHandlerFunction(
                            node,
                            i,
                            &Ospfv2InterfaceStatusHandler);

                    if (!NetworkIpGetRoutingProtocol(node,
                        ROUTING_PROTOCOL_OSPFv2))
                    {
                        Ospfv2Init(
                            node,
                            (Ospfv2Data **) &ip->
                                interfaceInfo[i]->routingProtocol,
                            nodeInput,
                            i);
                    }
                    else
                    {
                        NetworkIpUpdateUnicastRoutingProtocolAndRouterFunction(
                                                  node,
                                                  ROUTING_PROTOCOL_OSPFv2,
                                                  i);
                    }
                }
                else
                if (strcmp(tempProtocolString, "OSPFv2"))
                {
                    ERROR_ReportError("Need OSPFv2 as the underlying "
                        "unicast routing protocol to run MOSPF.\n");
                }

                NetworkIpAddMulticastRoutingProtocolType(node,
                                                     MULTICAST_PROTOCOL_MOSPF,
                                                     i);

                if (!NetworkIpGetMulticastRoutingProtocol(node,
                                                     MULTICAST_PROTOCOL_MOSPF))
                {
                    MospfInit(node, nodeInput, i);
                }
                else
                {
                    NetworkIpUpdateMulticastRoutingProtocolAndRouterFunction(
                                              node,
                                              MULTICAST_PROTOCOL_MOSPF,
                                              i);

                    /* Inform IGMP about multicast routing protocol */

                    if (ip->isIgmpEnable == TRUE)
                    {
                        IgmpSetMulticastProtocolInfo(
                                       node,
                                       i,
                                       &MospfLocalMembersJoinOrLeave);
                    }
                }
            } // if (strcmp...)

            else if (strcmp(protocolString, "PIM") == 0)
            {
                NetworkIpAddMulticastRoutingProtocolType(node,
                                                     MULTICAST_PROTOCOL_PIM,
                                                     i);

                if (!NetworkIpGetMulticastRoutingProtocol(node,
                                                     MULTICAST_PROTOCOL_PIM))
                {
                    RoutingPimInit(node, nodeInput, i);
                }
                else
                {
                    NetworkIpUpdateMulticastRoutingProtocolAndRouterFunction(
                                              node,
                                              MULTICAST_PROTOCOL_PIM,
                                              i);

                    /* Inform IGMP about multicast routing protocol */
                    if (ip->isIgmpEnable == TRUE)
                    {
                        IgmpSetMulticastProtocolInfo(
                                       node,
                                       i,
                                       &RoutingPimLocalMembersJoinOrLeave);
                    }
                }
            } // if (strcmp...)
            else
#endif // ENTERPRISE_LIB

#ifdef WIRELESS_LIB
            if (strcmp(protocolString, "ODMRP") == 0)
            {
                NetworkIpAddMulticastRoutingProtocolType(node,
                                                     MULTICAST_PROTOCOL_ODMRP,
                                                     i);

                if (!NetworkIpGetMulticastRoutingProtocol(node,
                                                     MULTICAST_PROTOCOL_ODMRP))
                {
                    OdmrpInit(node, nodeInput, i);
                }
                else
                {
                    NetworkIpUpdateMulticastRoutingProtocolAndRouterFunction(
                                              node,
                                              MULTICAST_PROTOCOL_ODMRP,
                                              i);
                }
            }
            else 
#endif // WIRELESS_LIB

#ifdef ADDON_MAODV
            if (strcmp(protocolString, "MAODV") == 0)
            {
                NetworkIpAddMulticastRoutingProtocolType(
                    node,
                    MULTICAST_PROTOCOL_MAODV,
                    i);

                if (!NetworkIpGetMulticastRoutingProtocol(
                    node,
                    MULTICAST_PROTOCOL_MAODV))
                {
                    MaodvInit(node, nodeInput, i);
                }
                else
                {
                    NetworkIpUpdateMulticastRoutingProtocolAndRouterFunction(
                        node,
                        MULTICAST_PROTOCOL_MAODV,
                        i);
                }
            }
            else
#endif // ADDON_MAODV
            {
                char errorString[MAX_STRING_LENGTH];
                sprintf(errorString,
                        "Unknown MULTICAST-PROTOCOL %s\n",
                        protocolString);
                ERROR_ReportError(errorString);
            }
        } // if (retval)
    }

    IO_ReadString(node->nodeId,
                     ANY_ADDRESS,
                     nodeInput,
                     "ICMP",
                     &retVal,
                     protocolString);

    if (retVal)
    {
        if (strcmp(protocolString, "YES") == 0)
        {
            NetworkIcmpInit(node, nodeInput);
        }
    }

    IO_ReadString(
             node->nodeId,
             ANY_ADDRESS,
             nodeInput,
             "MULTICAST-STATIC-ROUTE",
             &retVal,
             protocolString);

    if (retVal == TRUE && strcmp(protocolString, "YES") == 0)
    {
        RoutingMulticastStaticInit(
                    node,
                    nodeInput);
    }

#ifdef ENTERPRISE_LIB
    MobileIpInit(node, nodeInput);
    IpInitPerHopBehaviors(node, nodeInput);
#endif // ENTERPRISE_LIB

    ip->mftcInfo = NULL;
    ip->numMftcInfo = 0;
    ip->maxMftcInfo = 0;
    ip->ipMftcParameter = NULL;
    ip->numIpMftcParameters = 0;
    ip->maxIpMftcParameters = 0;
    ip->isEdgeRouter = FALSE;
    ip->mftcStatisticsEnabled = FALSE;

    // Read this node collect Diffserv statistics or not
    // The <variant> is one of YES | NO
    // Format is: DIFFSERV-EDGE-ROUTER-STATISTICS <variant>, for example
    //            DIFFSERV-EDGE-ROUTER-STATISTICS YES
    //            DIFFSERV-EDGE-ROUTER-STATISTICS NO
    // If not specified, default is NO

    IO_ReadString(
        node->nodeId,
        ANY_ADDRESS,
        nodeInput,
        "DIFFSERV-EDGE-ROUTER-STATISTICS",
        &retVal,
        buf);

    if (retVal && strcmp(buf, "YES") == 0)
    {
        ip->mftcStatisticsEnabled = TRUE;
    }

#ifdef ENTERPRISE_LIB
    // Read whether this node is Diffserv enabled Edge router or not
    // The <variant> is one of YES | NO
    // Format is: DIFFSERV-ENABLE-EDGE-ROUTER <variant>, for example
    //            DIFFSERV-ENABLE-EDGE-ROUTER YES
    //            DIFFSERV-ENABLE-EDGE-ROUTER NO
    //            [3 5]  DIFFSERV-ENABLE-EDGE-ROUTER YES
    // If not specified, default is NO

    IO_ReadString(
        node->nodeId,
        ANY_ADDRESS,
        nodeInput,
        "DIFFSERV-ENABLE-EDGE-ROUTER",
        &retVal,
        buf);

    if (retVal)
    {
        if (strcmp(buf, "YES") == 0)
        {
            ip->isEdgeRouter = TRUE;
            DIFFSERV_MFTrafficConditionerInit(node, nodeInput);
        }
        else if (strcmp(buf, "NO") == 0)
        {
            ip->isEdgeRouter = FALSE;
        }
        else
        {
            ERROR_ReportError(
                "DIFFSERV-ENABLE-EDGE-ROUTER: Unknown variant"
                " in configuration file.\n");
        }
    }

    // Initializing access list
    AccessListInit(node, nodeInput);

    // Initializing route map.  Make sure that it is initialized after the
    // access list initialization is over.
    RouteMapInit(node, nodeInput);

    // Initializing of PBR
    PbrInit(node, nodeInput);

    //Initialization of Route Redistribution.
    RouteRedistributeInit(node, nodeInput);
#endif // ENTERPRISE_LIB

    // Fragmentation id set to 0
    ip->ipFragmentId = 0;
    ip->fragmentListFirst = NULL;
}

//-----------------------------------------------------------------------------
// Layer function
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// this poriton of code for ATM (Start)
//-----------------------------------------------------------------------------

// /**
// FUNCTION :: Atm_IsMyPacket
// LAYER    :: Network
// PURPOSE ::  Check if this is own packet
// PARAMETERS ::
// + node : Node* : pointer to node.
// + destAddress : NodeAddress : Destination Address
// + incomingInterface : int : Incomming interface.
// RETURN :: BOOL : TRUE or FALSE
// **/
static  //inline//
BOOL Atm_IsMyPacket(Node *node,
                    NodeAddress destAddress,
                    int incomingInterface)
{
    int i;
    if (destAddress == ANY_DEST)
    {
        return TRUE;
    }

    // If the destination of the packet is end system.
    NodeAddress ipAddr = MAPPING_GetInterfaceAddressForInterface(
                                    node, node->nodeId, incomingInterface);

    if (destAddress == ipAddr)
    {
        return TRUE;
    }

    for (i = 0; i < node->numberInterfaces; i++)
    {
        if (destAddress == NetworkIpGetInterfaceAddress(node, i)
            || destAddress ==
                NetworkIpGetInterfaceBroadcastAddress(node, i))
        {
            return TRUE;
        }
    }

    // Now check the logical Interface
    const LogicalSubnet* myLogicalSubnet =
            AtmGetLogicalSubnetFromNodeId( node,
            node->nodeId, DEFAULT_INTERFACE);

    if (myLogicalSubnet->ipAddress == destAddress)
    {
        return TRUE;
    }

    return FALSE;
}


// /**
// FUNCTION :: Atm_RouteThePacketUsingLookupTable
// LAYER    :: Network
// PURPOSE ::  Routs the packet using lookup table.
// PARAMETERS ::
// + node : Node* : pointer to node.
// + destAddr : NodeAddress : Destination Address
// + outIntf : int* : Outgoing interface.
// + nextHop : NodeAddress* : Next hop
// RETURN :: void : None
// **/

void Atm_RouteThePacketUsingLookupTable(Node* node,
                                        NodeAddress destAddr,
                                        int* outIntf,
                                        NodeAddress* nextHop)
{
    int i;
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;


    // first check if I am the destination
    for (i = 0; i < node->numberInterfaces; i++)
    {
        if (ip->interfaceInfo[i]->ipAddress == destAddr)
        {
            // I am the destination
            *outIntf = 0;
            *nextHop = destAddr;
            return;
        }
    }

    // if not found
    // searching for matching entry in Network forwarding Table
    NetworkGetInterfaceAndNextHopFromForwardingTable(node,
        destAddr, outIntf, nextHop);

    if (*nextHop == 0)
    {
        *nextHop = destAddr;
    }
}


// /**
// FUNCTION :: NetworkIpReceivePacketFromAdaptationLayer
// LAYER    :: Network
// PURPOSE ::  Process packet after receiving from adaptation layer
// PARAMETERS ::
// + node : Node* : Pointer to node.
// + msg : Message * : Pointer to message.
// + incomingInterface : int : Incomming interface.
// RETURN :: void : None
// **/
void NetworkIpReceivePacketFromAdaptationLayer(
    Node *node,
    Message *msg,
    int incomingInterface)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    IpHeaderType* ipHeader = (IpHeaderType *) msg->packet;
    ip->stats.ipInReceives++;
    ip->stats.ipRecvdPktFromOtherNetwork++ ;

    if (Atm_IsMyPacket(node, ipHeader->ip_dst, incomingInterface))
    {
        IpHeaderType *ipHeader = (IpHeaderType *) msg->packet;
        NodeAddress src = ipHeader->ip_src;
        NodeAddress dst = ipHeader->ip_dst;
        TosType tos = ipHeader->ip_tos;

        MESSAGE_RemoveHeader(node, msg, IpHeaderSize(ipHeader), TRACE_IP);

        if (ipHeader->ip_p == IPPROTO_UDP)
        {
            SendToUdp(node, msg, tos, src, dst, incomingInterface);
        }
        else if (ipHeader->ip_p == IPPROTO_TCP)
        {
            BOOL aCongestionExperienced = FALSE;

            if (ipHeader->ip_tos & IPTOS_CE)
            {
                aCongestionExperienced = TRUE;
            }

            SendToTcp(node,
                msg,
                tos,
                src,
                dst,
                aCongestionExperienced);
        }
        else
        {
            printf(" Node %u receive Erronious pkt \n", node->nodeId);
        }

        return;
    }

    int outgoingInterface;
    NodeAddress nextHop;

    Atm_RouteThePacketUsingLookupTable(node,
        ipHeader->ip_dst,
        &outgoingInterface,
        &nextHop);

    if (nextHop != (unsigned) NETWORK_UNREACHABLE)
    {
        QueueUpIpFragmentForMacLayer(node, msg,
            outgoingInterface, nextHop);
    }
    else
    {
        // Increment stat for number of IP packets discarded because no
        // route could be found.
        ip->stats.ipOutNoRoutes++;

        // Free message.
        MESSAGE_Free(node, msg);
    }
}

// added for gateway


// /**
// FUNCTION :: NetworkIpGatewayHandleIPPacket
// LAYER    :: Network
// PURPOSE ::  Tries to route and send the packet using
//             othertype of interface available to taht node
// PARAMETERS ::
// + node : Node* : Pointer to node.
// + msg : Message * : Pointer to message.
// RETURN :: void : None
// **/
static
void NetworkIpGatewayHandleIPPacket(Node* node, Message* msg)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    // Search for associated physical interface other than IP
    // attached with this node, handed over the packet to
    // associated Adaptation layer for processing

    // Presently this case is valid for ATM node i.e.
    // only other type of interface connected is ATM interface
    IpHeaderType* ipHdr = (IpHeaderType *)(msg->packet);

    // differentiate between control & data packet

    if ((ipHdr->ip_p == IPPROTO_UDP)
        || (ipHdr->ip_p == IPPROTO_TCP))
    {
        if ((ipHdr->ip_p == IPPROTO_UDP)
            && (ipHdr->ip_tos == IPTOS_PREC_INTERNETCONTROL))
        {
            // It is a control packet
            // Control Packet must be confined within IP cloud
            // No need to send to ATM cloud

            MESSAGE_Free(node, msg);
            return;
        }

        // otherwise it is  a data packet
    }
    else
    {
        // It is a control packet
        // Control Packet must be confined within IP cloud
        // No need to send to ATM cloud

        MESSAGE_Free(node, msg);
        return;
    }

    // Now start processing for data packet

    // Search for associated Adaptation layer
    // And handed over the packet to that layer

    // At present only ATM adaptation layer available
    if (node->adaptationData.adaptationProtocol
        == ADAPTATION_PROTOCOL_NONE)
    {
        // so discard the packet
        MESSAGE_Free(node, msg);
        return;
    }

    // Handed over to adaptation layer
    ip->stats.ipSendPktToOtherNetwork++ ;

    ADAPTATION_ReceivePacketFromNetworkLayer(node, msg);
}


// /**
// FUNCTION :: NetworkIpRoutePacketThroughGateway
// LAYER    :: Network
// PURPOSE ::  Tries to route and send the packet using the node's
//              Configured Gateway.
// PARAMETERS ::
// + node : Node* : Pointer to node.
// + msg : Message * : Pointer to message.
// + incomingInterface : int : incoming interface of packet.
// RETURN :: void : None
// **/
static  //inline//
void NetworkIpRoutePacketThroughGateway(Node *node,
                                   Message *msg,
                                   int incomingInterface)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    RouterFunctionType routerFunction = NULL;
    BOOL packetWasRouted = FALSE;

    int intfForGt;
    NodeAddress nxtHopForGt;
    int i;

    NodeAddress gatewayAddr = ip->defaultGatewayId;

    ip->stats.ipRoutePktThruGt++;

    // check if any of my interface matches with gateway addr
    for (i = 0; i < node->numberInterfaces; i++)
    {
        if (ip->interfaceInfo[i]->ipAddress == ip->defaultGatewayId)
        {
            NetworkIpGatewayHandleIPPacket(node, msg);
            return;
        }
    }

    // try to route packet towards gateway using router function

    routerFunction = NetworkIpGetRouterFunction(node,
        DEFAULT_INTERFACE);

    if (routerFunction)
    {
        (routerFunction)(node,
            msg,
            gatewayAddr,
            NetworkIpGetInterfaceAddress(node, DEFAULT_INTERFACE),
            &packetWasRouted);
    }

    // route the packet to gateway using forwarding table
    if (!packetWasRouted)
    {
        // Otherwise route the packet to gateway
        NetworkGetInterfaceAndNextHopFromForwardingTable(
            node,
            gatewayAddr,
            &intfForGt,
            &nxtHopForGt);

        if ((int) nxtHopForGt == NETWORK_UNREACHABLE)
        {
            // gateway is not reachable from that node
            // Free message.
            MESSAGE_Free(node, msg);

            return;
        }
        else if (nxtHopForGt == 0)
        {
            // gateway resides in attached network

            NetworkIpSendPacketOnInterface(
                node,
                msg,
                incomingInterface,
                intfForGt,
                gatewayAddr);
        }
        else
        {
            // next hop to gateway is found

            NetworkIpSendPacketOnInterface(node,
                msg,
                incomingInterface,
                intfForGt,
                nxtHopForGt);
        }
    }   // end of if packet was not routed
}

// end for gateway

//-----------------------------------------------------------------------------
// this poriton of code for ATM (End)
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpLayer
// PURPOSE      Handle IP layer events, incoming messages and messages
//              sent to itself (timers, etc.).
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message.
// RETURN       None.
//----------------------------------------------------------------------------

void
NetworkIpLayer(Node *node, Message *msg)
{
    switch (msg->protocolType)
    {
        case GROUP_MANAGEMENT_PROTOCOL_IGMP:
        {
            IgmpLayer(node, msg);
            break;
        }
        case NETWORK_PROTOCOL_ICMP:
        {
            NetworkIcmpHandleProtocolEvent(node, msg);
            break;
        }
        case LINK_MANAGEMENT_PROTOCOL_CBQ:
        {
            CBQResourceManagerHandleProtocolEvent(node, msg);
            break;
        }
        case NETWORK_PROTOCOL_IP:
        {
            switch (msg->eventType)
            {
                case MSG_NETWORK_Backplane:
                {
                    NetworkIpReceiveFromBackplane(node, msg);
                    MESSAGE_Free(node, msg);
                    break;
                }
                case MSG_NETWORK_FromTransportOrRoutingProtocol:
                {
                    int *outgoingInterface = (int *)MESSAGE_ReturnInfo(msg);

                    RoutePacketAndSendToMac(node,
                                            msg,
                                            CPU_INTERFACE,
                                            *outgoingInterface,
                                            ANY_IP);
                    break;
                }
                case MSG_NETWORK_FromAdaptation:
                {
                    int intfId = ((AdaptationToNetworkInfo *)
                                 MESSAGE_ReturnInfo(msg))->intfId;
                    NodeAddress prevHop = ((AdaptationToNetworkInfo *)
                                          MESSAGE_ReturnInfo(msg))->prevHop;

                    NetworkIpReceivePacketFromAdaptationLayer(
                        node, msg, intfId);

                    break;
                }
                case MSG_NETWORK_DelayedSendToMac:
                {
                    ProcessDelayedSendToMac(node, msg);
                    break;
                }
                case MSG_NETWORK_JoinGroup:
                {
                    NodeAddress *mcastAddr = (NodeAddress *)
                                             MESSAGE_ReturnInfo(msg);

                    NetworkIpAddToMulticastGroupList(node, *mcastAddr);


#ifdef WIRELESS_LIB
                    if (NetworkIpCheckMulticastRoutingProtocol(
                            node, MULTICAST_PROTOCOL_ODMRP, ANY_INTERFACE))
                    {
                        OdmrpJoinGroup(node, *mcastAddr);
                    }
#endif // WIRELESS_LIB

                    MESSAGE_Free(node, msg);
                    break;
                }
                case MSG_NETWORK_LeaveGroup:
                {
                    NodeAddress *mcastAddr = (NodeAddress *)
                                             MESSAGE_ReturnInfo(msg);

                    NetworkIpRemoveFromMulticastGroupList(node, *mcastAddr);

#ifdef WIRELESS_LIB
                    if (NetworkIpCheckMulticastRoutingProtocol(
                            node, MULTICAST_PROTOCOL_ODMRP, ANY_INTERFACE))
                    {
                        OdmrpLeaveGroup(node, *mcastAddr);
                    }
#endif // WIRELESS_LIB

                    MESSAGE_Free(node, msg);
                    break;
                }
#ifdef ENTERPRISE_LIB
                case MSG_NETWORK_AccessList:
                {
                    AccessListHandleEvent(node, msg);
                    break;
                }
#endif // ENTERPRISE_LIB
                case MSG_NETWORK_IPsec:
                {
                    IPsecHandleEvent(node, msg);
                    break;
                }
                default:
                    ERROR_ReportError("Invalid switch value");
            }//switch//

            break;
        }

#ifdef WIRELESS_LIB
        case ROUTING_PROTOCOL_AODV:
        {
            AodvHandleProtocolEvent(node, msg);
            break;
        }
        case ROUTING_PROTOCOL_DYMO:
        {
            DymoHandleProtocolEvent(node, msg);
            break;
        }
        case ROUTING_PROTOCOL_DSR:
        {
            DsrHandleProtocolEvent(node, msg);
            break;
        }
        case ROUTING_PROTOCOL_FSRL:
        {
            FsrlHandleProtocolEvent(node, msg);
            break;
        }
        case ROUTING_PROTOCOL_LAR1:
        {
            Lar1HandleCheckTimeoutAlarm(node, msg);
            break;
        }
        case MULTICAST_PROTOCOL_ODMRP:
        {
            OdmrpHandleProtocolEvent(node, msg);
            break;
        }
        case ROUTING_PROTOCOL_BRP:
        {
            BrpProcessEvent(node, msg);
            break;
        }
        case ROUTING_PROTOCOL_ZRP:
        {
            ZrpProcessEvent(node,msg);
            break;
        }
        case ROUTING_PROTOCOL_IARP:
        {
            IarpProcessEvent(node,msg);
            break;
        }
        case ROUTING_PROTOCOL_IERP:
        {
            IerpProcessEvent(node,msg);
            break;
        }
        case NETWORK_PROTOCOL_NDP:
        {
            NdpHandleProtocolEvent(node, msg);
            break;
        }
//StartANTHOCNET
        case ROUTING_PROTOCOL_ANTHOCNET:
        {
            AnthocnetProcessEvent(node, msg);
            break;
        }
//EndANTHOCNET
#endif // WIRELESS_LIB

#ifdef ENTERPRISE_LIB
        case NETWORK_PROTOCOL_MOBILE_IP:
        {
            MobileIpLayer(node, msg);
            break;
        }
        case ROUTING_PROTOCOL_OSPFv2:
        {
            Ospfv2HandleRoutingProtocolEvent(node, msg);
            break;
        }
        case MULTICAST_PROTOCOL_DVMRP:
        {
            RoutingDvmrpHandleProtocolEvent(node, msg);
            break;
        }
        case MULTICAST_PROTOCOL_PIM:
        {
            RoutingPimHandleProtocolEvent(node, msg);
            break;
        }
        case ROUTING_PROTOCOL_IGRP:
        {
            IgrpHandleProtocolEvent(node, msg);
            break;
        }
        case ROUTING_PROTOCOL_EIGRP:
        {
            EigrpHandleProtocolEvent(node, msg);
            break;
        }
#endif // ENTERPRISE_LIB

#ifdef MILITARY_RADIOS_LIB
        case ROUTING_PROTOCOL_ODR:
        {
            OdrHandleProtocolEvent(node, msg);
            break;
        }
        case ROUTING_PROTOCOL_SDR:
        {
            SdrHandleProtocolEvent(node, msg);
            break;
        }
        case NETWORK_PROTOCOL_EPLRS:
        {
            EplrsProcessEvent(node, msg);
            break;
        }
#endif
#ifdef ADDON_MAODV
        case MULTICAST_PROTOCOL_MAODV:
        {
            MaodvHandleProtocolEvent(node, msg);
            break;
        }
#endif // ADDON_MAODV
        case ROUTING_PROTOCOL_ALL:
        {
            ERROR_Assert(FALSE, "IP event error");
            //HandleSpecialMacLayerStatusEvents(node, msg);
            break;
        }
//InsertPatch NETWORK_IP_LAYER
        default:
            ERROR_ReportError("Invalid switch value");
    }//switch//
}

//-----------------------------------------------------------------------------
// Finalize function
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpFinalize()
// PURPOSE      Finalize function for the IP model.  Finalize functions
//              for all network-layer IP protocols are called here.
// PARAMETERS   Node *node
//                  Pointer to node.
// RETURN       None.
//
// NOTES        All network-layer models are also finalized here.
//-----------------------------------------------------------------------------

void
NetworkIpFinalize(Node *node)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    Scheduler *schedulerPtr = NULL;
    int i = 0;

#ifdef MILITARY_RADIOS_LIB
    if (ip->eplrsData)
    {
        Eplrs_NodeFinalize(node);
        Eplrs_Finalize();
    }
#endif
    // NetworkIpLoopbackForwardingTableDisplay(node);
    //NetworkPrintForwardingTable(node);

    if (ip->isIgmpEnable == TRUE)
    {
        IgmpFinalize(node);
    }

    if (ip->isIcmpEnable == TRUE)
    {
        NetworkIcmpFinalize(node);
    }

#ifdef ENTERPRISE_LIB
    // Finalize for route map
    RouteMapFinalize(node);

    if (ip->isEdgeRouter == TRUE)
    {
        DIFFSERV_MFTrafficConditionerFinalize(node);
    }

    if (ip->mobileIpStruct)
    {
        MobileIpFinalize(node);
    }

#endif // ENTERPRISE_LIB

#ifdef WIRELESS_LIB
    if (ip->isNdpEnable)
    {
        NdpFinalize(node);
    }
#endif // WIRELESS_LIB

#ifdef MILITARY_RADIOS_LIB
    if (ip->sdrData)
    {
        SdrFinalize(node);
    }
#endif

    for (i = 0; i < node->numberInterfaces; i++)
    {
        switch (NetworkIpGetUnicastRoutingProtocolType(node, i))
        {
            case MULTICAST_PROTOCOL_STATIC:
            {
                RoutingMulticastStaticFinalize(node);
                break;
            }
#ifdef WIRELESS_LIB
            case ROUTING_PROTOCOL_LAR1:
            {
                Lar1Finalize(node);
                break;
            }
            case ROUTING_PROTOCOL_AODV:
            {
                AodvFinalize(node, i);
                break;
            }
            case ROUTING_PROTOCOL_DYMO:
            {
                DymoFinalize(node, i);
                break;
            }
            case ROUTING_PROTOCOL_DSR:
            {
                DsrFinalize(node);
                break;
            }
            case ROUTING_PROTOCOL_FSRL:
            {
                FsrlFinalize(node);
                break;
            }
            case ROUTING_PROTOCOL_STAR:
            {
                StarFinalize(node);
                break;
            }
            case ROUTING_PROTOCOL_BRP:
            {
                BrpFinalize(node);
                break;
            }
            case ROUTING_PROTOCOL_ZRP:
            {
                ZrpFinalize(node);
                break;
            }
            case ROUTING_PROTOCOL_IARP:
            {
                IarpFinalize(node);
                break;
            }
            case ROUTING_PROTOCOL_IERP:
            {
                IerpFinalize(node);
                break;
            }
//StartANTHOCNET
            case ROUTING_PROTOCOL_ANTHOCNET:
            {
                AnthocnetFinalize(node);
                break;
            }
//EndANTHOCNET
#endif // WIRELESS_LIB

#ifdef ENTERPRISE_LIB
            case ROUTING_PROTOCOL_OSPFv2:
            {
                Ospfv2Finalize(node, i);
                break;
            }
            case ROUTING_PROTOCOL_IGRP:
            {
                IgrpFinalize(node);
                break;
            }
            case ROUTING_PROTOCOL_EIGRP:
            {
                EigrpFinalize(node);
                break;
            }
#endif // ENTERPRISE_LIB

#ifdef MILITARY_RADIOS_LIB
            case ROUTING_PROTOCOL_ODR:
            {
                OdrFinalize(node);
                break;
            }
#endif
            case ROUTING_PROTOCOL_NONE:
            {
                // The user explicitly indicated that this node is not
                // running a network-layer routing protocol.  This code
                // is never reached, since this value is never set.
                // -Jeff
                break;
            }

//InsertPatch FINALIZE_FUNCTION

            default:

                // This routing protocol is not at the network layer
                // (it does its finalization at its own layer), so just
                // break.

                break;

        }//switch//
    }//for//

    for (i = 0; i < node->numberInterfaces; i++)
    {
        switch (ip->interfaceInfo[i]->multicastProtocolType)
        {
#ifdef WIRELESS_LIB
            case MULTICAST_PROTOCOL_ODMRP:
            {
                OdmrpFinalize(node);
                break;
            }
#endif // WIRELESS_LIB

#ifdef ENTERPRISE_LIB
            case MULTICAST_PROTOCOL_DVMRP:
            {
                RoutingDvmrpFinalize(node);
                break;
            }
            case MULTICAST_PROTOCOL_MOSPF:
            {
                MospfFinalize(node);
                break;
            }
            case MULTICAST_PROTOCOL_PIM:
            {
                RoutingPimFinalize(node);
                break;
            }
#endif // ENTERPRISE_LIB

#ifdef ADDON_MAODV
            case MULTICAST_PROTOCOL_MAODV:
            {
                MaodvFinalize(node);
                break;
            }
#endif // ADDON_MAODV

            default:
            {
                break;
            }
        }
    }

    if (node->networkData.networkStats == TRUE)
    {
        NetworkIpPrintStats(node);
    }

    // Print stats for IPsec
    if (ip->isIPsecEnabled == TRUE)
    {
        IPsecFinalize(node);
    }

#ifdef ENTERPRISE_LIB
    // Print access list statistics
    if (ip->isACLStatOn == TRUE)
    {
        AccessListFinalize(node);
    }

    // Print access list statistics
    if (ip->isPBRStatOn == TRUE)
    {
        PbrFinalize(node);
    }

    // Print Route Redistribute statistics
    if (ip->isRtRedistributeStatOn)
    {
        // Finalize for Route Redistribution
        RouteRedistributeFinalize(node);
    }
#endif // ENTERPRISE_LIB

    for (i = 0; i < node->numberInterfaces; i++)
    {
        int j;

        schedulerPtr = ip->interfaceInfo[i]->scheduler;

        schedulerPtr->finalize(node, "Network", i, "IP");

        for (j = 0; j < schedulerPtr->numQueue(); j++)
        {
            schedulerPtr->invokeQueueFinalize(node, "Network", i, j, "IP");
        }

        // Print INPUT queue statistics
        if (ip->backplaneThroughputCapacity !=
            NETWORK_IP_UNLIMITED_BACKPLANE_THROUGHPUT)
        {
            Scheduler* inputSchedulerPtr =
                ip->interfaceInfo[i]->inputScheduler;

            inputSchedulerPtr->finalize(node, "Network", i, "IP", "Input");

            for (j = 0; j < inputSchedulerPtr->numQueue(); j++)
            {
                inputSchedulerPtr->invokeQueueFinalize(
                                    node, "Network", i, j, "IP", "Input");
            }
        }
    }

    // Print CPU queue statistics
    if (ip->backplaneThroughputCapacity !=
        NETWORK_IP_UNLIMITED_BACKPLANE_THROUGHPUT)
    {
        Scheduler* cpuSchedulerPtr = ip->cpuScheduler;

        cpuSchedulerPtr->finalize(node, "Network", 0, "IP", "Cpu");
        cpuSchedulerPtr->invokeQueueFinalize(
                            node, "Network", 0, 0, "IP", "Cpu");
    }
}

//-----------------------------------------------------------------------------
// Transport layer to IP, sends IP packets out to network
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpReceivePacketFromTransportLayer()
// PURPOSE      Called by transport layer protocols (UDP, TCP) to send
//              UDP datagrams and TCP segments using IP.  Simply calls
//              NetworkIpSendRawMessage().
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message from transport layer containing
//                  payload data (UDP datagram, TCP segment) for an
//                  IP packet. (IP header needs to be added)
//              NodeAddress sourceAddress
//                  Source IP address.
//                  See NetworkIpSendRawMessage() for special
//                  values.
//              NodeAddress destinationAddress
//                  Destination IP address.
//              int outgoingInterface
//                  outgoing interface to use to transmit packet.
//              TosType priority
//                  TOS of packet.
//              unsigned char protocol
//                  IP protocol number.
//              BOOL isEcnCapable
//                  Is this node ECN capable?
// RETURN       None.
//
// NOTES        Currently, the transport layer cannot specify the TTL;
//              the default TTL is used.
//-----------------------------------------------------------------------------

void
NetworkIpReceivePacketFromTransportLayer(
    Node *node,
    Message *msg,
    NodeAddress sourceAddress,
    NodeAddress destinationAddress,
    int outgoingInterface,
    TosType priority,
    unsigned char protocol,
    BOOL isEcnCapable)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    ip->isPacketEcnCapable = isEcnCapable;

    NetworkIpSendRawMessage(
        node,
        msg,
        sourceAddress,
        destinationAddress,
        outgoingInterface,
        priority,
        protocol,
        IPDEFTTL);
}

void
NetworkIpReceivePacketFromTransportLayer(
    Node *node,
    Message *msg,
    Address sourceAddress,
    Address destinationAddress,
    int outgoingInterface,
    TosType priority,
    unsigned char protocol,
    BOOL isEcnCapable)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    ip->isPacketEcnCapable = isEcnCapable;
#ifdef CELLULAR_LIB
    if ((node->networkData.networkProtocol == NETWORK_PROTOCOL_IP ||
         node->networkData.networkProtocol == NETWORK_PROTOCOL_GSM ||
         node->networkData.networkProtocol == NETWORK_PROTOCOL_CELLULAR) &&
        (sourceAddress.networkType == NETWORK_IPV4) &&
        (destinationAddress.networkType == NETWORK_IPV4))
#else
    if ((node->networkData.networkProtocol == NETWORK_PROTOCOL_IP ) &&
        (sourceAddress.networkType == NETWORK_IPV4) &&
        (destinationAddress.networkType == NETWORK_IPV4))
#endif // CELLULAR_LIB
    {
        NetworkIpSendRawMessage(
            node,
            msg,
            GetIPv4Address(sourceAddress),
            GetIPv4Address(destinationAddress),
            outgoingInterface,
            priority,
            protocol,
            IPDEFTTL);
    }
    else if (((node->networkData.networkProtocol == NETWORK_PROTOCOL_IPV6) ||
              (IsDualIpEnabled(node))) &&
             (sourceAddress.networkType == NETWORK_IPV6) &&
             (destinationAddress.networkType == NETWORK_IPV6))
    {
        Ipv6SendRawMessage(
                node,
                msg,
                GetIPv6Address(sourceAddress),
                GetIPv6Address(destinationAddress),
                outgoingInterface,
                priority,
                protocol,
                IPDEFTTL);
    }
    else if ((node->networkData.networkProtocol == NETWORK_PROTOCOL_IP) &&
        ((sourceAddress.networkType == NETWORK_ATM)
         || (destinationAddress.networkType == NETWORK_ATM)))
    {
        NodeAddress srcAddr;
        NodeAddress dstAddr;
        const LogicalSubnet* srcLogicalSubnet;
        const LogicalSubnet* dstLogicalSubnet;

        // If source is pure ATM node
        if (sourceAddress.networkType == NETWORK_ATM)
        {
            srcLogicalSubnet =
                AtmGetLogicalSubnetFromNodeId(
                node,
                sourceAddress.interfaceAddr.atm.ESI_pt1,
                DEFAULT_INTERFACE);

            srcAddr = srcLogicalSubnet->ipAddress;

            // destination is pure ATM node
            if (destinationAddress.networkType == NETWORK_ATM)
            {
                dstLogicalSubnet =
                    AtmGetLogicalSubnetFromNodeId(
                    node,
                    destinationAddress.interfaceAddr.atm.ESI_pt1,
                    DEFAULT_INTERFACE);

                dstAddr = dstLogicalSubnet->ipAddress;
            }
            else
            {
                dstAddr = destinationAddress.interfaceAddr.ipv4;
            }

            // Add dummy IP header
            NetworkIpAddHeader(
                node,
                msg,
                srcAddr,
                dstAddr,
                priority,
                protocol,
                IPDEFTTL);

            // At present only ATM adaptation layer available
            if (node->adaptationData.adaptationProtocol
                == ADAPTATION_PROTOCOL_NONE)
            {
                // so discard the packet
                MESSAGE_Free(node, msg);
                return;
            }

            // Handed over to adaptation layer
            ip->stats.ipSendPktToOtherNetwork++ ;

            ADAPTATION_ReceivePacketFromNetworkLayer(node, msg);
            return;
        }
        else if (destinationAddress.networkType == NETWORK_ATM)
        {
            dstLogicalSubnet =
                AtmGetLogicalSubnetFromNodeId(
                node,
                destinationAddress.interfaceAddr.atm.ESI_pt1,
                DEFAULT_INTERFACE);

            dstAddr = dstLogicalSubnet->ipAddress;

            srcAddr = sourceAddress.interfaceAddr.ipv4;

            // Process as a simple IP node
            NetworkIpSendRawMessage(
                node,
                msg,
                srcAddr,
                dstAddr,
                outgoingInterface,
                priority,
                protocol,
                IPDEFTTL);
        }
    }
    else
    {
        ERROR_ReportError("Invalid network type.\n");
    }
}


//-----------------------------------------------------------------------------
// Network layer to MAC layer, sends IP packets out to network
//-----------------------------------------------------------------------------

//
// FUNCTION     NetworkIpSendRawMessage()
// PURPOSE      Called by NetworkIpReceivePacketFromTransportLayer() to
//              send to send UDP datagrams, TCP segments using IP.  Also
//              called by network-layer routing protocols (AODV, OSPF,
//              etc.) to send IP packets.  This function adds an IP
//              header and calls RoutePacketAndSendToMac().
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message with payload data for IP packet.
//                  (IP header needs to be added)
//              NodeAddress sourceAddress
//                  Source IP address. (See notes)
//              NodeAddress destinationAddress
//                  Destination IP address.
//              int outgoingInterface
//                  outgoing interface to use to transmit packet.
//              TosType priority
//                  TOS of packet.
//              unsigned char protocol
//                  IP protocol number.
//              unsigned ttl
//                  Time to live.
//                  See AddIpHeader() for special values.
// RETURN       None.
//
// NOTES        If sourceAddress is ANY_IP, lets IP assign the source
//              address (depends on the route).
//

void
NetworkIpSendRawMessage(
    Node *node,
    Message *msg,
    NodeAddress sourceAddress,
    NodeAddress destinationAddress,
    int outgoingInterface,
    TosType priority,
    unsigned char protocol,
    unsigned ttl)
{
    NodeAddress newSourceAddress;
    int interfaceIndex;

    // We shouldn't be getting the outgoing interface like this.
    // NetworkGetInterfaceIndexForDestAddress() actually checks the
    // lookup table, and we do this is RoutePacketAndSendToMac().
    // We should do all the routing in RoutePacketAndSendToMac(). -Jeff
    if (outgoingInterface == ANY_INTERFACE)
    {
        // Trying to figure the source address and outgoing interface to use
        if (sourceAddress != ANY_IP)
        {
            interfaceIndex =
                   NetworkIpGetInterfaceIndexFromAddress(node, sourceAddress);

            newSourceAddress = sourceAddress;
        }
        else
        {
            interfaceIndex =
                 NetworkGetInterfaceIndexForDestAddress(node,
                                                        destinationAddress);

            newSourceAddress = NetworkIpGetInterfaceAddress(node,
                                                            interfaceIndex);
        }
    }
    else
    {
        interfaceIndex = outgoingInterface;
        newSourceAddress = sourceAddress;
    }

    AddIpHeader(
        node,
        msg,
        newSourceAddress,
        destinationAddress,
        priority,
        protocol,
        ttl);


    // interfaceIndex represents outgoing interface.
    RoutePacketAndSendToMac(node, msg, CPU_INTERFACE, interfaceIndex, ANY_IP);
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpSendRawMessageWithDelay()
// PURPOSE      Same as NetworkIpSendRawMessage(), but schedules
//              event after a simulation delay.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message with payload data for IP packet.
//                  (IP header needs to be added)
//              NodeAddress sourceAddress
//                  Source IP address. (See notes)
//              NodeAddress destinationAddress
//                  Destination IP address.
//              int outgoingInterface
//                  outgoing interface to use to transmit packet.
//              TosType priority
//                  Priority of packet.
//              unsigned char protocol
//                  IP protocol number.
//              unsigned ttl
//                  Time to live.
//                  See AddIpHeader() for special values.
//              clocktype delay
//                  Delay.
// RETURN       None.
//
// NOTES        If sourceAddress is ANY_IP, lets IP assign the source
//              address (depends on the route).
//-----------------------------------------------------------------------------

void
NetworkIpSendRawMessageWithDelay(
    Node *node,
    Message *msg,
    NodeAddress sourceAddress,
    NodeAddress destinationAddress,
    int outgoingInterface,
    TosType priority,
    unsigned char protocol,
    unsigned ttl,
    clocktype delay)
{
    NodeAddress newSourceAddress;
    int interfaceIndex;
    int *info;

    if (outgoingInterface == ANY_INTERFACE)
    {
        if (sourceAddress != ANY_IP)
        {
            interfaceIndex =
                    NetworkIpGetInterfaceIndexFromAddress(node, sourceAddress);

            newSourceAddress = sourceAddress;
        }
        else
        {
            interfaceIndex =
                NetworkGetInterfaceIndexForDestAddress(node,
                                                       destinationAddress);

            newSourceAddress =
                        NetworkIpGetInterfaceAddress(node, interfaceIndex);
        }
    }
    else
    {
        interfaceIndex = outgoingInterface;
        newSourceAddress = sourceAddress;
    }

    AddIpHeader(
        node,
        msg,
        newSourceAddress,
        destinationAddress,
        priority,
        protocol,
        ttl);

    MESSAGE_InfoAlloc(node, msg, sizeof(int));

    info = (int *) MESSAGE_ReturnInfo(msg);
    *info = interfaceIndex;

    MESSAGE_SetEvent(msg, MSG_NETWORK_FromTransportOrRoutingProtocol);
    MESSAGE_SetLayer(msg, NETWORK_LAYER, NETWORK_PROTOCOL_IP);
    MESSAGE_Send(node, msg, delay);
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpSendRawMessageToMacLayer()
// PURPOSE      Called by network-layer routing protocols (AODV, OSPF,
//              etc.) to add an IP header to payload data, and with
//              the resulting IP packet, calls
//              NetworkIpSendPacketOnInterface().
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message with payload data for IP packet.
//                  (IP header needs to be added)
//              NodeAddress sourceAddress
//                  Source IP address.
//              NodeAddress destinationAddress
//                  Destination IP address.
//              TosType priority
//                  Priority of packet.
//              unsigned char protocol
//                  IP protocol number.
//              unsigned ttl
//                  Time to live.
//                  See AddIpHeader() for special values.
//              int outgoingInterface
//                  Index of outgoing interface.
//              NodeAddress nextHop
//                  Next hop IP address.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkIpSendRawMessageToMacLayer(
    Node *node,
    Message *msg,
    NodeAddress sourceAddress,
    NodeAddress destinationAddress,
    TosType priority,
    unsigned char protocol,
    unsigned ttl,
    int outgoingInterface,
    NodeAddress nextHop)
{
    AddIpHeader(
        node,
        msg,
        sourceAddress,
        destinationAddress,
        priority,
        protocol,
        ttl);

    NetworkIpSendPacketOnInterface(node,
                                  msg,
                                  CPU_INTERFACE,
                                  outgoingInterface,
                                  nextHop);
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpSendRawMessageToMacLayerWithDelay()
// PURPOSE      Same as NetworkIpSendRawMessageToMacLayer(),
//              but schedules the event after a simulation delay
//              by calling NetworkIpSendPacketOnInterfaceWithDelay().
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message with payload data for IP packet.
//                  (IP header needs to be added)
//              NodeAddress sourceAddress
//                  Source IP address.
//              NodeAddress destinationAddress
//                  Destination IP address.
//              TosType priority
//                  TOS of packet.
//              unsigned char protocol
//                  IP protocol number.
//              unsigned ttl
//                  Time to live.
//                  See AddIpHeader() for special values.
//              int outgoingInterface
//                  Index of outgoing interface.
//              NodeAddress nextHop
//                  Next hop IP address.
//              clocktype delay
//                  Delay.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkIpSendRawMessageToMacLayerWithDelay(
    Node *node,
    Message *msg,
    NodeAddress sourceAddress,
    NodeAddress destinationAddress,
    TosType priority,
    unsigned char protocol,
    unsigned ttl,
    int outgoingInterface,
    NodeAddress nextHop,
    clocktype delay)
{
    AddIpHeader(
        node,
        msg,
        sourceAddress,
        destinationAddress,
        priority,
        protocol,
        ttl);

    NetworkIpSendPacketOnInterfaceWithDelay(
        node, msg, CPU_INTERFACE, outgoingInterface, nextHop, delay);
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpSendPacketToMacLayer()
// PURPOSE      This function is called once the outgoing interface
//              index and next hop address to which to route an IP
//              packet are known.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message with IP packet.
//              int outgoingInterface
//                  Index of outgoing interface.
//              NodeAddress nextHop
//                  Next hop IP address.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkIpSendPacketToMacLayer(
    Node *node,
    Message *msg,
    int outgoingInterface,
    NodeAddress nextHop)
{
    NetworkIpSendPacketOnInterface(node,
                                  msg,
                                  CPU_INTERFACE,
                                  outgoingInterface,
                                  nextHop);
}



//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpSendPacketOnInterface()
// PURPOSE      This function is called once the outgoing interface
//              index and next hop address to which to route an IP
//              packet are known.  This queues an IP packet for delivery
//              to the MAC layer.  This functions calls
//              QueueUpIpFragmentForMacLayer().
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message with IP packet.
//              int incomingInterface
//                  Index of incoming interface.
//              int outgoingInterface
//                  Index of outgoing interface.
//              NodeAddress nextHop
//                  Next hop IP address.
// RETURN       None.
//
// NOTES        This function is used to initiate fragmentation if
//              required, but since fragmentation has been disabled, all
//              it does is assert false if the IP packet is too big
//              before calling the next function.
//-----------------------------------------------------------------------------

void
NetworkIpSendPacketOnInterface(
    Node *node,
    Message *msg,
    int incomingInterface,
    int outgoingInterface,
    NodeAddress nextHop)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    IpHeaderType *ipHeader = (IpHeaderType *) msg->packet;

    // Handle packet to security protocol if IpSec enable over this interface
    if (outgoingInterface != CPU_INTERFACE
        && ip->interfaceInfo[outgoingInterface]->spdOUT
        && !IsIPsecProcessed(node, msg))
    {
        if (IPsecHandleOutboundPacket(
                node, msg, incomingInterface, outgoingInterface))
        {
            return;
        }
    }

    // Increment stat for total number of IP datagrams passed to IP for
    // transmission.  Does not include those counted in ipForwardDatagrams.

    ip->stats.ipOutRequests++;

    // DUAL-IP: Check whether outgoing interface is v6-tunneled
    if (DualIpGetInterfaceTunnelType(node, outgoingInterface) == IPv6_TUNNEL)
    {
        DualIpEncapsulateV4Pkt(
            node,
            msg,
            incomingInterface,
            outgoingInterface,
            nextHop);

        return;
    }
    else if (outgoingInterface != CPU_INTERFACE
             && NetworkIpGetInterfaceType(node, outgoingInterface) ==
                NETWORK_IPV6)
    {
        // Drop the pkt if the interface is not v6-tunneled and
        // also not connected with an IPv4 network.

        MESSAGE_Free(node, msg);
        return;
    }

    #ifdef STK_INTERFACE
        StkUpdateIpStats(node, STK_STAT_IP_ORIGINATED,
                       ip->stats.ipOutRequests, getSimTime(node));
    #endif /* STK_INTERFACE */

   //-------------------------------------------------------------------//
   //          Removed due to fragmentation.
   /*
    ERROR_Assert(ipHeader->ip_len <= (unsigned) ip->maxPacketLength,
                 "IP datagram is too large, check IP-FRAGMENTATION-UNIT "
                 "in .config file");
    */
    //-----------------------------------------------------------------//


#ifdef ENTERPRISE_LIB
    if ((outgoingInterface != CPU_INTERFACE) &&
        (ip->interfaceInfo[outgoingInterface]->accessListOutPointer))
    {
        if (AccessListFilterPacket(
                    node,
                    msg,
                    ip->interfaceInfo[outgoingInterface]->
                        accessListOutPointer,
                    outgoingInterface,
                    ACCESS_LIST_OUT))
        {
            // Matched, so drop and return
            //Trace drop
            ActionData acnData;
            acnData.actionType = DROP;
            acnData.actionComment = DROP_ACCESS_LIST;
            TRACE_PrintTrace(node,
                             msg,
                             TRACE_NETWORK_LAYER,
                             PACKET_OUT,
                             &acnData);

            return;
        }
    }
#endif // ENTERPRISE_LIB

    // Receive a packet from Loopback Interface, using VT to display.
    if (ip->isLoopbackEnabled && (outgoingInterface == CPU_INTERFACE) &&
        (incomingInterface == CPU_INTERFACE) && node->guiOption)
    {
        GUI_Receive(node->nodeId,
                    node->nodeId,
                    GUI_NETWORK_LAYER,
                    GUI_DEFAULT_DATA_TYPE,
                    outgoingInterface,
                    incomingInterface,
                    TIME_getSimTime(node));
    }

#ifdef ENTERPRISE_LIB
    PbrDebug(node, ipHeader, incomingInterface, nextHop, outgoingInterface);
#endif // ENTERPRISE_LIB

    NetworkIpSendOnBackplane(node,
                             msg,
                             incomingInterface,
                             outgoingInterface,
                             nextHop);

}//NetworkIpSendPacketOnInterface//


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpSendPacketToMacLayerWithDelay()
// PURPOSE      Same as NetworkIpSendPacketOnInterface(), but schedules
//              event after a simulation delay.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message with IP packet.
//              int outgoingInterface
//                  Index of outgoing interface.
//              NodeAddress nextHop
//                  Next hop IP address.
//              clocktype delay
//                  Delay.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkIpSendPacketToMacLayerWithDelay(
    Node *node,
    Message *msg,
    int outgoingInterface,
    NodeAddress nextHop,
    clocktype delay)
{
    NetworkIpSendPacketOnInterfaceWithDelay(node,
                                            msg,
                                            CPU_INTERFACE,
                                            outgoingInterface,
                                            nextHop,
                                            delay);
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpSendPacketOnInterfaceWithDelay()
// PURPOSE      Same as NetworkIpSendPacketOnInterface(), but schedules
//              event after a simulation delay.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message with IP packet.
//              int incomingInterface
//                  Index of incoming interface.
//              int outgoingInterface
//                  Index of outgoing interface.
//              NodeAddress nextHop
//                  Next hop IP address.
//              clocktype delay
//                  Delay.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkIpSendPacketOnInterfaceWithDelay(
    Node *node,
    Message *msg,
    int incomingInterface,
    int outgoingInterface,
    NodeAddress nextHop,
    clocktype delay)
{
    DelayedSendToMacLayerInfoType *info;

    MESSAGE_InfoAlloc(node, msg, sizeof(DelayedSendToMacLayerInfoType));

    info = (DelayedSendToMacLayerInfoType *) MESSAGE_ReturnInfo(msg);
    info->incomingInterface = incomingInterface;
    info->outgoingInterface = outgoingInterface;
    info->nextHop = nextHop;

    MESSAGE_SetLayer(msg, NETWORK_LAYER, NETWORK_PROTOCOL_IP);
    MESSAGE_SetEvent(msg, MSG_NETWORK_DelayedSendToMac);
    MESSAGE_Send(node, msg, delay);
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpSendPacketToMacLayerWithNewStrictSourceRoute()
// PURPOSE      Tacks on a new source route to an existing IP packet and
//              then sends the packet to the MAC layer.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                   Pointer to message with IP packet.
//              NodeAddress newRouteAddresses[]
//                   Source route (address array).
//              int numNewRouteAddresses
//                   Number of array elements.
//              BOOL removeExistingRecordedRoute
// RETURN       None.
//
// NOTES        BSD modifies the source route in the IP header as the
//              packet traverses each node in the source route.  QualNet
//              doesn't do this.
//
//              There is not a "WithDelay" version of this function.
//-----------------------------------------------------------------------------

void
NetworkIpSendPacketToMacLayerWithNewStrictSourceRoute(
    Node *node,
    Message *msg,
    NodeAddress newRouteAddresses[],
    int numNewRouteAddresses,
    BOOL removeExistingRecordedRoute)
{
    IpHeaderType *ipHeader = (IpHeaderType *) msg->packet;
    char *NewRoutePositionPtr;

    IpOptionsHeaderType *ipOption = IpHeaderSourceRouteOptionField(ipHeader);

    if (ipOption == NULL)
    {
        //
        // Add a source route to a packet with no source route
        // currently. The next equation uses the fact that this
        // sum will always be divisable by 4.
        //
        AddIpOptionField(node, msg, IPOPT_SSRR,
                         ((numNewRouteAddresses *sizeof(NodeAddress)) +
                          IP_SOURCE_ROUTE_OPTION_PADDING));
    }
    else
    {
        //
        // Replace a source route to already existing source routing option
        // header.
        //
        if (removeExistingRecordedRoute)
        {
            ipOption->ptr = IPOPT_MINOFF;
        }//if//

        ExpandOrShrinkIpOptionField(node, msg, IPOPT_SSRR,
                                    ((ipOption->ptr - 1) +
                                     (numNewRouteAddresses *
                                      sizeof(NodeAddress)) +
                                     IP_SOURCE_ROUTE_OPTION_PADDING));
    }//if//

    // Copy the new route into the option header. Must be
    // byte copy because of alignment issues.

    ipHeader = (IpHeaderType *) msg->packet;
    ipOption = IpHeaderSourceRouteOptionField(ipHeader);
    NewRoutePositionPtr =
        (char *) IpHeaderSourceRouteOptionField(ipHeader) + (ipOption->ptr -
                                                             1);

    memcpy(NewRoutePositionPtr, newRouteAddresses,
           (sizeof(NodeAddress) *numNewRouteAddresses));

    // Special case for DSR. Allows for new replacement source route
    // whose first addresses is current node. It just moves the route
    // pointer to the correct next hop.

    if ((removeExistingRecordedRoute) &&
        (newRouteAddresses[0] == node->nodeId))
    {
        ipOption->ptr = (unsigned char) (ipOption->ptr + (sizeof(NodeAddress)));
    }//if//

    SourceRouteThePacket(node, msg);

}//NetworkIpSendPacketToMacLayerWithNewStrictSourceRoute//

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpRoutingProtocolBroadcastMessageWithDelay()
// PURPOSE      Broadcast packet with a specified delay.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *rawMessage
//                  Pointer to message with payload data for IP packet.
//                  (IP header needs to be added)
//              NetworkRoutingProtocolType routingProtocol
//
//              TosType priority
//                  TOS of packet.
//              unsigned char protocol
//                  IP protocol number.
//              unsigned ttl
//                  Time to live.
//                  See AddIpHeader() for special values.
//              clocktype delay
//                  Delay.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkIpRoutingProtocolBroadcastMessageWithDelay(
    Node *node,
    Message *rawMessage,
    NetworkRoutingProtocolType routingProtocol,
    TosType priority,
    unsigned char protocol,
    unsigned ttl,
    clocktype delay)
{
    Message *nextMessage;
    int i;
    int firstIndex = -1;

    for (i = 0; i < node->numberInterfaces; i++)
    {
        if (NetworkIpGetUnicastRoutingProtocolType(node, i) == routingProtocol)
        {
            if (firstIndex == -1)
            {
                firstIndex = i;
            }
            else
            {
                nextMessage = MESSAGE_Duplicate(node, rawMessage);

                AddIpHeader(
                    node,
                    nextMessage,
                    NetworkIpGetInterfaceAddress(
                        node, i),
                    NetworkIpGetInterfaceBroadcastAddress(
                        node, i),
                    priority,
                    protocol,
                    ttl);

                NetworkIpSendPacketOnInterfaceWithDelay(
                    node, nextMessage, CPU_INTERFACE, i, ANY_DEST, delay);
            }
        }
    }

    if (firstIndex > -1)
    {
        AddIpHeader(
           node,
           rawMessage,
           NetworkIpGetInterfaceAddress(
               node, firstIndex),
           NetworkIpGetInterfaceBroadcastAddress(
               node, firstIndex),
           priority,
           protocol,
           ttl);

         NetworkIpSendPacketOnInterfaceWithDelay(
             node, rawMessage, CPU_INTERFACE,
             firstIndex, ANY_DEST, delay);
    }
}

//-----------------------------------------------------------------------------
// MAC layer to IP, receives IP packets from the MAC layer
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpReceivePacketFromMacLayer()
// PURPOSE      IP received IP packet from MAC layer.  Determine whether
//              the packet is to be delivered to this node, or needs to
//              be forwarded.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message with IP packet.
//              NodeAddress previousHopAddress
//                  nodeId of the previous hop.
//              int incomingInterface
//                  Index of interface on which packet arrived.
// RETURN       None.
//
// NOTES        ipHeader->ip_ttl is decremented here, instead of the
//              way BSD TCP/IP does it, which is to decrement TTL right
//              before forwarding the packet.  QualNet's alternative
//              method suits its network-layer ad hoc routing protocols,
//              which may do their own forwarding.
//-----------------------------------------------------------------------------

void
NetworkIpReceivePacketFromMacLayer(
    Node *node,
    Message *msg,
    NodeAddress previousHopAddress,
    int incomingInterface)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    IpHeaderType *ipHeader = (IpHeaderType *) msg->packet;

#ifdef ENTERPRISE_LIB
    // check the packet with the access list criteria
    if (ip->interfaceInfo[incomingInterface]->accessListInPointer)
    {
        if (AccessListFilterPacket(
                node,
                msg,
                ip->interfaceInfo[incomingInterface]->accessListInPointer,
                incomingInterface,
                ACCESS_LIST_IN))
        {
            // has matched, so drop and return
            //Trace drop
            ActionData acnData;
            acnData.actionType = DROP;
            acnData.actionComment = DROP_ACCESS_LIST;
            TRACE_PrintTrace(node,
                             msg,
                             TRACE_NETWORK_LAYER,
                             PACKET_IN,
                             &acnData);

            return;
        }
    }
#endif // ENTERPRISE_LIB

    // Handle packet to security protocol if IpSec enable over this interface
    if (ip->interfaceInfo[incomingInterface]->spdIN
        && IPsecRequireProcessing(node, msg))
    {
        IPsecHandleInboundPacket(
            node, msg, incomingInterface, previousHopAddress);

        return;
    }

    #ifdef STK_INTERFACE
        if (ipHeader->ip_dst != ANY_DEST)
        {
            StkDrawLine(node, -1, previousHopAddress, "blue");
        }
    #endif /* STK_INTERFACE */

//GuiStart
    if ((node->guiOption == TRUE))
    {
        NodeAddress previousHopNodeId;

        previousHopNodeId = MAPPING_GetNodeIdFromInterfaceAddress(
                                                          node,
                                                          previousHopAddress);
        //
        // Receive a packet from MAC, using VT to display.

        // Due to the IPv6 related changes according to the GUI team proposal
        // this function will expect sending interface and receiving interface
        // Right now previousHopAddress is used at sending interface, yet to
        // be decided by GUI team
        GUI_Receive(previousHopNodeId, node->nodeId, GUI_NETWORK_LAYER,
                    GUI_DEFAULT_DATA_TYPE,
                    MAPPING_GetInterfaceIndexFromInterfaceAddress(
                        node,
                        previousHopAddress),
                    incomingInterface,
                    getSimTime(node));
    }
//GuiEnd

    // Increment stat for total number of received IP datagrams from all
    // interfaces.

    ip->stats.ipInReceives++;

    #ifdef STK_INTERFACE
        StkUpdateIpStats(node, STK_STAT_IP_RECEIVED,
                       ip->stats.ipInReceives, getSimTime(node));
    #endif /* STK_INTERFACE */

    // The IP header TTL field is decremented here, instead of the way
    // BSD TCP/IP does it, which is to decrement TTL right before
    // forwarding the packet.  QualNet's alternative method suits its
    // network-layer ad hoc routing protocols, which may do their own
    // forwarding.

    ipHeader->ip_ttl = (unsigned char) (ipHeader->ip_ttl - 1);

    // trace recd pkt
    ActionData acnData;
    acnData.actionType = RECV;
    acnData.actionComment = NO_COMMENT;
    TRACE_PrintTrace(node, msg, TRACE_NETWORK_LAYER, PACKET_IN, &acnData);

    // Whenever a multicast packet received at a node
    // the membership status of that node with specified group
    // is checked and if required the packet is delivered to
    // the upper layer

    if (NetworkIpIsMulticastAddress(node, ipHeader->ip_dst))
    {
        //For wireless scenarios the received packet may be duplicate
        // thus check the type of multicast protocl running
        // in incoming interface to prevent the duplicate delivery
        // to the upper layer.

        switch (ip->interfaceInfo[incomingInterface]->
            multicastProtocolType)
        {
#ifdef WIRELESS_LIB
            case MULTICAST_PROTOCOL_ODMRP:
            {
                if (OdmrpCheckIfItIsDuplicatePacket(node, msg))
                {
                    //Trace drop
                    ActionData acnData;
                    acnData.actionType = DROP;
                    acnData.actionComment = DROP_DUPLICATE_PACKET;
                    TRACE_PrintTrace(node,
                                     msg,
                                     TRACE_NETWORK_LAYER,
                                     PACKET_IN,
                                     &acnData);

                    // It is a duplicate packet no need to process it
                    MESSAGE_Free(node, msg);
                    return;
                }
                break;
            }
#endif // WIRELESS_LIB
            default:
                break;
        }
    }

    // IsMyPacket() determines whether the packet should be delivered or
    // forwarded.  IsMyPacket() checks multicast subscriptions as well
    // as the host and broadcast addresses of all local interfaces.

    if (IsMyPacket(node, ipHeader->ip_dst))
    {
        // Deliver IP packet to node.
        //
        // (We actually deliver a copy of the message instead of the
        // original.  This is in case the packet is multicast, one copy
        // of which is delivered, and the other copy, forwarded.)

        // Need to go through the backplane to the processor first before
        // deliverying data to higher layers...
        NetworkIpSendOnBackplane(node,
                                 MESSAGE_Duplicate(node, msg),
                                 incomingInterface,
                                 CPU_INTERFACE,
                                 previousHopAddress);

        //DeliverPacket(node, MESSAGE_Duplicate(node, msg),
        //              incomingInterface, previousHopAddress);

        if (NetworkIpIsMulticastAddress(node, ipHeader->ip_dst))
        {
            // Destination address is multicast address, so pass packet
            // to forwarding process for potential multicast routing.

            ForwardPacket(node, msg, incomingInterface, ANY_IP);
        }
        else
        {
            // Wasn't a multicast packet, so free message.
            //
            // [In my tests, the copy and free method happens to be 0.5%
            // faster than logic which copies only if necessary. -Jeff]

            MESSAGE_Free(node, msg);
        }
    }
    else
    {
        // Forward packet.
        ForwardPacket(node, msg, incomingInterface, previousHopAddress);
    }

}

//-----------------------------------------------------------------------------
// MAC-layer packet drop callbacks
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpNotificationOfPacketDrop()
// PURPOSE      Invoke callback functions when a packet is dropped.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message.
//              NodeAddress nextHopAddress,
//                  next hop address of dropped packet.
//              int interfaceIndex
//                  interface that experienced the packet drop.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkIpNotificationOfPacketDrop(Node *node,
                                  Message *msg,
                                  NodeAddress nextHopAddress,
                                  int interfaceIndex)
{
    // Totally Evil hack under new "subroutine" based layer
    // communication.

    MESSAGE_SetLayer(msg, NETWORK_LAYER, ROUTING_PROTOCOL_ALL);
    MESSAGE_SetEvent(msg, MSG_NETWORK_PacketDropped);

    HandleSpecialMacLayerStatusEvents(node,
                                      msg,
                                      nextHopAddress,
                                      interfaceIndex);
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpGetMacLayerStatusEventHandlerFunction()
// PURPOSE      Get the status event handler function pointer.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int interfaceIndex
//                  interface associated with the status handler function.
// RETURN       Status event handler function.
//-----------------------------------------------------------------------------

MacLayerStatusEventHandlerFunctionType
NetworkIpGetMacLayerStatusEventHandlerFunction(
    Node *node,
    int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    return ip->interfaceInfo[interfaceIndex]->
                                             macLayerStatusEventHandlerFunction;
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpSetMacLayerStatusEventHandlerFunction()
// PURPOSE      Allows the MAC layer to send status messages (e.g.,
//              packet drop, link failure) to a network-layer routing
//              protocol for routing optimization.
// PARAMETERS   Node *node
//                  Pointer to node.
//              MacLayerStatusEventHandlerFunctionType StatusEventHandlerPtr
//                  Status event handler function to call.
//              int interfaceIndex
//                  interface associated with the status event handler.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkIpSetMacLayerStatusEventHandlerFunction(
    Node *node,
    MacLayerStatusEventHandlerFunctionType StatusEventHandlerPtr,
    int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    ERROR_Assert(ip->interfaceInfo[interfaceIndex]->
                     macLayerStatusEventHandlerFunction == NULL,
                 "MAC-layer event handler function already set");

    ip->interfaceInfo[interfaceIndex]->macLayerStatusEventHandlerFunction =
                                                         StatusEventHandlerPtr;
}


//---------------------------------------------------------------------------
// Mac layer acknowledgement handlers
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// FUNCTION     NetworkIpGetMacLayerAckHandler
// PURPOSE      Get MAC layer ACK handler
// PARAMETERS   Node *node
//                  Pointer to node.
//              int interfaceIndex
//                  Interface associated with the ACK handler function.
// RETURN       ACK event handler function.
//---------------------------------------------------------------------------

MacLayerAckHandlerType
NetworkIpGetMacLayerAckHandler(
    Node *node,
    int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    return ip->interfaceInfo[interfaceIndex]->macAckHandler;
}


//---------------------------------------------------------------------------
// FUNCTION     NetworkIpReceiveMacAck
// PURPOSE      MAC received an ACK, so call ACK handler function.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int interfaceIndex
//                  Interface associated with the ACK handler function.
//              const Message *msg
//                  Message that was ACKed.
//              NodeAddress nextHop
//                  Next hop that sent the MAC layer ACK.
// RETURN       None
//---------------------------------------------------------------------------

void
NetworkIpReceiveMacAck(
    Node* node,
    int interfaceIndex,
    const Message* msg,
    NodeAddress nextHop)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    if (ip->checkMacAckHandler != FALSE)
    {
        MacLayerAckHandlerType macAckHandler
            = NetworkIpGetMacLayerAckHandler(
                  node, interfaceIndex);

        if (macAckHandler)
        {
            (macAckHandler)(node, interfaceIndex, msg, nextHop);
        }
    }
}

//---------------------------------------------------------------------------
// FUNCTION     NetworkIpSetMacLayerAckHandler
// PURPOSE      Set MAC layer ACK handler
// PARAMETERS   Node *node
//                  Pointer to node.
//              int interfaceIndex
//                  Interface associated with the ACK handler function.
// RETURN       None
//---------------------------------------------------------------------------

void
NetworkIpSetMacLayerAckHandler(
    Node *node,
    MacLayerAckHandlerType macAckHandlerPtr,
    int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    ERROR_Assert(ip->interfaceInfo[interfaceIndex]->
                     macAckHandler == NULL,
                     "Mac Ack handling function already set");

    ip->interfaceInfo[interfaceIndex]->macAckHandler = macAckHandlerPtr;

    ip->checkMacAckHandler = TRUE;
}


//-----------------------------------------------------------------------------
// MAC-layer promiscuous mode callbacks
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpSneakPeekAtMacPacket()
// PURPOSE      Called Directly by the MAC layer, this allows a routing
//              protocol to "sneak a peek" or "tap" messages it would not
//              normally see from the MAC layer.  This function will
//              possibly unfragment such packets and call the function
//              registered by the routing protocol to do the "Peek".
// PARAMETERS   Node *node
//                  Pointer to node.
//              const Message *msg
//                  The message being peeked at from the MAC layer.
//                  Must not be freed or modified!
//              int interface
//                  The interface of which the "peeked" message belongs to.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkIpSneakPeekAtMacPacket(Node *node,
                              const Message *msg,
                              int interfaceIndex,
                              NodeAddress prevHop)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    if (ip->checkMessagePeekFunction != FALSE)
    {
        PromiscuousMessagePeekFunctionType promiscuousMessagePeekFunction
            = NetworkIpGetPromiscuousMessagePeekFunction(
                  node, interfaceIndex);

        if (promiscuousMessagePeekFunction)
        {
            (promiscuousMessagePeekFunction)(node, msg, prevHop);
        }

    }//if//

}//NetworkIpSneakPeekAtMacPacket//

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpGetPromiscuousMessagePeekFunction()
// PURPOSE      Returns the network-layer function which will
//              promiscuously inspect packets.
//              See NetworkIpSneakPeekAtMacPacket().
// PARAMETERS   Node *node
//                  Pointer to node.
//              int interfaceIndex
//                  Interface associated with the peek function.
// RETURN       Peek function.
//-----------------------------------------------------------------------------

PromiscuousMessagePeekFunctionType
NetworkIpGetPromiscuousMessagePeekFunction(
    Node *node,
    int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    return ip->interfaceInfo[interfaceIndex]->promiscuousMessagePeekFunction;
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpSetPromiscuousMessagePeekFunction()
// PURPOSE      Sets the network-layer function which will
//              promiscuously inspect packets.
//              See NetworkIpSneakPeekAtMacPacket().
// PARAMETERS   Node *node
//                  Pointer to node.
//              PromiscuousMessagePeekFunctionType PeekFunctionPtr
//                  Peek function.
//              int interfaceIndex
//                  Interface associated with the peek function.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkIpSetPromiscuousMessagePeekFunction(
    Node *node,
    PromiscuousMessagePeekFunctionType PeekFunctionPtr,
    int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    ERROR_Assert(ip->interfaceInfo[interfaceIndex]->
                     promiscuousMessagePeekFunction == NULL,
                 "Promiscuous function already set");

    ip->interfaceInfo[interfaceIndex]->promiscuousMessagePeekFunction =
                                                               PeekFunctionPtr;

    ip->checkMessagePeekFunction = TRUE;
}

//-----------------------------------------------------------------------------
// Network layer to transport layer
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     SendToUdp()
// PURPOSE      Sends a UDP packet to UDP in the transport layer.
//              The source IP address, destination IP address, and
//              priority of the packet are also sent.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message with UDP packet.
//              TosType priority
//                  Priority of UDP packet.
//              NodeAddress sourceAddress
//                  Source IP address.
//              NodeAddress destinationAddress
//                  Destination IP address.
//              int incomingInterfaceIndex
//                  interface that received the packet.
// RETURN       None.
//-----------------------------------------------------------------------------

void
SendToUdp(
    Node *node,
    Message *msg,
    TosType priority,
    NodeAddress sourceAddress,
    NodeAddress destinationAddress,
    int incomingInterfaceIndex)
{
    NetworkToTransportInfo *infoPtr;

    MESSAGE_SetEvent(msg, MSG_TRANSPORT_FromNetwork);
    MESSAGE_SetLayer(msg, TRANSPORT_LAYER, TransportProtocol_UDP);
    MESSAGE_InfoAlloc(node, msg, sizeof(NetworkToTransportInfo));

    infoPtr = (NetworkToTransportInfo *) MESSAGE_ReturnInfo(msg);

    SetIPv4AddressInfo(&infoPtr->sourceAddr, sourceAddress);
    SetIPv4AddressInfo(&infoPtr->destinationAddr, destinationAddress);

    infoPtr->priority = priority;
    infoPtr->incomingInterfaceIndex = incomingInterfaceIndex;

    MESSAGE_Send(node, msg, PROCESS_IMMEDIATELY);
}

//-----------------------------------------------------------------------------
// FUNCTION     SendToTcp()
// PURPOSE      Same as SendToUdp(), except TCP packet.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message with TCP packet.
//              TosType priority
//                  Priority of UDP packet.
//              NodeAddress sourceAddress
//                  Source IP address.
//              NodeAddress destinationAddress
//                  Destination IP address.
//              BOOL aCongestionExperienced
//                  Determine if congestion is experienced (via ECN).
// RETURN       None.
//-----------------------------------------------------------------------------

void
SendToTcp(
    Node *node,
    Message *msg,
    TosType priority,
    NodeAddress sourceAddress,
    NodeAddress destinationAddress,
    BOOL aCongestionExperienced)
{
    NetworkToTransportInfo *infoPtr;

    MESSAGE_SetEvent(msg, MSG_TRANSPORT_FromNetwork);
    MESSAGE_SetLayer(msg, TRANSPORT_LAYER, TransportProtocol_TCP);
    MESSAGE_InfoAlloc(node, msg, sizeof(NetworkToTransportInfo));

    infoPtr = (NetworkToTransportInfo *) MESSAGE_ReturnInfo(msg);

    SetIPv4AddressInfo(&infoPtr->sourceAddr, sourceAddress);
    SetIPv4AddressInfo(&infoPtr->destinationAddr, destinationAddress);

    infoPtr->priority = priority;
    infoPtr->isCongestionExperienced = aCongestionExperienced;

    MESSAGE_Send(node, msg, PROCESS_IMMEDIATELY);
}

//-----------------------------------------------------------------------------
// FUNCTION     SendToRsvp()
// PURPOSE      Same as SendToUdp(), except RSVP packet.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message with RSVP packet.
//              TosType priority
//                  Priority of UDP packet.
//              NodeAddress sourceAddress
//                  Source IP address.
//              NodeAddress destinationAddress
//                  Destination IP address.
//              int interfaceIndex
//                  Incoming interface index
//              unsigned ttl
//                  Receiving TTL
// RETURN       None.
//-----------------------------------------------------------------------------

void
SendToRsvp(
    Node *node,
    Message *msg,
    TosType priority,
    NodeAddress sourceAddress,
    NodeAddress destinationAddress,
    int interfaceIndex,
    unsigned ttl)
{
    NetworkToTransportInfo *infoPtr;

    MESSAGE_SetEvent(msg, MSG_TRANSPORT_FromNetwork);
    MESSAGE_SetLayer(msg, TRANSPORT_LAYER, TransportProtocol_RSVP);
    MESSAGE_InfoAlloc(node, msg, sizeof(NetworkToTransportInfo));

    infoPtr = (NetworkToTransportInfo *) MESSAGE_ReturnInfo(msg);

    SetIPv4AddressInfo(&infoPtr->sourceAddr, sourceAddress);
    SetIPv4AddressInfo(&infoPtr->destinationAddr, destinationAddress);

    infoPtr->priority = priority;
    infoPtr->incomingInterfaceIndex = interfaceIndex;
    infoPtr->receivingTtl = ttl;

    MESSAGE_Send(node, msg, PROCESS_IMMEDIATELY);
}

//-----------------------------------------------------------------------------
// FUNCTION     SendToTransport()
// PURPOSE      Same as SendToUdp(), except designer transport protocol packet.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message with RSVP packet.
//              TosType priority
//                  Priority of UDP packet.
//              NodeAddress sourceAddress
//                  Source IP address.
//              NodeAddress destinationAddress
//                  Destination IP address.
//              int interfaceIndex
//                  Incoming interface index
//              unsigned ttl
//                  Receiving TTL
// RETURN       None.
//-----------------------------------------------------------------------------

void
SendToTransport(
    Node *node,
    Message *msg,
    TosType priority,
    NodeAddress sourceAddress,
    NodeAddress destinationAddress,
    int interfaceIndex,
    unsigned ttl,
    TransportProtocol protocolType)
{
    NetworkToTransportInfo *infoPtr;

    MESSAGE_SetEvent(msg, MSG_TRANSPORT_FromNetwork);
    MESSAGE_SetLayer(msg, TRANSPORT_LAYER, protocolType);
    MESSAGE_InfoAlloc(node, msg, sizeof(NetworkToTransportInfo));

    infoPtr = (NetworkToTransportInfo *) MESSAGE_ReturnInfo(msg);

    SetIPv4AddressInfo(&infoPtr->sourceAddr, sourceAddress);
    SetIPv4AddressInfo(&infoPtr->destinationAddr, destinationAddress);

    infoPtr->priority = priority;
    infoPtr->incomingInterfaceIndex = interfaceIndex;
    infoPtr->receivingTtl = ttl;

    MESSAGE_Send(node, msg, PROCESS_IMMEDIATELY);
}


//-----------------------------------------------------------------------------
// IP header
//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpRemoveIpHeader()
// PURPOSE      Removes the IP header from a message while also
//              returning all the fields of the header.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message.
//              NodeAddress *sourceAddress
//                  Storage for source IP address.
//              NodeAddress *destinationAddress
//                  Storage for destination IP address.
//              TosType *priority
//                  Storage for TosType.
//                  (values are not standard for "IP type of service field"
//                  but has correct function)
//              unsigned char *protocol
//                  Storage for IP protocol number.
//              unsigned *ttl
//                  Storage for time to live.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkIpRemoveIpHeader(
    Node *node,
    Message *msg,
    NodeAddress *sourceAddress,
    NodeAddress *destinationAddress,
    TosType *priority,
    unsigned char *protocol,
    unsigned *ttl)
{
    IpHeaderType *ipHeader = (IpHeaderType *) msg->packet;

    *priority = (TosType) ipHeader->ip_tos;
    *ttl = ipHeader->ip_ttl;
    *protocol = ipHeader->ip_p;
    *sourceAddress = ipHeader->ip_src;
    *destinationAddress = ipHeader->ip_dst;

    MESSAGE_RemoveHeader(node, msg, IpHeaderSize(ipHeader), TRACE_IP);
}

//-----------------------------------------------------------------------------
// IP header option field
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     AddIpOptionField()
// PURPOSE      Inserts an option field in the header of an IP packet.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message.
//              int optionCode
//                  The option code
//              int optionSize
//                  Size of the option
// RETURN       None.
//
// NOTES        optionCode should not conflict with standard ones, like
//              the codes for loose and strict source routing.
//              optionSize specifies the size of the option field beyond
//              the 3 byte option header.  Currently optionSize must
//              satisfy (optionSize + 3) % 4 == 0.  Header fields are
//              moved by this operation, so watch out for lingering
//              pointers into the message.
//-----------------------------------------------------------------------------

void
AddIpOptionField(
    Node *node,
    Message *msg,
    int optionCode,
    int optionSize)
{
    IpOptionsHeaderType *newIpOption = NULL;
    int oldHeaderSize = IpHeaderSize((IpHeaderType *) msg->packet);

    // Round up to nearest option size divisable by 4.
    int newIpOptionSize =
        4 * ((((sizeof(IpOptionsHeaderType) + optionSize) - 1) / 4) + 1);
    int newHeaderSize = oldHeaderSize + newIpOptionSize;

    ERROR_Assert(
        FindAnIpOptionField((IpHeaderType *) msg->packet, optionCode) == NULL,
        "Option already exists in IP header");

    ExpandOrShrinkIpHeader(node, msg, newHeaderSize);
    newIpOption =
        (IpOptionsHeaderType *) ((char *) msg->packet + oldHeaderSize);
    newIpOption->len = (unsigned char) newIpOptionSize;
    newIpOption->code = (unsigned char) optionCode;
    newIpOption->ptr = IPOPT_MINOFF;
}

//-----------------------------------------------------------------------------
// Source route
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     ExtractIpSourceAndRecordedRoute()
// PURPOSE      Retrieves a copy of the source and recorded route from
//              the options field in the header.
// PARAMETERS   Message *msg
//                  Pointer to message with IP packet.
//              NodeAddress RouteAddresses[]
//                  Storage for source/recorded route.
//              int *NumAddresses
//                  Storage for size of RouteAddresses[] array.
//              int *RouteAddressIndex
//                  The index of the first address of the source route;
//                  before this index is the recorded route.
// RETURN       None.
//-----------------------------------------------------------------------------

void
ExtractIpSourceAndRecordedRoute(
    Message *msg,
    NodeAddress RouteAddresses[],
    int *NumAddresses,
    int *RouteAddressIndex)
{
    IpHeaderType *ipHeader = (IpHeaderType *) msg->packet;
    IpOptionsHeaderType *ipOptions = IpHeaderSourceRouteOptionField(ipHeader);
    char *FirstAddress = ((char *) ipOptions + sizeof(IpOptionsHeaderType));

    ERROR_Assert(
        IpHeaderHasSourceRoute(ipHeader),
        "Cannot source route when IP header does not have source route");
    *NumAddresses =
        (ipOptions->len - sizeof(IpOptionsHeaderType) -
         IP_SOURCE_ROUTE_OPTION_PADDING) / sizeof(NodeAddress);
    *RouteAddressIndex =
        (ipOptions->ptr - sizeof(IpOptionsHeaderType) -
         IP_SOURCE_ROUTE_OPTION_PADDING) / sizeof(NodeAddress);
    memmove(RouteAddresses, FirstAddress,
            ((*NumAddresses) *sizeof(NodeAddress)));
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpGetRouterFunction()
// PURPOSE      Get the router function pointer.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int interfaceIndex
//                  interface associated with router function.
// RETURN       Router function.
//-----------------------------------------------------------------------------

RouterFunctionType
NetworkIpGetRouterFunction(
    Node *node,
    int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    return ip->interfaceInfo[interfaceIndex]->routerFunction;
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpSetRouterFunction()
// PURPOSE      Allows a routing protocol to set the "routing function"
//              (one of its functions) which is called when a packet
//              needs to be routed.
// PARAMETERS   Node *node
//                  Pointer to node.
//              RouterFunctionType RouterFunctionPtr
//                  Router function to set.
//              int interfaceIndex
//                  interface associated with router function.
// RETURN       None.
//
// NOTES        NetworkIpSetRouterFunction() allows a routing protocol
//              to define the routing function.  The routing function
//              is called by the network layer to ask the routing
//              protocol to route the packet.  The routing function is
//              given the packet and its destination.  The routing
//              protocol can route the packet and set "packetWasRouted"
//              to TRUE; or not route the packet and set to FALSE.  If
//              the packet, was not routed, then the network layer will
//              try to use the forwarding table or the source route the
//              source route in the IP header.  This function will also
//              be given packets for the local node the routing
//              protocols can look at packets for protocol reasons.  In
//              this case, the message should not be modified and
//              packetWasRouted must be set to FALSE.
//-----------------------------------------------------------------------------

void
NetworkIpSetRouterFunction(
    Node *node,
    RouterFunctionType routerFunctionPtr,
    int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
/*
    ERROR_Assert(ip->interfaceInfo[interfaceIndex]->routerFunction ==
                 NULL,
                 "Router function already set");
*/
    ip->interfaceInfo[interfaceIndex]->routerFunction = routerFunctionPtr;
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpAddUnicastRoutingProtocolType()
// PURPOSE      Add unicast routing protocol type to interface.
// PARAMETERS   Node *node
//                  Pointer to node.
//              NetworkRoutingProtocolType routingProtocolType
//                  Router function to add.
//              int interfaceIndex
//                  Interface associated with the router function.
// RETURN       None.
//-----------------------------------------------------------------------------
void
NetworkIpAddUnicastRoutingProtocolType(
    Node *node,
    NetworkRoutingProtocolType routingProtocolType,
    int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    if (ip->interfaceInfo[interfaceIndex]->interfaceType == NETWORK_IPV6)
    {
        ip->interfaceInfo[interfaceIndex]->ipv6InterfaceInfo->
            routerFunction = NULL;
        ip->interfaceInfo[interfaceIndex]->ipv6InterfaceInfo->
            routingProtocolType = routingProtocolType;
        ip->interfaceInfo[interfaceIndex]->ipv6InterfaceInfo->
            routingProtocol = NULL;
        return;
    }

    ip->interfaceInfo[interfaceIndex]->routerFunction = NULL;
    ip->interfaceInfo[interfaceIndex]->routingProtocolType =
                                         routingProtocolType;
    ip->interfaceInfo[interfaceIndex]->routingProtocol = NULL;
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpGetRoutingProtocol()
// PURPOSE      Get routing protocol structure associated with routing protocol
//              running on this interface.
// PARAMETERS   Node *node
//                  Pointer to node.
//              NetworkRoutingProtocolType routingProtocolType
//                  Routing protocol to retrieve.
// RETURN       Routing protocol structure requested.
//-----------------------------------------------------------------------------

void *
NetworkIpGetRoutingProtocol(
    Node *node,
    NetworkRoutingProtocolType routingProtocolType)
{
    int i;
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    for (i = 0; i < node->numberInterfaces; i++)
    {
        if (NetworkIpGetUnicastRoutingProtocolType(node, i)
            == routingProtocolType)
        {
            if (ip->interfaceInfo[i]->interfaceType == NETWORK_IPV6)
            {
                IPv6InterfaceInfo *interfaceInfo;
                interfaceInfo = ip->interfaceInfo[i]->ipv6InterfaceInfo;

                return interfaceInfo->routingProtocol;
            }
            return ip->interfaceInfo[i]->routingProtocol;
        }
    }

    return NULL; // Not reachable.
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpGetUnicastRoutingProtocolType()
// PURPOSE      Get unicast routing protocol type on this interface.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int interfaceIndex
//                  network interface for request
// RETURN       The unicast routing protocol type.
//-----------------------------------------------------------------------------

NetworkRoutingProtocolType
NetworkIpGetUnicastRoutingProtocolType(
    Node *node,
    int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    if (ip->interfaceInfo[interfaceIndex]->interfaceType == NETWORK_IPV6)
    {
        IPv6InterfaceInfo *interfaceInfo;
        interfaceInfo = ip->interfaceInfo[interfaceIndex]->ipv6InterfaceInfo;

        return interfaceInfo->routingProtocolType;
    }

    return ip->interfaceInfo[interfaceIndex]->routingProtocolType;
}


// FUNCTION   NetworkIpUpdateUnicastRoutingProtocolAndRouterFunction
// PURPOSE    Assign unicast routing protocol structure and router
//            function to an interface.  We are only allocating
//            the unicast routing protocol structure and router function
//            once by using pointers to the original structures.
// PARAMETERS node - this node.
//            routingProtocolType - unicast routing protocol to add.
//            interfaceIndex - interface associated with unicast protocol.
// RETURN     None.
void
NetworkIpUpdateUnicastRoutingProtocolAndRouterFunction(
    Node *node,
    NetworkRoutingProtocolType routingProtocolType,
    int interfaceIndex)
{
    int i;
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    for (i = 0; i < node->numberInterfaces; i++)
    {
        if (NetworkIpGetUnicastRoutingProtocolType(node, i)
            == routingProtocolType)
        {

            if (ip->interfaceInfo[interfaceIndex]->interfaceType
                == NETWORK_IPV6)
            {
                IPv6InterfaceInfo *interfaceInfo1;
                IPv6InterfaceInfo *interfaceInfo2;
                interfaceInfo1 = ip->interfaceInfo[interfaceIndex]
                                ->ipv6InterfaceInfo;

                interfaceInfo2 = ip->interfaceInfo[i]->ipv6InterfaceInfo;

                interfaceInfo1->routerFunction
                                    = interfaceInfo2->routerFunction;
                interfaceInfo1->routingProtocol
                                    = interfaceInfo2->routingProtocol;

                interfaceInfo1->macLayerStatusEventHandlerFunction =
                    interfaceInfo2->macLayerStatusEventHandlerFunction;

                return;
            }

            ip->interfaceInfo[interfaceIndex]->routerFunction =
                            ip->interfaceInfo[i]->routerFunction;

            ip->interfaceInfo[interfaceIndex]->routingProtocol =
                            ip->interfaceInfo[i]->routingProtocol;

            ip->interfaceInfo[interfaceIndex]->
                macLayerStatusEventHandlerFunction =
                      ip->interfaceInfo[i]->macLayerStatusEventHandlerFunction;

            ip->interfaceInfo[interfaceIndex]->
                promiscuousMessagePeekFunction =
                      ip->interfaceInfo[i]->promiscuousMessagePeekFunction;

            return;
        }
    }

    fprintf(stderr, "Could not find unicast router function %d\n",
            routingProtocolType);
    abort();
}


void NetworkIpSetHsrpOnInterface(Node *node, int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    ip->interfaceInfo[interfaceIndex]->hsrpEnabled = TRUE;
}

BOOL NetworkIpIsHsrpEnabled(Node *node, int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    if (interfaceIndex == ANY_INTERFACE)
    {
        int i;

        for (i = 0; i < node->numberInterfaces; i++)
        {
            if (ip->interfaceInfo[i]->hsrpEnabled)
            {
                return TRUE;
            }
        }

        return FALSE;
    }
    else
    {
        return ip->interfaceInfo[interfaceIndex]->hsrpEnabled;
    }
}



//-----------------------------------------------------------------------------
// Interface creation
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpAddNewInterface()
// PURPOSE      Add new interface to node.
// PARAMETERS   Node *node
//                  Pointer to node.
//              NodeAddress interfaceIpAddress
//                  Interface to add.
//              int numHostBits
//                  Number of host bits for the interface.
//              int *newInterfaceIndex
//                  The interface number of the new interface.
//              const NodeInput *nodeInput
//                  Provides access to configuration file.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkIpAddNewInterface(
    Node *node,
    NodeAddress interfaceIpAddress,
    int numHostBits,
    int *newInterfaceIndex,
    const NodeInput *nodeInput)
{
    BOOL wasFound;
    BOOL typeFound = FALSE;
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    int interfaceIndex = node->numberInterfaces;
    char intfString[MAX_STRING_LENGTH];

    node->numberInterfaces++;
    *newInterfaceIndex = interfaceIndex;

    ERROR_Assert(
        (interfaceIndex >= 0) && (interfaceIndex < MAX_NUM_INTERFACES),
        "Number of interfaces has exceeded MAX_NUM_INTERFACES or is < 0");

    ip->interfaceInfo[interfaceIndex] =
        (IpInterfaceInfoType*)MEM_malloc(sizeof(IpInterfaceInfoType));

    ip->interfaceInfo[interfaceIndex]->interfaceType = NETWORK_IPV4;

    ip->interfaceInfo[interfaceIndex]->ipAddress = interfaceIpAddress;
    ip->interfaceInfo[interfaceIndex]->numHostBits = numHostBits;


    ip->interfaceInfo[interfaceIndex]->routerFunction = NULL;
    ip->interfaceInfo[interfaceIndex]->routingProtocolType =
                                                         ROUTING_PROTOCOL_NONE;
    ip->interfaceInfo[interfaceIndex]->routingProtocol = NULL;

    ip->interfaceInfo[interfaceIndex]->multicastEnabled = FALSE;
    ip->interfaceInfo[interfaceIndex]->multicastRouterFunction = NULL;
    ip->interfaceInfo[interfaceIndex]->multicastProtocolType =
                                                         ROUTING_PROTOCOL_NONE;
    ip->interfaceInfo[interfaceIndex]->multicastRoutingProtocol = NULL;

    ip->interfaceInfo[interfaceIndex]->macLayerStatusEventHandlerFunction =
                                                                          NULL;
    ip->interfaceInfo[interfaceIndex]->promiscuousMessagePeekFunction = NULL;
    ip->interfaceInfo[interfaceIndex]->macAckHandler = NULL;

    ip->interfaceInfo[interfaceIndex]->backplaneStatus =
                                                     NETWORK_IP_BACKPLANE_IDLE;

    ip->interfaceInfo[interfaceIndex]->ipv6InterfaceInfo = NULL;

    ip->interfaceInfo[interfaceIndex]->hsrpEnabled = FALSE;

#ifdef ENTERPRISE_LIB
    // Access list initialization.
    ip->interfaceInfo[interfaceIndex]->accessListInPointer = NULL;
    ip->interfaceInfo[interfaceIndex]->accessListOutPointer = NULL;

    ip->interfaceInfo[interfaceIndex]->
                            accessListStat.packetDroppedByExtdAtIN = 0;
    ip->interfaceInfo[interfaceIndex]->
                            accessListStat.packetDroppedByExtdAtOUT = 0;
    ip->interfaceInfo[interfaceIndex]->
                            accessListStat.packetDroppedByStdAtOut = 0;
    ip->interfaceInfo[interfaceIndex]->
                            accessListStat.packetDroppedByStdAtIN = 0;
    ip->interfaceInfo[interfaceIndex]->
                            accessListStat.packetDroppedForMismatchAtOut = 0;
    ip->interfaceInfo[interfaceIndex]->
                            accessListStat.packetDroppedForMismatchAtIN = 0;

    ip->interfaceInfo[interfaceIndex]->routingTableUpdateFunction = NULL;

    ip->interfaceInfo[interfaceIndex]->rMapForPbr = NULL;

    // for PBR
    ip->interfaceInfo[interfaceIndex]->pbrStat.packetsPolicyRoutedLocal = 0;
    ip->interfaceInfo[interfaceIndex]->pbrStat.packetsNotPolicyRouted = 0;
    ip->interfaceInfo[interfaceIndex]->pbrStat.packetPrecSet = 0;
    ip->interfaceInfo[interfaceIndex]->pbrStat.packetsPolicyRouted = 0;
#endif // ENTERPRISE_LIB

    // Initialize Security Structures
    ip->interfaceInfo[interfaceIndex]->spdIN = NULL;
    ip->interfaceInfo[interfaceIndex]->spdOUT = NULL;

    sprintf(ip->interfaceInfo[interfaceIndex]->interfaceName,
            "interface%d", interfaceIndex);

    IO_ReadStringInstance(
            node->nodeId,
            ANY_DEST,
            nodeInput,
            "INTERFACE-NAME",
            interfaceIndex,    // parameterInstanceNumber
            FALSE,             // fallbackIfNoInstanceMatch
            &wasFound,
            ip->interfaceInfo[interfaceIndex]->interfaceName);

#ifdef ENTERPRISE_LIB
    // For Interface type and interface number follow the exact syntax:
    // [node-id] INTERFACE-TYPE interface-index
    //      interface-type interface-number
    IO_ReadStringInstance(
            node->nodeId,
            ANY_DEST,
            nodeInput,
            "INTERFACE-TYPE",
            interfaceIndex,
            FALSE,
            &typeFound,
            intfString);

    if (typeFound)
    {
        RtParseInterfaceTypeAndNumber(node, interfaceIndex, intfString);
    }
#endif // ENTERPRISE_LIB

}

//-----------------------------------------------------------------------------
// Queue setup
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpInitOutputQueueConfiguration()
// PURPOSE      Initializes queue parameters during startup.
// PARAMETERS   Node *node
//                  Pointer to node.
//              const NodeInput *nodeInput
//                  Pointer to node input.
//              int interfaceIndex
//                  interface associated with queue.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkIpInitOutputQueueConfiguration(
    Node *node,
    const NodeInput *nodeInput,
    int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    Address address;
    Scheduler *schedulerPtr = NULL;
    char schedulerTypeString[MAX_STRING_LENGTH] = {0};
    BOOL enableSchedulerStat = FALSE;
    int numPriorities = ALL_PRIORITIES;
    int i = 0;
    BOOL wasFound = FALSE;
    char buf[MAX_STRING_LENGTH] = {0};
    char graphDataStr[MAX_STRING_LENGTH] = "NA";

    NetworkGetInterfaceInfo(node, interfaceIndex, &address);

    // Read Number of Queues under the interface Scheduler
    // that chooses which packet to transmit next.
    IO_ReadInt(
        node->nodeId,
        &address,
        nodeInput,
        "IP-QUEUE-NUM-PRIORITIES",
        &wasFound,
        &numPriorities);

    if (!wasFound)
    {
        NetworkIpConfigurationError(node,
                                    "IP-QUEUE-NUM-PRIORITIES",
                                    interfaceIndex);
    }

    if ((numPriorities < 1) || (numPriorities > NUM_MAX_IP_QUEUE))
    {
        NetworkIpConfigurationError(node,
                                    "IP-QUEUE-NUM-PRIORITIES",
                                    interfaceIndex);
    }

    // Read if Scheduler Statistics is enabled
    IO_ReadString(
            node->nodeId,
            &address,
            nodeInput,
            "SCHEDULER-STATISTICS",
            &wasFound,
            buf);

    if (wasFound && (!strcmp(buf, "YES")))
    {
        enableSchedulerStat = TRUE;
    }

    // Initialize the SchedGraphStat Structure for each interface
    IO_ReadString(
        node->nodeId,
        &address,
        nodeInput,
        "SCHEDULER-GRAPH-STATISTICS",
        &wasFound,
        buf);

    if (wasFound && strcmp(buf, "YES") == 0)
    {
        BOOL retVal = FALSE;
        char sampleIntervalStr[MAX_STRING_LENGTH] = {0};

        //NetworkIpInterfaceInitForGraph(node, nodeInput, interfaceIndex);
        IO_ReadString(
                node->nodeId,
                &address,
                nodeInput,
                "SAMPLE-INTERVAL",
                &retVal,
                sampleIntervalStr);

        if (retVal)
        {
            char addressString[MAX_ADDRESS_STRING_LENGTH];
            char checkIntervalStr[MAX_STRING_LENGTH] = {0};

            strcpy(checkIntervalStr, sampleIntervalStr);
            IO_ConvertIpAddressToString(&address, addressString);
            strcat(addressString, ".out");

            if ((TIME_ConvertToClock(checkIntervalStr) < 0) ||
                (TIME_ConvertToClock(checkIntervalStr)
                        > TIME_ReturnMaxSimClock(node)))
            {
                NetworkIpConfigurationError(node,
                                        "SAMPLE-INTERVAL",
                                        interfaceIndex);
            }

            // graphDataStr: <sampleInterval> <interfaceAddress.out>
            sprintf(graphDataStr, "%s %s",sampleIntervalStr, addressString);
        }
        else
        {
            ERROR_ReportError("a SAMPLE-INTERVAL value should be specifed"
                " into config file and value must be positive\n");
        }
    }

    // Set the Scheduler
    IO_ReadString(
        node->nodeId,
        &address,
        nodeInput,
        "IP-QUEUE-SCHEDULER",
        &wasFound,
        schedulerTypeString);

    if (!wasFound)
    {
        NetworkIpConfigurationError(node,
                                    "IP-QUEUE-SCHEDULER",
                                    interfaceIndex);
    }
    else
    {
        if (!strcmp(schedulerTypeString, "CBQ"))
        {
            NodeInput lsrmInput;
            CBQResourceManager* resrcMngr = NULL;
            char pktSchedulerString[MAX_STRING_LENGTH] = {0};
            char resrcMngrGuideLine[MAX_STRING_LENGTH] = {0};
            int resrcMngrGuideLineLevel = 0;
            int interfaceBandwidth =
                    node->macData[interfaceIndex]->bandwidth;

            ReadCBQResourceManagerConfiguration(node,
                &address,
                nodeInput,
                &lsrmInput,
                resrcMngrGuideLine,
                &resrcMngrGuideLineLevel,
                pktSchedulerString);

            RESOURCE_MANAGER_Setup(node,
                                    lsrmInput,
                                    interfaceIndex,
                                    NETWORK_LAYER,
                                    interfaceBandwidth,
                                    &resrcMngr,
                                    resrcMngrGuideLine,
                                    resrcMngrGuideLineLevel,
                                    pktSchedulerString,
                                    enableSchedulerStat,
                                    TRUE,
                                    graphDataStr);

            schedulerPtr = (Scheduler*) resrcMngr;
        }
#ifdef ENTERPRISE_LIB
        else if (!strcmp(schedulerTypeString, "DIFFSERV-ENABLED"))
        {
            char secondSchedTypeString[MAX_STRING_LENGTH] = {0};
            DiffservScheduler* dsSchedulerPtr = NULL;

            ip->diffservEnabled = TRUE;

            ReadDiffservSchedulerConfiguration(node,
                                               &address,
                                               nodeInput,
                                               secondSchedTypeString);

            DIFFSERV_SCHEDULER_Setup(&dsSchedulerPtr,
                                     numPriorities,
                                     secondSchedTypeString,
                                     FALSE, // enableSchedulerStat,
                                     graphDataStr);

            schedulerPtr = (Scheduler*) dsSchedulerPtr;
        }
#endif // ENTERPRISE_LIB
        else
        {
            SCHEDULER_Setup(&schedulerPtr,
                            schedulerTypeString,
                            enableSchedulerStat,
                            graphDataStr);
        }

        ip->interfaceInfo[interfaceIndex]->scheduler = schedulerPtr;
    }


    // This for loop read queue weight and also assigne queue weight if it
    // is not specified by the user
    int unassignQueue = 0;
    double totalWeight = 0.0;
    double* queueWeight;

    queueWeight = (double*) MEM_malloc(sizeof(double) * numPriorities);
    memset(queueWeight, 0, (sizeof(double) * numPriorities));

    for (i = 0, totalWeight = 0.0; i < numPriorities; i++)
    {
        IO_ReadDoubleInstance(
            node->nodeId,
            &address,
            nodeInput,
            "QUEUE-WEIGHT", // IP-QUEUE-WEIGHT
            i,              // parameterInstanceNumber
            FALSE,          // fallbackIfNoInstanceMatch
            &wasFound,
            &queueWeight[i]);

        if (wasFound)
        {
            if (queueWeight[i] >= 1.0)
            {
                NetworkIpConfigurationError(node, "QUEUE-WEIGHT",
                                                    interfaceIndex);
            }
            totalWeight += queueWeight[i];
        }
        else
        {
            unassignQueue++;
            queueWeight[i] = 1.0;
        }
    }

    ERROR_Assert(totalWeight <= 1.0,"Total queue weight should be <= 1.0\n");
    if (unassignQueue > 0 && unassignQueue != numPriorities)
    {
        for (i = 0; i < numPriorities; i++)
        {
            if (queueWeight[i] == 1.0)
            {
                queueWeight[i] = (1 - totalWeight) / unassignQueue;
            }
        }
    }

    for (i = 0; i < numPriorities; i++)
    {
        Queue* queue = NULL;
        char queueTypeString[MAX_STRING_LENGTH] = {0};
        int queueSize = DEFAULT_NETWORK_OUTPUT_QUEUE_SIZE;
        BOOL enableQueueStat = FALSE;
        int priority = ALL_PRIORITIES;
        int addedQueuePriority = ALL_PRIORITIES;
        void* spConfigInfo = NULL; // Queue Specific configurations.


        IO_ReadStringInstance(
            node->nodeId,
            &address,
            nodeInput,
            "IP-QUEUE-TYPE",
            i,                 // parameterInstanceNumber
            TRUE,              // fallbackIfNoInstanceMatch
            &wasFound,
            queueTypeString);

        if (!wasFound)
        {
            NetworkIpConfigurationError(
                node, "IP-QUEUE-TYPE", interfaceIndex);
        }

        IO_ReadIntInstance(
            node->nodeId,
            &address,
            nodeInput,
            "IP-QUEUE-PRIORITY-QUEUE-SIZE", // IP-OUTPUT-QUEUE-SIZE
            i,                 // parameterInstanceNumber
            TRUE,              // fallbackIfNoInstanceMatch
            &wasFound,
            &queueSize);

        if (wasFound && (queueSize <= 0))
        {
            NetworkIpConfigurationError(node,
                                        "IP-QUEUE-PRIORITY-QUEUE-SIZE",
                                        interfaceIndex);
        }

        if (!ip->diffservEnabled)
        {
            IO_ReadIntInstance(
                node->nodeId,
                &address,
                nodeInput,
                "IP-QUEUE-PRIORITY",
                i,                 // parameterInstanceNumber
                TRUE,              // fallbackIfNoInstanceMatch
                &wasFound,
                &priority);

            if (wasFound && (priority < 0))
            {
                NetworkIpConfigurationError(
                    node, "IP-QUEUE-PRIORITY", interfaceIndex);
            }
        }

        IO_ReadString(
            node->nodeId,
            &address,
            nodeInput,
            "QUEUE-STATISTICS",
            &wasFound,
            buf);

        if (wasFound && (!strcmp(buf, "YES")))
        {
            enableQueueStat = TRUE;
        }

        if (!strcmp(queueTypeString, "FIFO"))
        {
            // No specific configuration for FIFO
        }
        else if (!strcmp(queueTypeString, "RED"))
        {
            IO_ReadString(
                node->nodeId,
                &address,
                nodeInput,
                "ECN",
                &wasFound,
                buf);

            if (wasFound && (!strcmp(buf, "YES")))
            {
                // Ecn enabled RED
                RedEcnParameters* redParams = NULL;

                ReadRed_EcnConfigurationParameters(node,
                                    &address,
                                    nodeInput,
                                    enableQueueStat,
                                    i,
                                    &redParams);
                memset(queueTypeString, 0, MAX_STRING_LENGTH);
                strcpy(queueTypeString, "RED-ECN");
                spConfigInfo = (void*)(redParams);
            }
            else
            {
                RedParameters* redParams = NULL;

                ReadRedConfigurationParameters(node,
                                    &address,
                                    nodeInput,
                                    enableQueueStat,
                                    i,
                                    &redParams);

                spConfigInfo = (void*)(redParams);
            }
        }
        else if (!strcmp(queueTypeString, "WRED"))
        {
            RedEcnParameters* redParams = NULL;

            ReadWred_EcnConfigurationThParameters(node,
                                    &address,
                                    nodeInput,
                                    enableQueueStat,
                                    i,
                                    &redParams);

            spConfigInfo = (void*)(redParams);
        }
        else if (!strcmp(queueTypeString, "RIO"))
        {
            RioParameters* rioParams = NULL;
            ReadRio_EcnConfigurationThParameters(node,
                            &address,
                            nodeInput,
                            enableQueueStat,
                            i,
                            &rioParams);
            spConfigInfo = (void*)(rioParams);
        }
        else
        {
            // Error :Unknown Queue Type
            ERROR_ReportError("Unknown queue type");
        }

        // Initialize Queue depending on queueTypeString specification
        QUEUE_Setup(node,
                    &queue,
                    queueTypeString,
                    queueSize,
                    interfaceIndex,
                    priority,
                    0, // infoFieldSize
                    enableQueueStat,
                    node->guiOption,
                    getSimTime(node),
                    spConfigInfo);

        addedQueuePriority = (*schedulerPtr).addQueue(queue, priority, queueWeight[i]);
        //GuiStart++-
        if (node->guiOption == TRUE)
        {
            GUI_AddInterfaceQueue(node->nodeId, GUI_NETWORK_LAYER,
                            interfaceIndex, addedQueuePriority, queueSize,
                            getSimTime(node));
        }
        //GuiEnd
    }
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpInitCpuQueueConfiguration()
// PURPOSE      Initializes cpu queue parameters during startup.
// PARAMETERS   Node *node
//                  Pointer to node.
//              const NodeInput *nodeInput
//                  Pointer to node input.
// RETURN       None.
//-----------------------------------------------------------------------------

//static
void
NetworkIpInitCpuQueueConfiguration(
    Node *node,
    const NodeInput *nodeInput)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    Scheduler *cpuSchedulerPtr = NULL;
    Queue* queue = NULL;
    queue = new Queue;
    BOOL enableQueueStat = FALSE;
    BOOL enableSchedulerStat = FALSE;
    char buf[MAX_STRING_LENGTH] = {0};
    BOOL wasFound = FALSE;
    int queueSize = DEFAULT_CPU_QUEUE_SIZE;

    if (ip->backplaneType == BACKPLANE_TYPE_CENTRAL)
    {
        //If router backplane type is CENTRAL then queue size is set as
        //DEFAULT_CPU_QUEUE_SIZE * Total number of interfaces of that node.
        queueSize = DEFAULT_CPU_QUEUE_SIZE * (node->numberInterfaces);
    }

    IO_ReadString(
        node->nodeId,
        ANY_ADDRESS,
        nodeInput,
        "INPUT-QUEUE-STATISTICS",
        &wasFound,
        buf);

    if (wasFound && (!strcmp(buf, "YES")))
    {
        enableQueueStat = TRUE;
    }

    IO_ReadString(
        node->nodeId,
        ANY_ADDRESS,
        nodeInput,
        "INPUT-SCHEDULER-STATISTICS",
        &wasFound,
        buf);

    if (wasFound && (!strcmp(buf, "YES")))
    {
        enableSchedulerStat = TRUE;
    }

    queue->SetupQueue(node,
                      "FIFO",
                      queueSize,
                      0,
                      0,
                      0,
                      enableQueueStat);

    SCHEDULER_Setup(&cpuSchedulerPtr,
                    "STRICT-PRIORITY",
                    enableSchedulerStat);

    ip->cpuScheduler = cpuSchedulerPtr;

    // Scheduler add Queue Functionality
    cpuSchedulerPtr->addQueue(queue);
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpInitInputQueueConfiguration()
// PURPOSE      Initializes input queue parameters during startup.
// PARAMETERS   Node *node
//                  Pointer to node.
//              const NodeInput *nodeInput
//                  Pointer to node input.
//              int interfaceIndex
//                  interface associated with queue.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkIpInitInputQueueConfiguration(
    Node *node,
    const NodeInput *nodeInput,
    int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    Address address;
    Scheduler* inputSchedulerPtr = NULL;
    int numPriorities = 0;
    int i = 0;
    BOOL enableQueueStat = FALSE;
    BOOL enableSchedulerStat = FALSE;
    char buf[MAX_STRING_LENGTH] = {0};
    BOOL wasFound = FALSE;

    NetworkGetInterfaceInfo(node, interfaceIndex, &address);

    IO_ReadString(
        node->nodeId,
        &address,
        nodeInput,
        "INPUT-SCHEDULER-STATISTICS",
        &wasFound,
        buf);

    if (wasFound && (!strcmp(buf, "YES")))
    {
        enableSchedulerStat = TRUE;
    }

    IO_ReadString(
        node->nodeId,
        &address,
        nodeInput,
        "INPUT-QUEUE-STATISTICS",
        &wasFound,
        buf);

    if (wasFound && (!strcmp(buf, "YES")))
    {
        enableQueueStat = TRUE;
    }

    // We assume priority scheduling for input queues.
    SCHEDULER_Setup(&inputSchedulerPtr, "STRICT-PRIORITY",
                        enableSchedulerStat);
    ip->interfaceInfo[interfaceIndex]->inputScheduler =
                        (Scheduler*) inputSchedulerPtr;

    // For input queues,
    // We assume only one FIFO queue for Input buffer.
    numPriorities = 1;

    // Leave for loop just in case we need multiple priority queues
    // later on.
    for (i = 0; i < numPriorities; i++)
    {
        Queue* queue = NULL;
        BOOL wasFound = FALSE;
        int inputQueueSize = DEFAULT_NETWORK_INPUT_QUEUE_SIZE;

        // Input queue size usually matches output queue size.
        IO_ReadIntInstance(
            node->nodeId,
            &address,
            nodeInput,
            "IP-QUEUE-PRIORITY-INPUT-QUEUE-SIZE", // IP-INPUT-QUEUE-SIZE
            i,                 // parameterInstanceNumber
            TRUE,              // fallbackIfNoInstanceMatch
            &wasFound,
            &inputQueueSize);

        if (wasFound && (inputQueueSize <= 0))
        {
            NetworkIpConfigurationError(node,
                                        "IP-QUEUE-PRIORITY-INPUT-QUEUE-SIZE",
                                        interfaceIndex);
        }

        queue = new Queue; // "FIFO"
        queue->SetupQueue(node,
                          "FIFO",
                          inputQueueSize,
                          interfaceIndex,
                          i,
                          0,
                          enableQueueStat);

        // Scheduler add Queue Functionality
        (*inputSchedulerPtr).addQueue(queue);
    }
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpCreateQueues()
// PURPOSE      Initializes input and output queue parameters during startup
// PARAMETERS   Node *node
//                  Pointer to node.
//              const NodeInput *nodeInput
//                  Pointer to node input.
//              int interfaceIndex
//                  interface associated with queue.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkIpCreateQueues(
    Node *node,
    const NodeInput *nodeInput,
    int interfaceIndex)
{
    NetworkIpInitOutputQueueConfiguration(node,
                                          nodeInput,
                                          interfaceIndex);

    NetworkIpInitInputQueueConfiguration(node,
                                         nodeInput,
                                         interfaceIndex);
}


//-----------------------------------------------------------------------------
// Network-layer dequeueing
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpQueueDequeuePacket()
// PURPOSE      Calls the packet scheduler for an interface to retrieve
//              an IP packet from a queue associated with the interface.
// PARAMETERS   Node *node
//                  Pointer to node.
//              SchedulerType *scheduler
//                  queue to dequeue from.
//              Message **msg
//                  Storage for pointer to message with IP packet.
//              NodeAddress *nextHopAddress
//                  Storage for packet's next hop address.
//              int outgoingInterface
//                  Used to determine where packet should go after passing
//                  through the backplane.
//              TosType *userPriority
//                  Storage for user priority of packet.
// RETURN       TRUE if dequeued successfully, FALSE otherwise.
//-----------------------------------------------------------------------------

static
BOOL NetworkIpQueueDequeuePacket(
    Node *node,
    Scheduler *scheduler,
    Message **msg,
    NodeAddress *nextHopAddress,
    int *outgoingInterface,
    int *networkType,
    TosType *userPriority)
{
    QueuePriorityType queuePriority = ALL_PRIORITIES;
    QueuedPacketInfo *infoPtr;

    if ((*scheduler).retrieve(ALL_PRIORITIES, DEQUEUE_NEXT_PACKET, msg,
                        &queuePriority, DEQUEUE_PACKET, getSimTime(node)))
    {
        ERROR_Assert(*msg != NULL, "Cannot dequeue packet");

        infoPtr = (QueuedPacketInfo *) MESSAGE_ReturnInfo((*msg));

        *userPriority = infoPtr->userTos;
        *nextHopAddress = infoPtr->nextHopAddress;
        *outgoingInterface = infoPtr->outgoingInterface;
        *networkType = infoPtr->networkType;

#ifdef DEBUG_IPV6
        printf("nextHop %x nodeId %d outgoingInterface %d\n",
            infoPtr->nextHopAddress,
            node->nodeId,
            infoPtr->outgoingInterface);
#endif
        return TRUE;
    }

    return FALSE; //Cannot dequeue packet;
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpCpuQueueDequeuePacket()
// PURPOSE      Calls the packet scheduler for an interface to retrieve
//              an IP packet from the cpu queue associated with the interface.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message **msg
//                  Storage for pointer to message with IP packet.
//              NodeAddress *nextHopAddress
//                  Storage for packet's next hop address.
//              int outgoingInterface
//                  Used to determine where packet should go after passing
//                  through the backplane.
//              QueuePriorityType *userPriority
//                  Storage for user priority of packet.
// RETURN       TRUE if dequeued successfully, FALSE otherwise.
//-----------------------------------------------------------------------------

BOOL NetworkIpCpuQueueDequeuePacket(
    Node *node,
    Message **msg,
    NodeAddress *nextHopAddress,
    int *outgoingInterface,
    int *networkType,
    QueuePriorityType *userPriority)

{
    BOOL dequeued = FALSE;
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    Scheduler *cpuScheduler;
    TosType userTos = ALL_PRIORITIES;

    cpuScheduler = ip->cpuScheduler;

    dequeued = NetworkIpQueueDequeuePacket(node,
                                           cpuScheduler,
                                           msg,
                                           nextHopAddress,
                                           outgoingInterface,
                                           networkType,
                                           &userTos);

    if (dequeued == FALSE) {
        return FALSE;
    }

    *userPriority = (unsigned char) (userTos >> 5);

    return TRUE;
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpInputQueueDequeuePacket()
// PURPOSE      Calls the packet scheduler for an interface to retrieve
//              an IP packet from the input queue associated with the interface.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int incomingInterface
//                  interface to dequeue from.
//              Message **msg
//                  Storage for pointer to message with IP packet.
//              NodeAddress *nextHopAddress
//                  Storage for packet's next hop address.
//              int outgoingInterface
//                  Used to determine where packet should go after passing
//                  through the backplane.
//              QueuePriorityType *userPriority
//                  Storage for priority of packet.
// RETURN       TRUE if dequeued successfully, FALSE otherwise.
//-----------------------------------------------------------------------------

BOOL NetworkIpInputQueueDequeuePacket(
    Node *node,
    int incomingInterface,
    Message **msg,
    NodeAddress *nextHopAddress,
    int *outgoingInterface,
    int *networkType,
    QueuePriorityType *userPriority)

{
    BOOL dequeued = FALSE;

    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    Scheduler *inputScheduler;
    TosType userTos = ALL_PRIORITIES;

    ERROR_Assert(
        incomingInterface >= 0 && incomingInterface < node->numberInterfaces,
        "Invalid interface index");

    inputScheduler = ip->interfaceInfo[incomingInterface]->inputScheduler;

    dequeued = NetworkIpQueueDequeuePacket(node,
                                           inputScheduler,
                                           msg,
                                           nextHopAddress,
                                           outgoingInterface,
                                           networkType,
                                           &userTos);

    if (dequeued == FALSE) {
        return FALSE;
    }

    *userPriority = (unsigned char) (userTos >> 5);

    return TRUE;
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpOutputQueueDequeuePacket()
// PURPOSE      Calls the packet scheduler for an interface to retrieve
//              an IP packet from a queue associated with the interface.
//              The dequeued packet, since it's already been routed,
//              has an associated next-hop IP address.  The packet's
//              priority value is also returned.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int interfaceIndex
//                  Index of interface.
//              Message **msg
//                  Storage for pointer to message with IP packet.
//              NodeAddress *nextHopAddress
//                  Storage for packet's next hop address.
//              QueuePriorityType *userPriority
//                  Storage for user priority of packet.
// RETURN       TRUE if dequeued successfully, FALSE otherwise.
//
// NOTES        This function is called by
//              MAC_OutputQueueDequeuePacket() (mac/mac.pc), which itself
//              is called from mac/mac_802_11.pc and other MAC protocol
//              source files.
//
//              This function will assert false if the scheduler cannot
//              return an IP packet for whatever reason.
//-----------------------------------------------------------------------------

BOOL NetworkIpOutputQueueDequeuePacket(
    Node *node,
    int interfaceIndex,
    Message **msg,
    NodeAddress *nextHopAddress,
    int *networkType,
    QueuePriorityType *userPriority)

{
    BOOL dequeued = FALSE;
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    Scheduler *scheduler = NULL;
    TosType userTos = ALL_PRIORITIES;
    int outgoingInterface;

    ERROR_Assert(
        interfaceIndex >= 0 && interfaceIndex < node->numberInterfaces,
        "Invalid interface index");

    scheduler = ip->interfaceInfo[interfaceIndex]->scheduler;

    dequeued = NetworkIpQueueDequeuePacket(node,
                                           scheduler,
                                           msg,
                                           nextHopAddress,
                                           &outgoingInterface,
                                           networkType,
                                           &userTos);

// ANTS01 begin (avgQTime):
	if (NetworkIpGetUnicastRoutingProtocolType(node,interfaceIndex) == ROUTING_PROTOCOL_ANTHOCNET)
	{
		int queueIndex;
		clocktype qTime = 0;
		Queue *tempQueue;
    for (queueIndex = scheduler->numQueue() - 1;queueIndex >= 0;queueIndex--)
    {
			tempQueue = scheduler->getQueue(queueIndex);
      qTime = tempQueue->averageTimeInQueue();
			if (qTime > 0)
			{
		  	ANTHOCNET_IPQueueTimeRegistration(node,interfaceIndex,qTime);
        tempQueue->resetPeriod(getSimTime(node));
				break;
			}
		}
	}
// ANTS01 end

    if (dequeued)
    {
        // Pass user priority (precedence - 3 bit field) to mac
        *userPriority = (TosType) (userTos >> 5);

        QueuePriorityType  queuePriority = 0;
        queuePriority = (QueuePriorityType) GetQueuePriorityFromUserTos(
                                        node, userTos, (*scheduler).numQueue());
        (*scheduler).collectGraphData(queuePriority,
                        MESSAGE_ReturnPacketSize((*msg)),
                        TIME_getSimTime(node));

        //Trace dequeue
        ActionData acn;
        acn.actionType = DEQUEUE;
        acn.actionComment = NO_COMMENT;
        acn.pktQueue.interfaceID = interfaceIndex;
        acn.pktQueue.queuePriority = queuePriority;
        TRACE_PrintTrace(node, *msg, TRACE_NETWORK_LAYER, PACKET_OUT, &acn);

        //GuiStart
        if (node->guiOption == TRUE)
        {
            unsigned queuePriority = GetQueuePriorityFromUserTos(
                                   node, userTos, (*scheduler).numQueue());
            GUI_QueueDequeuePacket(node->nodeId, GUI_NETWORK_LAYER,
                                   interfaceIndex, queuePriority,
                                   MESSAGE_ReturnPacketSize((*msg)),
                                   getSimTime(node));
        }
        //GuiEnd
    }

    return dequeued;
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpOutputQueueDequeuePacketForAPriority()
// PURPOSE      Same as NetworkIpOutputQueueDequeuePacket(), except the
//              packet dequeued is requested by a specific priority,
//              instead of leaving the priority decision up to the
//              packet scheduler.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int interfaceIndex
//                  Index of interface.
//              QueuePriorityType priority
//                  Priority of packet.
//              Message **msg
//                  Storage for pointer to message with IP packet.
//              NodeAddress *nextHopAddress
//                  Storage for packet's next hop address.
//              int *networkType
//                  Type of network (IP, Link-16, ...) used to route packet.
// RETURN       TRUE if dequeued successfully, FALSE otherwise.
//
// NOTES        This function is called by
//              MAC_OutputQueueDequeuePacketForAPriority() (mac/mac.pc),
//              which itself is called from mac/mac_802_11.pc and other
//              MAC protocol source files.
//
//              This function will assert false if the scheduler cannot
//              return an IP packet for whatever reason.
//-----------------------------------------------------------------------------

BOOL NetworkIpOutputQueueDequeuePacketForAPriority(
    Node *node,
    int interfaceIndex,
    QueuePriorityType priority,
    Message **msg,
    NodeAddress *nextHopAddress,
    int *networkType)
{
    NetworkDataIp *ip = (NetworkDataIp *)node->networkData.networkVar;
    Scheduler *scheduler = NULL;
    QueuePriorityType queuePriority = ALL_PRIORITIES;
    QueuedPacketInfo *infoPtr;

    ERROR_Assert(
        interfaceIndex >= 0 && interfaceIndex < node->numberInterfaces,
        "Invalid interface index");

    scheduler = ip->interfaceInfo[interfaceIndex]->scheduler;

    if ((*scheduler).retrieve(priority, DEQUEUE_NEXT_PACKET, msg,
                &queuePriority, DEQUEUE_PACKET, getSimTime(node)))
    {
        ERROR_Assert(*msg != NULL, "Cannot dequeue packet");

        infoPtr = (QueuedPacketInfo *) MESSAGE_ReturnInfo((*msg));

        *nextHopAddress = infoPtr->nextHopAddress;
        *networkType = infoPtr->networkType;

        //Trace dequeue
        ActionData acn;
        acn.actionType = DEQUEUE;
        acn.actionComment = NO_COMMENT;
        acn.pktQueue.interfaceID = interfaceIndex;
        acn.pktQueue.queuePriority = queuePriority;
        TRACE_PrintTrace(node, *msg, TRACE_NETWORK_LAYER, PACKET_OUT, &acn);

        (*scheduler).collectGraphData((int) priority,
                                MESSAGE_ReturnPacketSize((*msg)),
                                TIME_getSimTime(node));

        //GuiStart
        if (node->guiOption == TRUE)
        {
            GUI_QueueDequeuePacket(node->nodeId, GUI_NETWORK_LAYER,
                                   interfaceIndex, queuePriority,
                                   MESSAGE_ReturnPacketSize((*msg)),
                                   getSimTime(node));
        }
        //GuiEnd
        return TRUE;
    }
    return FALSE;
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpOutputQueueDropPacket
// PURPOSE      Drop a packet from the queue.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int interfaceIndex
//                  Index of interface.
//              Message **msg
//                  Packet to be dropped.
// RETURN       Next hop of dropped packet.
NodeAddress
NetworkIpOutputQueueDropPacket(Node* node, int interfaceIndex, Message **msg)
{
    NodeAddress nextHopAddress;
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    Scheduler *scheduler = NULL;
    QueuePriorityType queuePriority = ALL_PRIORITIES;

    scheduler = ip->interfaceInfo[interfaceIndex]->scheduler;

    (*scheduler).retrieve(ALL_PRIORITIES,
                        DEQUEUE_NEXT_PACKET,
                        msg,
                        &queuePriority,
                        DISCARD_PACKET,
                        getSimTime(node));

    nextHopAddress = ((QueuedPacketInfo *)
                     MESSAGE_ReturnInfo((*msg)))->nextHopAddress;
    return(nextHopAddress);
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpQueueDeleteOutboundPacketsToANode()
// PURPOSE      Deletes all packets in the queue going (probably broken),
//              to the specified next hop address.   There is option
//              to return all such packets back to the routing protocols.
//              via the usual mechanism (callback).
// PARAMETERS   Node *node
//                  Pointer to node.
//              const NodeAddress nextHopAddress
//                  Next hop associated with outbound packets.
//              const NodeAddress destinationAddress
//                  destination associated with outbound packets.
//              const BOOL returnPacketsToRoutingProtocol
//                  Determine whether or not dropped packets should be
//                  returned to the routing protocol for further processing.
// RETURN       None.
//-----------------------------------------------------------------------------

void NetworkIpDeleteOutboundPacketsToANode(
   Node *node,
   const NodeAddress nextHopAddress,
   const NodeAddress destinationAddress,
   const BOOL returnPacketsToRoutingProtocol)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    int interfaceIndex;

    for (interfaceIndex = 0;
         interfaceIndex < node->numberInterfaces;
         interfaceIndex++)
    {
        int queueIndex = DEQUEUE_NEXT_PACKET;
        Scheduler *scheduler = NULL;
        QueuePriorityType notUsed = 0;
        QueuedPacketInfo *infoPtr;

        scheduler = ip->interfaceInfo[interfaceIndex]->scheduler;

        while (queueIndex < NetworkIpOutputQueueNumberInQueue(
                                node,
                                interfaceIndex,
                                FALSE /* specificPriorityOnly */,
                                ALL_PRIORITIES))
        {
            Message *msg = NULL;
            QueuePriorityType msgPriority = ALL_PRIORITIES;
            NodeAddress currentNextHopAddress;
            NodeAddress currentDestinationAddress;

            if (!scheduler->retrieve(ALL_PRIORITIES,
                                     queueIndex,
                                     &msg,
                                     &msgPriority,
                                     PEEK_AT_NEXT_PACKET,
                                     getSimTime(node)))
            {
                ERROR_ReportError("Cannot retrieve packet");
            }

            infoPtr = (QueuedPacketInfo *) MESSAGE_ReturnInfo(msg);

            msgPriority = infoPtr->userTos;
            currentNextHopAddress = infoPtr->nextHopAddress;

            currentDestinationAddress =
                infoPtr->destinationAddress.ipv4DestAddr;

            if (((nextHopAddress == ANY_IP) ||
                 (currentNextHopAddress == nextHopAddress)) &&
                ((destinationAddress == ANY_IP) ||
                 (currentDestinationAddress == destinationAddress)))
            {
                if (!scheduler->retrieve(ALL_PRIORITIES,
                                         queueIndex,
                                         &msg,
                                         &msgPriority,
                                         DROP_PACKET,
                                         getSimTime(node)))
                {
                    ERROR_ReportError("Cannot retrieve packet");
                }

                msgPriority =
                    ((QueuedPacketInfo *) MESSAGE_ReturnInfo(msg))->userTos;

                if (returnPacketsToRoutingProtocol)
                {
                    HandleSpecialMacLayerStatusEvents(node,
                                                      msg,
                                                      nextHopAddress,
                                                      interfaceIndex);
                }
                else
                {
                    MESSAGE_Free(node, msg);
                }
            }
            else
            {
                queueIndex ++;
            }
        }
    }
}

//-----------------------------------------------------------------------------
// Network-layer queue information
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpQueueTopPacket()
// PURPOSE      Same as NetworkIpQueueDequeuePacket(), except the
//              packet is not actually dequeued.  Note that the message
//              containing the packet is not copied; the contents may
//              (inadvertently or not) be directly modified.
// PARAMETERS   Node *node
//                  Pointer to node.
//              SchedulerType *scheduler
//                  queue to get top packet from.
//              Message **msg
//                  Storage for pointer to message with IP packet.
//              NodeAddress *nextHopAddress
//                  Storage for packet's next hop address.
//              int *outgoingInterface
//                  Used to determine where packet should go after passing
//                  through the backplane.
//              int *networkType
//                  Whether packet is associated with an IP network, Link-16
//                  nework, etc...
//              QueuePriorityType *priority
//                  Storage for priority of packet.
// RETURN       TRUE if there is a packet, FALSE otherwise.
//-----------------------------------------------------------------------------

BOOL NetworkIpQueueTopPacket(
    Node *node,
    Scheduler *scheduler,
    Message **msg,
    NodeAddress *nextHopAddress,
    int *outgoingInterface,
    int *networkType,
    QueuePriorityType *priority)
{
    QueuePriorityType notUsed = ALL_PRIORITIES;
    QueuePriorityType queuePriority = ALL_PRIORITIES;
    BOOL isPktRetrieved = FALSE;
    QueuedPacketInfo *infoPtr;

    isPktRetrieved = (*scheduler).retrieve(notUsed,
                                            DEQUEUE_NEXT_PACKET,
                                            msg,
                                            &queuePriority,
                                            PEEK_AT_NEXT_PACKET,
                                            getSimTime(node));

    if (isPktRetrieved)
    {
        ERROR_Assert(*msg != NULL, "Cannot retrieve packet");

        infoPtr = (QueuedPacketInfo *) MESSAGE_ReturnInfo((*msg));

        // Retuning Queue priority
        *priority = queuePriority;
        *nextHopAddress = infoPtr->nextHopAddress;
        *outgoingInterface = infoPtr->outgoingInterface;
        *networkType = infoPtr->networkType;
    }

    return isPktRetrieved;
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpQueueTopPacketForUserPriority()
// PURPOSE      Same as NetworkIpQueueTopPacket(), except the priority
//              value returned is user priority instead of queue priority.
// PARAMETERS   Node *node
//                  Pointer to node.
//              SchedulerType *scheduler
//                  queue to get top packet from.
//              Message **msg
//                  Storage for pointer to message with IP packet.
//              NodeAddress *nextHopAddress
//                  Storage for packet's next hop address.
//              int *outgoingInterface
//                  Used to determine where packet should go after passing
//                  through the backplane.
//              int *networkType
//                  Whether packet is associated with an IP network, Link-16
//                  nework, etc...
//              TosType *priority
//                  Storage for user priority of packet.
// RETURN       TRUE if there is a packet, FALSE otherwise.
//-----------------------------------------------------------------------------

BOOL NetworkIpQueueTopPacketForUserPriority(
    Node *node,
    Scheduler *scheduler,
    Message **msg,
    NodeAddress *nextHopAddress,
    int *outgoingInterface,
    int *networkType,
    TosType *priority)
{
    QueuePriorityType notUsed = ALL_PRIORITIES;
    QueuePriorityType queuePriority = ALL_PRIORITIES;
    BOOL isPktRetrieved = FALSE;
    QueuedPacketInfo *infoPtr;

    isPktRetrieved = (*scheduler).retrieve(notUsed,
                                        DEQUEUE_NEXT_PACKET,
                                        msg,
                                        &queuePriority,
                                        PEEK_AT_NEXT_PACKET,
                                        getSimTime(node));
    if (isPktRetrieved)
    {
        ERROR_Assert(*msg != NULL, "Cannot retrieve packet");

        infoPtr = (QueuedPacketInfo *) MESSAGE_ReturnInfo((*msg));

        *priority = infoPtr->userTos;
        *nextHopAddress = infoPtr->nextHopAddress;
        *outgoingInterface = infoPtr->outgoingInterface;
        *networkType = infoPtr->networkType;
    }

    return isPktRetrieved;
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpCpuQueueTopPacket()
// PURPOSE      Same as NetworkIpCpuQueueDequeuePacket(), except the
//              packet is not actually dequeued.  Note that the message
//              containing the packet is not copied; the contents may
//              (inadvertently or not) be directly modified.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message **msg
//                  Storage for pointer to message with IP packet.
//              NodeAddress *nextHopAddress
//                  Storage for packet's next hop address.
//              int *outgoingInterface
//                  Used to determine where packet should go after passing
//                  through the backplane.
//              int *networkType
//                  Whether packet is associated with an IP network, Link-16
//                  nework, etc...
//              QueuePriorityType *priority
//                  Storage for priority of packet.
// RETURN       TRUE if there is a packet, FALSE otherwise.
//-----------------------------------------------------------------------------
BOOL NetworkIpCpuQueueTopPacket(
    Node *node,
    Message **msg,
    NodeAddress *nextHopAddress,
    int *outgoingInterface,
    int *networkType,
    QueuePriorityType *priority)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    Scheduler *cpuScheduler = NULL;

    cpuScheduler = ip->cpuScheduler;

    return NetworkIpQueueTopPacket(node,
                                   cpuScheduler,
                                   msg,
                                   nextHopAddress,
                                   outgoingInterface,
                                   networkType,
                                   priority);
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpInputQueueTopPacket()
// PURPOSE      Same as NetworkIpInputQueueDequeuePacket(), except the
//              packet is not actually dequeued.  Note that the message
//              containing the packet is not copied; the contents may
//              (inadvertently or not) be directly modified.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int incomingInterface
//                  interface to get top packet from.
//              Message **msg
//                  Storage for pointer to message with IP packet.
//              NodeAddress *nextHopAddress
//                  Storage for packet's next hop address.
//              int *outgoingInterface
//                  Used to determine where packet should go after passing
//                  through the backplane.
//              int *networkType
//                  Whether packet is associated with an IP network, Link-16
//                  nework, etc...
//              QueuePriorityType *priority
//                  Storage for priority of packet.
// RETURN       TRUE if there is a packet, FALSE otherwise.
//-----------------------------------------------------------------------------

BOOL NetworkIpInputQueueTopPacket(
    Node *node,
    int incomingInterface,
    Message **msg,
    NodeAddress *nextHopAddress,
    int *outgoingInterface,
    int *networkType,
    QueuePriorityType *priority)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    Scheduler *inputScheduler = NULL;

    ERROR_Assert(
        incomingInterface >= 0 && incomingInterface < node->numberInterfaces,
        "Invalid interface index");

    inputScheduler = ip->interfaceInfo[incomingInterface]->inputScheduler;

    return NetworkIpQueueTopPacket(node,
                                   inputScheduler,
                                   msg,
                                   nextHopAddress,
                                   outgoingInterface,
                                   networkType,
                                   priority);
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpOutputQueueTopPacket()
// PURPOSE      Same as NetworkIpOutputQueueDequeuePacket(), except the
//              packet is not actually dequeued.  Note that the message
//              containing the packet is not copied; the contents may
//              (inadvertently or not) be directly modified.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int interfaceIndex
//                  Index of interface.
//              Message **msg
//                  Storage for pointer to message with IP packet.
//              NodeAddress *nextHopAddress
//                  Storage for packet's next hop address.
//              int *networkType
//                  Whether packet is associated with an IP network, Link-16
//                  nework, etc...
//              QueuePriorityType *priority
//                  Storage for priority of packet.
// RETURN       TRUE if there is a packet, FALSE otherwise.
//
// NOTES        This function is called by MAC_OutputQueueTopPacket()
//              (mac/mac.pc), which itself is called from
//              mac/mac_802_11.pc and other MAC protocol source files.
//
//              This function will assert false if the scheduler cannot
//              return an IP packet for whatever reason.
//-----------------------------------------------------------------------------

BOOL NetworkIpOutputQueueTopPacket(
    Node *node,
    int interfaceIndex,
    Message **msg,
    NodeAddress *nextHopAddress,
    int* networkType,
    QueuePriorityType *priority)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    Scheduler *scheduler = NULL;
    int outgoingInterface;

    ERROR_Assert(
        interfaceIndex >= 0 && interfaceIndex < node->numberInterfaces,
        "Invalid interface index");

    scheduler = ip->interfaceInfo[interfaceIndex]->scheduler;

    return NetworkIpQueueTopPacket(node,
                                   scheduler,
                                   msg,
                                   nextHopAddress,
                                   &outgoingInterface,
                                   networkType,
                                   priority);
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpOutputQueueTopPacketForAPriority()
// PURPOSE      Same as NetworkIpOutputQueueDequeuePacketForAPriority(),
//              except the packet is not actually dequeued.  Note that
//              the message containing the packet is not copied; the
//              contents may (inadvertently or not) be directly
//              modified.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int interfaceIndex
//                  Index of interface.
//              QueuePriorityType priority
//                  Priority of the queue.
//              Message **msg
//                  Storage for pointer to message with IP packet.
//              NodeAddress *nextHopAddress
//                  Storage for packet's next hop address.
// RETURN       TRUE if there is a packet, FALSE otherwise.
//
// NOTES        This function is called by
//              MAC_OutputQueueTopPacketForAPriority() (mac/mac.pc),
//              which itself is called from mac/mac_802_11.pc and other
//              MAC protocol source files.
//
//              This function will assert false if the scheduler cannot
//              return an IP packet for whatever reason.
//-----------------------------------------------------------------------------

BOOL NetworkIpOutputQueueTopPacketForAPriority(
    Node *node,
    int interfaceIndex,
    QueuePriorityType priority,
    Message **msg,
    NodeAddress *nextHopAddress)
{
    NetworkDataIp *ip = (NetworkDataIp *)node->networkData.networkVar;
    Scheduler *scheduler = NULL;
    QueuePriorityType queuePriority = ALL_PRIORITIES;
    BOOL isPktRetrieved = FALSE;

    ERROR_Assert(
        interfaceIndex >= 0 && interfaceIndex < node->numberInterfaces,
        "Invalid interface index");

    scheduler = ip->interfaceInfo[interfaceIndex]->scheduler;

    isPktRetrieved = (*scheduler).retrieve(priority,
                                            DEQUEUE_NEXT_PACKET,
                                            msg,
                                            &queuePriority,
                                            PEEK_AT_NEXT_PACKET,
                                            getSimTime(node));

    if (isPktRetrieved)
    {
        ERROR_Assert(*msg != NULL, "Cannot retrieve packet");

        *nextHopAddress =
            ((QueuedPacketInfo *)MESSAGE_ReturnInfo((*msg)))->nextHopAddress;
    }

    return isPktRetrieved;
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpQueueIsEmpty()
// PURPOSE      Calls the packet scheduler for an interface to determine
//              whether the interface's cpu queue is empty.
// PARAMETERS   Node *node
//                  Pointer to node.
//              SchedulerType *scheduler
//                  Queue to determine empty from.
// RETURN       TRUE if the scheduler says the interface's cpu queue
//              is empty.
//              FALSE if the scheduler says the interface's cpu queue
//              is not empty.
//-----------------------------------------------------------------------------

static BOOL
NetworkIpQueueIsEmpty(Node *node,
                      Scheduler *scheduler)
{
    return ((*scheduler).isEmpty(ALL_PRIORITIES));
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpCpuQueueIsEmpty()
// PURPOSE      Calls the packet scheduler for an interface to determine
//              whether the interface's cpu queue is empty.
// PARAMETERS   Node *node
//                  Pointer to node.
// RETURN       TRUE if the scheduler says the interface's cpu queue
//              is empty.
//              FALSE if the scheduler says the interface's cpu queue
//              is not empty.
//-----------------------------------------------------------------------------


BOOL
NetworkIpCpuQueueIsEmpty(Node *node)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    Scheduler *cpuScheduler = NULL;

    cpuScheduler = ip->cpuScheduler;

    return NetworkIpQueueIsEmpty(node, cpuScheduler);
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpInputQueueIsEmpty()
// PURPOSE      Calls the packet scheduler for an interface to determine
//              whether the interface's input queue is empty.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int incomingInterface
//                  Index of interface.
// RETURN       TRUE if the scheduler says the interface's input queue
//              is empty.
//              FALSE if the scheduler says the interface's input queue
//              is not empty.
//-----------------------------------------------------------------------------

BOOL
NetworkIpInputQueueIsEmpty(Node *node, int incomingInterface)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    Scheduler *inputScheduler = NULL;

    ERROR_Assert(
        incomingInterface >= 0 && incomingInterface < node->numberInterfaces,
        "Invalid incoming interface");

    inputScheduler = ip->interfaceInfo[incomingInterface]->inputScheduler;

    return NetworkIpQueueIsEmpty(node, inputScheduler);
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpOutputQueueIsEmpty()
// PURPOSE      Calls the packet scheduler for an interface to determine
//              whether the interface's output queue is empty.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int interfaceIndex
//                  Index of interface.
// RETURN       TRUE if the scheduler says the interface's output queue
//              is empty.
//              FALSE if the scheduler says the interface's output queue
//              is not empty.
//-----------------------------------------------------------------------------

BOOL
NetworkIpOutputQueueIsEmpty(Node *node, int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    ERROR_Assert(
        interfaceIndex >= 0 && interfaceIndex < node->numberInterfaces,
        "Invalid interface index");

#ifdef ENTERPRISE_LIB
    MplsData *mpls = (MplsData *) node->macData[interfaceIndex]->mplsVar;
    if (mpls)
    {
        return (MplsOutputQueueIsEmpty(node, interfaceIndex));
    }
    else
#endif // ENTERPRISE_LIB
    {
        Scheduler *scheduler = NULL;

        scheduler = ip->interfaceInfo[interfaceIndex]->scheduler;

        return NetworkIpQueueIsEmpty(node, scheduler);
    }
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpOutputQueueNumberInQueue()
// PURPOSE      Calls the packet scheduler for an interface to determine
//              how many packets are in a queue.  There may be multiple
//              queues on an interface, so the priority of the desired
//              queue is also provided.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int interfaceIndex
//                  Index of interface.
//              BOOL specificPriorityOnly
//                  Should we only get the number of packets in queue for
//                  the specified priority only or for all priorities.
//              QueuePriorityType priority
//                  Priority of queue.
// RETURN       Number of packets in queue.
//-----------------------------------------------------------------------------

int
NetworkIpOutputQueueNumberInQueue(
    Node *node,
    int interfaceIndex,
    BOOL specificPriorityOnly,
    QueuePriorityType priority)
{
    NetworkDataIp *ip = (NetworkDataIp *)node->networkData.networkVar;
    Scheduler *scheduler = NULL;

    ERROR_Assert(
        interfaceIndex >= 0 && interfaceIndex < node->numberInterfaces,
        "Invalid interface index");

    scheduler = ip->interfaceInfo[interfaceIndex]->scheduler;
    return (*scheduler).numberInQueue(priority);
}


//-----------------------------------------------------------------------------
// FUNCTION: GetQueueNumberFromPriority
// PURPOSE:  Get queue number through which a given user priority
//           will be forwarded.
// PARAMETERS:  TosType userTos
//                  User priority.
//              int numQueues
//                  Maximum number of queue available.
// RETURN: Index of the queue.
//
// Note: This mapping is done based on ip scheduler mapping.
//       If there is any change in that mapping, this mapping
//       should also be changed accordingly.
//-----------------------------------------------------------------------------

unsigned GetQueueNumberFromPriority(
    TosType userTos,
    int numQueues)
{
    int i = 0;

    for (i = numQueues - 1; i >= 0; i--)
    {
        if ((TosType) i <= (userTos >> 5))
        {
            // Return the queue whose priority same as packet priority
            return i;
        }
    }
    return IPTOS_PREC_ROUTINE;
}

static
unsigned GetQueuePriorityFromUserTos(
    Node *node,
    TosType userTos,
    int numQueues)
{
    int i = 0;

    int queuePriority = ReturnPriorityForPHB(node, userTos);

    for (i = numQueues - 1; i >= 0; i--)
    {
        if (i <= queuePriority)
        {
            // Return the queue in or from which
            // packet has been queued or dequeued
            return i;
        }
    }
    return IPTOS_PREC_ROUTINE;
}


//-----------------------------------------------------------------------------
// Per hop behavior (PHB) routing
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     ReturnPriorityForPHB()
// PURPOSE
// PARAMETERS   Node *node
//                  Pointer to node.
//              TosType tos
//
//
// RETURNS
//-----------------------------------------------------------------------------

QueuePriorityType
ReturnPriorityForPHB(
    Node *node,
    TosType tos)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    int i;

#ifdef DEBUG
    printf("#%d: ReturnPriorityForPHB(%d) = ", node->nodeId, tos);
#endif

    for (i = 0; i < ip->numPhbInfo; i++)
    {
        if (ip->phbInfo[i].ds == tos >> 2)
        {
#ifdef DEBUG
    printf("%d\n", ip->phbInfo[i].priority);
#endif
            return ip->phbInfo[i].priority;
        }
    }

#ifdef ENTERPRISE_LIB
    // look for the default priority queue for best effort traffic
    if (ip->diffservEnabled)
    {
        for (i = 0; i < ip->numPhbInfo; i++)
        {
            if (ip->phbInfo[i].ds == DIFFSERV_DS_CLASS_BE)
            {
#ifdef DEBUG
                printf("%d\n", ip->phbInfo[i].priority);
#endif
                return ip->phbInfo[i].priority;
            }
        }
    }
#endif // ENTERPRISE_LIB

#ifdef DEBUG
    printf("%d\n", (QueuePriorityType) tos >> 5);
#endif
    return ((QueuePriorityType) (tos >> 5));
}

//-----------------------------------------------------------------------------
// Routing table (forwarding table)
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     NetworkGetInterfaceAndNextHopFromForwardingTable()
// PURPOSE      Do a lookup on the routing table with a destination IP
//              address to obtain a route (index of an outgoing
//              interface and a next hop Ip address).
// PARAMETERS   Node *node
//                  Pointer to node.
//              NodeAddress destinationAddress
//                  Destination IP address.
//              int *interfaceIndex
//                  Storage for index of outgoing interface.
//              NodeAddress *nextHopAddress
//                  Storage for next hop IP address.
//                  If no route can be found, *nextHopAddress will be
//                  set to NETWORK_UNREACHABLE.
// RETURN       None.
//-----------------------------------------------------------------------------

void NetworkGetInterfaceAndNextHopFromForwardingTable(
    Node *node,
    NodeAddress destinationAddress,
    int *interfaceIndex,
    NodeAddress *nextHopAddress)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    NetworkForwardingTable *forwardTable = &(ip->forwardTable);
    int i;

    *interfaceIndex = NETWORK_UNREACHABLE;
    *nextHopAddress = (unsigned) NETWORK_UNREACHABLE;

    //NetworkPrintForwardingTable(node);

    for (i=0; i < forwardTable->size; i++) {
        NodeAddress maskedDestinationAddress =
            MaskIpAddress(
                destinationAddress, forwardTable->row[i].destAddressMask);

        if (forwardTable->row[i].destAddress == maskedDestinationAddress
            && forwardTable->row[i].nextHopAddress != (unsigned) NETWORK_UNREACHABLE
            && forwardTable->row[i].interfaceIsEnabled != FALSE)
        {
            *interfaceIndex = forwardTable->row[i].interfaceIndex;
            *nextHopAddress = forwardTable->row[i].nextHopAddress;
            break;
        }
    }
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpGetInterfaceIndexForNextHop()
// PURPOSE      This function looks at the network address of each of a
//              node's network interfaces.  When nextHopAddress is
//              matched to a network, the interface index corresponding
//              to the network is returned.
//              (used by NetworkUpdateForwardingTable() and ospfv2.pc)
// PARAMETERS   Node *node
//                  Pointer to node.
//              NodeAddress nextHopAddress
//                  IP address
//
// RETURN       Index of outgoing interface, if nextHopAddress is on a
//              directly connected network.
//              -1, otherwise.
//-----------------------------------------------------------------------------

int
NetworkIpGetInterfaceIndexForNextHop(
    Node *node,
    NodeAddress nextHopAddress)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    int i;

    if (nextHopAddress == (unsigned) NETWORK_UNREACHABLE)
    {
        // Return bad value early if the IP address is NETWORK_UNREACHABLE.

        return -1;
    }

    // Given a next hop IP address, return the index of the corresponding
    // outgoing interface.

    for (i = 0; i < node->numberInterfaces; i++)
    {
        NodeAddress subnetMask = NetworkIpGetInterfaceSubnetMask(node, i);
        NodeAddress maskedAddress = MaskIpAddress(nextHopAddress, subnetMask);
        NodeAddress interfaceNetworkAddress =
            MaskIpAddress(ip->interfaceInfo[i]->ipAddress, subnetMask);

        if (maskedAddress == interfaceNetworkAddress)
        {
            return i;
        }
    }

    // Couldn't find IP address on a directly connected network, so
    // return -1 as the interface index.

    return -1;
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkGetInterfaceIndexForDestAddress()
// PURPOSE      Get interface for the destination address.
// PARAMETERS   Node *node
//                  Pointer to node.
//              NodeAddress destAddress
//                  destination associated with the interface.
// RETURN       interface index associated with destination.
//-----------------------------------------------------------------------------

int
NetworkGetInterfaceIndexForDestAddress(
    Node *node,
    NodeAddress destAddress)
{
    int interfaceIndex;
    NodeAddress nextHop;

    if (node->numberInterfaces == 1 || destAddress == ANY_DEST)
    {
        return DEFAULT_INTERFACE;
    }

    NetworkGetInterfaceAndNextHopFromForwardingTable(
        node, destAddress, &interfaceIndex, &nextHop);

    if (nextHop == (unsigned) NETWORK_UNREACHABLE)
    {
        return DEFAULT_INTERFACE;
    }
    else
    {
        return interfaceIndex;
    }
}


// FUNCTION   NetworkIpGetInterfaceIndexFromAddress
// PURPOSE    Get the interface index from an IP address.
// PARAMETERS node - this node.
//            address - address to determine interface index for.
// RETURN     interface index associated with specified address.
int NetworkIpGetInterfaceIndexFromAddress(Node *node, NodeAddress address)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    int i;

    for (i = 0; i < node->numberInterfaces; i++)
    {
        if (ip->interfaceInfo[i]->ipAddress == address)
        {
            return i;
        }
    }

    return -1;
}


// FUNCTION   NetworkIpGetInterfaceIndexFromSubnetAddress
// PURPOSE    Get the interface index from an IP subnet address.
// PARAMETERS node - this node.
//            address - subnet address to determine interface index for.
// RETURN     interface index associated with specified subnet address.
int NetworkIpGetInterfaceIndexFromSubnetAddress(Node *node,
                                                NodeAddress address)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    int i;
    NodeAddress networkAddr, interfaceNetworkAddr;

    for (i = 0; i < node->numberInterfaces; i++)
    {
        networkAddr = MaskIpAddress(address,
                                    ConvertNumHostBitsToSubnetMask(
                                           ip->interfaceInfo[i]->numHostBits));

        interfaceNetworkAddr = MaskIpAddress(ip->interfaceInfo[i]->ipAddress,
                                    ConvertNumHostBitsToSubnetMask(
                                           ip->interfaceInfo[i]->numHostBits));

        if (networkAddr == interfaceNetworkAddr)
        {
            return i;
        }
    }

    return -1;
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkGetMetricForDestAddress()
// PURPOSE      Get the cost metric for a destination from the forwarding table.
// PARAMETERS   Node *node
//                  Pointer to node.
//              NodeAddress destAddress
//                  destination to get cost metric from.
// RETURN       Cost metric associated with destination.
//-----------------------------------------------------------------------------

int
NetworkGetMetricForDestAddress(
    Node *node,
    NodeAddress destAddress)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    NetworkForwardingTable *forwardTable = &(ip->forwardTable);
    int i;
    int metric = 0xFFFFFFFF;

    for (i = 0; i < forwardTable->size; i++) {
        NodeAddress maskedDestinationAddress =
            MaskIpAddress(
                destAddress, forwardTable->row[i].destAddressMask);

        if (forwardTable->row[i].destAddress == maskedDestinationAddress
            && forwardTable->row[i].nextHopAddress != (unsigned) NETWORK_UNREACHABLE
            && forwardTable->row[i].interfaceIsEnabled != FALSE)
        {
            metric = forwardTable->row[i].cost;
            break;
        }
    }
    return metric;
}



//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpSetRouteUpdateEventFunction()
// PURPOSE      Set a callback fuction when a route changes from the forwarding
//              table.
// PARAMETERS   Node *node
//                  Pointer to node.
//              NetworkRouteUpdateEventType routeUpdateFunctionPtr)
//                  Route update callback function to set.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkIpSetRouteUpdateEventFunction(
    Node *node,
    NetworkRouteUpdateEventType routeUpdateFunctionPtr)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    ERROR_Assert(ip->routeUpdateFunction == NULL,
                 "Error while setting route update function\n");

    ip->routeUpdateFunction = routeUpdateFunctionPtr;
}



//-----------------------------------------------------------------------------
// FUNCTION     NetworkRouteUpdateEventType()
// PURPOSE      Get the route update callback function.
// PARAMETERS   Node *node
//                  Pointer to node.
// RETURN       Route update function.
//-----------------------------------------------------------------------------

NetworkRouteUpdateEventType
NetworkIpGetRouteUpdateEventFunction(
    Node *node)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    return (ip->routeUpdateFunction);
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkRoutingGetAdminDistance()
// PURPOSE      Get the administrative distance of a routing protocol.
// PARAMETERS   Node *node
//                  Pointer to node.
//              NetworkRoutingProtocolType type
//                  Type value of routing protocol.
// RETURN       The administrative distance of the routing protocol.
//
// NOTES        These values don't quite match those recommended by
//              Cisco.
//-----------------------------------------------------------------------------

NetworkRoutingAdminDistanceType
NetworkRoutingGetAdminDistance(
    Node *node,
    NetworkRoutingProtocolType type)
{
    switch (type)
    {
        case ROUTING_PROTOCOL_STATIC:
        {
            return ROUTING_ADMIN_DISTANCE_STATIC;
        }

        case ROUTING_PROTOCOL_DEFAULT:
        {
            return ROUTING_ADMIN_DISTANCE_DEFAULT;
        }

        case EXTERIOR_GATEWAY_PROTOCOL_EBGPv4:
        {
            return ROUTING_ADMIN_DISTANCE_EBGPv4;
        }

        case EXTERIOR_GATEWAY_PROTOCOL_IBGPv4:
        {
            return ROUTING_ADMIN_DISTANCE_IBGPv4;
        }

        case EXTERIOR_GATEWAY_PROTOCOL_BGPv4_LOCAL:
        {
            return ROUTING_ADMIN_DISTANCE_BGPv4_LOCAL;
        }

        case ROUTING_PROTOCOL_OSPFv2:
        {
            return ROUTING_ADMIN_DISTANCE_OSPFv2;
        }
        case ROUTING_PROTOCOL_OSPFv3:
        {
            return ROUTING_ADMIN_DISTANCE_OSPFv3;
        }

        case ROUTING_PROTOCOL_STAR:
        {
            return ROUTING_ADMIN_DISTANCE_STAR;
        }

        case ROUTING_PROTOCOL_BELLMANFORD:
        {
            return ROUTING_ADMIN_DISTANCE_BELLMANFORD;
        }

        case ROUTING_PROTOCOL_FISHEYE:
        {
            return ROUTING_ADMIN_DISTANCE_FISHEYE;
        }

        case ROUTING_PROTOCOL_SDR:
        {
            return ROUTING_ADMIN_DISTANCE_SDR;
        }

        case ROUTING_PROTOCOL_OLSR_INRIA:
        {
            return ROUTING_ADMIN_DISTANCE_OLSR;
        }

        case ROUTING_PROTOCOL_OLSRv2_NIIGATA:
        {
            return ROUTING_ADMIN_DISTANCE_OLSRv2_NIIGATA;
        }

        case ROUTING_PROTOCOL_IGRP:
        {
            return ROUTING_ADMIN_DISTANCE_IGRP;
        }

        case ROUTING_PROTOCOL_EIGRP:
        {
            return ROUTING_ADMIN_DISTANCE_EIGRP;
        }
//StartRIP
        case ROUTING_PROTOCOL_RIP:
        {
            return ROUTING_ADMIN_DISTANCE_RIP;
        }
//EndRIP
//StartRIPng
        case ROUTING_PROTOCOL_RIPNG:
        {
            return ROUTING_ADMIN_DISTANCE_RIPNG;
        }
//EndRIPng
//StartANTHOCNET
        case ROUTING_PROTOCOL_ANTHOCNET:
        {
            return ROUTING_ADMIN_DISTANCE_ANTHOCNET;
        }
//EndANTHOCNET
//InsertPatch ROUTING_ADMIN_DISTANCE
        default:
            ERROR_ReportError("Invalid switch value");
            return (NetworkRoutingAdminDistanceType) 0;  // Not reachable.
    }
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkInitForwardingTable()
// PURPOSE      Initialize the IP fowarding table, allocate enough
//              memory for number of rows.
// PARAMETERS   Node *node
//                  Pointer to node.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkInitForwardingTable(Node *node)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    ip->forwardTable.size = 0;
    ip->forwardTable.allocatedSize = 0;
    ip->forwardTable.row = NULL;
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkUpdateForwardingTable()
// PURPOSE      Update or add entry to IP routing table.  Search the
//              routing table for an entry with an exact match for
//              destAddress, destAddressMask, and routing protocol.
//              Update this entry with the specified nextHopAddress
//              (the outgoing interface is automatically determined
//              from the nextHopAddress -- see code).  If no matching
//              entry found, then add a new route.
// PARAMETERS   Node *node
//                  Pointer to node.
//              NodeAddress destAddress
//                  IP address of destination network or host.
//              NodeAddress destAddressMask
//                  Netmask.
//              NodeAddress nextHopAddress
//                  Next hop IP address.
//              int cost,
//                  Cost metric associated with the route.
//              NetworkRoutingProtocolType type
//                  Type value of routing protocol.
// RETURN       None.
//
// NOTES        The type field implies that the routing table can
//              simultaneously have entries with the same destination
//              addresses and netmasks, these entries added by different
//              routing protocols.
//
//              This function should have an interfaceIndex field, if
//              the protocol wishes to specify the outgoing interface
//              directly.
//-----------------------------------------------------------------------------

void
NetworkUpdateForwardingTable(
    Node *node,
    NodeAddress destAddress,
    NodeAddress destAddressMask,
    NodeAddress nextHopAddress,
    int interfaceIndex,
    int cost,
    NetworkRoutingProtocolType type)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    NetworkForwardingTable *forwardTable = &(ip->forwardTable);

    NetworkRoutingAdminDistanceType adminDistance
        = NetworkRoutingGetAdminDistance(node, type);
    int i;

    NetworkRouteUpdateEventType routeUpdateFunction = NULL;

    if (interfaceIndex == ANY_INTERFACE)
    {
        if (nextHopAddress == (unsigned) NETWORK_UNREACHABLE)
        {
            interfaceIndex = DEFAULT_INTERFACE;
        }
        else
        {
            interfaceIndex = NetworkIpGetInterfaceIndexForNextHop(
                node,
                nextHopAddress);
        }
    }

    if (interfaceIndex < 0)
    {
        char err[MAX_STRING_LENGTH];
        char addr[MAX_STRING_LENGTH];

        IO_ConvertIpAddressToString(nextHopAddress, addr);
        sprintf(err, "Node %u: Next hop %s is not connected to this node\n",
            node->nodeId, addr);
        ERROR_ReportError(err);
    }

#ifdef ENTERPRISE_LIB
    // Will proceed if Redistribution is enabled
    if (ip->rtRedistributeIsEnabled == TRUE)
    {
        RouteRedistributeAddHook(
            node,
            destAddress,
            destAddressMask,
            nextHopAddress,
            interfaceIndex,
            cost,
            type);
    }
#endif // ENTERPRISE_LIB

    for (i = 0; i < forwardTable->size
                && (forwardTable->row[i].destAddress != destAddress
                    || forwardTable->row[i].destAddressMask != destAddressMask
                    || forwardTable->row[i].protocolType != type); i++)
    {
        // Loop until match.
    }

    if (i == forwardTable->size)
    {
        forwardTable->size++;

        if (forwardTable->size > forwardTable->allocatedSize)
        {
            if (forwardTable->allocatedSize == 0)
            {
                forwardTable->allocatedSize = FORWARDING_TABLE_ROW_START_SIZE;
                forwardTable->row = (NetworkForwardingTableRow*)
                    MEM_malloc(
                        forwardTable->allocatedSize *
                        sizeof(NetworkForwardingTableRow));
            }
            else
            {
                int newSize = (forwardTable->allocatedSize * 2);

                NetworkForwardingTableRow* newTableRow =
                    (NetworkForwardingTableRow*)MEM_malloc(
                        newSize * sizeof(NetworkForwardingTableRow));

                memcpy(newTableRow, forwardTable->row,
                       (forwardTable->allocatedSize *
                        sizeof(NetworkForwardingTableRow)));

                MEM_free(forwardTable->row);
                forwardTable->row = newTableRow;
                forwardTable->allocatedSize = newSize;
            }//if//
        }//if//

        while (i > 0 &&
               (destAddress > forwardTable->row[i - 1].destAddress
                || (destAddress == forwardTable->row[i - 1].destAddress
                    && destAddressMask > forwardTable->row[i - 1].destAddressMask)
                || (destAddress == forwardTable->row[i - 1].destAddress
                    && destAddressMask == forwardTable->row[i - 1].destAddressMask
                    && adminDistance < forwardTable->row[i - 1].adminDistance)))
        {
            forwardTable->row[i] = forwardTable->row[i - 1];
            i--;
        }//while//
    }//if//

    forwardTable->row[i].destAddress = destAddress;
    forwardTable->row[i].destAddressMask = destAddressMask;
    forwardTable->row[i].interfaceIndex = interfaceIndex;
    forwardTable->row[i].nextHopAddress = nextHopAddress;
    forwardTable->row[i].protocolType = type;
    forwardTable->row[i].adminDistance = adminDistance;

    forwardTable->row[i].cost = cost;

    routeUpdateFunction = NetworkIpGetRouteUpdateEventFunction(node);

    if (routeUpdateFunction)
    {
        (routeUpdateFunction)(node, destAddress, destAddressMask,
                nextHopAddress, interfaceIndex, cost, adminDistance);
    }

    if (MAC_InterfaceIsEnabled(node, interfaceIndex))
    {
        forwardTable->row[i].interfaceIsEnabled = TRUE;
    }
    else
    {
        forwardTable->row[i].interfaceIsEnabled = FALSE;
    }
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkEmptyFowardingTable()
// PURPOSE      Remove entries in the routing table corresponding to a
//              given routing protocol.
// PARAMETERS   Node *node
//                  Pointer to node.
//              NetworkRoutingProtocolType type
//                  Type of routing protocol whose entries are to be
//                  removed.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkEmptyForwardingTable(
    Node *node,
    NetworkRoutingProtocolType type)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    NetworkForwardingTable *rt = &ip->forwardTable;

    int i = 0;

    // Go through the routing table...
    while (i < rt->size)
    {
        // Delete entries that corresponds to the routing protocol used
        if (rt->row[i].protocolType == type)
        {
            int j = i + 1;

            // Move all other entries down
            while (j < rt->size)
            {
                rt->row[j - 1] = rt->row[j];
                j++;
            }

            // Update forwarding table size.
            rt->size--;
        }
        else
        {
            i++;
        }
    }
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkPrintForwardingTable()
// PURPOSE      Display all entries in node's routing table.
// PARAMETERS   Node *node
//                  Pointer to node.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkPrintForwardingTable(Node *node)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    NetworkForwardingTable *rt = &ip->forwardTable;

    int i;
    char clockStr[MAX_CLOCK_STRING_LENGTH];

    ctoa((getSimTime(node) / SECOND), clockStr);

    printf("Forwarding Table for node %u at time %s\n", node->nodeId, clockStr);
    printf("---------------------------------------------------------------"
        "--------------------\n");
    printf("          dest          mask        intf       nextHop    protocol"
        "    admin    Flag\n");
    printf("---------------------------------------------------------------"
        "--------------------\n");
    for (i = 0; i < rt->size; i++)
    {
        char address[20];
        IO_ConvertIpAddressToString(rt->row[i].destAddress, address);
        printf("%15s  ", address);
        IO_ConvertIpAddressToString(rt->row[i].destAddressMask, address);
        printf("%15s  ", address);
        printf("%5u", rt->row[i].interfaceIndex);
        IO_ConvertIpAddressToString(rt->row[i].nextHopAddress, address);
        printf("%15s   ", address);
        printf("%5u      ", rt->row[i].protocolType);
        printf("%5u", rt->row[i].adminDistance);

        if (rt->row[i].interfaceIsEnabled) {
            printf("       U\n");
        } else {
            printf("       D\n");
        }
    }

    printf ("\n");
}


//-----------------------------------------------------------------------------
// Interface IP addresses
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpGetInterfaceAddress()
// PURPOSE      Get interface address on this interface.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int interfaceIndex
//                  interface to get address from.
// RETURN       Interface address associated with interface.
//-----------------------------------------------------------------------------

NodeAddress
NetworkIpGetInterfaceAddress(
    Node *node,
    int interfaceIndex)
{
    if (INTERFACE_DEBUG) {
        printf("partition %d checking interface %d address for node %d\n", node->partitionData->partitionId, interfaceIndex, node->nodeId);
        fflush(stdout);
    }
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    assert(ip != NULL);

    if (interfaceIndex == (int)ANY_INTERFACE)
    {
        return (NodeAddress)ANY_ADDRESS;
    }

    return (ip->interfaceInfo[interfaceIndex]->ipAddress);
}


NodeAddress
NetworkIpGetLinkLayerAddress(Node* node, int interfaceIndex)
{
    NetworkType networkType =
        NetworkIpGetInterfaceType(node, interfaceIndex);

    if (networkType == NETWORK_IPV4)
    {
        return NetworkIpGetInterfaceAddress(node, interfaceIndex);
    }
    else if (networkType == NETWORK_IPV6)
    {
        return Ipv6GetLinkLayerAddress(node, interfaceIndex, NULL);
    }

    // When dual IP at interface level will be supported, this function
    // may need to be modified to consider both interface. Though it will
    // be problem, because our link layer address is different for IPv4
    // and IPv6, which is not supposed to be.

    return 0;
}

int
NetworkIpGetInterfaceIndexFromLinkAddress(Node* node,
                          NodeAddress ownMacAddr)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    int i;

    for (i = 0; i < node->numberInterfaces; i++)
    {
        if (ip->interfaceInfo[i]->interfaceType == NETWORK_IPV4)
        {
            if (ip->interfaceInfo[i]->ipAddress == ownMacAddr)
            {
                return i;
            }
        }
        else if (ip->interfaceInfo[i]->interfaceType == NETWORK_IPV6)
        {
            if (ip->interfaceInfo[i]->ipv6InterfaceInfo->linkLayerAddr
                == ownMacAddr)
            {
                return i;
            }
        }
    }
    return -1;
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpGetInterfaceName()
// PURPOSE      Get interface name on this interface.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int interfaceIndex
//                  interface to get address from.
// RETURN       Interface name associated with interface.
//-----------------------------------------------------------------------------

char*
NetworkIpGetInterfaceName(
    Node *node,
    int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    return (ip->interfaceInfo[interfaceIndex]->interfaceName);
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpGetInterfaceNetworkAddress()
// PURPOSE      Get network interface address on this interface.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int interfaceIndex
//                  interface to get network address from.
// RETURN       network address associated with interface.
//-----------------------------------------------------------------------------

NodeAddress
NetworkIpGetInterfaceNetworkAddress(
    Node *node,
    int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    return (ip->interfaceInfo[interfaceIndex]->ipAddress
            & ConvertNumHostBitsToSubnetMask(
                  ip->interfaceInfo[interfaceIndex]->numHostBits));
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpGetInterfaceSubnetMask()
// PURPOSE      Get network interface subnet mask on this interface.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int interfaceIndex
//                  interface to get subnet mask from
// RETURN       Subnet mask of specified interface.
//-----------------------------------------------------------------------------

NodeAddress
NetworkIpGetInterfaceSubnetMask(
    Node *node,
    int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    return (ConvertNumHostBitsToSubnetMask(
                ip->interfaceInfo[interfaceIndex]->numHostBits));
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpGetInterfaceNumHostBits()
// PURPOSE      Get number of host nits on this interface.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int interfaceIndex
//                  interface to get number of host bits from.
// RETURN       Number of host bits associated with the interface.
//-----------------------------------------------------------------------------

int
NetworkIpGetInterfaceNumHostBits(
    Node *node,
    int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    return (ip->interfaceInfo[interfaceIndex]->numHostBits);
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpGetInterfaceBroadcastAddress()
// PURPOSE      Get broadcast address on this interface.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int interfaceIndex
//                  interface to get broadcast address from.
// RETURN       Broadcast address of the interface.
//-----------------------------------------------------------------------------

NodeAddress
NetworkIpGetInterfaceBroadcastAddress(
    Node *node,
    int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    //
    // ((sizeof(NodeAddress)*8) - numHostBits) is the number of bits that should
    // be set to '1' for broadcasting
    //
    // ANY_DEST is 0xffffffff: shifting this by the above number of bits
    // provides the basis for the bitwise OR
    //
    // Performing bitwise OR with the above value and the interface IP gives
    // the Interface Broadcast Address
    //

    if (NetworkIpGetInterfaceType(node, interfaceIndex) == NETWORK_IPV6)
    {
        return ANY_DEST;
    }

    int bMaskBits = ((sizeof(NodeAddress)*8)-
                 ip->interfaceInfo[interfaceIndex]->numHostBits);

    if(bMaskBits >= (sizeof(ANY_DEST)*8))
    {
            return ip->interfaceInfo[interfaceIndex]->ipAddress;
    }
    else
    {
        return (ip->interfaceInfo[interfaceIndex]->ipAddress |
            (ANY_DEST >> bMaskBits));

    }

}

//-----------------------------------------------------------------------------
// Miscellaneous
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpIsMyIP()
// PURPOSE      Calls IsMyPacket().
// PARAMETERS   Node *node
//                  Pointer to node.
//              NodeAddress ipAddr
//                  An IP address.
//
// RETURN       Calls IsMyPacket().
//
// NOTES        This is only called by application/mpls_ldp.pc.  Should
//              probably eliminate this function. -Jeff
//-----------------------------------------------------------------------------

BOOL
NetworkIpIsMyIP(Node *node, NodeAddress ipAddress)
{
    return IsMyPacket(node, ipAddress);
}

//-----------------------------------------------------------------------------
// Debugging
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpConfigurationError()
// PURPOSE      Print out IP configuration error.
// PARAMETERS   Node *node
//                  Pointer to node.
//              char parameterName[]
//                  Error message to print out.
//              int interfaceIndex
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkIpConfigurationError(
    Node *node,
    char parameterName[],
    int interfaceIndex)
{
    char buf[MAX_STRING_LENGTH];

    sprintf(buf,
            "node %d (interface %d) cannot find \"%s\" in the CONFIG.IN file"
            " or the value specified is invalid",
            node->nodeId, interfaceIndex, parameterName);
    ERROR_ReportError(buf);
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkPrintIpHeader()
// PURPOSE      Display IP header for message containing an IP header
//              or IP packet.
// PARAMETERS   Message *msg
//                  Pointer to message with IP header or IP packet.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkPrintIpHeader(Message *msg)
{
    IpHeaderType *ipHdr;
    struct ip_options *opt;
    char *dataptr = msg->packet;
    int i;

    ipHdr = (IpHeaderType *) (msg->packet);

    printf("IP header\n");
    printf("totalLength %d headerLength %d moreFragment %d fragmentOffset %d"
           "  timeToLive %d ",
           ipHdr->ip_len, ipHdr->ip_hl, ipHdr->ip_more_fragments,
           ipHdr->ip_fragment_offset,
           ipHdr->ip_ttl);
    printf("protocol %d sourceId %d destId %d identity %d\n",
           ipHdr->ip_p, ipHdr->ip_src, ipHdr->ip_dst, ipHdr->ip_id);

    dataptr += sizeof(IpHeaderType);
    opt = (struct ip_options *) dataptr;

    printf("code %d, len %d, ptr %d\n", opt->code, opt->len, opt->ptr);

    if (ipHdr->ip_hl * 4 > sizeof(IpHeaderType))
    {
        if (opt->code == IPOPT_SSRR)
        {
            for (i = 1; i <= (opt->len - 3) / 4; i++)
            {
                NodeAddress nodePtr;
                memcpy(&nodePtr, (dataptr + i*4 - 1), sizeof(NodeAddress));
                printf("%d, ", nodePtr);

            }
            printf ("\n");
        }
    }

    //printf("payload %s\n", (dataptr + opt->len));
}

//-----------------------------------------------------------------------------
// Statistics
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpRunTimeStat()
// PURPOSE      Print IP runtime statistics.
// PARAMETERS   Node *node
//                  Pointer to node.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkIpRunTimeStat(Node *node)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    clocktype      now = getSimTime(node);

    if (node->guiOption) {
        GUI_SendUnsignedData(node->nodeId, ip->stats.ipInReceivesId,
                             ip->stats.ipInReceives, now);
        GUI_SendUnsignedData(node->nodeId, ip->stats.ipInHdrErrorsId,
                             ip->stats.ipInHdrErrors, now);
    }
    else {
        printf("Node: %u\n", node->nodeId);
        printf("This time period:\n");

        printf("ipInReceives         -> %u\n",
               ip->stats.ipInReceives -
               ip->stats.ipInReceivesLastPeriod);
        ip->stats.ipInReceivesLastPeriod = ip->stats.ipInReceives;

        printf("ipInHdrErrors        -> %u\n",
               ip->stats.ipInHdrErrors -
               ip->stats.ipInHdrErrorsLastPeriod);
        ip->stats.ipInHdrErrorsLastPeriod = ip->stats.ipInHdrErrors;

        printf("ipInForwardDatagrams -> %u\n",
               ip->stats.ipInForwardDatagrams -
               ip->stats.ipInForwardDatagramsLastPeriod);
        ip->stats.ipInForwardDatagramsLastPeriod =
            ip->stats.ipInForwardDatagrams;

        printf("ipInDelivers         -> %u\n",
               ip->stats.ipInDelivers -
               ip->stats.ipInDeliversLastPeriod);
        ip->stats.ipInDeliversLastPeriod = ip->stats.ipInDelivers;

        printf("ipOutRequests        -> %u\n",
               ip->stats.ipOutRequests -
               ip->stats.ipOutRequestsLastPeriod);
        ip->stats.ipOutRequestsLastPeriod = ip->stats.ipOutRequests;

        printf("ipOutDiscards        -> %u\n",
               ip->stats.ipOutDiscards -
               ip->stats.ipOutDiscardsLastPeriod);
        ip->stats.ipOutDiscardsLastPeriod = ip->stats.ipOutDiscards;

        printf("ipOutNoRoutes        -> %u\n",
               ip->stats.ipOutNoRoutes -
               ip->stats.ipOutNoRoutesLastPeriod);
        ip->stats.ipOutNoRoutesLastPeriod = ip->stats.ipOutNoRoutes;

        printf("ipReasmReqds         -> %u\n",
               ip->stats.ipReasmReqds -
               ip->stats.ipReasmReqdsLastPeriod);
        ip->stats.ipReasmReqdsLastPeriod = ip->stats.ipReasmReqds;

        printf("ipReasmOKs           -> %u\n",
               ip->stats.ipReasmOKs -
               ip->stats.ipReasmOKsLastPeriod);
        ip->stats.ipReasmOKsLastPeriod = ip->stats.ipReasmOKs;

        printf("ipReasmFails         -> %u\n",
               ip->stats.ipReasmFails -
               ip->stats.ipReasmFailsLastPeriod);
        ip->stats.ipReasmFailsLastPeriod = ip->stats.ipReasmFails;

        printf("ipFragOKs            -> %u\n",
               ip->stats.ipFragOKs -
               ip->stats.ipFragOKsLastPeriod);
        ip->stats.ipFragOKsLastPeriod = ip->stats.ipFragOKs;
    }
}



//-----------------------------------------------------------------------------
// Multicast
//-----------------------------------------------------------------------------

// FUNCTION   NetworkIpAddToMulticastGroupList
// PURPOSE    Add group to multicast group list.
// PARAMETERS node - this node.
//            groupAddress - group to add to multicast group list.
// RETURN     None.
void NetworkIpAddToMulticastGroupList(Node *node,
                                      NodeAddress groupAddress)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    ListItem *item = ip->multicastGroupList->first;

    NetworkIpMulticastGroupEntry *entry;

    // Go through the group list...
    while (item)
    {
        entry = (NetworkIpMulticastGroupEntry *) item->data;

        // Group already exists, so incrememt member count.
        if (entry->groupAddress == groupAddress)
        {
            entry->memberCount++;
            return;
        }

        item = item->next;
    }

    // Group doesn't exist, so add to multicast group list.

    entry = (NetworkIpMulticastGroupEntry *)
            MEM_malloc(sizeof(NetworkIpMulticastGroupEntry));

    entry->groupAddress = groupAddress;
    entry->memberCount = 1;

    ListInsert(node, ip->multicastGroupList, getSimTime(node), entry);
}


// FUNCTION   NetworkIpRemoveFromMulticastGroupList
// PURPOSE    Remove group from multicast group list.
// PARAMETERS node - this node.
//            groupAddress - group to remove from multicast group list.
// RETURN     None.
void NetworkIpRemoveFromMulticastGroupList(Node *node,
                                           NodeAddress groupAddress)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    ListItem *item = ip->multicastGroupList->first;
    NetworkIpMulticastGroupEntry *entry;

    // Go through the multicast group list...
    while (item)
    {
        entry = (NetworkIpMulticastGroupEntry *) item->data;

        // Found it...
        if (entry->groupAddress == groupAddress)
        {
            // If only no one else belongs to the group, remove from group list.
            if (entry->memberCount == 1)
            {
                ListGet(node,
                        ip->multicastGroupList,
                        item,
                        TRUE,
                        FALSE);
            }
            // Someone else also belongs to the group,
            // so decrement member count.
            else if (entry->memberCount > 1)
            {
                entry->memberCount--;
            }
            else
            {
                assert(FALSE);
            }

            return;
        }

        item = item->next;
    }
}


// FUNCTION   NetworkIpPrintMulticastGroupList
// PURPOSE    Print the multicast group list.
// PARAMETERS node - this node.
// RETURN     None.
void NetworkIpPrintMulticastGroupList(Node *node)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    ListItem *item = ip->multicastGroupList->first;
    NetworkIpMulticastGroupEntry *entry;

    printf("Node %d multicast group list\n", node->nodeId);

    while (item)
    {
        entry = (NetworkIpMulticastGroupEntry *) item->data;

        printf("    group address = %u, member count = %u\n",
                entry->groupAddress, entry->memberCount);

        item = item->next;
    }
}


// FUNCTION   NetworkIpIsPartOfMulticastGroup
// PURPOSE    Check if we are part of the multicast group.
// PARAMETERS node - this node.
//            groupAddress - group to check if node is part of multicast group.
// RETURN     TRUE if node is part of multicast group, FALSE otherwise.
BOOL NetworkIpIsPartOfMulticastGroup(Node *node, NodeAddress groupAddress)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    ListItem *item = ip->multicastGroupList->first;
    NetworkIpMulticastGroupEntry *entry;

    // Go through list and see if we belong to this multicast group.
    while (item)
    {
        entry = (NetworkIpMulticastGroupEntry *) item->data;

        // Found it!
        if (entry->groupAddress == groupAddress)
        {
            return TRUE;
        }

        item = item->next;
    }

    // Not part of multicast group...
    return FALSE;
}


// FUNCTION   NetworkIpJoinMulticastGroup
// PURPOSE    Join a multicast group.
// PARAMETERS node - this node.
//            mcastAddr, multicast group to join.
//            delay - when to join group.
// RETURN     None.
void NetworkIpJoinMulticastGroup(Node *node,
                          NodeAddress mcastAddr,
                          clocktype delay)
{
    char errStr[MAX_STRING_LENGTH];
    char grpStr[MAX_ADDRESS_STRING_LENGTH];

    if (mcastAddr < IP_MIN_MULTICAST_ADDRESS
        || mcastAddr > IP_MAX_MULTICAST_ADDRESS)
    {
        IO_ConvertIpAddressToString(mcastAddr, grpStr);
        sprintf(errStr, "Node %u: Unable to set JoinTimer\n"
            "    Group address %s is not a valid multicast address\n",
            node->nodeId, grpStr);

        ERROR_ReportError(errStr);
    }
    else if (mcastAddr <= IP_MAX_RESERVED_MULTICAST_ADDRESS)
    {
        IO_ConvertIpAddressToString(mcastAddr, grpStr);
        sprintf(errStr, "Node %u: Unable to set JoinTimer\n"
            "    Group address %s falls in reserve multicast address "
            "space\n",
            node->nodeId, grpStr);

        ERROR_ReportError(errStr);
    }
    else
    {
        NetworkIpSetMulticastTimer(
            node, MSG_NETWORK_JoinGroup, mcastAddr, delay);


        if (!NetworkIpCheckMulticastRoutingProtocol(
#ifdef ADDON_MAODV
                node, MULTICAST_PROTOCOL_ODMRP, ANY_INTERFACE)
            && !NetworkIpCheckMulticastRoutingProtocol(
                node, MULTICAST_PROTOCOL_MAODV, ANY_INTERFACE))
#else // ADDON_MAODV
                node, MULTICAST_PROTOCOL_ODMRP, ANY_INTERFACE))
#endif // ADDON_MAODV
        {
            // Send group joining request
            IgmpJoinGroup(node, mcastAddr, delay);
        }

    }
}


// FUNCTION   NetworkIpLeaveMulticastGroup
// PURPOSE    Leave a multicast group.
// PARAMETERS node - this node.
//            mcastAddr, multicast group to leave.
//            delay - when to leave group.
// RETURN     None.
void NetworkIpLeaveMulticastGroup(Node *node,
                                  NodeAddress mcastAddr,
                                  clocktype delay)
{
    char errStr[MAX_STRING_LENGTH];
    char grpStr[MAX_ADDRESS_STRING_LENGTH];

    if (mcastAddr < IP_MIN_MULTICAST_ADDRESS
        || mcastAddr > IP_MAX_MULTICAST_ADDRESS)
    {
        IO_ConvertIpAddressToString(mcastAddr, grpStr);
        sprintf(errStr, "Node %u: Unable to set LeaveTimer\n"
            "    Group address %s is not a valid multicast address\n",
            node->nodeId, grpStr);

        ERROR_ReportError(errStr);
    }
    else if (mcastAddr <= IP_MAX_RESERVED_MULTICAST_ADDRESS)
    {
        IO_ConvertIpAddressToString(mcastAddr, grpStr);
        sprintf(errStr, "Node %u: Unable to set LeaveTimer\n"
            "    Group address %s falls in reserve multicast address "
            "space\n",
            node->nodeId, grpStr);

        ERROR_ReportError(errStr);
    }
    else
    {
        NetworkIpSetMulticastTimer(
            node, MSG_NETWORK_LeaveGroup, mcastAddr, delay);


        if (!NetworkIpCheckMulticastRoutingProtocol(
#ifdef ADDON_MAODV
                node, MULTICAST_PROTOCOL_ODMRP, ANY_INTERFACE)
            && !NetworkIpCheckMulticastRoutingProtocol(
                node, MULTICAST_PROTOCOL_MAODV, ANY_INTERFACE))
#else // ADDON_MAODV
                node, MULTICAST_PROTOCOL_ODMRP, ANY_INTERFACE))
#endif // ADDON_MAODV
        {
            // Send group leaving request
            IgmpLeaveGroup(node, mcastAddr, delay);
        }

    }
}


// FUNCTION   NetworkIpSetMulticastTimer
// PURPOSE    Set timer to join and leave multicast groups.
// PARAMETERS node - this node.
//            mcastAddr, multicast group to join or leave.
//            delay - when to join or leave group.
// RETURN     None.
void NetworkIpSetMulticastTimer(Node *node,
                          Int32 eventType,
                          NodeAddress mcastAddr,
                          clocktype delay)

{
    Message *msg;
    NodeAddress *info;

    msg = MESSAGE_Alloc(node,
                         NETWORK_LAYER,
                         NETWORK_PROTOCOL_IP,
                         eventType);

    MESSAGE_InfoAlloc(node, msg, sizeof(NodeAddress));

    info = (NodeAddress *) MESSAGE_ReturnInfo(msg);

    memcpy(info, &mcastAddr, sizeof(NodeAddress));

    MESSAGE_Send(node, msg, delay);
}


// FUNCTION   NetworkIpSetMulticastRoutingProtocol
// PURPOSE    Assign a multicast routing protocol structure to an interface.
// PARAMETERS node - this node.
//            multicastRoutingProtocol - multicast routing protocol to set.
//            interfaceIndex - interface associated with multicast protocol.
// RETURN     None.
void
NetworkIpSetMulticastRoutingProtocol(
    Node *node,
    void *multicastRoutingProtocol,
    int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    ERROR_Assert(ip->interfaceInfo[interfaceIndex]->multicastRoutingProtocol ==
                 NULL,
                 "Multicast router function already set");

    ip->interfaceInfo[interfaceIndex]->multicastRoutingProtocol=
                                                      multicastRoutingProtocol;
}


// FUNCTION   NetworkIpGetMulticastRoutingProtocol
// PURPOSE    Get the multicast routing protocol structure.
// PARAMETERS node - this node.
//            routingProtcolType - the multicast protocol to get.
// RETURN     Multicast routing protocol structure.
void *
NetworkIpGetMulticastRoutingProtocol(
    Node *node,
    NetworkRoutingProtocolType routingProtocolType)
{
    int i;
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    char errStr[MAX_STRING_LENGTH];

    for (i = 0; i < node->numberInterfaces; i++)
    {
        if (ip->interfaceInfo[i]->multicastProtocolType
            == routingProtocolType)
        {
            return ip->interfaceInfo[i]->multicastRoutingProtocol;
        }
    }

    sprintf(errStr,
            "Could not find multicast routing protocol %d at node %d\n",
            routingProtocolType, node->nodeId);
    ERROR_Assert(FALSE, errStr);

    return NULL; // Not reachable.
}


// FUNCTION   NetworkIpAddMulticastRoutingProtocolType
// PURPOSE    Assign a multicast protocol type to an interface.
// PARAMETERS node - this node.
//            multicastProtcolType - the multicast protocol to add.
//            interfaceIndex - interface associated with multicast protocol.
// RETURN     None
void NetworkIpAddMulticastRoutingProtocolType(
    Node *node,
    NetworkRoutingProtocolType multicastProtocolType,
    int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    ip->interfaceInfo[interfaceIndex]->multicastEnabled = TRUE;
    ip->interfaceInfo[interfaceIndex]->multicastRouterFunction = NULL;
    ip->interfaceInfo[interfaceIndex]->multicastProtocolType =
                                                          multicastProtocolType;
    ip->interfaceInfo[interfaceIndex]->multicastRoutingProtocol = NULL;
}


// FUNCTION   NetworkIpSetMulticastRouterFunction
// PURPOSE    Set the multicast router function to an interface.
// PARAMETERS node - this node.
//            routerFunctionPtr - multicast router function to set on this
//                                interface.
//            interfaceIndex - interface associated with multicast protocol.
// RETURN     None.
void
NetworkIpSetMulticastRouterFunction(
    Node *node,
    MulticastRouterFunctionType routerFunctionPtr,
    int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    assert(ip->interfaceInfo[interfaceIndex]->multicastRouterFunction == NULL);

    ip->interfaceInfo[interfaceIndex]->multicastRouterFunction =
                                                              routerFunctionPtr;
}


// FUNCTION   NetworkIpGetMulticastRouterFunction
// PURPOSE    Get the multicast router function from an interface.
// PARAMETERS node - this node.
//            interfaceIndex - interface associated with multicast protocol.
// RETURN     Multicast router function on this interface.
MulticastRouterFunctionType
NetworkIpGetMulticastRouterFunction(
    Node *node,
    int interfaceIndex)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    return ip->interfaceInfo[interfaceIndex]->multicastRouterFunction;
}


// FUNCTION  NetworkIpUpdateMulticastRoutingProtocolAndRouterFunction
// PURPOSE   Assign multicast routing protocol structure and router
//           function to an interface.  We are only allocating
//           the multicast routing protocol structure and router function
//           once by using pointers to the original structures.
void
NetworkIpUpdateMulticastRoutingProtocolAndRouterFunction(
    Node *node,
    NetworkRoutingProtocolType routingProtocolType,
    int interfaceIndex)
{
    int i;
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    for (i = 0; i < node->numberInterfaces; i++)
    {
        if (ip->interfaceInfo[i]->multicastProtocolType
            == routingProtocolType)
        {
            ip->interfaceInfo[interfaceIndex]->multicastRouterFunction =
                            ip->interfaceInfo[i]->multicastRouterFunction;

            ip->interfaceInfo[interfaceIndex]->multicastRoutingProtocol =
                            ip->interfaceInfo[i]->multicastRoutingProtocol;
            return;
        }
    }

    fprintf(stderr, "Could not find multicast router function %d\n",
            routingProtocolType);
    abort();
}


// FUNCTION  NetworkIpIsMulticastAddress
// PURPOSE   Check if an address is a multicast address.
// PARAMETERS node - this node.
//            address - address to determine if multicast address.
// RETURN     TRUE if address is multicast address, FALSE, otherwise.
BOOL NetworkIpIsMulticastAddress(Node *node, NodeAddress address)
{
    if (address >= IP_MIN_MULTICAST_ADDRESS &&
        address <= IP_MAX_MULTICAST_ADDRESS)
    {
        return TRUE;
    }

    return FALSE;
}

/*
 *  FUNCTION    NetworkInitMulticastForwardingTable
 *  PURPOSE     initialize the multicast fowarding table, allocate enough
 *              memory for number of rows, used by ip
 *  PARAMETER   node - this node.
 *  RETURN      None.
 */
void NetworkInitMulticastForwardingTable(Node *node)
{
    NetworkDataIp *ip;

    ip = (NetworkDataIp *) node->networkData.networkVar;

    ip->multicastForwardingTable.size = 0;
    ip->multicastForwardingTable.allocatedSize = 0;
    ip->multicastForwardingTable.row = NULL;
}


/*
 *  FUNCTION    NetworkEmptyMulticastFowardingTable
 *  PURPOSE     empty out all the entries in the multicast forwarding table.
 *              basically set the size of table back to 0.
 *  PARAMETER   node - this node.
 *  RETURN      None.
 */
void NetworkEmptyMulticastForwardingTable(Node *node)
{
    NetworkDataIp *ip;
    int i;

    ip = (NetworkDataIp *) node->networkData.networkVar;

    for (i=0; i < ip->multicastForwardingTable.size; i++)
    {
        ListFree(node,
                 ip->multicastForwardingTable.row[i].outInterfaceList,
                 FALSE);
    }

    ip->multicastForwardingTable.size = 0;
}



/*  FUNCTION    NetworkGetOutgoingInterfaceFromMulticastForwardingTable
 *  PURPOSE     get the interface Id node that lead to the
 *              (source, multicast group) pair.
 *  PARAMETER   node - its own node.
 *              sourceAddress - multicast source address.
 *              groupAddress - multicast group address to foward to.
 *  RETURN      interface Id from node to (source, multicast group), or
 *              NETWORK_UNREACHABLE (no such entry is found)
 */
LinkedList* NetworkGetOutgoingInterfaceFromMulticastForwardingTable(
    Node *node,
    NodeAddress sourceAddress,
    NodeAddress groupAddress)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    NetworkMulticastForwardingTable *multicastForwardingTable =
                                        &(ip->multicastForwardingTable);
    int i;

    // Look for (source, group) pair.  If exist, use the indicated interface.
    // If not, return NETWORK_UNREACHABLE;
    for (i=0; i < multicastForwardingTable->size; i++) {

        if ((multicastForwardingTable->row[i].sourceAddress ==
            sourceAddress) &&
            (multicastForwardingTable->row[i].multicastGroupAddress ==
            groupAddress))
        {
            return multicastForwardingTable->row[i].outInterfaceList;
            break;
        }
    }

    return NULL;
}


/*  FUNCTION    NetworkUpdateMulticastForwardingTable
 *  PURPOSE     update entry with (sourceAddress, multicastGroupAddress) pair.
 *              search for the row with (sourceAddress, multicastGroupAddress)
 *              and update its interface.
 *              if no row is found, add a new row and increase table size.
 *  PARAMETER   node - its own node.
 *              sourceAddress - multicast source
 *              multicastGroupAddress - multicast group
 *              interfaceIndex - interface to use for
 *                               (sourceAddress, multicastGroupAddress).
 */
void NetworkUpdateMulticastForwardingTable(
                            Node *node,
                            NodeAddress sourceAddress,
                            NodeAddress multicastGroupAddress,
                            int interfaceIndex)
{
    BOOL newInsert = FALSE;
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    NetworkMulticastForwardingTable *multicastForwardingTable =
                                        &(ip->multicastForwardingTable);

    int i;

    if (interfaceIndex < 0)
    {
        printf("Error node %u: invalid interface index %d is invalid.\n",
                node->nodeId, interfaceIndex);
        abort();
    }

    // See if there's a match in the table already and get the index
    // of the table where we want to update (match or new entry).

    for (i=0; i < multicastForwardingTable->size
        && (multicastForwardingTable->row[i].sourceAddress != sourceAddress
            || multicastForwardingTable->row[i].multicastGroupAddress !=
               multicastGroupAddress); i++)
    {
        /* No match. */
    }

    // If where we are going to insert is at the end, make sure the
    // table is big enough.  If not, make it bigger.
    if (i == multicastForwardingTable->size)
    {
        multicastForwardingTable->size++;

        newInsert = TRUE;

        // Increase the table size if we ran out of space.
        if (multicastForwardingTable->size >
            multicastForwardingTable->allocatedSize)
        {
            if (multicastForwardingTable->allocatedSize == 0)
            {
                multicastForwardingTable->allocatedSize =
                                                FORWARDING_TABLE_ROW_START_SIZE;

                multicastForwardingTable->row =
                                (NetworkMulticastForwardingTableRow *)
                                MEM_malloc(
                                    multicastForwardingTable->allocatedSize *
                                    sizeof(NetworkMulticastForwardingTableRow));
            }
            else
            {
                int newSize = (multicastForwardingTable->allocatedSize * 2);

                NetworkMulticastForwardingTableRow* newTableRow =
                        (NetworkMulticastForwardingTableRow*)
                        MEM_malloc(
                          newSize * sizeof(NetworkMulticastForwardingTableRow));

                memcpy(newTableRow,
                       multicastForwardingTable->row,
                       (multicastForwardingTable->allocatedSize *
                                sizeof(NetworkMulticastForwardingTableRow)));

                MEM_free(multicastForwardingTable->row);

                multicastForwardingTable->row = newTableRow;
                multicastForwardingTable->allocatedSize = newSize;
            }//if//
        }//if//

        while (i > 0 &&
               (sourceAddress >
                            multicastForwardingTable->row[i - 1].sourceAddress))
        {
            multicastForwardingTable->row[i] =
                                            multicastForwardingTable->row[i-1];
            i--;
        }//while//

    }//if//

    multicastForwardingTable->row[i].sourceAddress = sourceAddress;
    multicastForwardingTable->row[i].multicastGroupAddress =
                                                        multicastGroupAddress;

    if (newInsert)
    {
        ListInit(node, &multicastForwardingTable->row[i].outInterfaceList);
    }

    // Only insert interface if interface not in the interface list already.
    if (!NetworkInMulticastOutgoingInterface(node,
                         multicastForwardingTable->row[i].outInterfaceList,
                         interfaceIndex))
    {
        int *outInterfaceIndex = (int *) MEM_malloc(sizeof(int));

        *outInterfaceIndex = interfaceIndex;

        ListInsert(node,
                   multicastForwardingTable->row[i].outInterfaceList,
                   getSimTime(node),
                   outInterfaceIndex);
    }
}


/*
 *  FUNCTION    NetworkPrintMulticastForwardingTable
 *  PURPOSE     display all entries in multicast forwarding table of the node.
 *  PARAMETER   node - this node.
 *  RETURN      None.
 */
void NetworkPrintMulticastForwardingTable(Node *node)
{
    int i;
    NetworkDataIp *ip;
    NetworkMulticastForwardingTable *rt;

    ip = (NetworkDataIp *) node->networkData.networkVar;
    rt = &(ip->multicastForwardingTable);

    printf ("Multicast Forwarding Table for node %u\n", node->nodeId);

    for (i=0; i<rt->size; i++) {
        printf("sourceId %u "
               "multicastGroupAddress %u\n",
                rt->row[i].sourceAddress,
                rt->row[i].multicastGroupAddress);

        NetworkPrintMulticastOutgoingInterface(node,
                                               rt->row[i].outInterfaceList);
        printf ("\n");
    }
}


/*
 *  FUNCTION    NetworkPrintMulticastOutgoingInterface
 *  PURPOSE     Print mulitcast outgoing interfaces.
 *  PARAMETER   node - this node.
 *              list - list of outgoing interfaces.
 *  RETURN      None.
 */
void NetworkPrintMulticastOutgoingInterface(Node *node, LinkedList* list)
{
    ListItem *item = list->first;

    printf("    interface =");

    while (item)
    {
        int *interfaceIndex =  (int *)item->data;

        printf(" %u", *interfaceIndex);

        item = item->next;
    }
}


/*
 *  FUNCTION    NetworkInMulticastOutgoingInterface
 *  PURPOSE     Determine if interface is in multicast outgoing interface list.
 *  PARAMETER   node - this node.
 *              list - list of outgoing interfaces.
 *              interfaceIndex - interface to determine if in outgoing
 *              interface list.
 *  RETURN      TRUE if interface is in multicast outgoing interface list,
 *              FALSE otherwise.
 */

BOOL NetworkInMulticastOutgoingInterface(Node *node,
                                         LinkedList* list,
                                         int interfaceIndex)
{
    ListItem *item = list->first;

    while (item)
    {
        int *outInterfaceIndex =  (int *)item->data;

        if (interfaceIndex == *outInterfaceIndex)
        {
            return TRUE;
        }

        item = item->next;
    }

    return FALSE;
}



//-----------------------------------------------------------------------------
// FUNCTIONS WITH INTERNAL LINKAGE
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Packet delivery and forwarding
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     DeliverPacket()
// PURPOSE      Deliver IP packet from MAC layer to the appropriate
//              transport-layer or network-layer protocol.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message with IP packket.
//              int interfaceIndex
//                  Index of interface from which packet was received.
//-----------------------------------------------------------------------------

static void //inline//
DeliverPacket(Node *node, Message *msg,
              int interfaceIndex, NodeAddress previousHopAddress)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    NodeAddress sourceAddress = 0;
    NodeAddress destinationAddress =0;
    unsigned char ipProtocolNumber;
    unsigned ttl =0;
    TosType priority;

    BOOL aCongestionExperienced = FALSE;

    IpHeaderType *ipHeader = (IpHeaderType *) msg->packet;

    if ((int) ipHeader->ip_len > MESSAGE_ReturnPacketSize(msg))
    {
        BOOL isReassembled = FALSE;
        Message* joinedMsg = NULL;
        joinedMsg = IpFragmentInput(node, msg, interfaceIndex, &isReassembled);
        if (isReassembled)
        {
            msg = joinedMsg;
            ipHeader = (IpHeaderType *) msg->packet;
        }
        else
        {
            return;
        }
        // ERROR_ReportError("Received fragmented packet. So u need to join it\n");
    }
    ipProtocolNumber = ipHeader->ip_p;

#ifdef ENTERPRISE_LIB
    if (!MplsReturnStateSpace(node))
    {
        if ((ipHeader->ip_dst != ANY_DEST) &&
            (interfaceIndex != CPU_INTERFACE) &&
            !NetworkIpIsMulticastAddress(node, ipHeader->ip_dst))
        {
            // Let the routing protocol get a look at the original packet
            // before it is given to the transport layer or a
            // routing protocol.

            BOOL packetWasRouted = FALSE;
            RouterFunctionType routerFunction =
                NetworkIpGetRouterFunction(node, interfaceIndex);

            if (routerFunction)
            {
                (routerFunction)(node, msg, ipHeader->ip_dst,
                                 previousHopAddress, &packetWasRouted);
            }
            ERROR_Assert(!packetWasRouted,
                         "Router function routed packet it should only "
                         "inspect");
            // ANTS01 begin: (to deal with duplicate packets at the destination)
            if (msg->packetCreationTime > getSimTime(node))
            {
              MESSAGE_Free(node,msg);
              return;
            }
            // ANTS01 end
        }
    }
#endif // ENTERPRISE_LIB

    if (ipHeader->ip_tos & IPTOS_CE) {
          aCongestionExperienced = TRUE;
    }

#ifdef ADDON_MAODV
    if (FindAnIpOptionField(ipHeader, IPOPT_MAODV) != NULL)
    {
        //Trace drop
        ActionData acnData;
        acnData.actionType = DROP;
        acnData.actionComment = DROP_MAODV_OPTION_FIELD;
        TRACE_PrintTrace(node,
                         msg,
                         TRACE_NETWORK_LAYER,
                         PACKET_IN,
                         &acnData);

        MESSAGE_Free (node, msg);
        return;
    }
#endif // ADDON_MAODV

#ifdef IPNE_INTERFACE
    // If this is an external node then send the packet to the
    // operational network if we are doing true emulation
    if (node->isIpneNode)
    {
        if (IPNE_ForwardFromNetworkLayer(node, interfaceIndex, msg))
        {
            return;
        }
    }
#endif

    NetworkIpRemoveIpHeader(
                            node,
                            msg,
                            &sourceAddress,
                            &destinationAddress,
                            &priority,
                            &ipProtocolNumber,
                            &ttl);

    // Increment stat for number of IP datagrams delivered to
    // appropriate protocol module.

    ip->stats.ipInDelivers++;

    // Increment stat for total value of TTLs for packets delivered to
    // this node.  This is used to calculate an "ipInDelivers TTL-based
    // average hop count" metric.

    ip->stats.deliveredPacketTtlTotal += ttl;

    switch (ipProtocolNumber)
    {
        // Delivery to transport layer protocols.

        case IPPROTO_UDP:
        {
            SendToUdp(node, msg, priority, sourceAddress, destinationAddress,
                      interfaceIndex);
            break;
        }
        case IPPROTO_TCP:
        {
            SendToTcp(node,
                      msg,
                      priority,
                      sourceAddress,
                      destinationAddress,
                      aCongestionExperienced);
            break;
        }
//InsertPatch TRANSPORT_SEND_TO_TRANS

        // Delivery to generic network layer protocols.

        case IPPROTO_IGMP:
        {
            IgmpHandleProtocolPacket(
               node, msg, sourceAddress, destinationAddress, interfaceIndex);
            break;
        }

        case IPPROTO_ICMP:
        {
            if (ip->isIcmpEnable)
            {
                NetworkIcmpHandleProtocolPacket(node,
                                                msg,
                                                sourceAddress,
                                                destinationAddress,
                                                interfaceIndex);
            }
            else
            {
                //Trace drop
                ActionData acnData;
                acnData.actionType = DROP;
                acnData.actionComment = DROP_ICMP_NOT_ENABLE;
                TRACE_PrintTrace(node,
                                 msg,
                                 TRACE_NETWORK_LAYER,
                                 PACKET_IN,
                                 &acnData);
                MESSAGE_Free(node, msg);
            }

            break;
        }

        case IPPROTO_IPV6:
        {
             // When a packet is received with this type of protocol,
             // it means an IPv4-tunneled IPv6 pkt is encapsulated within
             // the IPv4 pkt.

            if (IsDualIpEnabled(node))
            {
                DualIpHandleTunneledIPv6Pkt(
                    node,
                    msg,
                    interfaceIndex,
                    previousHopAddress);
            }
            else
            {
                // Since the node is not Dual-IP enabled, drop the packet.
                ip->stats.ipInHdrErrors++;
                MESSAGE_Free(node, msg);
            }

            break;
        }

        // Delivery to network-layer routing protocols.

#ifdef WIRELESS_LIB
        case IPPROTO_AODV:
        {
            if (NetworkIpGetUnicastRoutingProtocolType(node, interfaceIndex) ==
                ROUTING_PROTOCOL_AODV
                || DualIpGetInterfaceTunnelType(node, interfaceIndex) == IPv6_TUNNEL)
            {
                Address srcAddress;
                Address destAddress;

                SetIPv4AddressInfo(&srcAddress, sourceAddress);

                SetIPv4AddressInfo(&destAddress, destinationAddress);

                AodvHandleProtocolPacket(
                    node,
                    msg,
                    srcAddress,
                    destAddress,
                    ttl,
                    interfaceIndex);
            }
            else
            {
                //Trace drop
                ActionData acnData;
                acnData.actionType = DROP;
                acnData.actionComment =
                                    DROP_PROTOCOL_UNAVAILABLE_AT_INTERFACE;
                TRACE_PrintTrace(node,
                                 msg,
                                 TRACE_NETWORK_LAYER,
                                 PACKET_IN,
                                 &acnData);

                MESSAGE_Free(node, msg);
            }

            break;
        }
        case IPPROTO_DYMO:
        {
            if (NetworkIpGetUnicastRoutingProtocolType(node, interfaceIndex) ==
                ROUTING_PROTOCOL_DYMO
                || DualIpGetInterfaceTunnelType(node, interfaceIndex) == IPv6_TUNNEL)
            {
                Address srcAddress;
                Address destAddress;

                SetIPv4AddressInfo(&srcAddress, sourceAddress);
                SetIPv4AddressInfo(&destAddress, destinationAddress);

                DymoHandleProtocolPacket(
                    node,
                    msg,
                    srcAddress,
                    destAddress,
                    interfaceIndex);
            }
            else
            {
                //Trace drop
                ActionData acnData;
                acnData.actionType = DROP;
                acnData.actionComment =
                    DROP_PROTOCOL_UNAVAILABLE_AT_INTERFACE;
                TRACE_PrintTrace(node,
                                 msg,
                                 TRACE_NETWORK_LAYER,
                                 PACKET_IN,
                                 &acnData);
            }
            MESSAGE_Free(node, msg);
            break;
        }

        case IPPROTO_DSR:
        {
            if (NetworkIpGetUnicastRoutingProtocolType(node, interfaceIndex) ==
                ROUTING_PROTOCOL_DSR)
            {
                DsrHandleProtocolPacket(node, msg, sourceAddress,
                    destinationAddress, ttl, previousHopAddress);
            }
            else
            {
                //Trace drop
                ActionData acnData;
                acnData.actionType = DROP;
                acnData.actionComment =
                                    DROP_PROTOCOL_UNAVAILABLE_AT_INTERFACE;
                TRACE_PrintTrace(node,
                                 msg,
                                 TRACE_NETWORK_LAYER,
                                 PACKET_IN,
                                 &acnData);

                MESSAGE_Free(node, msg);
            }

            break;
        }
        case IPPROTO_FSRL:
        {
            if (NetworkIpGetUnicastRoutingProtocolType(node, interfaceIndex) ==
                ROUTING_PROTOCOL_FSRL)
            {
                FsrlHandleProtocolPacket(node, msg, sourceAddress,
                    destinationAddress, interfaceIndex, ttl);
            }
            else
            {
                //Trace drop
                ActionData acnData;
                acnData.actionType = DROP;
                acnData.actionComment =
                                    DROP_PROTOCOL_UNAVAILABLE_AT_INTERFACE;
                TRACE_PrintTrace(node,
                                 msg,
                                 TRACE_NETWORK_LAYER,
                                 PACKET_IN,
                                 &acnData);

                MESSAGE_Free(node, msg);
            }

            break;
        }
        case IPPROTO_LAR1:
        {
            if (NetworkIpGetUnicastRoutingProtocolType(node, interfaceIndex) ==
                ROUTING_PROTOCOL_LAR1)
            {
                Lar1HandleProtocolPacket(node, msg);
            }
            else
            {
                //Trace drop
                ActionData acnData;
                acnData.actionType = DROP;
                acnData.actionComment =
                                    DROP_PROTOCOL_UNAVAILABLE_AT_INTERFACE;
                TRACE_PrintTrace(node,
                                 msg,
                                 TRACE_NETWORK_LAYER,
                                 PACKET_IN,
                                 &acnData);

                MESSAGE_Free(node, msg);
            }

            break;
        }
        case IPPROTO_STAR:
        {
            if (NetworkIpGetUnicastRoutingProtocolType(node, interfaceIndex) ==
                ROUTING_PROTOCOL_STAR)
            {
                StarHandleProtocolPacket(node, msg, sourceAddress);
            }
            else
            {
                //Trace drop
                ActionData acnData;
                acnData.actionType = DROP;
                acnData.actionComment =
                                    DROP_PROTOCOL_UNAVAILABLE_AT_INTERFACE;
                TRACE_PrintTrace(node,
                                 msg,
                                 TRACE_NETWORK_LAYER,
                                 PACKET_IN,
                                 &acnData);

                MESSAGE_Free(node, msg);
            }

            break;
        }
        case IPPROTO_BRP:
        {
            BrpDeliver(node,
                       msg,
                       sourceAddress,
                       destinationAddress,
                       interfaceIndex,
                       ttl);
            break;
        }
        case IPPROTO_IARP:
        {
            IarpHandleProtocolPacket(node, msg, sourceAddress,
                                     destinationAddress, interfaceIndex,
                                     ttl);
            break;
        }
        case IPPROTO_ZRP:
        {
            ZrpHandleProtocolPacket(node, msg, sourceAddress,
                                    destinationAddress, interfaceIndex,
                                    ttl);
            break;
        }
        case IPPROTO_IERP:
        {
            IerpHandleProtocolPacket(node, msg, sourceAddress,
                                     destinationAddress, interfaceIndex,
                                     ttl);
            break;
        }
        case IPPROTO_ODMRP:
        {
            OdmrpHandleProtocolPacket(node, msg, ipHeader->ip_src,
                ipHeader->ip_dst);
            break;
        }
        case IPPROTO_NDP:
        {
            NdpHandleProtocolPacket(
                node,
                msg,
                sourceAddress,
                interfaceIndex);
            break;
        }
//StartANTHOCNET
        case IPPROTO_ANTHOCNET:
        {
            if (NetworkIpGetUnicastRoutingProtocolType(node, interfaceIndex) ==
                ROUTING_PROTOCOL_ANTHOCNET
                || DualIpGetInterfaceTunnelType(node, interfaceIndex) == IPv6_TUNNEL)
            {
                AnthocnetHandleProtocolPacket(node, msg, sourceAddress, 
                  destinationAddress, interfaceIndex, ttl);
            }
            else
            {
                MESSAGE_Free(node, msg);
            }
            break;
        }
//EndANTHOCNET
#endif // WIRELESS_LIB
#ifdef ENTERPRISE_LIB
        case IPPROTO_RSVP:
        {
            SendToRsvp(node, msg, priority, sourceAddress,
                       destinationAddress, interfaceIndex, ttl);
            break;
        }
        case IPPROTO_IPIP:
        {
            /*
             * This is for IP in IP encapsulation. When a packet is received
             * with this type of protocol, it means the original IP packet is
             * tunneled. Decapsulate it
             */
             MobileIpDecapsulateDatagram(node, msg);
             break;
        }
        case IPPROTO_MOBILE_IP:
        {
            MobileIpHandleProtocolPacket(node,
                                         msg,
                                         sourceAddress,
                                         destinationAddress,
                                         interfaceIndex);
            break;
        }
        case IPPROTO_OSPF:
        {
            if (NetworkIpGetUnicastRoutingProtocolType(node, interfaceIndex) ==
                ROUTING_PROTOCOL_OSPFv2)
            {
                Ospfv2HandleRoutingProtocolPacket(node,
                                                  msg,
                                                  sourceAddress,
                                                  interfaceIndex);
            }
            else
            {
                //Trace drop
                ActionData acnData;
                acnData.actionType = DROP;
                acnData.actionComment =
                                    DROP_PROTOCOL_UNAVAILABLE_AT_INTERFACE;
                TRACE_PrintTrace(node,
                                 msg,
                                 TRACE_NETWORK_LAYER,
                                 PACKET_IN,
                                 &acnData);

                MESSAGE_Free(node, msg);
            }

            break;
        }
        case IPPROTO_PIM:
        {
            RoutingPimHandleProtocolPacket(node, msg, sourceAddress);
            break;
        }
        case IPPROTO_IGRP:
        {
            if (NetworkIpGetUnicastRoutingProtocolType(node, interfaceIndex) ==
                ROUTING_PROTOCOL_IGRP)
            {
                 IgrpHandleProtocolPacket(
                     node,
                     msg,
                     sourceAddress,
                     interfaceIndex);
            }
            else
            {
                //Trace drop
                ActionData acnData;
                acnData.actionType = DROP;
                acnData.actionComment =
                                    DROP_PROTOCOL_UNAVAILABLE_AT_INTERFACE;
                TRACE_PrintTrace(node,
                                 msg,
                                 TRACE_NETWORK_LAYER,
                                 PACKET_IN,
                                 &acnData);

                MESSAGE_Free(node, msg);
            }

            break;
        }
        case IPPROTO_EIGRP:
        {
           EigrpHandleProtocolPacket(
                node,
                msg,
                sourceAddress,
                interfaceIndex,
                previousHopAddress);

            break;
        }
#endif // ENTERPRISE_LIB

#ifdef MILITARY_RADIOS_LIB
        case IPPROTO_ODR:
        {
            //if (NetworkIpGetUnicastRoutingProtocolType(node, interfaceIndex) ==
            //    ROUTING_PROTOCOL_ODR
            //    || DualIpGetInterfaceTunnelType(node, interfaceIndex) == IPv6_TUNNEL)
            if (NetworkIpGetUnicastRoutingProtocolType(node, interfaceIndex) ==
                ROUTING_PROTOCOL_ODR)
            {
                OdrHandleProtocolPacket(node, msg, sourceAddress,
                                            destinationAddress, ttl,
                                            interfaceIndex);
            }

            else
            {
                //Trace drop
                ActionData acnData;
                acnData.actionType = DROP;
                acnData.actionComment =
                                    DROP_PROTOCOL_UNAVAILABLE_AT_INTERFACE;
                TRACE_PrintTrace(node,
                                 msg,
                                 TRACE_NETWORK_LAYER,
                                 PACKET_IN,
                                 &acnData);

                MESSAGE_Free(node, msg);
            }

            break;
        }
        case IPPROTO_SDR:
        {
            if (SdrActiveOnInterface(node, interfaceIndex) == TRUE)
            {
                SdrHandleProtocolPacket(node,
                                        msg,
                                        sourceAddress,
                                        destinationAddress,
                                        interfaceIndex,
                                        ttl);
            }
            else{
                //Trace drop
                ActionData acnData;
                acnData.actionType = DROP;
                acnData.actionComment =
                                    DROP_PROTOCOL_UNAVAILABLE_AT_INTERFACE;
                TRACE_PrintTrace(node,
                                 msg,
                                 TRACE_NETWORK_LAYER,
                                 PACKET_IN,
                                 &acnData);
                MESSAGE_Free(node, msg);
            }

            break;
        }
#endif // MILITARY_RADIOS_LIB
#ifdef ADDON_MAODV
        case IPPROTO_MAODV:
        {
            MaodvHandleProtocolPacket(
                node,
                msg,
                sourceAddress,
                destinationAddress,
                ttl,
                interfaceIndex);

            break;
        }
#endif // ADDON_MAODV

#ifdef CELLULAR_LIB
        case IPPROTO_GSM:
        {
             GsmLayer3ReceivePacketOverIp(node, msg, sourceAddress);
             break;
        }
        case IPPROTO_CELLULAR:
        {
             CellularLayer3ReceivePacketOverIp(node, msg, sourceAddress);
             break;
        }
#endif // CELLULAR_LIB

#ifdef ADVANCED_WIRELESS_LIB
        case IPPROTO_DOT16:
        {
            Dot16BackboneReceivePacketOverIp(node, msg, sourceAddress);
             break;
        }
#endif
        //InsertPatch NETWORK_HANDLE_PACKET
        default:
            ERROR_ReportError("Invalid switch value");

    }//switch//
}//DeliverPacket//


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpUseBackplaneIfPossible()
// PURPOSE      Determine if there's any packet that needs to go through the
//              the backplane and if so, pass the packet through the backplane.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int incomingInterface
//                  Index of interface which packet comes from.
//-----------------------------------------------------------------------------

void
NetworkIpUseBackplaneIfPossible(Node *node,
                                int incomingInterface)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    int outgoingInterface;
    int networkType = 0;
    NodeAddress hopAddr = 0;
    TosType priority = 0;
    Message *newMsg = NULL;
    int packetSize;
    BOOL isEmpty = TRUE;
    NetworkIpBackplaneStatusType *backplaneStatus;

    if (incomingInterface == CPU_INTERFACE ||
        ip->backplaneType == BACKPLANE_TYPE_CENTRAL)
    {
        backplaneStatus = &ip->backplaneStatus;
    }
    else
    {
        backplaneStatus =
                     &ip->interfaceInfo[incomingInterface]->backplaneStatus;
    }

    // If the interface is busy sending on the backplane, then wait...
    if (*backplaneStatus != NETWORK_IP_BACKPLANE_IDLE)
    {
       return;
    }

    // If we are here, then the interface is free to send on
    // the backplane...

    // Get the message information according to the interface queue...
    if (incomingInterface == CPU_INTERFACE ||
        ip->backplaneType == BACKPLANE_TYPE_CENTRAL)
    {
        if (!NetworkIpCpuQueueIsEmpty(node))
        {
            NetworkIpCpuQueueTopPacket(node,
                                 &newMsg,
                                 &hopAddr,
                                 &outgoingInterface,
                                 &networkType,
                                 &priority);

            isEmpty = FALSE;
        }
    }
    else
    {
        if (!NetworkIpInputQueueIsEmpty(node, incomingInterface))
        {
            NetworkIpInputQueueTopPacket(node,
                                 incomingInterface,
                                 &newMsg,
                                 &hopAddr,
                                 &outgoingInterface,
                                 &networkType,
                                 &priority);

            isEmpty = FALSE;
        }
    }

    // Only send on backplane if queue is not empty.
    if (!isEmpty)
    {
        Message *backplaneMsg;
        clocktype backplaneDelay;
        NetworkIpBackplaneInfo *backplaneInfo;

        packetSize = MESSAGE_ReturnPacketSize(newMsg);

        if (incomingInterface == CPU_INTERFACE
            || ip->backplaneType == BACKPLANE_TYPE_CENTRAL)
        {
            backplaneDelay = (clocktype) (packetSize * 8 * SECOND /
                                      ip->backplaneThroughputCapacity);
        }
        else
        {
            backplaneDelay = (clocktype) (packetSize * 8 * SECOND /
                ip->interfaceInfo[incomingInterface]->disBackplaneCapacity);
        }

        if (networkType == NETWORK_PROTOCOL_IPV6)
        {
            backplaneMsg = MESSAGE_Alloc(node,
                                         NETWORK_LAYER,
                                         NETWORK_PROTOCOL_IPV6,
                                         MSG_NETWORK_Backplane);
        }
        else
        {
            backplaneMsg = MESSAGE_Alloc(node,
                                     NETWORK_LAYER,
                                     NETWORK_PROTOCOL_IP,
                                     MSG_NETWORK_Backplane);
        }
        MESSAGE_InfoAlloc(node,
                          backplaneMsg,
                          sizeof(NetworkIpBackplaneInfo));

        backplaneInfo = (NetworkIpBackplaneInfo *)
                         MESSAGE_ReturnInfo(backplaneMsg);

        backplaneInfo->incomingInterface = incomingInterface;
        backplaneInfo->hopAddr = hopAddr;

        *backplaneStatus = NETWORK_IP_BACKPLANE_BUSY;

        MESSAGE_Send(node, backplaneMsg, backplaneDelay);
    }
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpReceiveFromBackplane()
// PURPOSE      Process packets that have passed through the backplane.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg,
//                  packet that just passed through the backplane.
//-----------------------------------------------------------------------------

void
NetworkIpReceiveFromBackplane(Node *node, Message *msg)
{
    Message *queueMsg;
    int outgoingInterface;
    int networkType;
    NodeAddress hopAddr;
    TosType priority;
    NetworkIpBackplaneStatusType *backplaneStatus;

    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    NetworkIpBackplaneInfo *info = (NetworkIpBackplaneInfo *)
                                   MESSAGE_ReturnInfo(msg);

    if (info->incomingInterface == CPU_INTERFACE
        || ip->backplaneType == BACKPLANE_TYPE_CENTRAL)
    {
        backplaneStatus = &ip->backplaneStatus;

        NetworkIpCpuQueueDequeuePacket(node,
                                 &queueMsg,
                                 &hopAddr,
                                 &outgoingInterface,
                                 &networkType,
                                 &priority);
    }
    else
    {
        backplaneStatus =
                 &ip->interfaceInfo[info->incomingInterface]->backplaneStatus;

        NetworkIpInputQueueDequeuePacket(node,
                                 info->incomingInterface,
                                 &queueMsg,
                                 &hopAddr,
                                 &outgoingInterface,
                                 &networkType,
                                 &priority);
    }

    assert(*backplaneStatus == NETWORK_IP_BACKPLANE_BUSY);

    *backplaneStatus = NETWORK_IP_BACKPLANE_IDLE;

    if (outgoingInterface == CPU_INTERFACE)
    {
        QueuedPacketInfo* packetInfo =
            (QueuedPacketInfo*) MESSAGE_ReturnInfo(queueMsg);

        if (networkType == NETWORK_PROTOCOL_IPV6)
        {
            ip6_hdr* ipv6Header = (ip6_hdr *)MESSAGE_ReturnPacket(queueMsg);

            if (ip->backplaneType == BACKPLANE_TYPE_CENTRAL)
            {
                ip6_deliver(node, queueMsg, 0,
                    packetInfo->incomingInterface, ipv6Header->ip6_nxt);
            }
            else
            {
                ip6_deliver(node, queueMsg, 0,
                    info->incomingInterface, ipv6Header->ip6_nxt);
            }
        }
        else // Ipv4
        {
            if (ip->backplaneType == BACKPLANE_TYPE_CENTRAL)
            {
                DeliverPacket(node, queueMsg,
                    packetInfo->incomingInterface, info->hopAddr);
            }
            else
            {
                DeliverPacket(node, queueMsg,
                    info->incomingInterface, info->hopAddr);
            }
        }
    }
    else
    {
        if (networkType == NETWORK_PROTOCOL_IPV6)
        {
            QueueUpIpv6FragmentForMacLayer(node,
                                     queueMsg,
                                     outgoingInterface,
                                     info->hopAddr);
        }
        else
        {
            QueueUpIpFragmentForMacLayer(node,
                                         queueMsg,
                                         outgoingInterface,
                                         info->hopAddr);
        }
    }

    NetworkIpUseBackplaneIfPossible(node,
                                    info->incomingInterface);
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpSendOnBackplane()
// PURPOSE      Simulates packets going to the router backplane before
//              they are further processed.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message with IP packket.
//              int incomingInterface
//                  Index of interface from which packet was received.
//              int outgoingInterface
//                  Index of interface packet is to be sent to.
//              NodeAddress hopAddr
//                  address of the next hop node (for forwarding the packet)
//                  or the last hop node (for delivering the packet to upper
//                  layers.
//-----------------------------------------------------------------------------
void //inline//
NetworkIpSendOnBackplane(
     Node *node,
     Message *msg,
     int incomingInterface,
     int outgoingInterface,
     NodeAddress hopAddr)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    IpHeaderType *ipHeader = (IpHeaderType *) MESSAGE_ReturnPacket(msg);

    // If we assume unlimited backplane throughput,
    // then we skip the backplane overhead altogther.

    if (ip->backplaneThroughputCapacity ==
        NETWORK_IP_UNLIMITED_BACKPLANE_THROUGHPUT)
    {
        if (outgoingInterface == CPU_INTERFACE)
        {
             DeliverPacket(node, msg, incomingInterface, hopAddr);
        }
        else
        {
            QueueUpIpFragmentForMacLayer(node,
                                         msg,
                                         outgoingInterface,
                                         hopAddr);
        }
    }
    else
    {
        BOOL queueIsFull = FALSE;

        if (incomingInterface == CPU_INTERFACE ||
            ip->backplaneType == BACKPLANE_TYPE_CENTRAL)
        {
            NetworkIpCpuQueueInsert(node,
                               msg,
                               hopAddr,
                               ipHeader->ip_dst,
                               outgoingInterface,
                               NETWORK_PROTOCOL_IP,
                               &queueIsFull,
                               incomingInterface);
        }
        else
        {
            NetworkIpInputQueueInsert(node,
                               incomingInterface,
                               msg,
                               hopAddr,
                               ipHeader->ip_dst,
                               outgoingInterface,
                               NETWORK_PROTOCOL_IP,
                               &queueIsFull);
        }

        // If queue is full, then just drop the packet.  No need to go
        // through the backplane.
        if (queueIsFull)
        {
           // Keep stats on how many packets are dropped due to
           // over backplane throughput limit.
           ip->stats.ipNumDroppedDueToBackplaneLimit++;

            //Trace drop
            ActionData acnData;
            acnData.actionType = DROP;
            acnData.actionComment = DROP_QUEUE_OVERFLOW;
            if (outgoingInterface == CPU_INTERFACE)
            {
                TRACE_PrintTrace(node,
                                 msg,
                                 TRACE_NETWORK_LAYER,
                                 PACKET_IN,
                                 &acnData);
            }
            else
            {
                TRACE_PrintTrace(node,
                                 msg,
                                 TRACE_NETWORK_LAYER,
                                 PACKET_OUT,
                                 &acnData);
            }


           // Backplane capacity has been reached, so drop packet.
           MESSAGE_Free(node, msg);
        }
        else
        {
            NetworkIpUseBackplaneIfPossible(node,
                                            incomingInterface);
        }
    }
}


//-----------------------------------------------------------------------------
// FUNCTION     ForwardPacket()
// PURPOSE      Forward IP packet.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message with IP packket.
//              int incomingInterface
//                  Index of interface from which packet was received.
//              NodeAddress previousHopAddress
//                  Previous hop of packet.
//
// NOTES        This routine is only called for packets which arrive
//              from the MAC layer.  Packets from a network-layer or
//              transport-layer protocol are not "forwarded", per se.
//-----------------------------------------------------------------------------

static void //inline//
ForwardPacket(
    Node *node,
    Message *msg,
    int incomingInterface,
    NodeAddress previousHopAddress)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    IpHeaderType *ipHeader;
    ipHeader = (IpHeaderType *) MESSAGE_ReturnPacket(msg);

#ifdef ADDON_MAODV
    BOOL PacketWasRouted = FALSE;
    MulticastRouterFunctionType routerFunction =
               NetworkIpGetMulticastRouterFunction(node, incomingInterface);

    // if multicast router function is defined let it have a look in
    // the packet

    if ( routerFunction != NULL
        && NetworkIpIsMulticastAddress(node, ipHeader->ip_dst)
        && !NetworkIpIsMyIP(node, ipHeader->ip_src)
        && ( ip->interfaceInfo[incomingInterface]->multicastProtocolType
        == MULTICAST_PROTOCOL_MAODV))
    {
         (routerFunction)(node,
                         msg,
                         ipHeader->ip_dst,
                         incomingInterface,
                         &PacketWasRouted,
                         previousHopAddress);

        if (PacketWasRouted)
        {
            return;
        }
    }

#endif // ADDON_MAODV

    if (ip->ipForwardingEnabled == FALSE)
    {
        //Trace drop
        ActionData acnData;
        acnData.actionType = DROP;
        acnData.actionComment = DROP_IPFORWARD_NOT_ENABLE;
        TRACE_PrintTrace(node,
                         msg,
                         TRACE_NETWORK_LAYER,
                         PACKET_OUT,
                         &acnData);

        MESSAGE_Free(node, msg);
        return;
    }

    if (ipHeader->ip_ttl == 0)
    {
        // Increment stat for number of IP datagrams discarded because
        // of header errors (e.g., checksum error, version number
        // mismatch, TTL exceeded, etc.).

        ip->stats.ipInHdrErrors++;
        //Trace drop
        ActionData acnData;
        acnData.actionType = DROP;
        acnData.actionComment = DROP_TTL_ZERO;
        TRACE_PrintTrace(node,
                         msg,
                         TRACE_NETWORK_LAYER,
                         PACKET_OUT,
                         &acnData);

                // send ICMP packet with time exceeded type
                NetworkIcmpCreateTimeExceedMessage(node,
                                        msg,
                                        ipHeader->ip_src,
                                        incomingInterface);

        // Free message.

        MESSAGE_Free(node, msg);
    }
    else
    {
        if (NetworkIpIsMulticastAddress(node, ipHeader->ip_dst))
        {
            /* Self originating packet should not be forwarded */
            if (ipHeader->ip_src ==
                    NetworkIpGetInterfaceAddress(node, incomingInterface))
            {
                //Trace drop
                ActionData acnData;
                acnData.actionType = DROP;
                acnData.actionComment = DROP_MULTICAST_ADDR_SELF_PACKET;
                TRACE_PrintTrace(node,
                                 msg,
                                 TRACE_NETWORK_LAYER,
                                 PACKET_OUT,
                                 &acnData);

                /* Free message */
                MESSAGE_Free(node, msg);
            }
            else
            {
                ((NetworkDataIp *) node->networkData.networkVar)
                        ->stats.ipInForwardDatagrams++;

                #ifdef STK_INTERFACE
                    StkUpdateIpStats(
                       node,
                       STK_STAT_IP_FORWARDED,
                       ip->stats.ipInForwardDatagrams,
                       getSimTime(node));
                #endif /* STK_INTERFACE */

                RoutePacketAndSendToMac(node,
                                        msg,
                                        incomingInterface,
                                        ANY_INTERFACE,
                                        previousHopAddress);
            }
        }
        else
        {
            ((NetworkDataIp *) node->networkData.networkVar)
                    ->stats.ipInForwardDatagrams++;

            #ifdef STK_INTERFACE
                StkUpdateIpStats(
                   node,
                   STK_STAT_IP_FORWARDED,
                   ip->stats.ipInForwardDatagrams,
                   getSimTime(node));
            #endif /* STK_INTERFACE */

             RoutePacketAndSendToMac(node,
                                     msg,
                                     incomingInterface,
                                     ANY_INTERFACE,
                                     previousHopAddress);
        }
    }
}//ForwardPacket//

//-----------------------------------------------------------------------------
// FUNCTION     RoutePacketAndSendToMac()
// PURPOSE      "Route" a packet by determining the next hop IP address
//              and the outgoing interface.  It is assumed that the IP
//              packet already has an IP header at this point.  Figure
//              out the next hop in the route and send the packet.
//              First the "routing function" is checked and if that fails
//              the default source route or lookup table route is used.
//              [needs updating]
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message with IP packet.
//              int incomingInterface
//                  Index of interface on which packet arrived.
//                  [This is different if the packet originated from
//                  the network or transport layers.  This value is only
//                  for multicast packets, currently.]
//              int outgoingInterface
//                  Used only when the application specifies a specific
//                  interface to use to transmit packet.
//-----------------------------------------------------------------------------

void //inline//
RoutePacketAndSendToMac(Node *node,
                        Message *msg,
                        int incomingInterface,
                        int outgoingInterface,
                        NodeAddress previousHopAddress)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    IpHeaderType *ipHeader = (IpHeaderType *) msg->packet;
    int outgoingInterfaceToUse;
    int interfaceIndex;
    NodeAddress outgoingBroadcastAddress;

    assert(incomingInterface != ANY_INTERFACE);

#ifdef IPNE_INTERFACE
    // If this is an external node then send the packet to the
    // operational network if we are doing true emulation.  Make sure the
    // packet was not received on the CPU_INTERFACE (ie, this node is
    // originating the packet)
    if (node->isIpneNode && incomingInterface != CPU_INTERFACE)
    {
        if (IPNE_ForwardFromNetworkLayer(node, incomingInterface, msg))
        {
            return;
        }
    }
#endif

    // Used to determine what routing protocol to use on a interface.
    if (incomingInterface == CPU_INTERFACE)
    {
        // If sent by this node, then routing protocol should be associated
        // with the outgoing interface.
        interfaceIndex = outgoingInterface;
    }
    else
    {
        // If packet is being forwarded, then routing protocol should be
        // associated with the incoming interface.
        interfaceIndex = incomingInterface;
    }


    // Fragmentation code here.
    // IP Fragmentation code.
    if (MESSAGE_ReturnPacketSize(msg) > GetNetworkIPFragUnit(node))
    {
        ipFragmetedMsg* fragHead = NULL;
        ipFragmetedMsg* tempFH = NULL;

        if (IpHeaderHasSourceRoute(ipHeader))
        {
            ERROR_ReportError("Source routing packet is not allowed for "
                "Ip fragmentation\n");
        }

        if (IpFragmentPacket(node,
                msg,
                GetNetworkIPFragUnit(node),
                &fragHead))
        {
            while (fragHead)
            {
//-------------------------------------------------------------------------//
// Sending fragmented packets
//-------------------------------------------------------------------------//

                    // If broadcast is of type ANY_DEST, then outgoing interface
                    // should have already been determined before this.

                    if (ipHeader->ip_dst == ANY_DEST)
                    {
                        NetworkIpSendPacketOnInterface(
                                          node,
                                          fragHead->msg,
                                          incomingInterface,
                                          outgoingInterface,
                                          ipHeader->ip_dst);
                    }
                    // Check if it's a broadcast.
                    else if (IsOutgoingBroadcast(node,
                             ipHeader->ip_dst,
                             &outgoingInterfaceToUse,
                             &outgoingBroadcastAddress))
                    {
                        NetworkIpSendPacketOnInterface(node,
                                                      fragHead->msg,
                                                      incomingInterface,
                                                      outgoingInterfaceToUse,
                                                      outgoingBroadcastAddress);

                    }
                    // Use the multicast forwarding table if we need to
                    // send a multicast packet.
                    else if (NetworkIpIsMulticastAddress(node, ipHeader->ip_dst))
                    {
                        BOOL packetWasRouted = FALSE;

                        MulticastRouterFunctionType routerFunction =
                                NetworkIpGetMulticastRouterFunction(node, interfaceIndex);

                        if (routerFunction != NULL)
                        {
                            (routerFunction)(node,
                                             fragHead->msg,
                                             ipHeader->ip_dst,
                                             interfaceIndex,
                                             &packetWasRouted,
                                             previousHopAddress);
                        }

                        if (!packetWasRouted)
                        {
                            RouteThePacketUsingMulticastForwardingTable(
                                node,
                                                                        fragHead->msg,
                                incomingInterface,
                                NETWORK_IPV4);
                        }
                    }
                    else
                    {
                        // First check if MPLS is enabled, and if so, transfer the packet
                        // to MPLS.  Next, try to route with the routing protocol supplied
                        // routing function.  If the function doesn't exists or
                        // fails to find a route, then try standard lookup table
                        // or source routing.

                        BOOL packetWasRouted = FALSE;

#ifdef ENTERPRISE_LIB
                        /*
                         * Check whether the packet comes from Qos application. If so
                         * then pass the packet to the Q-OSPF where the path will be
                         * calculated (or if the path is calculated previously by other
                         * node) and forwarded by Q-OSPF.
                         */
                        if (NetworkIpGetUnicastRoutingProtocolType(node, interfaceIndex)
                            == ROUTING_PROTOCOL_OSPFv2 &&
                            QospfIsPacketDemandedQos(node, fragHead->msg))
                        {

                            IpOptionsHeaderType *ipOptions =
                                IpHeaderSourceRouteOptionField((IpHeaderType *) fragHead->msg->packet);

                            if (ipOptions == NULL)
                            {
                                /*
                                 * I.e. address list not yet assigned in IP header's option
                                 * field. So call the path calculation algorithm, which will
                                 * search the path and if the path can be assigned, it will be
                                 * included into the IP header option field and packet will be
                                 * forwarded.
                                 */

                                QospfForwardQosApplicationPacket(node, fragHead->msg, &packetWasRouted);
                            }
                            else
                            {
                                /*
                                 * Path present in option field. So forward the packet by
                                 * taking the next hop from the IP header option field.
                                 */

                                ERROR_ReportError("Source routing packet is not allowed for "
                                    "Ip fragmentation\n");
                            }

                        }

                        MplsData *mpls = MplsReturnStateSpace(node);
                        if (mpls)
                        {
                            MplsRoutePacket(node,
                                            mpls,
                                            fragHead->msg,
                                            ipHeader->ip_dst,
                                            &packetWasRouted);
                        }
#endif // ENTERPRISE_LIB

                        if (!packetWasRouted)
                        {
                            // Router functions are present only for wireless routing
                            // protocols.  Wireless routing protocols currently support
                            // only single-interface nodes.
                            //
                            // For these reasons, just return the routerFunction of the
                            // first interface.  This was a change from the old code.
                            // The old code would not support routing across different
                            // wireless subnets; this change makes it possible (given
                            // that each node in the route path is single-interface and
                            // runs the routing protocol).

                            RouterFunctionType routerFunction = NULL;

                            routerFunction = NetworkIpGetRouterFunction(node,
                                                                        interfaceIndex);

                            if (routerFunction)
                            {
                                (routerFunction)(node,
                                                 fragHead->msg,
                                                 ipHeader->ip_dst,
                                                 previousHopAddress,
                                                 &packetWasRouted);
                            }

#ifdef MILITARY_RADIOS_LIB
                            if (!packetWasRouted)
                            {
                                 if (EplrsActiveOnInterface(node, interfaceIndex)){

                                    EplrsIncQueuePacketFromNetworkLayer(node,
                                            interfaceIndex,
                                            fragHead->msg,
                                            previousHopAddress,
                                            &packetWasRouted);
                                 }
                            }
#endif

                            if (!packetWasRouted)
                            {
                                RouteThePacketUsingLookupTable(node,
                                                               fragHead->msg,
                                                               incomingInterface);
                            }//if//
                        }//if//
                }//if//

//-------------------------------------------------------------------------//
// End of sending fragmented packets
//-------------------------------------------------------------------------//
                tempFH = fragHead;
                fragHead = fragHead->next;
                MEM_free(tempFH);

             }
                MESSAGE_Free(node, msg);
            }
            else
            {
                ERROR_ReportError("IP header with option header, is not allowed "
                    "for IP fragmentation in this version\n");
            }
            return;
        }// End fragmentation code.


//-------------------------------------------------------------------------//
// Non fragmented packet same as before.
//-------------------------------------------------------------------------//
#ifdef ENTERPRISE_LIB
    if (ip->mobileIpStruct)
    {
        MobileIpEncapsulateDatagram(node, msg);
    }

    // For PBR analysis
    // No particular check is available to ascertain whether its a data
    //  or a control packet(we can do a precedence check, but thats not
    // a foolproof test). So all the packets are policy routed.

    // 15 June, 2004, Added check to avoid (OLSR) control packet matching
    if (ipHeader->ip_tos != IPTOS_PREC_INTERNETCONTROL &&
        !NetworkIpIsLoopbackInterfaceAddress(ipHeader->ip_src) &&
        !NetworkIpIsLoopbackInterfaceAddress(ipHeader->ip_dst))
    {
        if (incomingInterface != CPU_INTERFACE)
        {
            if (!(RtParseIsPktFrmMe(node, ipHeader->ip_src)) &&
               (ip->interfaceInfo[incomingInterface]->rMapForPbr))
            {
               // If the packet is not from me and PBR is enabled
               if (PbrInvoke(node,
                  msg,
                  previousHopAddress,
                  incomingInterface,
                  ip->interfaceInfo[incomingInterface]->rMapForPbr,
                  FALSE,
                  FALSE))
               {
                   return;
               }
             }
         }
         else
         if ((RtParseIsPktFrmMe(node, ipHeader->ip_src)) && (ip->local) &&
             (ip->rMapForPbr))
         {
             // If the packet is from me and local PBR is enabled
             if (PbrInvoke(node, msg, 0, PBR_NO_INTERFACE, ip->rMapForPbr,
                 FALSE, TRUE))
             {
                 return;
             }
         }
    }
#endif // ENTERPRISE_LIB

    if (ip->isLoopbackEnabled && (incomingInterface == CPU_INTERFACE) &&
        NetworkIpLoopbackLoopbackUnicastsToSender(node, msg))
    {
        // do nothing. Ip datagram is already Looped back.
    }

    // If broadcast is of type ANY_DEST, then outgoing interface
    // should have already been determined before this.
    else if (ipHeader->ip_dst == ANY_DEST)
    {
        // Datagrams sent to a broadcast address are copied to
        // the loopback interface
        //if (ip->isLoopbackEnabled && (incomingInterface == CPU_INTERFACE))
        //{
        //    NetworkIpLoopbackBroadcastAndMulticastToSender(node, msg);
        //}

        NetworkIpSendPacketOnInterface(
                          node,
                          msg,
                          incomingInterface,
                          outgoingInterface,
                          ipHeader->ip_dst);
    }
    // Check if it's a broadcast.
    else if (IsOutgoingBroadcast(node,
             ipHeader->ip_dst,
             &outgoingInterfaceToUse,
             &outgoingBroadcastAddress))
    {
        // Datagrams sent to a broadcast address are copied to
        // the loopback interface
        //if (ip->isLoopbackEnabled && (incomingInterface == CPU_INTERFACE))
        //{
        //    NetworkIpLoopbackBroadcastAndMulticastToSender(node, msg);
        //}

        NetworkIpSendPacketOnInterface(node,
                                      msg,
                                      incomingInterface,
                                      outgoingInterfaceToUse,
                                      outgoingBroadcastAddress);

    }
    // Use the multicast forwarding table if we need to
    // send a multicast packet.
    else if (NetworkIpIsMulticastAddress(node, ipHeader->ip_dst))
    {
        BOOL packetWasRouted = FALSE;

        // Datagrams sent to a broadcast address are copied to
        // the loopback interface
        //if (ip->isLoopbackEnabled && (incomingInterface == CPU_INTERFACE))
        //{
        //    NetworkIpLoopbackBroadcastAndMulticastToSender(node, msg);
        //}

        MulticastRouterFunctionType routerFunction =
                NetworkIpGetMulticastRouterFunction(node, interfaceIndex);

        if (routerFunction != NULL)
        {
            (routerFunction)(node,
                             msg,
                             ipHeader->ip_dst,
                             interfaceIndex,
                             &packetWasRouted,
                             previousHopAddress);
        }

        if (!packetWasRouted)
        {
            RouteThePacketUsingMulticastForwardingTable(
                node,
                msg,
                incomingInterface,
                NETWORK_IPV4);
        }
    }
    else
    {
        // First check if MPLS is enabled, and if so, transfer the packet
        // to MPLS.  Next, try to route with the routing protocol supplied
        // routing function.  If the function doesn't exists or
        // fails to find a route, then try standard lookup table
        // or source routing.

        BOOL packetWasRouted = FALSE;

        /*
         * Check whether the packet comes from Qos application. If so
         * then pass the packet to the Q-OSPF where the path will be
         * calculated (or if the path is calculated previously by other
         * node) and forwarded by Q-OSPF.
         */
#ifdef ENTERPRISE_LIB
        if (NetworkIpGetUnicastRoutingProtocolType(node, interfaceIndex)
            == ROUTING_PROTOCOL_OSPFv2 &&
            QospfIsPacketDemandedQos(node, msg))
        {

            IpOptionsHeaderType *ipOptions =
                IpHeaderSourceRouteOptionField((IpHeaderType *) msg->packet);

            if (ipOptions == NULL)
            {
                /*
                 * I.e. address list not yet assigned in IP header's option
                 * field. So call the path calculation algorithm, which will
                 * search the path and if the path can be assigned, it will be
                 * included into the IP header option field and packet will be
                 * forwarded.
                 */

                QospfForwardQosApplicationPacket(node, msg, &packetWasRouted);
            }
            else
            {
                /*
                 * Path present in option field. So forward the packet by
                 * taking the next hop from the IP header option field.
                 */
                SourceRouteThePacket(node, msg);

                packetWasRouted = TRUE;
            }

        }

        MplsData *mpls = MplsReturnStateSpace(node);
        if (mpls)
        {
            MplsRoutePacket(node,
                            mpls,
                            msg,
                            ipHeader->ip_dst,
                            &packetWasRouted);
        }
#endif // ENTERPRISE_LIB

        if (!packetWasRouted)
        {
            // Router functions are present only for wireless routing
            // protocols.  Wireless routing protocols currently support
            // only single-interface nodes.
            //
            // For these reasons, just return the routerFunction of the
            // first interface.  This was a change from the old code.
            // The old code would not support routing across different
            // wireless subnets; this change makes it possible (given
            // that each node in the route path is single-interface and
            // runs the routing protocol).

            RouterFunctionType routerFunction = NULL;

            routerFunction = NetworkIpGetRouterFunction(node,
                                                        interfaceIndex);

            if (routerFunction)
            {
                (routerFunction)(node,
                                 msg,
                                 ipHeader->ip_dst,
                                 previousHopAddress,
                                 &packetWasRouted);
            }
#ifdef MILITARY_RADIOS_LIB
            if (!packetWasRouted)
            {
                if (EplrsActiveOnInterface(node, interfaceIndex)){

                    EplrsIncQueuePacketFromNetworkLayer(node,
                                            interfaceIndex,
                                            msg,
                                            previousHopAddress,
                                            &packetWasRouted);
                }
            }
#endif
            if (!packetWasRouted)
            {
                if (IpHeaderHasSourceRoute(ipHeader))
                {
                    SourceRouteThePacket(node, msg);
                }
                else
                {
                    RouteThePacketUsingLookupTable(node,
                                                   msg,
                                                   incomingInterface);
                }//if//
            }//if//
        }//if//
    }//if//
}//RoutePacketAndSendToMac//


//-----------------------------------------------------------------------------
// FUNCTION     RouteThePacketUsingLookupTable()
// PURPOSE      Tries to route and send the packet using the node's
//              forwarding table.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message with IP packet.
//              int incomingInterface
//                  incoming interface of packet.
//-----------------------------------------------------------------------------

void
RouteThePacketUsingLookupTable(Node *node, Message *msg, int incomingInterface)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    IpHeaderType *ipHeader = (IpHeaderType *) msg->packet;

    int outgoingInterface;
    NodeAddress nextHop;

    NetworkGetInterfaceAndNextHopFromForwardingTable(
        node, ipHeader->ip_dst, &outgoingInterface, &nextHop);

    if (nextHop == (unsigned) NETWORK_UNREACHABLE)
    {
        // Increment stat for number of IP packets discarded because no
        // route could be found.
        //
        // Individual routing modules like AODV, LAR1, DSR, and ZRP may
        // or may not increment this value.

        // For PBR analysis
        // No particular check is available to ascertain whether its a data
        //  or a control packet(we can do a precedence check, but thats not
        // a foolproof test). So all the packets are policy routed.

#ifdef ENTERPRISE_LIB
        // 15 June, 2004, Added check to avoid (OLSR) control packet matching
        if (ipHeader->ip_tos != IPTOS_PREC_INTERNETCONTROL)
        {
            BOOL pbrResult = FALSE;

            // If the packet is not from me.
            if (!RtParseIsPktFrmMe(node, ipHeader->ip_src))
            {
                if ((incomingInterface != CPU_INTERFACE) &&
                    (ip->interfaceInfo[incomingInterface]->rMapForPbr))
                {
                    RouteMap* rMap =
                        ip->interfaceInfo[incomingInterface]->rMapForPbr;

                    if (rMap->hasDefault)
                    {
                        pbrResult = PbrInvoke(node, msg, 0, incomingInterface,  rMap,
                            TRUE, FALSE);
                        if (pbrResult)
                        {
                            // the packet is already routed
                            return;
                        }
                    }
                }
            }
            else
            if (ip->local)
            {
                // Packet is from me and local PBR enabled
                RouteMap* rMap = ip->rMapForPbr;

                if (rMap->hasDefault)
                {
                    pbrResult = PbrInvoke(node, msg, 0, PBR_NO_INTERFACE, rMap,
                        TRUE, TRUE);
                    if (pbrResult)
                    {
                        // the packet is already routed
                        return;
                    }
                }
            }
        }
#endif // ENTERPRISE_LIB

#ifdef WIRELESS_LIB
// For dymo gateway
        if (ip->isManetGateway)
        {
            Address temp;

            SetIPv4AddressInfo(
                &temp,
                ipHeader->ip_dst);

            if (DymoIsPrefixMatch(&temp,
                                  &ip->manetPrefixAddr,
                                  ip->manetPrefixlength))
            {
                BOOL packetWasRouted = FALSE;

                Dymo4RouterFunction(
                    node,
                    msg,
                    ipHeader->ip_dst,
                    ANY_DEST,
                    &packetWasRouted);

                if (packetWasRouted) {
                    return;
                }
            }
        }// end of if
// end for dymo gateway
#endif // WIRELESS_LIB

        // added for gateway
        // Check if any default gateway is present for that node

        if (ip->gatewayConfigured)
        {
            // Route the packet thru that default gateway

            NetworkIpRoutePacketThroughGateway(node,
                msg,
                incomingInterface);

            return;
        }

        // end for gateway

        ip->stats.ipOutNoRoutes++;

        //Trace drop
        ActionData acnData;
        acnData.actionType = DROP;
        acnData.actionComment = DROP_NO_ROUTE;
        TRACE_PrintTrace(node,
                         msg,
                         TRACE_NETWORK_LAYER,
                         PACKET_OUT,
                         &acnData);


        // Free message.

        MESSAGE_Free(node, msg);

        return;
    }

    // Found route (outgoing interface and next hop address) in
    // forwarding table.  Queue packet on interface with next hop
    // address.

    // nextHop == 0 for switched ethernet routes
    if (nextHop == 0)
    {
        NetworkIpSendPacketOnInterface(
            node,
            msg,
            incomingInterface,
            outgoingInterface,
            ipHeader->ip_dst);
    }
    else
    {
        NetworkIpSendPacketOnInterface(node,
                                      msg,
                                      incomingInterface,
                                      outgoingInterface,
                                      nextHop);
    }
}

//-----------------------------------------------------------------------------
// FUNCTION     QueueUpIpFragmentForMacLayer()
// PURPOSE      Called by NetworkIpSendPacketOnInterface().  Checks if
//              the output queue(s) for the specified interface is empty
//              or full, and calls NetworkIpOutputQueueInsert() to queue
//              the IP packet.  Drops packet if the queue was full.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message with IP packet.
//              int interfaceIndex
//                  Index of outgoing interface.
//              NodeAddress nextHop
//                  Next hop address.
//
// NOTES        This is one of the places where
//              MAC_NetworkLayerHasPacketToSend() may be called,
//              which starts the MAC packet-sending process.  See the
//              comments for the necessary state.
//
//              MAC_NetworkLayerHasPacketToSend() is also called
//              in mpls.pc.
//-----------------------------------------------------------------------------

static void //inline//
QueueUpIpFragmentForMacLayer(
    Node *node,
    Message *msg,
    int interfaceIndex,
    NodeAddress nextHop)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    Scheduler *scheduler = ip->interfaceInfo[interfaceIndex]->scheduler;
    IpHeaderType* ipHeader = (IpHeaderType *) MESSAGE_ReturnPacket(msg);
    BOOL queueIsFull;
    BOOL queueWasEmpty;

#ifdef ENTERPRISE_LIB
    MplsData *mpls = (MplsData *) node->macData[interfaceIndex]->mplsVar;

    // If the Diffserv Multi-Field Traffic Conditioner is Enabled
    if (ip->isEdgeRouter == TRUE)
    {
        // Check whether or not the Diffserv Multi-Field Traffic Conditioner
        // will drop this packet
        DIFFSERV_TrafficConditionerProfilePacketAndMarkOrDrop(
            node,
            ip,
            msg,
            interfaceIndex,
            &queueIsFull);

        if (queueIsFull)
        {
            // DiffServ Multi-Field Traffic Conditioner dropped this packet
            // Free message and return early.

            // Increment stat for number of output IP packets discarded
            // because of a lack of buffer space.

            ip->stats.ipOutDiscards++;

            //Trace drop
            ActionData acnData;
            acnData.actionType = DROP;
            acnData.actionComment = DROP_QUEUE_OVERFLOW;
            TRACE_PrintTrace(node,
                             msg,
                             TRACE_NETWORK_LAYER,
                             PACKET_OUT,
                             &acnData);

            MESSAGE_Free(node, msg);

            #ifdef STK_INTERFACE
                StkUpdateIpStats(
                       node,
                       STK_STAT_IP_DROPPED,
                       ip->stats.ipOutDiscards,
                       getSimTime(node));
            #endif /* STK_INTERFACE */

// GuiStart
            if (node->guiOption == TRUE)
            {
                unsigned int priority = GetQueuePriorityFromUserTos(
                                            node,
                                            (TosType)ipHeader->ip_tos,
                                            (*scheduler).numQueue());
                GUI_QueueDropPacket(node->nodeId, GUI_NETWORK_LAYER,
                                    interfaceIndex, priority,
                                    getSimTime(node));
            }
//GuiEnd

            return;
        }
    }

    // Check the emptiness of the interface's output queue(s) before
    // attempting to queue packet.
    if (mpls)
    {
        queueWasEmpty = MplsOutputQueueIsEmpty(node, interfaceIndex);
    }
    else
#endif // ENTERPRISE_LIB
    {
        queueWasEmpty = NetworkIpOutputQueueIsEmpty(node, interfaceIndex);
    }

    // Queue packet on output queue of interface.

    NetworkIpOutputQueueInsert(node,
                               interfaceIndex,
                               msg,
                               nextHop,
                               ipHeader->ip_dst,
                               NETWORK_PROTOCOL_IP,
                               &queueIsFull);

// GuiStart
    if (!queueIsFull && node->guiOption == TRUE)
    {
        unsigned int priority = GetQueuePriorityFromUserTos(
                                    node,
                                    (TosType)ipHeader->ip_tos,
                                    (*scheduler).numQueue());

        GUI_QueueInsertPacket(node->nodeId, GUI_NETWORK_LAYER,
                              interfaceIndex, priority,
                              MESSAGE_ReturnPacketSize(msg),
                              getSimTime(node));
    }
//GuiEnd

    if (queueIsFull)
    {
        // Increment stat for number of output IP packets discarded
        // because of a lack of buffer space.

        ip->stats.ipOutDiscards++;


        #ifdef STK_INTERFACE
            StkUpdateIpStats(
                       node,
                       STK_STAT_IP_DROPPED,
                       ip->stats.ipOutDiscards,
                       getSimTime(node));
        #endif /* STK_INTERFACE */

        // Free message and return early.

// GuiStart
        if (node->guiOption == TRUE)
        {
            int priority = GetQueuePriorityFromUserTos(
                                node,
                                (TosType)ipHeader->ip_tos,
                                (*scheduler).numQueue());

            GUI_QueueDropPacket(node->nodeId, GUI_NETWORK_LAYER,
                                interfaceIndex, priority,
                                getSimTime(node));
        }
//GuiEnd
        //Trace drop
        ActionData acnData;
        acnData.actionType = DROP;
        acnData.actionComment = DROP_QUEUE_OVERFLOW;
        TRACE_PrintTrace(node,
                         msg,
                         TRACE_NETWORK_LAYER,
                         PACKET_OUT,
                         &acnData);

        MESSAGE_Free(node, msg);
        return;
    }

    // Did not have to drop packet because of lack of buffer space.
    //
    // Start the MAC packet-sending process if the interface's output
    // queue(s) was empty, and after the insert attempt, now has a
    // packet to send.

    if (queueWasEmpty)
    {
#ifdef ENTERPRISE_LIB
        if (mpls)
        {
            if (!MplsOutputQueueIsEmpty(node, interfaceIndex))
            {
                MAC_NetworkLayerHasPacketToSend(node, interfaceIndex);
            }
        }
        else
#endif // ENTERPRISE_LIB
        {
            if (!NetworkIpOutputQueueIsEmpty(node, interfaceIndex))
            {
                MAC_NetworkLayerHasPacketToSend(node, interfaceIndex);
            }
        }
    }
}

//-----------------------------------------------------------------------------
// FUNCTION     ProcessDelayedSendToMac()
// PURPOSE
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message with IP packet.
//-----------------------------------------------------------------------------

static void //inline//
ProcessDelayedSendToMac(Node *node, Message *msg)
{
    DelayedSendToMacLayerInfoType *info =
        (DelayedSendToMacLayerInfoType *) MESSAGE_ReturnInfo(msg);

    NetworkIpSendPacketOnInterface(
        node,
        msg,
        info->incomingInterface,
        info->outgoingInterface,
        info->nextHop);
}

//-----------------------------------------------------------------------------
// Source route
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     IpHeaderSourceRouteOptionField()
// PURPOSE      Returns the source route contained in IP packet.
// PARAMETERS   IpHeaderType *ipHeader
//                  Pointer to IP header.
// RETURNS      Pointer to header of source route option field, if
//              source route is present.
//              NULL, if source route is not present.
//-----------------------------------------------------------------------------

static IpOptionsHeaderType * //inline//
IpHeaderSourceRouteOptionField(IpHeaderType *ipHeader)
{
    return (FindAnIpOptionField(ipHeader, IPOPT_SSRR));
}

//-----------------------------------------------------------------------------
// FUNCTION     SourceRouteThePacket()
// PURPOSE      Extract the next hop IP address from a packet with a
//              source route in the IP header, and call
//              NetworkIpSendPacketOnInterface().
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message with IP packet.
//-----------------------------------------------------------------------------

static void //inline//
SourceRouteThePacket(Node *node, Message *msg)
{
    IpOptionsHeaderType *ipOptions =
        IpHeaderSourceRouteOptionField((IpHeaderType *) msg->packet);
    int outgoingInterface;

    if (ipOptions->ptr < ipOptions->len)
    {
        // Extract Next Address
        // The definition of "ptr" seems to number 1 as the
        // first byte of the the options field.

        char *routeRawAddressBytes = (char *) ipOptions + ipOptions->ptr - 1;
        NodeAddress nextHop;

        // Done because the address bytes may be out of alignment
        // on certain processors.

        memcpy(&nextHop, routeRawAddressBytes, sizeof(NodeAddress));

        // Record Route (Replace next address with this node's address).

        ipOptions->ptr += sizeof(NodeAddress);

        // See DEFAULT_INTERFACE?
        // Source routing only works on single interface nodes. (i.e., single
        // subnet wireless).  This should be fixed! -Jeff

        outgoingInterface = NetworkIpGetInterfaceIndexForNextHop(node,
                                                                 nextHop);

        NetworkIpSendPacketOnInterface(node,
                                      msg,
                                      CPU_INTERFACE,
                                      outgoingInterface,
                                      nextHop);
    }
    else
    {
        // QualNet should not create corrupted source routes.

        ERROR_ReportError("Source route option in IP header corrupted");

        // Re-enable if corrupted source routes are possible.
        // NetworkDataIp *ip = (NetworkDataIp *)
        //                           node->networkData.networkVar;
        //
        // Increment stat for number of IP datagrams discarded because
        // of header errors (e.g., checksum error, version number
        // mismatch, TTL exceeded, etc.).
        //
        //ip->stats.ipInHdrErrors++;
        //
        //MESSAGE_Free(node, msg);
    }//if//
}

//-----------------------------------------------------------------------------
// Boolean utility routines for packet forwarding process
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     IsMyPacket()
// PURPOSE      Determines from the destination IP address whether a
//              packet should be delivered (instead of forwarded).
// PARAMETERS   Node *node
//                  Pointer to node.
//              NodeAddress destAddress
//                  An IP packet's destination IP address.
// RETURNS      TRUE if packet should be delivered to node.
//              FALSE if packet should not be delivered to node.
//
// NOTES        TRUE is returned if destAddress matches any local
//              interface's IP address or broadcast address.  TRUE is
//              also returned if destAddress is 255.255.255.255
//              (ANY_DEST).
//
//              If destAddress is a multicast address, it's checked
//              against any multicast groups the node is subscribed to.
//              [BSD has interfaces subscribing to multicast groups;
//              QualNet has nodes subscribing.]
//-----------------------------------------------------------------------------

static BOOL //inline
IsMyPacket(Node *node, NodeAddress destAddress)
{
    int i;

    if (destAddress == ANY_DEST)
    {
        return TRUE;
    }

    // Check if we are registered for any multicast group.
    // If we are, then accept this packet.
    if (NetworkIpIsMulticastAddress(node, destAddress))
    {
        if (NetworkIpIsPartOfMulticastGroup(node, destAddress))
        {
            return TRUE;
        }

        return FALSE;
    }

    for (i = 0; i < node->numberInterfaces; i++)
    {
        if (destAddress == NetworkIpGetInterfaceAddress(node, i)
            || destAddress == NetworkIpGetInterfaceBroadcastAddress(node, i))
        {
            return TRUE;
        }
    }

    return FALSE;
}

//-----------------------------------------------------------------------------
// FUNCTION     IsOutgoingBroadcast()
// PURPOSE
// PARAMETERS   Node *node
//                  Pointer to node.
//              NodeAddress destAddress
//                  An IP packet's destination IP address.
//              int *outgoingInterface
//                  Outgoing interface index.
//              NodeAddress *outgoingBroadcastAddress
//
// RETURNS
//-----------------------------------------------------------------------------

BOOL
IsOutgoingBroadcast(
    Node *node,
    NodeAddress destAddress,
    int *outgoingInterface,
    NodeAddress *outgoingBroadcastAddress)
{
    int i;

    ERROR_Assert(
        outgoingInterface && outgoingBroadcastAddress,
        "NULL pointer passed to IsOutgoingBroadcast()");

    if (destAddress == ANY_DEST)
    {
        *outgoingInterface = DEFAULT_INTERFACE;
        *outgoingBroadcastAddress = ANY_DEST;
        return TRUE;
    }

    for (i = 0; i < node->numberInterfaces; i++)
    {
        if (destAddress == NetworkIpGetInterfaceBroadcastAddress(node, i))
        {
            *outgoingInterface = i;
            *outgoingBroadcastAddress = ANY_DEST;
            return TRUE;
        }
    }

    return FALSE;
}

//-----------------------------------------------------------------------------
// Per hop behavior (PHB)
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     IpInitPerHopBehaviors()
// PURPOSE
// PARAMETERS   Node *node
//                  Pointer to node.
//              const NodeInput *nodeInput
//
// RETURN       None.
//-----------------------------------------------------------------------------

//static
void //inline//
IpInitPerHopBehaviors(
    Node *node,
    const NodeInput *nodeInput)
{
#ifdef ENTERPRISE_LIB
    NodeInput phbInput;
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    int i, items;
    BOOL retVal;

    IO_ReadCachedFile(
        node->nodeId,
        ANY_ADDRESS,
        nodeInput,
        "PER-HOP-BEHAVIOR-FILE",
        &retVal,
        &phbInput);

    ip->numPhbInfo = 0;


    if (retVal != TRUE)
    {
        return;
    }

    ip->phbInfo = (IpPerHopBehaviorInfoType *)
                  MEM_malloc(sizeof(IpPerHopBehaviorInfoType) *
                         NUM_INITIAL_PHB_INFO_ENTRIES);
    ip->maxPhbInfo = NUM_INITIAL_PHB_INFO_ENTRIES;

    for (i = 0; i < phbInput.numLines; i++)
    {
        char identifier[MAX_STRING_LENGTH];
        int tempDs = 0;
        int tempPriority = 0;
        unsigned char ds = 0;
        QueuePriorityType priority = 0;

        sscanf(phbInput.inputStrings[i], "%s", identifier);

        if (strcmp(identifier, "DS-TO-PRIORITY-MAP") == 0)
        {
            items = sscanf(phbInput.inputStrings[i], "%*s %d %d",
                       &tempDs, &tempPriority);

            if (items != 2)
            {
                ERROR_ReportError("DS-TO-PRIORITY-MAP expects <ds> <prio>\n");
            }
            ds = (unsigned char) tempDs;
        }
        else
        if (strcmp(identifier, "DEFAULT-DS-TO-PRIORITY-MAP") == 0)
        {
            items = sscanf(phbInput.inputStrings[i], "%*s %d",&tempPriority);

            if (items != 1)
            {
                ERROR_ReportError("DEFAULT-DS-TO-PRIORITY-MAP expects only"
                    " default priority queue <prio>\n");
            }

            ds = (unsigned char) DIFFSERV_DS_CLASS_BE;
        }
        priority = (QueuePriorityType) tempPriority;

        AddPHBEntry(node, ds, priority);
    }
#endif // ENTERPRISE_LIB
}

//-----------------------------------------------------------------------------
// FUNCTION     AddPHBEntry()
// PURPOSE
// PARAMETERS   Node *node
//                  Pointer to node.
//              unsigned char ds
//
//              QueuePriorityType priority
//
// RETURN       None.
//-----------------------------------------------------------------------------

static void //inline//
AddPHBEntry(
    Node *node,
    unsigned char ds,
    QueuePriorityType priority)
{
#ifdef ENTERPRISE_LIB
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    if (ip->numPhbInfo == ip->maxPhbInfo)
    {
        IpPerHopBehaviorInfoType* newPhbInfo =
            (IpPerHopBehaviorInfoType*)
            MEM_malloc(sizeof(IpPerHopBehaviorInfoType) *
                (ip->maxPhbInfo + NUM_INITIAL_PHB_INFO_ENTRIES));

        memcpy(newPhbInfo, ip->phbInfo,
               sizeof(IpPerHopBehaviorInfoType) * ip->numPhbInfo);

        ip->maxPhbInfo += NUM_INITIAL_PHB_INFO_ENTRIES;
        MEM_free(ip->phbInfo);
        ip->phbInfo = newPhbInfo;
    }

    ip->phbInfo[ip->numPhbInfo].ds = ds;
    ip->phbInfo[ip->numPhbInfo].priority = priority;

#ifdef DEBUG
    printf("#%d: Add PHB for ds %d => prio %d\n", node->nodeId, ds, priority);
#endif
    ip->numPhbInfo++;
#endif // ENTERPRISE_LIB
}

//-----------------------------------------------------------------------------
// Callbacks into IP made by the MAC layer, helper functions
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     HandleSpecialMacLayerStatusEvents()
// PURPOSE      Give the routing protocol the special status messages
//              that come from the MAC layer that notify when special
//              events at that level occur.  Used by the routing protocols
//              for routing optimization.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message with MAC-layer event.
//              int interfaceIndex
//                  interface associated with MAC events.
// RETURN       None.
//-----------------------------------------------------------------------------

//ANTS01: used to be static
void HandleSpecialMacLayerStatusEvents(Node *node,
                                  Message *msg,
                                  const NodeAddress nextHopAddress,
                                  int interfaceIndex)
{

    MacLayerStatusEventHandlerFunctionType macLayerStatusEventHandlerFunction
        = NetworkIpGetMacLayerStatusEventHandlerFunction(
              node, interfaceIndex);

    if (macLayerStatusEventHandlerFunction)
    {
        (macLayerStatusEventHandlerFunction)(node,
                                             msg,
                                             nextHopAddress,
                                             interfaceIndex);
    }//if//
    MESSAGE_Free(node, msg);
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpRemoveOutputQueue()
// PURPOSE
// PARAMETERS   Node *node
//                  Pointer to node.
//              int interfaceIndex
//                  Interface associated with queue.
//              QueuePriorityType priority
//                  Priority queue to remove packet from.
// RETURN       None.
//-----------------------------------------------------------------------------

/*void
NetworkIpRemoveOutputQueue(
    Node *node,
    int interfaceIndex,
    QueuePriorityType priority)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    Scheduler *schedulerPtr =
        ip->interfaceInfo[interfaceIndex]->scheduler;
    int i;

    for (i = 0; i < schedulerPtr->numQueues; i++)
    {
        if (schedulerPtr->queue[i].priority == priority)
        {
            int afterEntries = schedulerPtr->numQueues - (i + 1);

            if (afterEntries > 0)
            {
                memmove(&schedulerPtr->queue[i], &schedulerPtr->queue[i+1],
                        afterEntries);
            }
            schedulerPtr->numQueues--;
            return;
        }
    }
}

*/
//-----------------------------------------------------------------------------
// Network-layer enqueueing
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpQueueInsert()
// PURPOSE      Calls the packet scheduler for an interface to retrieve
//              an IP packet from a queue associated with the interface.
//              The dequeued packet, since it's already been routed,
//              has an associated next-hop IP address.  The packet's
//              priority value is also returned.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int incomingInterface
//                  interface of input queue.
//              Message *msg
//                  Pointer to message with IP packet.
//              NodeAddress nextHopAddress
//                  Packet's next hop address.
//              NodeAddress destinationAddress
//                  Packet's destination address.
//              int outgoingInterface
//                  Used to determine where packet should go after passing
//                  through the backplane.
//              int networkType
//                  Type of network packet is using (IP, Link-16, ...)
//              BOOL *queueIsFull
//                  Storage for boolean indicator.
//                  If TRUE, packet was not queued because scheduler
//                  reported queue was (or queues were) full.
//              int incomingInterface
//                  Incoming interface, Default argument used by backplane.
// RETURN       None.
//-----------------------------------------------------------------------------

//static void : TBD must be static
void
NetworkIpQueueInsert(
    Node *node,
    Scheduler *scheduler,
    Message *msg,
    NodeAddress nextHopAddress,
    NodeAddress destinationAddress,
    int outgoingInterface,
    int networkType,
    BOOL *queueIsFull,
    int incomingInterface = ANY_INTERFACE)
{
    int queueIndex = ALL_PRIORITIES;
    IpHeaderType *ipHeader = (IpHeaderType *) MESSAGE_ReturnPacket(msg);
    QueuedPacketInfo *infoPtr;

    // Tack on the nextHopAddress to the message using the insidious "info"
    // field.

    MESSAGE_InfoAlloc(node, msg, sizeof(QueuedPacketInfo));

    infoPtr = (QueuedPacketInfo *) MESSAGE_ReturnInfo(msg);

    infoPtr->nextHopAddress = nextHopAddress;
    infoPtr->destinationAddress.ipv4DestAddr = destinationAddress;
    infoPtr->outgoingInterface = outgoingInterface;
    infoPtr->networkType = networkType;
    infoPtr->userTos = (TosType)ipHeader->ip_tos;
    infoPtr->incomingInterface = incomingInterface;

    // Call the "insertFunction" via function pointer.

    // Call the Scheduler "insertFunction"
    queueIndex = GenericPacketClassifier(scheduler,
        (int) ReturnPriorityForPHB(node, (TosType) ipHeader->ip_tos));

    (*scheduler).insert(msg,
                        queueIsFull,
                        queueIndex,
                        NULL, //const void* infoField,
                        getSimTime(node));

    //Trace Enqueue
    ActionData acn;
    acn.actionType = ENQUEUE;
    acn.actionComment = NO_COMMENT;
    acn.pktQueue.interfaceID = outgoingInterface;
    acn.pktQueue.queuePriority = (TosType) ipHeader->ip_tos;
    if (outgoingInterface != CPU_INTERFACE)
    {
        TRACE_PrintTrace(node, msg, TRACE_NETWORK_LAYER, PACKET_OUT, &acn);
    }
    else
    {
        TRACE_PrintTrace(node, msg, TRACE_NETWORK_LAYER, PACKET_IN, &acn);
    }
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpCpuQueueInsert()
// PURPOSE      Calls the cpu packet scheduler for an interface to retrieve
//              an IP packet from a queue associated with the interface.
//              The dequeued packet, since it's already been routed,
//              has an associated next-hop IP address.  The packet's
//              priority value is also returned.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message with IP packet.
//              NodeAddress nextHopAddress
//                  Packet's next hop address.
//              NodeAddress destinationAddress
//                  Packet's destination address.
//              int outgoingInterface
//                  Used to determine where packet should go after passing
//                  through the backplane.
//              int networkType
//                  Type of network packet is using (IP, Link-16, ...)
//              BOOL *queueIsFull
//                  Storage for boolean indicator.
//                  If TRUE, packet was not queued because scheduler
//                  reported queue was (or queues were) full.
//              int incomingInterface
//                  Incoming interface of packet.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkIpCpuQueueInsert(
    Node *node,
    Message *msg,
    NodeAddress nextHopAddress,
    NodeAddress destinationAddress,
    int outgoingInterface,
    int networkType,
    BOOL *queueIsFull,
    int incomingInterface)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    Scheduler *cpuScheduler = ip->cpuScheduler;

    NetworkIpQueueInsert(
                        node,
                        cpuScheduler,
                        msg,
                        nextHopAddress,
                        destinationAddress,
                        outgoingInterface,
                        networkType,
                        queueIsFull,
                        incomingInterface);
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpInputQueueInsert()
// PURPOSE      Calls the input packet scheduler for an interface to retrieve
//              an IP packet from a queue associated with the interface.
//              The dequeued packet, since it's already been routed,
//              has an associated next-hop IP address.  The packet's
//              priority value is also returned.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int incomingInterface
//                  interface of input queue.
//              Message *msg
//                  Pointer to message with IP packet.
//              NodeAddress nextHopAddress
//                  Packet's next hop address.
//              NodeAddress destinationAddress
//                  Packet's destination address.
//              int outgoingInterface
//                  Used to determine where packet should go after passing
//                  through the backplane.
//              int networkType
//                  Type of network packet is using (IP, Link-16, ...)
//              BOOL *queueIsFull
//                  Storage for boolean indicator.
//                  If TRUE, packet was not queued because scheduler
//                  reported queue was (or queues were) full.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkIpInputQueueInsert(
    Node *node,
    int incomingInterface,
    Message *msg,
    NodeAddress nextHopAddress,
    NodeAddress destinationAddress,
    int outgoingInterface,
    int networkType,
    BOOL *queueIsFull)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    Scheduler *inputScheduler = NULL;

    ERROR_Assert(
        incomingInterface >= 0 && incomingInterface < node->numberInterfaces,
        "Invalid incoming interface");

    inputScheduler = ip->interfaceInfo[incomingInterface]->inputScheduler;
    NetworkIpQueueInsert(
                        node,
                        inputScheduler,
                        msg,
                        nextHopAddress,
                        destinationAddress,
                        outgoingInterface,
                        networkType,
                        queueIsFull);
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpOutputQueueInsert()
// PURPOSE      Calls the output packet scheduler for an interface to retrieve
//              an IP packet from a queue associated with the interface.
//              The dequeued packet, since it's already been routed,
//              has an associated next-hop IP address.  The packet's
//              priority value is also returned.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int outgoingInterface
//                  interface of output queue.
//              Message *msg
//                  Pointer to message with IP packet.
//              NodeAddress nextHopAddress
//                  Packet's next hop address.
//              NodeAddress destinationAddress
//                  Packet's destination address.
//              int networkType
//                  Type of network packet is using (IP, Link-16, ...)
//              BOOL *queueIsFull
//                  Storage for boolean indicator.
//                  If TRUE, packet was not queued because scheduler
//                  reported queue was (or queues were) full.
// RETURN       None.
// NOTES        Called by QueueUpIpFragmentForMacLayer().
//-----------------------------------------------------------------------------

void
NetworkIpOutputQueueInsert(
    Node *node,
    int outgoingInterface,
    Message *msg,
    NodeAddress nextHopAddress,
    NodeAddress destinationAddress,
    int networkType,
    BOOL *queueIsFull)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;

    Scheduler *scheduler = NULL;

#ifdef ENTERPRISE_LIB
    MplsData *mpls = (MplsData *) node->macData[outgoingInterface]->mplsVar;
    if (mpls)
    {
        MplsIpOutputQueueInsert(node, mpls, outgoingInterface, msg,
                                nextHopAddress, queueIsFull, TRUE);
        return;
    }
#endif // ENTERPRISE_LIB

    ERROR_Assert(
        outgoingInterface >= 0 && outgoingInterface < node->numberInterfaces,
        "Invalid outgoing interface");

    scheduler = ip->interfaceInfo[outgoingInterface]->scheduler;
    NetworkIpQueueInsert(
                        node,
                        scheduler,
                        msg,
                        nextHopAddress,
                        destinationAddress,
                        outgoingInterface,
                        networkType,
                        queueIsFull);
}



#ifdef ADDON_LINK16

void
NetworkQueueInsert(
    Node *node,
    Scheduler *scheduler,
    Message *msg,
    NodeAddress nextHopAddress,
    NodeAddress destinationAddress,
    int outgoingInterface,
    int networkType,
    BOOL *queueIsFull,
    int incomingInterface = ANY_INTERFACE)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    QueuedPacketInfo *infoPtr;

    // Tack on the nextHopAddress to the message using the insidious "info"
    // field.

    MESSAGE_InfoAlloc(node, msg, sizeof(QueuedPacketInfo));

    infoPtr = (QueuedPacketInfo *) MESSAGE_ReturnInfo(msg);

    infoPtr->nextHopAddress = nextHopAddress;
    infoPtr->destinationAddress.ipv4DestAddr = destinationAddress;
    infoPtr->outgoingInterface = outgoingInterface;
    infoPtr->networkType = networkType;
    infoPtr->userTos = (TosType) IPTOS_PREC_ROUTINE;
    infoPtr->incomingInterface = incomingInterface;

    // Call the "insert" function
    int queueIndex = GenericPacketClassifier(scheduler,
            (int) ReturnPriorityForPHB(node, (TosType) IPTOS_PREC_ROUTINE));
    (*scheduler).insert(msg,
        queueIsFull,
        queueIndex,
        NULL, //infoField,
        getSimTime(node));

    //Trace Enqueue
    if (!(*queueIsFull))
    {
        IpHeaderType *ipHeader
                    = (IpHeaderType *) MESSAGE_ReturnPacket(msg);

        //Trace Enqueue
        ActionData acn;
        acn.actionType = ENQUEUE;
        acn.actionComment = NO_COMMENT;
        acn.pktQueue.interfaceID = outgoingInterface;
        acn.pktQueue.queuePriority = (TosType) ipHeader->ip_tos;
        TRACE_PrintTrace(node, msg, TRACE_NETWORK_LAYER, PACKET_IN, &acn);
    }
}


void
NetworkOutputQueueInsert(
    Node *node,
    int outgoingInterface,
    Message *msg,
    NodeAddress nextHopAddress,
    NodeAddress destinationAddress,
    int networkType,
    BOOL *queueIsFull)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    MplsData *mpls = (MplsData *) node->macData[outgoingInterface]->mplsVar;

    Scheduler *scheduler = NULL;

    if (mpls)
    {
        MplsIpOutputQueueInsert(node, mpls, outgoingInterface, msg,
                                nextHopAddress, queueIsFull, TRUE);
        return;
    }

    ERROR_Assert(
        outgoingInterface >= 0 && outgoingInterface < node->numberInterfaces,
        "Invalid outgoing interface");

    scheduler = ip->interfaceInfo[outgoingInterface]->scheduler;
    NetworkQueueInsert(
                       node,
                       scheduler,
                       msg,
                       nextHopAddress,
                       destinationAddress,
                       outgoingInterface,
                       networkType,
                       queueIsFull,
                       ANY_INTERFACE);
}

#endif // ADDON_LINK16


//-----------------------------------------------------------------------------
// IP header
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpAddHeader()
// PURPOSE      Add an IP packet header to a message.
//              Just calls AddIpHeader.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message.
//              NodeAddress sourceAddress
//                  Source IP address.
//              NodeAddress destinationAddress
//                  Destination IP address.
//              TosType priority
//                  Currently a TosType.
//                  (values are not standard for "IP type of service field"
//                  but has correct function)
//              unsigned char protocol
//                  IP protocol number.
//              unsigned ttl
//                  Time to live.
//                  If 0, uses default value IPDEFTTL, as defined in
//                  include/ip.h.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkIpAddHeader(
    Node *node,
    Message *msg,
    NodeAddress sourceAddress,
    NodeAddress destinationAddress,
    TosType priority,
    unsigned char protocol,
    unsigned ttl)
{
    AddIpHeader(node,
                msg,
                sourceAddress,
                destinationAddress,
                priority,
                protocol,
                ttl);
}


//-----------------------------------------------------------------------------
// FUNCTION     AddIpHeader()
// PURPOSE      Add an IP packet header to a message.
//              The new message has an IP packet.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message.
//              NodeAddress sourceAddress
//                  Source IP address.
//              NodeAddress destinationAddress
//                  Destination IP address.
//              TosType priority
//                  Currently a TosType.
//                  (values are not standard for "IP type of service field"
//                  but has correct function)
//              unsigned char protocol
//                  IP protocol number.
//              unsigned ttl
//                  Time to live.
//                  If 0, uses default value IPDEFTTL, as defined in
//                  include/ip.h.
// RETURN       None.
//-----------------------------------------------------------------------------

void
AddIpHeader(
    Node *node,
    Message *msg,
    NodeAddress sourceAddress,
    NodeAddress destinationAddress,
    TosType priority,
    unsigned char protocol,
    unsigned ttl)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    IpHeaderType *ipHeader;
    int hdrSize = sizeof(IpHeaderType);

    MESSAGE_AddHeader(node, msg, hdrSize, TRACE_IP);

    ipHeader = (IpHeaderType *) msg->packet;
    memset(ipHeader, 0, hdrSize);

    ipHeader->ip_v = IPVERSION4;
    ipHeader->ip_id = ip->packetIdCounter;
    ip->packetIdCounter++;
    ipHeader->ip_src = sourceAddress;
    ipHeader->ip_dst = destinationAddress;

    if (ttl == 0)
    {
        ipHeader->ip_ttl = IPDEFTTL;
    }
    else
    {
        ipHeader->ip_ttl = (unsigned char) ttl;
    }

    // TOS field (8 bit) in the IPV4 header
    ipHeader->ip_tos = priority;


    if (ip->isPacketEcnCapable)
    {
        // Bits 6 and 7 of TOS field in the IPV4 header are used by ECN
        // and proposed respectively for the ECT and CE bits.
        // So before assign the value of priority to ip_tos, leave bits 6 and 7.

        if (ipHeader->ip_tos & 0x03)
        {
            // User TOS specification conflicts with an ~enabled~ ECN
            char errorString[MAX_STRING_LENGTH];
            sprintf(errorString,
                    "~enabled~ ECN!!! ECN bits of TOS field in"
                    " application Input should contain zero values\n");
            ERROR_ReportError(errorString);
        }

        ipHeader->ip_tos |=  IPTOS_ECT;
        ip->isPacketEcnCapable = FALSE;
    }
#ifdef ECN_DEBUG_TEST
    {
        /*
         * Mark the CE bit of some specific data packets (for testing)
         */
        int markCount = ECN_TEST_PKT_MARK;
        UInt32 markValues[] = { 14002};
        static int markFlag[] = { 1};
        struct tcphdr *aTcpHdr = (struct tcphdr *)((char*)ipHeader +
                                                    hdrSize);
        if (markCount) {
            int counter;
            for (counter = 0; counter < markCount; counter++) {
                if (aTcpHdr->th_seq == markValues[counter]
                            && markFlag[counter]) {
                    markFlag[counter]--;
                    ipHeader->ip_tos |=  IPTOS_CE;
                    printf ("\nSequence number of CE (specific for test)"
                             " marked packet is %u\n\n",(unsigned) aTcpHdr->th_seq);
                }
            }
        }
    }
#endif /* ECN_DEBUG_TEST */

    ipHeader->ip_p = protocol;

    ERROR_Assert(MESSAGE_ReturnPacketSize(msg) <= IP_MAXPACKET,
                 "IP datagram (including header) exceeds IP_MAXPACKET bytes");

    ipHeader->ip_len = MESSAGE_ReturnPacketSize(msg);
    SetIpHeaderSize(ipHeader, hdrSize);

}

//-----------------------------------------------------------------------------
// FUNCTION     ExpandOrShrinkIpHeader()
// PURPOSE      Increases the size of an IP header so as to allow for
//              adding a new source route or other options.
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message.
//              int newHeaderSize
//                  New IP header size in bytes.
//                  (This value is checked.)
// RETURN       None.
//-----------------------------------------------------------------------------

static void //inline//
ExpandOrShrinkIpHeader(
    Node *node,
    Message *msg,
    int newHeaderSize)
{
    IpHeaderType *ipHeader = (IpHeaderType *) msg->packet;

    ERROR_Assert(newHeaderSize <= IP_MAX_HEADER_SIZE,
                 "IP header exceeds IP_MAX_HEADER_SIZE bytes");

    if (IpHeaderSize(ipHeader) != (unsigned) newHeaderSize)
    {
        IpHeaderType *origIpHeader = ipHeader;
        int oldIpHeaderSize = IpHeaderSize(ipHeader);

        MESSAGE_RemoveHeader(node, msg, oldIpHeaderSize, TRACE_IP);

        MESSAGE_AddHeader(node, msg, newHeaderSize, TRACE_IP);

        ipHeader = (IpHeaderType *) msg->packet;
        memmove((char *) ipHeader, (char *) origIpHeader,
                MIN(oldIpHeaderSize, newHeaderSize));
        ipHeader->ip_len = MESSAGE_ReturnPacketSize(msg);
        SetIpHeaderSize(ipHeader, newHeaderSize);
    }//if//
}

//-----------------------------------------------------------------------------
// IP header option field
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     ExpandOrShrinkIpOptionField()
// PURPOSE
// PARAMETERS   Node *node
//                  Pointer to node.
//              Message *msg
//                  Pointer to message.
//              const int int optionCode
//                  option code
//              const int newIpOptionSize
//                  size of option
// RETURN       None.
//-----------------------------------------------------------------------------

static void //inline//
ExpandOrShrinkIpOptionField(
    Node *node,
    Message *msg,
    const int optionCode,
    const int newIpOptionSize)
{
    IpHeaderType *ipHeader = (IpHeaderType *) msg->packet;
    IpOptionsHeaderType *ipOption = FindAnIpOptionField(ipHeader, optionCode);
    int oldHeaderSize = IpHeaderSize(ipHeader);
    int oldIpOptionSize = ipOption->len;
    int deltaOptionSize = newIpOptionSize - ipOption->len;
    int newHeaderSize = oldHeaderSize + deltaOptionSize;
    int bytesAfterOption = (int) (((char *) ipHeader + oldHeaderSize) -
                                  ((char *) ipOption + oldIpOptionSize));

    ERROR_Assert(newIpOptionSize % 4 == 0,
                 "IP option size must be a multiple of 4 bytes in QualNet");

    ExpandOrShrinkIpHeader(node, msg, newHeaderSize);

    ipHeader = (IpHeaderType *) msg->packet;
    ipOption = FindAnIpOptionField(ipHeader, optionCode);
    ipOption->len = (unsigned char) newIpOptionSize;

    // Move the header data after this option field to make room for
    // the new option data.

    memmove(((char *) ipOption + newIpOptionSize),
            ((char *) ipOption + oldIpOptionSize), bytesAfterOption);
}

//-----------------------------------------------------------------------------
// Statistics
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpInitStats()
// PURPOSE      Initialize statistics.
// PARAMETERS   Node* node
//              NetworkIpStatsType *stats
//                  Pointer to stats struct.
// RETURN       None.
//-----------------------------------------------------------------------------

static void //inline//
NetworkIpInitStats(Node* node, NetworkIpStatsType *stats)
{
    stats->deliveredPacketTtlTotal = 0;

    if (node->guiOption) {
        stats->ipInReceivesId = GUI_DefineMetric("IP In Receives",
                                                 node->nodeId,
                                                 GUI_NETWORK_LAYER, 0,
                                                 GUI_UNSIGNED_TYPE,
                                                 GUI_CUMULATIVE_METRIC);

        stats->ipInHdrErrorsId = GUI_DefineMetric("IP Header Errors",
                                                  node->nodeId,
                                                  GUI_NETWORK_LAYER, 0,
                                                  GUI_UNSIGNED_TYPE,
                                                  GUI_CUMULATIVE_METRIC);
    }

    stats->ipInReceives = 0;
    stats->ipInHdrErrors = 0;
    stats->ipInAddrErrors = 0;
    stats->ipInForwardDatagrams = 0;
    stats->ipInDiscards = 0;
    stats->ipInDelivers = 0;
    stats->ipOutRequests = 0;
    stats->ipOutDiscards = 0;
    stats->ipOutNoRoutes = 0;
    stats->ipReasmReqds = 0;
    stats->ipReasmOKs = 0;
    stats->ipReasmFails = 0;
    stats->ipFragOKs = 0;

    stats->ipInReceivesLastPeriod = 0;
    stats->ipInHdrErrorsLastPeriod = 0;
    stats->ipInAddrErrorsLastPeriod = 0;
    stats->ipInForwardDatagramsLastPeriod = 0;
    stats->ipInDiscardsLastPeriod = 0;
    stats->ipInDeliversLastPeriod = 0;
    stats->ipOutRequestsLastPeriod = 0;
    stats->ipOutDiscardsLastPeriod = 0;
    stats->ipOutNoRoutesLastPeriod = 0;
    stats->ipReasmReqdsLastPeriod = 0;
    stats->ipReasmOKsLastPeriod = 0;
    stats->ipReasmFailsLastPeriod = 0;
    stats->ipFragOKsLastPeriod = 0;

    stats->ipNumDroppedDueToBackplaneLimit = 0;

    // Fragmentation statistics.
    stats->ipFragsCreated = 0;
    stats->ipPacketsAfterFragsReasm = 0;
    stats->ipFragsInBuff = 0;

    //ATM : statistics added for gateway
    stats->ipRoutePktThruGt = 0;
    stats->ipSendPktToOtherNetwork = 0;
    stats->ipRecvdPktFromOtherNetwork = 0;
    //ATM
}


//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpPrintStats()
// PURPOSE      Print IP statistics using IO_PrintStat().
// PARAMETERS   Node *node
//                  Pointer to node.
// RETURN       None.
//-----------------------------------------------------------------------------

void
NetworkIpPrintStats(Node *node)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    NetworkIpStatsType *stats = &ip->stats;
    char buf[MAX_STRING_LENGTH];
    double averageHopCount;

    sprintf(buf, "ipInReceives = %u", stats->ipInReceives);
    IO_PrintStat(node, "Network", "IP", ANY_DEST, -1 /* instance Id */, buf);
    sprintf(buf, "ipInHdrErrors = %u", stats->ipInHdrErrors);
    IO_PrintStat(node, "Network", "IP", ANY_DEST, -1 /* instance Id */, buf);
    sprintf(buf, "ipInForwardDatagrams = %u", stats->ipInForwardDatagrams);
    IO_PrintStat(node, "Network", "IP", ANY_DEST, -1 /* instance Id */, buf);
    sprintf(buf, "ipInDelivers = %u", stats->ipInDelivers);
    IO_PrintStat(node, "Network", "IP", ANY_DEST, -1 /* instance Id */, buf);
    sprintf(buf, "ipOutRequests = %u", stats->ipOutRequests);
    IO_PrintStat(node, "Network", "IP", ANY_DEST, -1 /* instance Id */, buf);
    sprintf(buf, "ipOutDiscards = %u", stats->ipOutDiscards);
    IO_PrintStat(node, "Network", "IP", ANY_DEST, -1 /* instance Id */, buf);
    sprintf(buf, "ipOutNoRoutes = %u", stats->ipOutNoRoutes);
    IO_PrintStat(node, "Network", "IP", ANY_DEST, -1 /* instance Id */, buf);

    sprintf(buf, "Packets fragmented = %u", stats->ipFragOKs);
    IO_PrintStat(node, "Network", "IP", ANY_DEST, -1 /* instance Id */, buf);
    sprintf(buf, "Fragments created = %u", stats->ipFragsCreated);
    IO_PrintStat(node, "Network", "IP", ANY_DEST, -1, buf);
    sprintf(buf, "Fragments received = %u", stats->ipReasmReqds);
    IO_PrintStat(node, "Network", "IP", ANY_DEST, -1 /* instance Id */, buf);
    sprintf(buf, "Fragments dropped = %u", stats->ipReasmFails);
    IO_PrintStat(node, "Network", "IP", ANY_DEST, -1 /* instance Id */, buf);
    sprintf(buf, "Fragments in Buffer = %u", stats->ipFragsInBuff);
    IO_PrintStat(node, "Network", "IP", ANY_DEST, -1, buf);
    sprintf(buf, "Fragments reassembled = %u", stats->ipReasmOKs);
    IO_PrintStat(node, "Network", "IP", ANY_DEST, -1 /* instance Id */, buf);
    sprintf(buf, "Packets created after reassembling = %u", stats->ipPacketsAfterFragsReasm);
    IO_PrintStat(node, "Network", "IP", ANY_DEST, -1, buf);
    sprintf(buf, "ipInDelivers TTL sum = %u", stats->deliveredPacketTtlTotal);
    IO_PrintStat(node, "Network", "IP", ANY_DEST, -1 /* instance Id */, buf);

    // ATM : statistics added for gateway
    if (ip->gatewayConfigured)
    {
        sprintf(buf, "ipRoutePktThruGt = %u",
        stats->ipRoutePktThruGt);
        IO_PrintStat(node, "Network", "IP", ANY_DEST, -1 /* instance Id */, buf);

        sprintf(buf, "Send pkt to Other Net = %u",
        stats->ipSendPktToOtherNetwork);
        IO_PrintStat(node, "Network", "IP", ANY_DEST, -1 /* instance Id */, buf);

        sprintf(buf, "Recvd pkt from Other Net  = %u",
        stats->ipRecvdPktFromOtherNetwork);
        IO_PrintStat(node, "Network", "IP", ANY_DEST, -1 /* instance Id */, buf);
    }
    //ATM

    // Report TTL-based average hop count.  Note that the average hop
    // count is not accurate at all when there are many IP packets with
    // initial TTLs not equal to IPDEFTTL, as defined in include/ip.h.
    // This inaccuracy is present for AODV, for example, since many
    // AODV packets have an initial TTL of 1.

    if (stats->ipInDelivers == 0)
    {
        averageHopCount = 0;
    }
    else
    {
        averageHopCount =
            (double)
            ((IPDEFTTL * stats->ipInDelivers) - stats->deliveredPacketTtlTotal)
            / stats->ipInDelivers;
    }

    sprintf(buf, "ipInDelivers TTL-based average hop count = %.2f",
            floor(averageHopCount * 100 + 0.5) / 100);

    IO_PrintStat(node, "Network", "IP", ANY_DEST, -1 /* instance Id */, buf);

    if (ip->backplaneThroughputCapacity !=
        NETWORK_IP_UNLIMITED_BACKPLANE_THROUGHPUT)
    {
        sprintf(buf, "ipNumDroppedDueToBackplaneLimit = %u",
                stats->ipNumDroppedDueToBackplaneLimit);
        IO_PrintStat(node, "Network", "IP", ANY_DEST, -1, buf);
    }
}

//-----------------------------------------------------------------------------
// Multicast
//-----------------------------------------------------------------------------

// /**
// API                 :: RouteThePacketUsingMulticastForwardingTable
// LAYER               :: Network
// PURPOSE             :: Tries to route the multicast packet using the
//                        multicast forwarding table.
// PARAMETERS          ::
// + node               : Node*     : this node
// + msg                : Message*  : Pointer to Message
// + incomingInterface  : int       : Incomming Interface
// RETURN              :: void      : NULL.
// **/
void
RouteThePacketUsingMulticastForwardingTable(
    Node* node,
    Message *msg,
    int incomingInterface,
    NetworkType netType)
{
    IpHeaderType* ipHeader = NULL;
    ip6_hdr* ipv6Header = NULL;
    NodeAddress nextHop;
    LinkedList* interfaceList;
    ListItem *item;

    if (netType == NETWORK_IPV4)
    {
        ipHeader = (IpHeaderType* ) MESSAGE_ReturnPacket(msg);

        // Get all the interfaces that we must use to forward the
        // multicast packets.
        interfaceList =
            NetworkGetOutgoingInterfaceFromMulticastForwardingTable(
                node,
                ipHeader->ip_src,
                ipHeader->ip_dst);
    }
    else
    {
        ipv6Header = (ip6_hdr* ) MESSAGE_ReturnPacket(msg);

        if (ipv6Header->ip6_hlim < IPTTLDEC)
        {
            MESSAGE_Free(node, msg);

            return;
        }

        ipv6Header->ip6_hlim -= IPTTLDEC;

        // Get all the interfaces that we must use to forward the
        // multicast packets.
        interfaceList = Ipv6GetOutgoingInterfaceFromMulticastTable(
                            node,
                            ipv6Header->ip6_src,
                            ipv6Header->ip6_dst);
    }

    // No interfaces for multicast packet, so drop...
    if (interfaceList == NULL)
    {
        IPv6Data* ipv6  = (IPv6Data*) node->networkData.networkVar->ipv6;

        // Packet Trace for IPv6 not supported
        if (netType == NETWORK_IPV4)
        {
            //Trace drop
            ActionData acnData;
            acnData.actionType = DROP;
            acnData.actionComment = DROP_NO_ROUTE;
            TRACE_PrintTrace(node,
                            msg,
                            TRACE_NETWORK_LAYER,
                            PACKET_OUT,
                            &acnData);
        }
        else
        {
            ipv6->ip6_stat.ip6s_noroute++;
        }

        MESSAGE_Free(node, msg);

        return;
    }


    // Send packet out to each interface that we need to forward onto.

    item = interfaceList->first;

    while (item)
    {
        int *outgoingInterface = (int *) item->data;

        nextHop = ANY_DEST;

        if (netType == NETWORK_IPV4)
        {
            NetworkIpSendPacketOnInterface(
                node,
                MESSAGE_Duplicate(node, msg),
                incomingInterface,
                *outgoingInterface,
                nextHop);
        }
        else
        {
            route rt;

            COPY_ADDR6(ipv6Header->ip6_dst, rt.ro_dst);

            ip6_output(
                node,
                MESSAGE_Duplicate(node, msg),
                (optionList*)NULL,
                &rt,
                0,
                (ip_moptions*)NULL,
                *outgoingInterface);
        }

        item = item->next;
    }

    MESSAGE_Free(node, msg);
}

//---------------------------------------------------------------------------
//-------------------------------------------------------------------------//
//          IP FRAGMENTATION IS END-TO-END ONLY.
//-------------------------------------------------------------------------//

//---------------------------------------------------------------------------
// FUNCTION             : IpFragmentPacket
// PURPOSE             :: Fragment an IP packet
// PARAMETERS          ::
// +node                : Node* node: pointer to Node
// +msg                 : Message* msg: Pointer to Message Structure
// +mtu                 : int mtu: Maximum Transmission Unit of the circuit.
//                                  list header.
// RETURN               : int : status of fragmentation.
// NOTES                : IP fragmented packet processing function
//---------------------------------------------------------------------------
int
IpFragmentPacket(
    Node* node,
    Message* msg,
    int mtu,
    ipFragmetedMsg** fragmentHead)
{
    int off;
    int hlen;
    int fragmentedLen;

    IpHeaderType* originalIpHdr = NULL;
    IpHeaderType *ipHeader = NULL;

    Message* tmpMsg = NULL;
    ipFragmetedMsg* fragmentChain = NULL;

    char* tempPayload = NULL;
    unsigned short offset;
    unsigned int curPayloadLen;

    char*  payload = MESSAGE_ReturnPacket(msg);
    int packetLen = 0;
    NetworkDataIp* ip = (NetworkDataIp *) node->networkData.networkVar;

    NodeAddress sourceAddress;
    NodeAddress destinationAddress;
    TosType priority = 0;
    unsigned char protocol;
    unsigned hLim;


    ipHeader = (IpHeaderType*) payload;
    if (ipHeader->ip_hl * 4 > sizeof(IpHeaderType))
    {
        return FALSE;
    }
    hlen = sizeof(IpHeaderType);


    // Now Calculate the fragmentedLen
    // The Fragmentable Part of the original packet is divided into
    // fragments, each, except possibly the last one,
    // being an integer multiple of 8 octets long.

    fragmentedLen = ((int)(mtu - hlen) / 8) * 8;

    // Get the copy of the ip header.
    originalIpHdr = (IpHeaderType*) MEM_malloc(sizeof(IpHeaderType));

    memcpy(originalIpHdr, payload, sizeof(IpHeaderType));

    packetLen = ipHeader->ip_len;
    // Remove the Ip header from the original packet.
    NetworkIpRemoveIpHeader(
        node,
        msg,
        &sourceAddress,
        &destinationAddress,
        &priority,
        &protocol,
        &hLim);

    // Get the new payload after removing the ip header.
    payload = MESSAGE_ReturnPacket(msg);


    // Make the first fragment Packet.
    tmpMsg = MESSAGE_Alloc(node,
                           NETWORK_LAYER,
                           NETWORK_PROTOCOL_IP,
                           MSG_NETWORK_Ip_Fragment);

    // allocate packet for fragmented part by taking care of original
    // packet and virtual packet.
    if (msg->packetSize >= fragmentedLen)
    {
        MESSAGE_PacketAlloc(
            node,
            tmpMsg,
            fragmentedLen ,
            TRACE_IP);

            // Now Make the fragmented Packet. with out virtual packet
            tempPayload = MESSAGE_ReturnPacket(tmpMsg);
            memcpy(tempPayload, payload, fragmentedLen);
    }
    else
    {
        MESSAGE_PacketAlloc(
            node,
            tmpMsg,
            msg->packetSize,
            TRACE_IP);

            tempPayload = MESSAGE_ReturnPacket(tmpMsg);
            memcpy(tempPayload, payload, msg->packetSize);

            // Now Make the fragmented Packet. with virtual packet.
            MESSAGE_AddVirtualPayload(
                node,
                tmpMsg,
                fragmentedLen - msg->packetSize);
    }

    //------------------------------------------------------------------------//
    // QULNET'S EXTRA OVERHEAD TO MANAGE BROKEN MESSAGE.
    //------------------------------------------------------------------------//
    tmpMsg->sequenceNumber = msg->sequenceNumber;
    tmpMsg->originatingProtocol = msg->originatingProtocol;
    tmpMsg->protocolType = msg->protocolType;
    tmpMsg->layerType = msg->layerType;
    tmpMsg->numberOfHeaders = msg->numberOfHeaders;

    for (int headerCounter = 0;
        headerCounter < msg->numberOfHeaders;
        headerCounter++)
    {
        tmpMsg->headerProtocols[headerCounter] =
            msg->headerProtocols[headerCounter];
        tmpMsg->headerSizes[headerCounter] = msg->headerSizes[headerCounter];
    }
    MESSAGE_CopyInfo(node, tmpMsg, msg);
    //------------------------------------------------------------------------//
    // END OF QUALNET SPECIFIC WORK.
    //------------------------------------------------------------------------//

    // Unfragmented part added here first fragment header and then ipv6 header

    NetworkIpAddHeader(
        node,
        tmpMsg,
        sourceAddress,
        destinationAddress,
        priority,
        protocol,
        hLim);

    ipHeader = (IpHeaderType *)MESSAGE_ReturnPacket(tmpMsg);
    ipHeader->ip_more_fragments = 1;
    ipHeader->ip_id = ip->ipFragmentId;

    ipHeader->ip_len = packetLen;
    ipHeader->ip_fragment_offset = 0;
    ip->stats.ipFragsCreated++;

     // Now put it into the fragmented list.

    (*fragmentHead) = (ipFragmetedMsg*) MEM_malloc(sizeof(ipFragmetedMsg));
    (*fragmentHead)->next = NULL;
    (*fragmentHead)->msg = tmpMsg;
    fragmentChain = (*fragmentHead);

    // Loop through length of segment after first fragment,
    // make new header and copy data of each part and link onto chain.
    for (off = fragmentedLen;
        off < (packetLen - hlen); off += fragmentedLen)
    {
        BOOL ip_more_fragments = FALSE;
        tmpMsg = MESSAGE_Alloc(
                    node,
                    NETWORK_LAYER,
                    NETWORK_PROTOCOL_IP,
                    MSG_NETWORK_Ip_Fragment);

        offset = (unsigned short) off;
        offset = (unsigned short) (offset >> 3);
        if ((packetLen - hlen) <= off + fragmentedLen)
        {
            curPayloadLen = packetLen - hlen - off;
        }
        else
        {
            ip_more_fragments = TRUE;
            curPayloadLen = fragmentedLen;
        }

        // packet to curPayloadLen - hlen, because fragment and V6 header
        // size are allocated later but it will allocated when there is
        // no virtual packet else 0 byte is allocated
        // virtual packet is added.
        if (msg->packetSize > off)
        {
            MESSAGE_PacketAlloc(node, tmpMsg,
                curPayloadLen,
                TRACE_IP);
            // Now Make the fragmented Packet. with out virtual packet
            tempPayload = MESSAGE_ReturnPacket(tmpMsg);
            memcpy(tempPayload, payload + off, curPayloadLen);
        }
        else
        {
            MESSAGE_PacketAlloc(node, tmpMsg,
                0,
                TRACE_IP);
            // Now Make the fragmented Packet. with virtual packet.
            MESSAGE_AddVirtualPayload(node, tmpMsg, curPayloadLen);
        }

        // unfragmented part added here first fragment header and then ipv6 header

        NetworkIpAddHeader(
                node,
                tmpMsg,
                sourceAddress,
                destinationAddress,
                priority,
                protocol,
                hLim);

        // Set the ip headers fragment option porperly.
        ipHeader = (IpHeaderType *)MESSAGE_ReturnPacket(tmpMsg);

        if (ip_more_fragments)
        {
            ipHeader->ip_more_fragments = ip_more_fragments;
        }
        ipHeader->ip_id = ip->ipFragmentId;

        ipHeader->ip_fragment_offset = offset;
        ipHeader->ip_len = packetLen;

        // Now put it in the fragment chain.
        fragmentChain->next = (ipFragmetedMsg*)
                                MEM_malloc(sizeof(ipFragmetedMsg));
        fragmentChain = fragmentChain->next;
        fragmentChain->msg = tmpMsg;
        fragmentChain->next = NULL;

        ip->stats.ipFragsCreated++;
        // QULNET'S EXTRA OVERHEAD TO MANAGE BROKEN MESSAGE.
        tmpMsg->sequenceNumber = msg->sequenceNumber;

    } // end of creating all fragments

    MEM_free(originalIpHdr);
    ip->ipFragmentId++;
    ip->stats.ipFragOKs++;
    return TRUE;
} // end of fragmentation.


//---------------------------------------------------------------------------
// FUNCTION             : IpDeleteFragmentedPacket
// PURPOSE             :: This function deletes all the fragmented packet.
// PARAMETERS          ::
// +node                : Node* node    : Pointer to node
// +ipv6                : NetworkDataIp* ip: Ip data pointer of node.
// +prevFp              : IpFragQueue** prevFp: Pointer to the pointer of
//                                      previous fragment queue.
// +fp                  : IpFragQueue** fp: Pointer to the pointer of
//                                      fragment queue.
// RETURN               : None
// NOTES                : fragmented header processing function
//---------------------------------------------------------------------------

static void
IpDeleteFragmentedPacket(
    Node* node,
    NetworkDataIp* ip,
    IpFragQueue** prevFp,
    IpFragQueue** fp)
{
    IpFragQueue* tempFp = NULL;
    IpFragData* tempFrg = (*fp)->firstMsg;
    IpFragData* grbFrg = NULL;

    while (tempFrg)
    {
        grbFrg = tempFrg;
        tempFrg = tempFrg->nextMsg;
        MESSAGE_Free(node, grbFrg->msg);
        MEM_free(grbFrg);
        ip->stats.ipReasmFails++;
        ip->stats.ipFragsInBuff--;
    }
    if ((*prevFp))
    {
        (*prevFp)->next = (*fp)->next;
    }
    else
    {
        ip->fragmentListFirst = (*fp)->next;
    }
    tempFp = (*fp);
    (*fp) = (*fp)->next;
    MEM_free(tempFp);
}


//---------------------------------------------------------------------------
// FUNCTION             : IpFragmentInput
// PURPOSE             :: Fragment header input processing
// PARAMETERS          ::
// +node                : Node* node    : Pointer to node
// +msg                 : Message* msg  : Pointer to Message Structure
// +interfaceId         : int interfaceId  : Value of the Interface index
//
// RETURN               : None
// NOTES                : fragmented header processing function
//---------------------------------------------------------------------------

Message*
IpFragmentInput(Node* node, Message* msg, int interfaceId, BOOL* isReassembled)
{
    char* payload = MESSAGE_ReturnPacket(msg);
    int packetLen = MESSAGE_ReturnPacketSize(msg);

    NetworkDataIp* ip = (NetworkDataIp *) node->networkData.networkVar;

    IpHeaderType* ipHeader = (IpHeaderType *) payload;

    IpFragQueue* prevFp = NULL;
    IpFragQueue* fp = ip->fragmentListFirst;

    Message* joinedMsg = NULL;

    ip->stats.ipReasmReqds++;

    // Reassemble, first pullup headers.
    if ((unsigned) packetLen < sizeof(IpHeaderType))
    {
        // Fragment Error.
        ip->stats.ipReasmFails++;
        *isReassembled = FALSE;
        //Trace drop
        ActionData acnData;
        acnData.actionType = DROP;
        acnData.actionComment = DROP_START_FRAG_LENG_LESSTHAN_IPHEDEARTYPE;
        TRACE_PrintTrace(node,
                         msg,
                         TRACE_NETWORK_LAYER,
                         PACKET_IN,
                         &acnData);

        MESSAGE_Free(node, msg);
        return NULL;
    }

    // Make sure that fragments have a data length
    // that's a non-zero multiple of 8 bytes.
    if ((ipHeader->ip_more_fragments) &&
        ((ipHeader->ip_len <= sizeof(IpHeaderType)) ||
        (((packetLen - sizeof(IpHeaderType)) & 0x7) != 0)))
    {
         // Fragmentation Error.
         ip->stats.ipReasmFails++;
         *isReassembled = FALSE;
        //Trace drop
        ActionData acnData;
        acnData.actionType = DROP;
        acnData.actionComment = DROP_MORE_FRAG_IPLENG_LESSEQ_IPHEDEARTYPE;
        TRACE_PrintTrace(node,
                         msg,
                         TRACE_NETWORK_LAYER,
                         PACKET_IN,
                         &acnData);

         MESSAGE_Free(node, msg);
         return NULL;
    }

    while (fp)
    {
        // Look for queue of fragments of this datagram.
        if (ipHeader->ip_id == fp->ipFrg_id &&
            (ipHeader->ip_src == fp->ipFrg_src) &&
            (ipHeader->ip_dst == fp->ipFrg_dst))
        {
            break;
        }
        // Delete all the fragmented packets whose time has expired.
        if (getSimTime(node) >= fp->fragHoldTime)
        {
            IpDeleteFragmentedPacket(node, ip, &prevFp, &fp);
        }// end of deleting.
        else
        {
            prevFp = fp;
            fp = fp->next;
        }
    }

    // Now add the fragment in the buffer.
    IpFragData* newFragMsg = NULL;
    if (fp)
    {
        IpFragData* temp = fp->firstMsg;
        char* tempPayload = MESSAGE_ReturnPacket(temp->msg);
        IpHeaderType* tempIpHeader = (IpHeaderType *)tempPayload;
        int msgFragOff = ipHeader->ip_fragment_offset << 3;
        int currentFragOff = tempIpHeader->ip_fragment_offset << 3;

        // if this fragment offset is less than the starting fragment
        // offset so put it in the front
        if (currentFragOff > msgFragOff)
        {
            newFragMsg = (IpFragData*) MEM_malloc(sizeof(IpFragData));
            newFragMsg->nextMsg = fp->firstMsg;
            fp->firstMsg = newFragMsg;
        }
        else
        {
            IpFragData* prevTemp = temp;
            for (;
                temp != NULL; temp = temp->nextMsg)
               {
                   tempPayload = MESSAGE_ReturnPacket(temp->msg);
                   tempIpHeader = (IpHeaderType *)tempPayload;
                   msgFragOff = ipHeader->ip_fragment_offset << 3;
                   currentFragOff = tempIpHeader->ip_fragment_offset << 3;

                   if (currentFragOff == msgFragOff)
                   {
                       MESSAGE_Free(node, msg);
                       return NULL;
                   }

                   if (currentFragOff > msgFragOff)
                   {
                       break;
                   }
                   prevTemp = temp;
               }
               newFragMsg = (IpFragData*) MEM_malloc(sizeof(IpFragData));
               newFragMsg->nextMsg = prevTemp->nextMsg;
               prevTemp->nextMsg = newFragMsg;
         }

    }
    else
    {
        if (!prevFp)
        {
            ip->fragmentListFirst =
                (IpFragQueue*) MEM_malloc(sizeof(IpFragQueue));
            fp = ip->fragmentListFirst;
        }
        else
        {
            prevFp->next =
                (IpFragQueue*) MEM_malloc(sizeof(IpFragQueue));
            fp = prevFp->next;
        }

        fp->firstMsg = (IpFragData*) MEM_malloc(sizeof(IpFragData));
        fp->firstMsg->nextMsg = NULL;
        newFragMsg = fp->firstMsg;
        fp->ipFrg_id = ipHeader->ip_id;
        fp->ipFrg_src = ipHeader->ip_src;
        fp->ipFrg_dst = ipHeader->ip_dst;

        fp->actualacketSize = ipHeader->ip_len;
        fp->totalFragmentSize = 0;

        fp->fragHoldTime = getSimTime(node) + IP_FRAGMENT_HOLD_TIME;
        fp->next = NULL;
    }

    ip->stats.ipFragsInBuff++;
    newFragMsg->msg = msg;
    fp->totalFragmentSize += MESSAGE_ReturnPacketSize(msg) -
            sizeof(IpHeaderType);

    // Not all the fragment packet received yet so wait for all to
    // come till atleast hold time
    if ((fp->actualacketSize -  sizeof(IpHeaderType)) != fp->totalFragmentSize)
    {
          *isReassembled = FALSE;
          return NULL;
    }
    else // This is the last fragment packet so try to ressemble it.
    {
        if (fp)
        {
            // Join all the  fragmented packets then return.
            joinedMsg = IpFragementReassamble(node, msg, fp);

            // Delete the fragment queue head
            if (ip->fragmentListFirst == fp)
            {
                ip->fragmentListFirst = fp->next;
            }
            else
            {
                prevFp->next = fp->next;
            }
            MEM_free(fp);

            if (!joinedMsg)
            {
                *isReassembled = FALSE;
                return NULL;
            }
        }

    }
    ip->stats.ipPacketsAfterFragsReasm++;
    *isReassembled = TRUE;
    return joinedMsg;
}


//---------------------------------------------------------------------------
// FUNCTION             : IpFragementReassamble
// PURPOSE             :: Take incoming datagram fragment and try to
//                        reassemble it into whole datagram.  If a chain for
//                        reassembly of this datagram already exists, then it
//                        is given as fp; otherwise have to make a chain.
// PARAMETERS          ::
// +node                : Node* node    : Pointer to Node
// +msg                 : Message* msg  : Pointer to Message Sturcture
// +fp                  : Ipv6FragQueue* fp: Pointer to fragmentation queue.
// RETURN               : Message*  : Pointer to the Message Structure
// NOTES                : Ressambling message function
//---------------------------------------------------------------------------

Message*
IpFragementReassamble(Node* node, Message* msg, IpFragQueue* fp)
{
    NetworkDataIp* ip = (NetworkDataIp *) node->networkData.networkVar;
    char* payload = NULL;
    int packetLen = 0;

    int totalLength = 0;
    int totalVirtualPackeLength = 0;
    char* joinedPayload = NULL;
    int lengthToCopy = 0;
    int reassError = 0;
    int hLen = sizeof(IpHeaderType);

    IpFragData* tempFragData = NULL;
    IpFragData* prevFragData = NULL;
    Message* joinedMsg = NULL;

    NodeAddress sourceAddress;
    NodeAddress destinationAddress;
    TosType priority = 0;
    unsigned char protocol;
    unsigned int hLim;

    // First calculate the total packet length to produce the large packet.

    tempFragData = fp->firstMsg;
    // Calculate the first fragment Information.
    payload = MESSAGE_ReturnPacket(tempFragData->msg);
    IpHeaderType* ipHeader = (IpHeaderType*)payload;

    // Starting fragment offset should be zero.
    if ((ipHeader->ip_fragment_offset << 3) != 0)
    {
        reassError = 1;
        //drop all the fragment packets and return
        tempFragData = fp->firstMsg;
        while (tempFragData)
        {
            prevFragData = tempFragData;
            tempFragData = tempFragData->nextMsg;
            ip->stats.ipReasmFails++;
            ip->stats.ipFragsInBuff--;

            //Trace drop
            ActionData acnData;
            acnData.actionType = DROP;
            acnData.actionComment = DROP_START_FRAGMENT_OFFSET_NOT_ZERO;
            TRACE_PrintTrace(node,
                             prevFragData->msg,
                             TRACE_NETWORK_LAYER,
                             PACKET_IN,
                             &acnData);

            MESSAGE_Free(node, prevFragData->msg);
            MEM_free(prevFragData);
        }
        return NULL;
    }

    totalLength = 0;
    totalVirtualPackeLength = 0;

    // Then calculate packetsize rest of the fragmented packets.
    while (tempFragData->nextMsg != NULL)
    {
        payload = MESSAGE_ReturnPacket(tempFragData->msg);
        ipHeader = (IpHeaderType*)payload;
        int currentFragOff = ipHeader->ip_fragment_offset << 3;

        // Check for the proper offset.
        if (currentFragOff != (totalLength + totalVirtualPackeLength))
        {
            reassError = 1;
            break;
        }

        totalLength += tempFragData->msg->packetSize - sizeof(IpHeaderType);

        totalVirtualPackeLength += tempFragData->msg->virtualPayLoadSize;

        tempFragData = tempFragData->nextMsg;
    }

    // Now add the last packet and virtual packet length
    payload = MESSAGE_ReturnPacket(tempFragData->msg);
    ipHeader = (IpHeaderType*)payload;

    if ((int) (ipHeader->ip_fragment_offset << 3) !=  (totalLength + totalVirtualPackeLength))
    {
        reassError = 1;
    }
    // Last fragment should have more faragment flag as zero.
    int mff = ipHeader->ip_more_fragments;
    if (mff)
    {
        reassError = 1;
    }

    totalLength += tempFragData->msg->packetSize - sizeof(IpHeaderType);
    totalVirtualPackeLength += tempFragData->msg->virtualPayLoadSize;

    if (reassError)
    {
        // Drop all the fragment packets and return
        tempFragData = fp->firstMsg;
        while (tempFragData)
        {
            prevFragData = tempFragData;
            tempFragData = tempFragData->nextMsg;
            ip->stats.ipReasmFails++;
            ip->stats.ipFragsInBuff--;
            //Trace drop
            ActionData acnData;
            acnData.actionType = DROP;
            acnData.actionComment = DROP_ALLFRAGMENT_NOT_COLLECTED;
            TRACE_PrintTrace(node,
                             prevFragData->msg,
                             TRACE_NETWORK_LAYER,
                             PACKET_IN,
                             &acnData);

            MESSAGE_Free(node, prevFragData->msg);
            MEM_free(prevFragData);
        }
        return NULL;
    }

    //Now allocate joined Message data;
    joinedMsg = MESSAGE_Alloc(node,
                           NETWORK_LAYER,
                           NETWORK_PROTOCOL_IP,
                           MSG_NETWORK_Ip_Fragment);

    MESSAGE_PacketAlloc(node, joinedMsg, totalLength, TRACE_IP);

    // Now Make the reassemble Packet. with virtual packet.
    MESSAGE_AddVirtualPayload(node, joinedMsg, totalVirtualPackeLength);

    joinedPayload = MESSAGE_ReturnPacket(joinedMsg);

    // Now copy all the messages.
    tempFragData = fp->firstMsg;

    // Remove the Ip header from the original packet.
    NetworkIpRemoveIpHeader(
        node,
        tempFragData->msg,
        &sourceAddress,
        &destinationAddress,
        &priority,
        &protocol,
        &hLim);

    payload = MESSAGE_ReturnPacket(tempFragData->msg);
    //packetLen = MESSAGE_ReturnPacketSize(tempFragData->msg);
    packetLen = tempFragData->msg->packetSize;

    lengthToCopy = packetLen;
    memcpy(joinedPayload, payload, lengthToCopy);
    ip->stats.ipReasmOKs++;
    ip->stats.ipFragsInBuff--;
//------------------------------------------------------------------------//
// QULNET'S EXTRA OVERHEAD TO MANAGE TO JOIN MESSAGE.
//------------------------------------------------------------------------//
    joinedMsg->sequenceNumber = tempFragData->msg->sequenceNumber;
    joinedMsg->originatingProtocol = tempFragData->msg->originatingProtocol;
    joinedMsg->protocolType = tempFragData->msg->protocolType;
    joinedMsg->layerType = tempFragData->msg->layerType;
    joinedMsg->numberOfHeaders = tempFragData->msg->numberOfHeaders;

    for (int headerCounter = 0;
        headerCounter < tempFragData->msg->numberOfHeaders;
        headerCounter++)
    {
        joinedMsg->headerProtocols[headerCounter] =
            tempFragData->msg->headerProtocols[headerCounter];
        joinedMsg->headerSizes[headerCounter] =
            tempFragData->msg->headerSizes[headerCounter];

    }
    MESSAGE_CopyInfo(node, joinedMsg, tempFragData->msg);
//------------------------------------------------------------------------//
// END OF QUALNET SPECIFIC WORK.
//------------------------------------------------------------------------//

    // Go for next processing.
    joinedPayload += lengthToCopy;
    prevFragData = tempFragData;
    tempFragData = tempFragData->nextMsg;
    MESSAGE_Free(node, prevFragData->msg);
    MEM_free(prevFragData);

    while (tempFragData != NULL)
    {
         payload = MESSAGE_ReturnPacket(tempFragData->msg);

         packetLen = tempFragData->msg->packetSize;
         lengthToCopy = packetLen - hLen;

         if (lengthToCopy > 0)
         {
             memcpy(joinedPayload, payload + hLen, lengthToCopy);
             joinedPayload += lengthToCopy;
         }

         ip->stats.ipReasmOKs++;
         ip->stats.ipFragsInBuff--;
         prevFragData = tempFragData;
         tempFragData = tempFragData->nextMsg;

         MESSAGE_Free(node, prevFragData->msg);
         MEM_free(prevFragData);
    }

    NetworkIpAddHeader(
        node,
        joinedMsg,
        sourceAddress,
        destinationAddress,
        priority,
        protocol,
        hLim);

    //MESSAGE_Free(node, msg);
    return joinedMsg;
}

//-----------------------------------------------------------------------------
// End of fragmentation related code.
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpQueuePeekWithIndex()
// PURPOSE      Like NetworkIpQueueTopPacket(), except a index to the
//            . packet to peek at is given. Note that the message
//              containing the packet is not copied; the contents may
//              (inadvertently or not) be directly modified.
// PARAMETERS   Node *node
//                  Pointer to node.
//              SchedulerType *scheduler
//                  queue to get top packet from.
//              int msgIndex
//                  index to message
//              Message **msg
//                  Storage for pointer to message with IP packet.
//              NodeAddress *nextHopAddress
//                  Storage for packet's next hop address.
//              int *outgoingInterface
//                  Used to determine where packet should go after passing
//                  through the backplane.
//              int *networkType
//                  Whether packet is associated with an IP network, Link-16
//                  nework, etc...
//              QueuePriorityType *priority
//                  Storage for priority of packet.
// RETURN       TRUE if there is a packet, FALSE otherwise.
//-----------------------------------------------------------------------------

BOOL NetworkIpQueuePeekWithIndex(
    Node *node,
    Scheduler *scheduler,
    int msgIndex,
    Message **msg,
    NodeAddress *nextHopAddress,
    int *outgoingInterface,
    int *networkType,
    QueuePriorityType *priority)
{
    QueuePriorityType notUsed = ALL_PRIORITIES;
    QueuePriorityType queuePriority = ALL_PRIORITIES;
    BOOL isPktRetrieved = FALSE;
    QueuedPacketInfo *infoPtr;

    isPktRetrieved = (*scheduler).retrieve(ALL_PRIORITIES,
                                            msgIndex,
                                            msg,
                                            &queuePriority,
                                            PEEK_AT_NEXT_PACKET,
                                            getSimTime(node));

    if (isPktRetrieved)
    {
        ERROR_Assert(*msg != NULL, "Cannot retrieve packet");

        infoPtr = (QueuedPacketInfo *) MESSAGE_ReturnInfo((*msg));

        // Retuning Queue priority
        *priority = queuePriority;
        *nextHopAddress = infoPtr->nextHopAddress;
        *outgoingInterface = infoPtr->outgoingInterface;
        *networkType = infoPtr->networkType;
    }

    return isPktRetrieved;
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpOutputQueuePeekWithIndex()
// PURPOSE      Like NetworkIpQueueTopPacket(), except a index to the
//            . packet to peek at is given. Note that the message
//              containing the packet is not copied; the contents may
//              (inadvertently or not) be directly modified.
// PARAMETERS   Node *node
//                  Pointer to node.
//              int interfaceIndex
//                  Index of interface.
//              int msgIndex
//                  Index of message in queue.
//              Message **msg
//                  Storage for pointer to message with IP packet.
//              NodeAddress *nextHopAddress
//                  Storage for packet's next hop address.
//              QueuePriorityType *priority
//                  Storage for priority of packet.
// RETURN       TRUE if there is a packet, FALSE otherwise.
//
// NOTES        This function is called by MAC_OutputQueueTopPacket()
//              (mac/mac.pc), which itself is called from
//              mac/mac_802_11.pc and other MAC protocol source files.
//
//              This function will assert false if the scheduler cannot
//              return an IP packet for whatever reason.
//-----------------------------------------------------------------------------

BOOL NetworkIpOutputQueuePeekWithIndex(
    Node *node,
    int interfaceIndex,
    int msgIndex,
    Message **msg,
    NodeAddress *nextHopAddress,
    QueuePriorityType *priority)
{
    NetworkDataIp *ip = (NetworkDataIp *) node->networkData.networkVar;
    Scheduler *scheduler = NULL;
    int outgoingInterface;
    int networkType;

    ERROR_Assert(
        interfaceIndex >= 0 && interfaceIndex < node->numberInterfaces,
        "Invalid interface index");

    scheduler = ip->interfaceInfo[interfaceIndex]->scheduler;

    return NetworkIpQueuePeekWithIndex(node,
                                   scheduler,
                                   msgIndex,
                                   msg,
                                   nextHopAddress,
                                   &outgoingInterface,
                                   &networkType,
                                   priority);
}

//-----------------------------------------------------------------------------
// FUNCTION     NetworkIpOutputQueueDequeuePacketWithIndex()
// PURPOSE      Same as NetworkIpOutputQueueDequeuePacket(), except the
//              packet dequeued is specified by an index,
// PARAMETERS   Node *node
//                  Pointer to node.
//              int interfaceIndex
//                  Index of interface.
//              int msgIndex
//                  Index of packet.
//              Message **msg
//                  Storage for pointer to message with IP packet.
//              NodeAddress *nextHopAddress
//                  Storage for packet's next hop address.
//              int *networkType
//                  Type of network (IP, Link-16, ...) used to route packet.
// RETURN       TRUE if dequeued successfully, FALSE otherwise.
//
// NOTES        This function is called by
//              MAC_OutputQueueDequeuePacketWithIndex() (mac/mac.pc),
//              which itself is called from mac/mac_802_11.pc and other
//              MAC protocol source files.
//
//              This function will assert false if the scheduler cannot
//              return an IP packet for whatever reason.
//-----------------------------------------------------------------------------

BOOL NetworkIpOutputQueueDequeuePacketWithIndex(
    Node *node,
    int interfaceIndex,
    int msgIndex,
    Message **msg,
    NodeAddress *nextHopAddress,
    int *networkType)
{
    NetworkDataIp *ip = (NetworkDataIp *)node->networkData.networkVar;
    Scheduler *scheduler = NULL;
    QueuePriorityType queuePriority = ALL_PRIORITIES;
    QueuedPacketInfo *infoPtr;

    ERROR_Assert(
        interfaceIndex >= 0 && interfaceIndex < node->numberInterfaces,
        "Invalid interface index");

    scheduler = ip->interfaceInfo[interfaceIndex]->scheduler;

    if ((*scheduler).retrieve(ALL_PRIORITIES, msgIndex, msg,
                &queuePriority, DEQUEUE_PACKET, getSimTime(node)))
    {
        ERROR_Assert(*msg != NULL, "Cannot dequeue packet");

        infoPtr = (QueuedPacketInfo *) MESSAGE_ReturnInfo((*msg));

        *nextHopAddress = infoPtr->nextHopAddress;
        *networkType = infoPtr->networkType;

        //Trace dequeue
        ActionData acn;
        acn.actionType = DEQUEUE;
        acn.actionComment = NO_COMMENT;
        acn.pktQueue.interfaceID = interfaceIndex;
        acn.pktQueue.queuePriority = queuePriority;
        TRACE_PrintTrace(node, *msg, TRACE_NETWORK_LAYER, PACKET_OUT, &acn);

        (*scheduler).collectGraphData((int) queuePriority,
                                MESSAGE_ReturnPacketSize((*msg)),
                                TIME_getSimTime(node));

#if 0
        //GuiStart
        if (node->guiOption == TRUE)
        {
            GUI_QueueDequeuePacket(node->nodeId, GUI_NETWORK_LAYER,
                                   interfaceIndex, queuePriority,
                                   MESSAGE_ReturnPacketSize((*msg)),
                                   getSimTime(node) + getSimStartTime(node));
        }
        //GuiEnd
#endif
        return TRUE;
    }
    return FALSE;
}
